/*
 * include/revirt/perfctr.h - function prototypes for perfctr functions
 *
 * Copyright (C) 2002 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef __REVIRT_PERFCTR_H
#define __REVIRT_PERFCTR_H

#include <linux/types.h>
#include <revirt/rlog_min.h>
/* 
 * Abstract stuff
 *
 * REVIRT_NR_PFCTRS -- in rlog_min.h
 * revirt_pfctr {
 *    .overflow_msr
 *    .overflow_bit -- right now this assumes 1==enabled
 *    .overflow_clear_mask -- assumes 0 == clear
 *    .count
 *    .list[REVIRT_NR_PFCTRS]
 *       .counter_msr
 *       .arch
 *          [arch-dependent stuff]
 */
struct revirt_pfctr_s {
	int count;
	struct {
		unsigned int control_msr; /* MSR to enable overflow */
		unsigned int enable_bit;  /* Bits in ^ to set and clear */
		unsigned int clear_mask;  /* Bits in ^ to clear after ovf(P4)*/
		int counter_index;	  /* Index of .list[] to set - */
	} overflow;
	int subtract_exceptions;  /* Does the main counter count exceptions? */
	struct {
		unsigned int counter_msr;
		union {
		    struct {
			unsigned int cccr_msr;
			s64 cccr_val;
			unsigned int escr_msr;
			s64 escr_val;
		    } p4;
		    struct {
			unsigned int evtsel_msr;
			s64 evtsel_val;
		    } athlon;
		    struct {
			unsigned int evtsel_msr;
			s64 evtsel_val;
		    } p6;
		} arch;
	} list[REVIRT_NR_PFCTRS];
};

extern struct revirt_pfctr_s revirt_pfctr;

void perfctr_enable_counters(void);
void perfctr_clear_counters(void);
void save_perfctr_regs(struct task_struct * prev);
void restore_perfctr_regs(struct task_struct * next, struct task_struct * prev);
void perfctr_clear_ovf(void);
void perfcntr_enable_overflow(void);
void perfcntr_disable_overflow(void);

extern void perfctr_set_msr(unsigned int addr, s64 x);
extern void perfctr_get_msr(unsigned int addr, s64 * x);
extern void perfctr_set_ctr(unsigned int addr, s64 x);
/* Sign-extend counter */
extern void perfctr_get_ctr(unsigned int addr, s64 * x);
#endif // def __REVIRT_PERFCT_H
