/*
 * include/revirt/revirt.h - options for whole replay system
 *
 * Copyright (C) 2002 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 */
#ifndef _REPLAY_REPLAY_H
#define _REPLAY_REPLAY_H

/* 
 * LOG_SIGNAL_MASK is which signals to log; while REPLAY_SIGNAL_MASK
 * is which signals to squash automatically if they happen on replay.  I.e.,
 * squash SIGIOs and SIGALRMS on replay, because we only want ours to get in;
 * but don't stop SIGKILLs or SIGTERMs.
 *
 * We log SIGSTOP because it causes the process to block.  UML sometimes
 * sends itself a SIGSTOP to 'yield' to the ptracing thread, but the 
 * yield doesn't happen until the signal handling code -- i.e., after the
 * process is returning from the kill(self) system call.  The signal code
 * is used to dealing with this stuff, so I think adding it is the best
 * way to do this.
 * 
 * We don't block additional STOPs in replay, tho, because we want to be
 * able to stop the whole thing from outside.
 */
#define REPLAY_LOG_SIGNAL_MASK ((1 << SIGALRM) | (1 << SIGIO) | (1 << SIGTERM) | (1 << SIGVTALRM) | (1 << SIGWINCH))
#define REPLAY_REPLAY_SIGNAL_MASK ((1 << SIGIO) | (1 << SIGALRM) | (1 << SIGVTALRM) | (1 << SIGWINCH))
#define REPLAY_REPLAY_EXPECTED_SIGS ((1 << SIGSEGV) | (1 << SIGUSR1) \
	| (1 << SIGALRM) | (1 << SIGIO) | (1 << SIGTRAP) | (1 << SIGTERM) \
	| (1 << SIGKILL) | (1 << SIGSTOP) | (1 << SIGVTALRM) | (1 << SIGWINCH))

#endif
