/*
 * include/revirt/rlog.h
 *
 * This file contains definitions relating to the revirt logging system.
 */
#ifndef _RLOG_H
#define _RLOG_H

//#include <linux/sched.h>
#ifdef __KERNEL__
# include <linux/types.h>
# include <asm/current.h>
#else
# include <sys/types.h>
#endif

#include <linux/types.h>

#include <revirt/rlog_min.h>
#include <revirt/replay.h>
#include <revirt/syscalls.h>
/*
 * Common definitions, used by both kernel and userland
 */
#define RLOG_SCHEDULE_ENTRY (0x80)

#define ENTRY_SCHED  (0 | RLOG_SCHEDULE_ENTRY)
#define ENTRY_FORK   (1 | RLOG_SCHEDULE_ENTRY)
#define ENTRY_EXIT   (2 | RLOG_SCHEDULE_ENTRY)
#define ENTRY_TOP    (3 | RLOG_SCHEDULE_ENTRY)
#define ENTRY_SIGNAL	(4 | RLOG_SCHEDULE_ENTRY)
#define ENTRY_EXECPARM (5 | RLOG_SCHEDULE_ENTRY)

//#define ENTRY_PUT_USER	15 // in rlog_min.h
#define ENTRY_COPY_USER	16
#define ENTRY_SYSRET	17
#define ENTRY_CLEAR_USER 18
#define ENTRY_COPY_USER_NODATA 19
#define ENTRY_SIGFRAME 20
#define ENTRY_RT_SIGFRAME 21
#define ENTRY_MARKER 22
//#define ENTRY_SENDERIP 23
#define ENTRY_CHKPT_MARKER 23
#define ENTRY_RDTSC 24
#define ENTRY_END 25
#define ENTRY_IN 26
#define ENTRY_DMA 27
#define ENTRY_GENERIC 28

#define REPLAY_PBUF_MAX_SIZE 10

struct revirt_options_xreplay_st {
    int packetrate;
    int packetsize;
};

#ifdef __KERNEL__
typedef  __kernel_fd_set __revirt_syscall_set;
#else
typedef struct {
	unsigned long fds_bits[32];
} __revirt_syscall_set;
#endif

#ifdef __KERNEL__
typedef  __kernel_fd_set __revirt_fd_set;
#else
typedef struct {
	unsigned long fds_bits[8];
} __revirt_fd_set;
#endif
struct revirt_options_st { 
	/* FIXME -- there's really no need for two of these */
    //unsigned int logflags;
    unsigned int replayflags;
    struct revirt_options_xreplay_st xreplay;
    int network_fd;
    __revirt_fd_set unlogged_fds;
    __revirt_syscall_set allowed_syscalls;
    __revirt_syscall_set unlogged_syscalls;
    __revirt_syscall_set use_unlogged_fds;
    struct {
	int size;
	unsigned int mask;
    } logbuf;
    int replay_debug;
    int ztrace_debug;
};

#define REPLAY_OPT_LOG_WHOLE_SIGFRAME	(1 << 0)
#define REPLAY_OPT_LOG_WHOLE_PTREGS		(1 << 1)
#define REPLAY_OPT_LOG_WHOLE_SIGINFO	(1 << 2)
#define REPLAY_OPT_LOG_SYSCALL_LOGS_BRANCHCOUNT	(1 << 3)
#define REPLAY_OPT_LOG_ALL_SYSCALLS		(1 << 4)
#define REPLAY_OPT_LOG_LOG_SENDERIP		(1 << 5)
#define REPLAY_OPT_REPLAY_DEBUG		(1 << 8)
#define	REPLAY_OPT_ZTRACE_DEBUG		(1 << 9)
#define REPLAY_OPT_THROTTLE_NETWORK	(1 << 10)
#define REPLAY_OPT_WARNING_HALTS	(1 << 11)
#define REPLAY_OPT_RESTRICT_SYSCALLS	(1 << 12)
#define REPLAY_OPT_LOGSYS_ON_REPLAY	(1 << 13)
#define REPLAY_OPT_LOG_SYNCBLOCK	(1 << 14)
#define REPLAY_OPT_NO_SYNCBLOCK		(1 << 15)
#define REPLAY_OPT_EMULATE_RDTSC	(1 << 16)
#define REPLAY_OPT_TRAP_OK		(1 << 17)
#define REPLAY_OPT_NOCHKPT_ZR		(1 << 18)
#define REPLAY_OPT_STK_NOCOW		(1 << 19)
#define REPLAY_OPT_ALWAYS_WRITE_MARKER  (1 << 20)
#define REPLAY_OPT_MARK_CHECKPOINT 	(1 << 21)
#define REPLAY_OPT_CHECKPOINT_DEBUG	(1 << 22)
#define REPLAY_OPT_STK_NOWRITE		(1 << 23)
#define REPLAY_OPT_WAITEVENT		(1 << 24)
#define REPLAY_OPT_DEBUGLOG		(1 << 25)
#define REPLAY_OPT_LOG_NOWRITE		(1 << 26)
#define REPLAY_OPT_LOG_NULL_SCHED	(1 << 27)

#define REPLAY_OPT_VALUE		(1 << 30)
#define REPLAY_OPT_PACKETRATE		(REPLAY_OPT_VALUE | 1)
#define REPLAY_OPT_PACKETSIZE		(REPLAY_OPT_VALUE | 2)
#define REPLAY_OPT_NETWORKFD		(REPLAY_OPT_VALUE | 3)
#define REPLAY_OPT_UNLOGGEDFDS		(REPLAY_OPT_VALUE | 4)
#define REPLAY_OPT_LOGBUFSIZE		(REPLAY_OPT_VALUE | 5)
#define REPLAY_OPT_LOG_SYSCALL		(REPLAY_OPT_VALUE | 6)
#define REPLAY_OPT_UNLOG_SYSCALL	(REPLAY_OPT_VALUE | 7)

//#define REPLAY_OPT_LOGMASK (0xff)
#define REPLAY_OPT_REPLAYMASK (0x3fffffff)

#ifdef __KERNEL__
extern struct revirt_options_st revirt_options;
#ifdef CONFIG_REVIRT_DEBUG
# define Rprintk(x...) \
	revirt_debug_log_printk('R', x)
//#define Cprintk(x...)
# define Cprintk(x...) \
	revirt_debug_log_printk('C', x)
#else
#define Rprintk(x...)
#define Cprintk(x...)
#endif

#endif

//#define REPLAY_RECORD_ALL_EXCEPTIONS # DEFINED IN syscalls.h
#ifdef CONFIG_REVIRT_ALL_SYSCALLS
# warning "Logging all system calls"
# define REPLAY_ALL_SYSCALLS
#endif
//#define REPLAY_SYSCALL_LOGS_BRANCHCOUNT
#ifdef CONFIG_REVIRT_SYSCALL_LOGS_BRANCHCOUNT
# define REPLAY_SYSCALL_LOGS_BRANCHCOUNT
#endif
//#define REPLAY_WHOLE_PTREGS
#ifdef CONFIG_REVIRT_WHOLE_PTREGS
# define REPLAY_WHOLE_PTREGS
#endif
//#define REPLAY_WHOLE_SIGFRAME
#ifdef CONFIG_REVIRT_WHOLE_SIGFRAME
# define REPLAY_WHOLE_SIGFRAME
#endif
//#define REPLAY_WHOLE_SIGINFO
#ifdef CONFIG_REVIRT_WHOLE_SIGINFO
# define REPLAY_WHOLE_SIGINFO
#endif
//#define REPLAY_RECORD_ALL_PFCTRS
#ifdef CONFIG_REVIRT_RECORD_ALL_PFCTRS
# define REPLAY_RECORD_ALL_PFCTRS
#endif
//#define REPLAY_LOG_SENDERIP
#define REPLAY_THROTTLE_NETWORK
#define	REPLAY_UNIQID
#define	REPLAY_ONLY_USR_EXCEPTIONS
#define REPLAY_CHECK_FPOS

#define IS_SCHED(x) ( (x) & RLOG_SCHEDULE_ENTRY )

//#define UML_NETWORK_FD 0x44 // 2.2.20 guest
//#define UML_NETWORK_FD 0x56 // 2.4.18 guest
#define UML_NETWORK_FD 87 // 2.4.18 guest
#define UML_IS_NETWORK_FD(fd) ((fd)==(revirt_options.network_fd))
#define UML_IS_UNLOGGED_FD(fd) (FD_ISSET((fd), &revirt_options.unlogged_fds))

//#define REPLAY_DEFAULT_LOGFLAGS REPLAY_OPT_RESTRICT_SYSCALLS
/* FIXME add RESTRICT_SYSCALLS */
#define REPLAY_DEFAULT_REPLAYFLAGS \
   (REPLAY_OPT_WARNING_HALTS \
    |REPLAY_OPT_EMULATE_RDTSC \
    |REPLAY_OPT_NOCHKPT_ZR)


//#define REVIRT_EFLAGS_MASK (0x40DD5)
/* 
 * This doesn't include the TRAP flag... is that a problem?
 */
#define REVIRT_EFLAGS_MASK (0x40CD5)


struct revirt_ctrdata {
	__s64 hw_branch;
#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
	int exceptions[REVIRT_NR_EX];
#endif
#ifdef REPLAY_RECORD_ALL_PFCTRS
	unsigned int hw_intr;
	unsigned int ex_noerror;
	unsigned int ex_error;
	__s64 counter[REVIRT_NR_PFCTRS];
#endif
};

/*
 * This is all calculated in copy_branchdata now.
 */
#define ADJ_BC(x) ((__s64)(x).hw_branch )
#define ADJ_BC_R(x) ADJ_BC(x)

//#ifndef __KERNEL__
//typedef s64 loff_t;
//#endif

#define REPLAY_GENERIC_SIZE 16
struct rlog_entry_generic {
    __u8 data[REPLAY_GENERIC_SIZE];
    __u8 type;
#ifdef REPLAY_SYSCALL_LOGS_BRANCHCOUNT
    struct revirt_ctrdata ctr;
#endif
};

struct rlog_entry_in {
    __u32 byte;
    __u16 port;
    __u32 eip;
};

struct rlog_entry_end {
    __u16 pid;
};


struct rlog_header {
	__u8 type;
	int len;
	unsigned long id;
	loff_t pos;
};

struct rlog_rdtsc {
	int hi, lo;
	unsigned long eip;
#ifdef REPLAY_RDTSC_LOGS_SID 
	__s64 syscall_id;
#endif
};

struct revirt_senderip {
	unsigned long ip;
};

/*
 * We need this to be three so that we're not forced to log when we have two...
 * 
 * FIXME -- this needs to be setup so that we can deal with having only 2...
 */
#define REVIRT_MAX_ACTIVE_PIDS 3

struct revirt_active_pid_st {
    __u8 i;
    __u16 pid[REVIRT_MAX_ACTIVE_PIDS];
    int epochs_skipped;
    __u8 runnable;
};

struct rlog_exit_entry {
	__u16 pid;
	struct revirt_active_pid_st active;
};

struct rlog_schedule_entry {
	unsigned int flags;
	__u16 pid;
	unsigned long eip;
	int data;
	struct revirt_ctrdata ctr;
	unsigned int edi;
	__u16 nrxuser;
	unsigned long wchan;
	__s64 syscall_id;
	enum {
		STYPE_SCHEDULE,
		STYPE_SIGNAL,
		STYPE_SYNCBLOCK,
		STYPE_END,
		STYPE_DMA,
		STYPE_LAST,
		STYPE_NULL
	} stype;
	struct revirt_active_pid_st active;
	__u8 extra;
};

#define STYPE_MAX 7
#define STYPE_NAMES \
	{ "SCHEDULE", "SIGNAL", "SYNCBLOCK", "END", "DMA", "LAST", "NULL" }

struct rlog_entry_fork {
	__u16 pid;
	__u16 child;
};

struct rlog_put_user_entry {
        __u64 xx;
        unsigned long pptr;
        __s8 ssize, rretval;
        __s16 pid;
};

struct rlog_copy_user_entry {
	unsigned long tto;
	unsigned long ssize;
	__u16 pid;
	__u16 syscall;
	__s64 syscall_id;
};

struct rlog_dma_entry {
	unsigned long dest;
	unsigned long size;
	__u16 pid;
	unsigned long eip;
	int seq;
};

struct rlog_clear_user_entry {
	unsigned long to;
	unsigned long n;
	__u16 pid, syscall;
};

# include <asm/ptrace.h>

struct __min_pt_regs { unsigned int eax; };

struct rlog_entry_sysret {
	__u16 pid;
	__u16 nrxuser;
	__u16 syscall;
	__s64 syscall_id;
#ifdef REPLAY_WHOLE_PTREGS
	struct pt_regs regs;
#else
	struct __min_pt_regs regs;
#endif
#ifdef REPLAY_SYSCALL_LOGS_BRANCHCOUNT
	struct revirt_ctrdata ctr;
#endif
};


struct ztrace_ctl_struct {
    int flags;
    int yield;

    int pid;
#ifdef __KERNEL__
    struct task_struct * tsk;
#else
    void * tsk;
#endif
    struct ztrace_breturn *bret;

    unsigned long rhid;
    unsigned long last_sched_id;
    __s64 perferr;
    __s64 perfctr_save;

    struct {
	unsigned long nrxuser;
	int edi;
    } xstop;

    struct {
	unsigned eip;
	unsigned int wchan;
	int flags;
	__s64 count;
	//int epochs_skipped; /* This is in active */
    } brk;
    struct revirt_active_pid_st active;

    int can_checkpoint;
};

#ifdef __KERNEL__

struct revirt_tsk_s {
	/* Our links to the rest of the world */
	struct task_struct * next;
	struct revirt_group_struct * group;
	struct task_struct * sched;

	int current_syscall;
	__s64 syscall_id;

	/* All things that count */
	struct perfctr_s perfctr[REVIRT_NR_PFCTRS];
	int ecount[REVIRT_NR_EX];
	/* replace with swcounter[] or something */
	/*
	 * 0: interrupts
	 * 1: exceptions w/o an error_code
	 * 2: exceptions w/ an error code
	 */
	unsigned int swctr[REVIRT_NR_SWCTR];
	unsigned long nrxuser;

	/* Per-process replay info */
	struct {
	    struct rlog_header rh;
	    struct rlog_entry_sysret le;
	    __s64 syscall_id;
	} sysret_cache;
	int syserror_replay_unlogged;
	int syserror_write_cleanup;
	unsigned long syserror_save;
};

#define REVIRT_MAX_PATHNAME 256

struct revirt_group_struct {
    struct task_struct * sched;
    struct task_struct * procs;
    int ref;

    struct ztrace_ctl_struct zt_ctl;

    /* Current state */
    struct revirt_active_pid_st active;

    /* Replay file state -- setup in revirt_read_open */
    /* MANUAL RESTORE */
    struct file * scallf; /* == NULL */
    struct {
    	struct rlog_header rh;
	int valid; /* == 0 */
	int l;
    } cache;
    loff_t replay_pos;	/* == 0 */
    loff_t log_pos;
    unsigned int log_id;
    char filename[REVIRT_MAX_PATHNAME];
};

static inline struct revirt_group_struct * get_rgroup(struct task_struct * tsk) {
	if(tsk && tsk->revirt_tsk)
		return tsk->revirt_tsk->group;
	else
		return NULL;
}

/* 
 * This macro sets SUPPRESS_SCHED, saving the old value in flags (so that
 * you can call this in a nested fashion).  To maintian the invariant that
 * SUPPRESS_SCHED requires -- namely, that this process will be the only
 * process in the replay group to be scheduled until SUPPRESS_SCHED is
 * cleared -- we REPLAY_WAIT all the other processes.
 *
 * ASSUMPTION: This assumes that during logging, the "natural" state of
 * a process is NOT to have REPLAY_WAIT set.  If that is not true, we need
 * to find some way to save / restore the state.  
 */
#define revirt_suppress_sched(tsk, flags)				\
    do {								\
    	flags=0; /* Save some compiler warnings */			\
	if(tsk->revirt_flags & REPLAY_REPLAY) {				\
	    flags=tsk->revirt_flags & REPLAY_NO_YIELD_ON_BLOCK;		\
	    tsk->revirt_flags |= REPLAY_NO_YIELD_ON_BLOCK;		\
	} else {							\
	    struct task_struct *p;					\
	    /* SMP: Spin lock here? */					\
	    flags=tsk->revirt_flags & REPLAY_SUPPRESS_SCHED;		\
	    tsk->revirt_flags |= REPLAY_SUPPRESS_SCHED;			\
	    if(tsk->revirt_tsk) {					\
		p=tsk->revirt_tsk->next;				\
		while(p!=tsk) {						\
		    p->revirt_flags |= REPLAY_WAIT;			\
		    p=p->revirt_tsk->next;				\
		}							\
	    }								\
	} 								\
    } while(0)

#define revirt_restore_sched(tsk, flags)				\
    do {								\
	if(tsk->revirt_flags & REPLAY_REPLAY) {				\
	    tsk->revirt_flags&=~REPLAY_NO_YIELD_ON_BLOCK;		\
	    tsk->revirt_flags|=flags;					\
	} else {							\
	    struct task_struct *p;					\
	    tsk->revirt_flags&=~REPLAY_SUPPRESS_SCHED;			\
	    tsk->revirt_flags|=flags;					\
	    if(tsk->revirt_tsk && !(flags & REPLAY_SUPPRESS_SCHED)){	\
		p=tsk->revirt_tsk->next;				\
		while(p != tsk) {					\
		    p->revirt_flags &= ~REPLAY_WAIT;			\
		    p=p->revirt_tsk->next;				\
		} 							\
	    }								\
	}								\
    }while(0)
    
#endif 



/*
 * Stupid, stupid, stupid...
 */
#ifdef __KERNEL__

#include <linux/types.h>
#include <linux/time.h>
#include <linux/times.h>
#include <asm/signal.h>

#else

#include <linux/posix_types.h>

#define _NSIG_K         64
#define _NSIG_BPW_K     32
#define _NSIG_WORDS_K   (_NSIG_K / _NSIG_BPW_K)

typedef struct {
        unsigned long sig[_NSIG_WORDS_K];
} __kernel_sigset_t;

/* from linux/time.h */
/*
 * Needed by struct rlog_marker_entry
 */
struct __kernel_timeval {
	__kernel_time_t		tv_sec;
	__kernel_suseconds_t	tv_usec;
};

#endif

struct rlog_marker_entry {
	//short type;
	int mark;
#ifdef __KERNEL__
	struct timeval tv;
#else
	struct __kernel_timeval tv;
#endif
};


#ifdef __KERNEL__
# include <asm/sigcontext.h>
# include <asm/signal.h>
# include <asm/siginfo.h>
# include <asm/ucontext.h>

#else  /* !__KERNEL__ */
# include <signal.h>
//# define _NSIG           64
# define _NSIG_BPW       32
# define _NSIG_WORDS     (_NSIG / _NSIG_BPW)

# define SI_MAX_SIZE     128
# define SI_PAD_SIZE     ((SI_MAX_SIZE/sizeof(int)) - 3)

typedef union __kernel_sigval {
	int sival_int;
	void * sival_ptr;
} __kernel_sigval_t;

struct __kernel_siginfo {
	int si_signo;
	int si_errno;
	int si_code;

	union {
		int _pad[SI_PAD_SIZE];
		struct {
			__kernel_pid_t _pid;
			__kernel_uid_t _uid;
		} _kill;

		struct {
			unsigned int _timer1;
			unsigned int _timer2;
		} _timer;

		struct {
			__kernel_pid_t _pid;
			__kernel_uid_t _uid;
			__kernel_sigval_t _sigval;
		} _rt;

		struct {
			__kernel_pid_t _pid;
			__kernel_uid_t _uid;
			int _status;
			__kernel_clock_t _utime;
			__kernel_clock_t _stime;
		} _sigchld;

		struct {
			void * _addr;
		} _sigfault;

		struct {
			int _band;
			int _fd;
		} _sigpoll;
	} _sifields;
};

typedef struct __kernel_sigaltstack {
	void * ss_sp;
	int ss_flags;
	__kernel_size_t ss_size;
} __kernel_stack_t;

struct __kernel_ucontext {
	unsigned long uc_flags;
	struct __kernel_ucontext * uc_link;
	__kernel_stack_t	uc_stack;
	struct sigcontext uc_mcontext;
	__kernel_sigset_t uc_sigmask; // defined above
};

#endif /* __KERNEL__ */


/* from arch/i386/kernel/signal.c */
struct __kernel_sigframe {
	char * pretcode;
	int sig;
	struct sigcontext sc;
	struct _fpstate fpstate;
	unsigned long extramask[_NSIG_WORDS-1];
	char retcode[8];
};

struct __kernel_rt_sigframe {
	char *pretcode;
        int sig;
#ifdef __KERNEL__
        struct siginfo *pinfo;
#else
        struct __kernel_siginfo *pinfo;
#endif
        void *puc;
#ifdef __KERNEL__
        struct siginfo info;
        struct ucontext uc;
#else
        struct __kernel_siginfo info;
        struct __kernel_ucontext uc;
#endif
        struct _fpstate fpstate;
        char retcode[8];
};

/*
 * struct sigrame is declared inside signal.c, so even the kernel needs
 * to re-declare it here...
 */
struct rlog_entry_sigframe {
	//short type;
	__u16 pid;
#ifdef REPLAY_WHOLE_SIGFRAME
	struct __kernel_sigframe frame;
#else
	struct {
	    __u8 sig;
	    unsigned long trapno; // sigcontext.trapno
	} frame;
#endif
};

struct rlog_entry_rt_sigframe {
	//short type;
	__u16 pid;
#ifdef REPLAY_WHOLE_SIGFRAME
	struct __kernel_rt_sigframe frame;
#else
	struct {
	    __u8 sig;
	    unsigned long trapno; // frame.uc.uc_mcontext.trapno
	} frame;
#endif
};

#if 1 // REPLAY_LOG_SIGNAL_AT_HANDLING
struct rlog_entry_signal {
	//short type;
	__u16 pid;
//	int n;		/* Number of signals in the queue */
#ifndef REPLAY_WHOLE_SIGINFO
	int si_signo, si_errno, si_code;
	int _pad[2];
#else
# ifdef __KERNEL__
	struct siginfo info;
# else
        struct __kernel_siginfo info;
# endif
#endif

#ifdef __KERNEL__	
	sigset_t blocked;
	sigset_t pending_signal;
	sigset_t dfl;
#else
	__kernel_sigset_t blocked;
	__kernel_sigset_t pending_signal;
	__kernel_sigset_t dfl;
#endif
};
#endif

#if 0 // REPLAY_LOG_SIGNAL_AT_DELIVERY
struct rlog_entry_signal {
	short type;
	int pid;
	int sig;
	int infoval;
#ifndef REPLAY_WHOLE_SIGINFO
	int si_signo, si_errno, si_code;
	int _pad[2];
#else
# ifdef __KERNEL__
	struct siginfo info;
# else
        struct __kernel_siginfo *pinfo;
# endif
#endif

};
#endif


struct rlog_entry_scall_lexecve {
	//int type;
	__u16 pid;
	int fname_index;
	int argv_index;
	int env_index;
};

/*
 * Kernel-only definitions
 */

#ifdef __KERNEL__

#define LOGPAGESIZE 4096
#define NLOGSPERPAGE (LOGPAGESIZE/sizeof(struct log_schedule_entry))

#include <revirt/syscalls.h>

int create_new_rtsk(struct task_struct *, struct task_struct *tsk, int);
int put_rtsk(struct task_struct *);

void save_perfctr(struct task_struct *prev);
void revirt_exit(struct task_struct *);

void revirt_log_schedule(struct task_struct* , int, struct task_struct *);
void revirt_flush_schedule(struct task_struct *);
void revirt_log_exit(void);
void revirt_log_sysret(struct pt_regs  * regs);
void revirt_log_sigframe(struct __kernel_sigframe * frame);
//void revirt_log_signal(struct task_struct *, struct siginfo *);
void revirt_log_sendsig(int sig, struct siginfo *, struct task_struct *);
void revirt_log_rt_sigframe(struct __kernel_rt_sigframe * frame);
void revirt_log_dma(struct task_struct *, void *, void *, int, int);

asmlinkage int revirt_replay_generic(int type, int len, char * data);
void revirt_log_generic(int type, int len, char * data);

void ztrace_tsk_gets_signal(void);
void ztrace_tsk_debug(void);
void ztrace_tsk_perf(void);

int revirt_replay_sigframe(struct __kernel_sigframe *, 
			struct __kernel_sigframe *);
int revirt_replay_rt_sigframe(struct __kernel_rt_sigframe *, 
			struct __kernel_rt_sigframe *);
void revirt_replay_rdtsc(int *, int *, unsigned long);
void revirt_log_rdtsc(int, int, unsigned long);

int revirt_read_open(char *, struct revirt_group_struct *);
int revirt_read_close(struct revirt_group_struct *);
int revirt_read_log(struct rlog_header *);
void clear_branchdata(struct task_struct *);
void copy_branchdata(struct task_struct *, struct revirt_ctrdata *);
unsigned int revirt_getwchan(struct task_struct *);

u32 revirt_replay_in(u16 port, u32 eip);
void revirt_log_in(u32 byte, u16 port, u32 eip);

int revirt_alloc_log_buf(int);
asmlinkage int revirt_write_log(char *buf, int type, int len, 
	struct revirt_group_struct *rgroup);
asmlinkage int revirt_writev_log(int type, const struct iovec *v, int n, 
	struct revirt_group_struct *rgroup);

#ifdef REPLAY_UNIQID
int rlog_resetid(unsigned int i);
#endif

#include <linux/uio.h>
extern unsigned int revirt_interrupts;
extern unsigned int revirt_exceptions1;
extern unsigned int revirt_exceptions2;
extern unsigned int revirt_current_syscall;
#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
extern unsigned int revirt_ecount[REVIRT_NR_EX];
#endif
extern struct revirt_options_st revirt_options;

extern unsigned int revirt_edi;

extern __s64 zt_perferr;

extern char revirt_scall_buf[];

#ifdef CONFIG_REVIRT_DEBUG
int revirt_debug_log_printk(char,
	const char *, 
	unsigned long,
	unsigned long,
	unsigned long,
	unsigned long,
	unsigned long);
#else
#define revirt_debug_log_printk(x...)
#endif

# endif /* __KERNEL__ */

#endif /* _RLOG_H */
