/*
 * checkpoint.c -- Functions relating to revirt and checkpointing.
 */

#include <linux/sched.h>
#include <linux/slab.h>
#include <revirt/checkpoint.h>

#include <linux/module.h>

#include <linux/signal.h>

int revirt_free_checkpoint(struct revirt_checkpoint_struct * chkpt) {
	kfree(chkpt);
	return 0;
}

EXPORT_SYMBOL_NOVERS(revirt_free_checkpoint);


struct revirt_checkpoint_struct * revirt_take_replay_checkpoint(struct task_struct *tsk) {
    struct revirt_checkpoint_struct * chkpt;
    struct task_struct * p;
    struct revirt_tsk_s * rtsk;
    struct revirt_group_struct * rgroup;
    int i=0;

    if(!(rtsk=tsk->revirt_tsk) || !(rgroup=tsk->revirt_tsk->group)) {
    	printk("revirt_take_replay_checkpoint: !(revirt_tsk && revirt_group)\n");
	return NULL;
    }

    if(!rgroup->zt_ctl.can_checkpoint) {
    	printk("revirt_take_replay_checkpoint: zt_ctl.can_checkpoint not set.\n");
	return NULL;
    }

    if((chkpt=kmalloc(sizeof(*chkpt), GFP_KERNEL))==NULL) {
    	printk("revirt_take_replay_checkpoint: unable to malloc!\n");
	return NULL;
    }

    chkpt->group = *rgroup;
    Cprintk("chkpt: rgroup replay_pos %lld (%llx) log_pos %lld (%llx)\n",
    		rgroup->replay_pos,
    		rgroup->replay_pos,
    		rgroup->log_pos,
    		rgroup->log_pos,0);
    Cprintk("    zt_ctl fl %x pid %d rhid %lu\n",
		rgroup->zt_ctl.flags,
		rgroup->zt_ctl.pid,
		rgroup->zt_ctl.rhid,0,0);
    if(rgroup->zt_ctl.flags != ZTRACE_EPOCH_STOP) {
      //printk("      WARNING, unusual replay chkpt state, please inspect manually. [pause]\n");
	//ztrace_debug_pause();
    }
    /* 
     * Make sure we have the most current stuff
     */
    save_perfctr_regs(current);
    p=tsk;
    do {
	if(i>= MAX_CHECKPOINT_PIDS) {
	    printk("revirt_take_replay_checkpoint: %d pids, max %d!\n",
	    		i, MAX_CHECKPOINT_PIDS);
	    goto out_free;
	}
	if(!(p->revirt_flags & REPLAY_REPLAY)) {
	    printk("revirt_take_replay_checkpoint: task %d is not replaying!\n", p->pid);
    	    goto out_free;
       	}
	Cprintk("chkpt: pid %d, flags 0x%x\n", p->pid, p->revirt_flags,0,0,0);
	Cprintk("  rtsk: current_syscall %d sid %lld replay_unlogged %d\n",
			rtsk->current_syscall,
			rtsk->syscall_id,
			rtsk->syserror_replay_unlogged,0,0);
	chkpt->proc[i].revirt_flags = p->revirt_flags;
	chkpt->proc[i].pid = p->pid;
	chkpt->proc[i].rtsk = *rtsk;

	if(!(p=rtsk->next)) {
	    printk("revirt_take_replay_checkpoint: rtsk->next==NULL, ring broken!\n");
	    goto out_free;
	}
	if(!(rtsk=p->revirt_tsk)) {
	    printk("revirt_take_replay_checkpoint: tsk %d has no revirt_tsk, ring broken!\n",
	    	p->pid);
	    goto out_free;
	}
	i++;
    } while(p!=tsk);
    chkpt->nrpids=i;

    return chkpt;

   out_free:
    kfree(chkpt);
    return NULL;
}

EXPORT_SYMBOL_NOVERS(revirt_take_replay_checkpoint);

struct revirt_checkpoint_struct * revirt_take_log_checkpoint(struct task_struct *tsk) {
    struct revirt_checkpoint_struct * chkpt;
    struct task_struct * p;
    struct revirt_tsk_s * rtsk;
    struct revirt_group_struct * rgroup;
    int i=0;

    if(!(rtsk=tsk->revirt_tsk) || !(rgroup=tsk->revirt_tsk->group)) {
    	printk("revirt_take_log_checkpoint: !(revirt_tsk && revirt_group)\n");
	return NULL;
    }


    if((chkpt=kmalloc(sizeof(*chkpt), GFP_KERNEL))==NULL) {
    	printk("revirt_take_log_checkpoint: unable to malloc!\n");
	return NULL;
    }

    chkpt->group = *rgroup;

    chkpt->group.scallf=NULL;
    chkpt->group.cache.valid=0;

    chkpt->group.replay_pos = chkpt->group.log_pos;

    /* 
     * These are values that will probably actually be used...
     */

    chkpt->group.zt_ctl.flags=ZTRACE_RESTORE_CHKPT;
    chkpt->group.zt_ctl.yield=ZTRACE_RESTORE_CHKPT;
    chkpt->group.zt_ctl.pid=tsk->pid;
    chkpt->group.zt_ctl.tsk = tsk;
    chkpt->group.zt_ctl.bret=NULL;
    chkpt->group.zt_ctl.rhid = chkpt->group.log_id;
    /* last_sched_id should be modified in-place by log_schedule */
    chkpt->group.zt_ctl.perferr=0;
    chkpt->group.zt_ctl.perfctr_save=0; 
    chkpt->group.zt_ctl.active.epochs_skipped=0;
    /* FIXME -- eventually need to make logging fill in zt_ctl.xstop info */

    Cprintk("chkpt: rgroup log_pos %lld zt_ctl fl %x pid %d rhid %lu epochs_skipped %d\n",
    		rgroup->log_pos,
		rgroup->zt_ctl.flags,
		rgroup->zt_ctl.pid,
		rgroup->zt_ctl.rhid,
		rgroup->active.epochs_skipped);
    /* 
     * Make sure we have the most current stuff
     */
    save_perfctr_regs(current);
    p=tsk;
    i=0;
    do {
	if(i>= MAX_CHECKPOINT_PIDS) {
	    printk("revirt_take_log_checkpoint: %d pids, max %d!\n",
	    		i, MAX_CHECKPOINT_PIDS);
	    goto out_free;
	}
	if(!(p->revirt_flags & REPLAY_LOG)) {
	    printk("revirt_take_replay_checkpoint: task %d is not replaying!\n", p->pid);
    	    goto out_free;
       	}
	Cprintk("chkpt: pid %d, flags 0x%x\n", p->pid, p->revirt_flags, 0,0,0);
	Cprintk("  rtsk: current_syscall %d sid %lld replay_unlogged %d\n",
			rtsk->current_syscall,
			rtsk->syscall_id,
			rtsk->syserror_replay_unlogged, 0, 0);
	chkpt->proc[i].revirt_flags = p->revirt_flags;
	chkpt->proc[i].pid = p->pid;
	chkpt->proc[i].rtsk = *rtsk;

	/*
	 * FIXME -- do we have to do anything with these?
	 */
	chkpt->proc[i].rtsk.syserror_replay_unlogged=0;
	chkpt->proc[i].rtsk.syserror_write_cleanup=0;
	chkpt->proc[i].rtsk.syserror_save=0;


	if(!(p=rtsk->next)) {
	    printk("revirt_take_replay_checkpoint: rtsk->next==NULL, ring broken!\n");
	    goto out_free;
	}
	if(!(rtsk=p->revirt_tsk)) {
	    printk("revirt_take_replay_checkpoint: tsk %d has no revirt_tsk, ring broken!\n",
	    	p->pid);
	    goto out_free;
	}
	i++;
    } while(p!=tsk);
    chkpt->nrpids=i;

    return chkpt; 
   out_free:
    kfree(chkpt);
    return NULL;

}

struct revirt_checkpoint_struct * revirt_take_checkpoint(struct task_struct *tsk) {
	if(tsk->revirt_flags & REPLAY_REPLAY) {
	    return revirt_take_replay_checkpoint(tsk);
	} else if(tsk->revirt_flags & REPLAY_LOG) {
	    return revirt_take_log_checkpoint(tsk);
	}
	else return NULL;
}

EXPORT_SYMBOL_NOVERS(revirt_take_checkpoint);


int revirt_restore_checkpoint(struct task_struct *tsk,  
	struct task_struct *active_tsk,
	struct revirt_checkpoint_struct * chkpt) {
    int i;
    struct task_struct *p, *sched=NULL;
    struct revirt_group_struct * rgroup=NULL;
    struct revirt_tsk_s * rtsk, *crtsk;
    int flags;
    int log_to_replay=0;

    if(!chkpt) {
    	printk("revirt_restore_checkpoint: NULL pointer, not restoring\n");
	return 0;
    }

    Cprintk("restore: chkpt->group.log_pos %lld (%llx)\n", 
    	chkpt->group.log_pos,
    	chkpt->group.log_pos, 0,0,0);

    if(tsk->sigpending) {
      	printk("revirt_restore_checkpoint: signals in queue!  Removing...\n");
	while(1) {
	    int signr;
	    siginfo_t info;
	    spin_lock_irq(&tsk->sigmask_lock);
	    signr=dequeue_signal(&tsk->blocked, &info);
	    spin_unlock_irq(&tsk->sigmask_lock);
	    if(!signr)
	    	break;
	    else
	    	printk("revirt_restore_checkpoint: removing signal %d\n", signr);
	}

    }

    /* 
     * If we are currently logging and are restoring to a checkpoint, we must
     * be wanting to switch into replay mode.
     *
     * Be advised that if the logging is "paused", we won't be doing this.
     *
     * (I suppose technically someone might want the functionality of 
     * checkpointing, going back, and taking a different path of logging.)
     * 
     * FIXME -- we should probably make this an explicit move... i.e., if you
     * want to start replaying, actually switch it to replay on purpose... 
     */
    if(tsk->revirt_flags & REPLAY_LOG) {
	struct rlog_entry_end le;
    	printk("revirt_restore_checkpoint: WARNING, switching over to replay from logging!\n");
	revirt_log_schedule(tsk, STYPE_END, NULL);
	le.pid = tsk->pid;
	revirt_write_log((char *)&le, ENTRY_END, sizeof(le), NULL);
	log_to_replay=1;
    }

    for(i=0; i<chkpt->nrpids; i++) {
	if((p=find_task_by_pid(chkpt->proc[i].pid))==NULL) {
	    printk("revirt_restore_checkpoint: can't find pid %d\n",
    		    chkpt->proc[i].pid);
    	    return -ESRCH;
       	}
	if(!(rtsk=p->revirt_tsk)) {
	    printk("restore: pid %d has no revirt_tsk!\n", p->pid);
	    return -EBUSY;
	}

	if(rgroup && rgroup != rtsk->group) {
	    printk("restore:pid %d rgroup (%x)!= other groups (%x)\n", p->pid,
	    		rtsk->group, rgroup);
	    return -EBUSY;
	} else {
	    rgroup=rtsk->group;
	    if(!rgroup) {
		printk("restore:pid %d no rgroup!\n", p->pid);
		return -EBUSY;
	    }
	    if((p->revirt_flags & REPLAY_REPLAY) 
	    		&& (!rgroup->zt_ctl.can_checkpoint)) {
	    	printk("restore: zt_ctl.can_checkpoint not set!\n");
		return -EBUSY;
	    }
	}

	flags = chkpt->proc[i].revirt_flags;

	/* It's not a given that we're switching to replay mode */
	if(log_to_replay) {
	    flags &= ~REPLAY_LOG;
	    flags |= REPLAY_REPLAY;
	    if(p != active_tsk) {
		Cprintk("restore: WAITing non-active pid %d\n", p->pid,
			0,0,0,0);
		flags |= REPLAY_WAIT;
	    }
	}

	Cprintk("restoring pid %d:  flags current %x chkpt %x restored %x\n", 
			p->pid, p->revirt_flags, 
			chkpt->proc[i].revirt_flags,
			flags,0);

	p->revirt_flags =flags; 

	if(flags & (REPLAY_DELAY_PERF|REPLAY_ZOMBIE|REPLAY_SSTEP)) {
	    printk("WARNING, unexpected flags %d\n",
	    	flags & (REPLAY_DELAY_PERF|REPLAY_ZOMBIE|REPLAY_SSTEP));
	    //ztrace_debug_pause();
	}

	crtsk= &chkpt->proc[i].rtsk;

	if(!crtsk->next) {
		Cprintk("  tsk %d chkpt next NULL, using current values\n", p->pid,0,0,0,0);
		crtsk->next = rtsk->next;
	}
	if(!crtsk->group) {
		crtsk->group = rtsk->group;
	}
	if(!crtsk->sched) {
		Cprintk("  tsk %d chkpt sched NULL, using current value\n", p->pid,0,0,0,0);
		crtsk->sched = rtsk->sched;
	}

	if(rtsk->next == crtsk->next
	    && rtsk->group == crtsk->group
	    && rtsk->sched == crtsk->sched) {
	    /* 
	     * The values should've been massaged in log_checkpoint
	     */
	    *rtsk = *crtsk;
	} else {
		printk("mismatch! %x %x, %x %x, %x %x\n",
			rtsk->next,
			crtsk->next,
			rtsk->group,
			crtsk->group,
			rtsk->sched,
			crtsk->sched);
		return -EBUSY;
	}
    }
    if(rgroup) {
	if(log_to_replay) {
	    if(!rgroup->sched) {
		printk("restore: waiting for scheduler...\n");
		ztrace_debug_pause();
	    }
	    if(!rgroup->sched) {
		printk("Still no scheduler, return -EBUSY\n");
		return -EBUSY;
	    }
	}

	if(!chkpt->group.sched) {
		Cprintk("  rgroup sched NULL, using current values\n",0,0,0,0,0);
		chkpt->group.sched=rgroup->sched;
	}
	if(!chkpt->group.procs) {
		chkpt->group.procs=rgroup->procs;
	}
	if(!chkpt->group.scallf) {
		chkpt->group.scallf=rgroup->scallf;
	}

	if( rgroup->sched == chkpt->group.sched
     		&& rgroup->procs == chkpt->group.procs
     		&& rgroup->scallf == chkpt->group.scallf) {
	    *rgroup = chkpt->group;
	    Cprintk("restore: moving rgroup->zt_ctl.yield from %x to %x\n",
		    rgroup->zt_ctl.yield, ZTRACE_RESTORE_CHKPT,0,0,0);
	    rgroup->zt_ctl.yield=ZTRACE_RESTORE_CHKPT;
	    Cprintk("restore: replay_pos %lld (%llx), log_pos %lld (%llx) zt_ctl.rhid %lu\n", 
	    		rgroup->replay_pos, 
	    		rgroup->replay_pos, 
			rgroup->log_pos, 
			rgroup->log_pos, 
			rgroup->zt_ctl.rhid);

	    if(log_to_replay) {
	    	int error;
		printk("restore: trying to open scallf %s\n", rgroup->filename);
		error=revirt_read_open(rgroup->filename, rgroup);
		if(error < 0) {
		    printk("restore: couldn't open scallf (%d)!\n", error);
		    return error;
		}
	    }
	    	
	} else {
	    printk("restore: mismatch somewhere!\n");
    	    printk(" sched: %x %x\n",
			rgroup->sched, chkpt->group.sched);
    	    printk(" procs: %x %x\n",
		rgroup->procs, chkpt->group.procs);
    	    printk(" scallf: %x %x\n",
			rgroup->scallf, chkpt->group.scallf);
	    return -EBUSY;
	}
    } else {
    	printk("restore: didn't find rgroup!\n");
	return -EBUSY;
    }

    restore_perfctr_regs(current, NULL);
	
    ztrace_sync_ztc(&rgroup->zt_ctl);

    return 0;
}

EXPORT_SYMBOL_NOVERS(revirt_restore_checkpoint);
