/*
 * log.c - Functions associated with ReVirt logging system
 *
 * Copyright (C) 2002-2003 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <linux/linkage.h>
#include <linux/sched.h>
#include <asm/current.h>
#include <linux/smp_lock.h>
#include <asm/semaphore.h>
#include <linux/tqueue.h>
#include <linux/slab.h>
#include <asm/perfctr.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/processor.h>
#include <asm/debugreg.h>
#include <asm/user.h>
#include <revirt/rlog.h>
#include <revirt/ztrace.h>
#include <revirt/perfctr.h>
#include <asm/unistd.h>

#include <asm/fixmap.h>
#include <asm/apic.h>
#include <asm/mtrr.h>
#include <asm/siginfo.h>

#include <linux/module.h>

/*
 * FIXME -- inline, put in .h if necessary
 */
void copy_branchdata(struct task_struct *from, 
			    struct revirt_ctrdata *to) {
	struct revirt_tsk_s * rtsk = from->revirt_tsk;
	struct ztrace_ctl_struct * ztc;
	if(!rtsk)
		return;

	ztc=get_ztc(from);

	if(from == current) {
		save_perfctr_regs(from);
	}

#ifdef REPLAY_DEBUG
	if(rtsk->perfctr[0].ctr >= (1LL<<32)) {
		printk("copy_branchdata: warning: large count: 0x%llx\n",
			rtsk->perfctr[0].ctr);
	}
#endif

	to->hw_branch=rtsk->perfctr[revirt_pfctr.overflow.counter_index].ctr;
	if(revirt_pfctr.subtract_exceptions)
		to->hw_branch-=(rtsk->swctr[0]+rtsk->swctr[1]+rtsk->swctr[2]);
	if(ztc && from->revirt_flags & REPLAY_REPLAY 
		&& from->pid == ztc->pid)
		to->hw_branch+=ztc->perferr;
#ifdef REPLAY_RECORD_ALL_PFCTRS
        to->hw_intr=rtsk->swctr[0];
	to->ex_noerror = rtsk->swctr[1];
	to->ex_error= rtsk->swctr[2];

	do {
	    int i;
	    for(i=0; i<REVIRT_NR_PFCTRS; i++)
		to->counter[i]=rtsk->perfctr[i].ctr;
	} while(0);
#endif

#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
	do {
	    int i;
	    for(i=0; i<REVIRT_NR_EX; i++)
	    	to->exceptions[i]=rtsk->ecount[i];
	} while(0);
#endif
}

EXPORT_SYMBOL_NOVERS(copy_branchdata);

void clear_branchdata(struct task_struct *tsk) {
	int i;
	struct revirt_tsk_s * rtsk = tsk->revirt_tsk;
	if(rtsk==NULL)
		return;
	/* FIXME -- optimize here? */
	for(i=0; i<revirt_pfctr.count; i++)
		rtsk->perfctr[i].ctr=0;
	if(revirt_pfctr.subtract_exceptions) {
	    for(i=0; i<REVIRT_NR_SWCTR; i++)
		rtsk->swctr[i]=0;
	}
#if 0
	for(i=0; i<REVIRT_NR_PFCTRS; i++)
		rtsk->perfctr[i].ctr=0;
	for(i=0; i<REVIRT_NR_SWCTR; i++)
		rtsk->swctr[i]=0;
#endif
#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
	do { 
		int i; 
		for(i=0; i<REVIRT_NR_EX; i++)  {
    		    rtsk->ecount[i]=0; 
		}
	} while(0);
#endif
	/* 
	 * No, we don't want to clear nrxuer here... if it actually does get interrupted
	 * between to xusers, we want the nrxuser thing to remain consistent.
	 */
	//rtsk->nrxuser=0;

	if(tsk == current) {
		restore_perfctr_regs(tsk, NULL);
	}
}

void clear_branchdata_group(struct task_struct * tsk) {
    struct task_struct * p=tsk;
    do {
	clear_branchdata(p);
	if(p->revirt_tsk)
	    p=p->revirt_tsk->next;
	else
	    printk("clear_branchdata_group: pid %d has no revirt_tsk, ring broken!\n", 
	    		p->pid);
    } while (p!=tsk);
}


/*
 * Copied from arch/i386/kernel/process.c:get_wchan()
 */
extern void scheduling_functions_start_here(void);
extern void scheduling_functions_end_here(void);
#define first_sched     ((unsigned long) scheduling_functions_start_here)
#define last_sched      ((unsigned long) scheduling_functions_end_here)

unsigned int revirt_getwchan(struct task_struct * prev) {
          unsigned long ebp, esp, eip;
          unsigned long stack_page;
          int count=0;

	  stack_page = (unsigned long) prev;
          esp=prev->thread.esp;
          if(!stack_page || esp < stack_page || esp > 8188+stack_page)
                count=16;
          ebp = * (unsigned long *) esp;
          do {
                if(ebp < stack_page || ebp > 8184+stack_page)
                        return 0;
                eip = *(unsigned long *) (ebp + 4);
                if(eip < first_sched || eip >= last_sched) {
                        return eip;
                }
                ebp = * (unsigned long *) ebp;
                        
          } while(count++<16);
	  return 0;
}

spinlock_t revirt_schedule_spinlock = SPIN_LOCK_UNLOCKED;

void revirt_log_putuser(u64 x, unsigned long ptr, int size, int retval) {
    struct rlog_put_user_entry le = {
    	//.type = ENTRY_PUT_USER,
	.xx = x,
	.pptr = ptr,
	.ssize = size,
	.rretval = retval,
	.pid = current->pid,
	//.syscall = revirt_current_syscall
    };
    revirt_write_log((char *)&le, ENTRY_PUT_USER, sizeof(le), NULL);
    current->revirt_tsk->nrxuser++;
}

//struct revirt_active_pid_st revirt_active_pids = { .i=0, .pid={0, }, .epochs_skipped=0 };
//int revirt_runnable = 1;
void update_active_pids(struct task_struct * tsk,
		struct revirt_active_pid_st * active) {
    int i;
   /*
    * Look through the list of active pids for the current pid.  If it's not
    * found, add it in.
    */
    for(i=0; i<active->i; i++) {
	if(active->pid[i]==tsk->pid) 
	    break;
    }

    if(i == active->i) {
	active->pid[i]=tsk->pid;
	active->i++;
    }
}

void clear_active_pids(struct revirt_active_pid_st * active) {
    int i;
    /*
     * Clear active pids
     */
    active->epochs_skipped=0;
    active->i=0;
    for(i=0; i<REVIRT_MAX_ACTIVE_PIDS; i++) {
	active->pid[i]=0;
    }
}

/* 
 * Normally, schedule entries are made when a logged process is scheduled
 * out.  However, it's not uncommon (and sometimes it's incredibly common)
 * for the same pid to be scheduled again right away.  In this case, the 
 * first schedule entry is redundant.
 * 
 * However, at the time we take the schedule, we don't know who will be
 * scheduled next, so we don't know if the entry is redundant or not.
 *
 * So, if it's a candidate (STYPE_SCHEDULE only), we write it to a temporary
 * buffer and *don't* clear the branch counts or active pid list.
 *
 * So when do we write it?  We need to write it before any other log is
 * written, and we need to clear the branch counts before anyone else
 * copies them into some other log.  We could, I suppose, call 
 * revirt_flush_schedule from do_schedule() when we're scheduling in some
 * other process.
 */
static struct rlog_schedule_entry sched_tmp = { .stype = STYPE_NULL };
static struct task_struct * sched_tmp_tsk=NULL;

/* This will be called by schedule() if last is not logging but next is */
void revirt_flush_schedule(struct task_struct * next) {
    if(sched_tmp.stype == STYPE_SCHEDULE) {
	if(next->pid == sched_tmp.pid) {
	    sched_tmp.stype=STYPE_NULL;
	    /* For debugging, write it anyway? */
	    if(revirt_options.replayflags & REPLAY_OPT_LOG_NULL_SCHED) 
		revirt_write_log((char *)&sched_tmp, ENTRY_SCHED, 
			sizeof(sched_tmp), sched_tmp_tsk->revirt_tsk->group);
	} else {
	    struct task_struct * tsk= sched_tmp_tsk;
	    struct revirt_active_pid_st * active=NULL;
    	    active = &tsk->revirt_tsk->group->active;

	    /* Clear counters, active structs, etc. */
    	    clear_active_pids(active);
    	    if(sched_tmp.active.i >1) {
    		clear_branchdata_group(tsk);
    	    } else {
    		clear_branchdata(tsk);
    	    }
	    tsk->revirt_flags &= ~(REPLAY_FIRST_AFTER_SIG);

	    /* Write log */
	    if(tsk->revirt_tsk->group) {
		tsk->revirt_tsk->group->zt_ctl.last_sched_id 
			= tsk->revirt_tsk->group->log_id;
	    }
	    revirt_write_log((char *)&sched_tmp, ENTRY_SCHED, sizeof(sched_tmp), tsk->revirt_tsk->group);
	}

	/* In any case, clear it */
	sched_tmp.stype = STYPE_NULL;
	sched_tmp_tsk=NULL;
    }
}

void revirt_log_schedule(struct task_struct* tsk, int stype, struct task_struct * next) {
        struct rlog_schedule_entry le;
	int flags;
	struct task_struct * p;
	struct revirt_active_pid_st * active;
	int runnable=0;
	int clear_pids=1;
	int write_log=1;


	if(!(tsk->revirt_flags & REPLAY_LOG)) {
	    printk("revirt_log_schedule called for unlogged process. \n pid: 0x%x\n" \
	    	   " revirt_flags: 0x%x\n", 
	    		tsk->pid, tsk->revirt_flags);
	    BUG();
	}


	/*
	 * Don't bother logging schedules after a process has called
	 * do_exit.
	 */
	if(tsk->revirt_flags & (REPLAY_ZOMBIE|REPLAY_SUPPRESS_SCHED)){
		if(stype==STYPE_SCHEDULE) {
		    return;
		}
	}

	if(!tsk->revirt_tsk) {
		printk("revirt_log_schedule: pid %d has no revirt_tsk!\n",
				tsk->pid);
		return;
	}

	if(sched_tmp.stype == STYPE_SCHEDULE) {
		printk("revirt_log_schedule: sched_tmp.stype SCHED! pid %d tsk %d\n",
			sched_tmp.pid, tsk->pid);
			
	}

	le.extra=0;

	if(tsk->revirt_tsk->group) {
    	    active = &tsk->revirt_tsk->group->active;

    	    update_active_pids(tsk, active);

	/*
	 * See how many processes are runnable.  We need to do this every time because
	 * we need to log if:
	 *  There are currently 2 processes runnable
	 *  At the last schedule, there were 2 procs runnable (even if now one of those
	 *   is syncronously blocked).
	 */
    	    p=tsk;
    	    do {
    		if(p->state == TASK_RUNNING)
    		    runnable++;
       		if(p->revirt_tsk)
		    p=p->revirt_tsk->next;
		else
		    break;
	    } while (p!=tsk);

	/*
	 * See if we can skip this log.  We can skip syncronous blocks if:
	 *  - if the option isn't disabled
	 *  - we have a scheduling log (all signals must be logged)
	 *  - the task is blocked
	 *  - it's blocked in a designated "syncronous block" section
	 *  - it is the only pid runnable
	 *  - the last schedule, there was only one pid runnable
	 *  - the 'active pids' array won't be full if some new process runs next 	
	 *	scheduling epoch.
	 *
	 * FIXME -- it might be nice to not need to leave the extra array entry that's
	 * always open, as we do now...
	 */
	    if(!(revirt_options.replayflags & REPLAY_OPT_NO_SYNCBLOCK)
		&& stype==STYPE_SCHEDULE
		&& tsk->state != TASK_RUNNING
		&& tsk->revirt_flags & REPLAY_SYNC_BLOCK
		&& runnable==1
		&& active->runnable==1
		&& active->i < REVIRT_MAX_ACTIVE_PIDS) {

		stype = STYPE_SYNCBLOCK;
		clear_pids=0;
	    } 
	    else {
		if(stype != STYPE_SCHEDULE) 
			le.extra |= 1;
		if(tsk->state == TASK_RUNNING) 
			le.extra |= 2;
		if(!(tsk->revirt_flags & REPLAY_SYNC_BLOCK))
			le.extra |= 4;
		if(runnable != 1)
			le.extra |= 8;
		if(active->runnable != 1)
			le.extra |= 16;
		if(active->i >= REVIRT_MAX_ACTIVE_PIDS)
			le.extra |= 32;
	    }
	
	    active->runnable = runnable;
	    le.active = *active;
	} else {
	    active=NULL;
	    le.extra = 64;
	 }
	
	/*
	 * FIXME -- we probably need stronger locking here... if we get a schedule
	 * between the unlock and the write, things are going to be out of order!
	 *
	 * Hmm... I wonder if we could separate out getting the log id, and then lock
	 * the entire log creation...
	 */

	/* FIXME - Do we need spinlocking? */
	copy_branchdata(tsk,&le.ctr);
	
        le.pid= tsk->pid;
	le.stype = stype;

	le.eip=zreg(tsk,eip);
	/* Plato and GDB insert int3's which won't be there during replay.
	 * Logically, any "schedule" that happens after one of their int3's
	 * happens at the int3 itself. */
	if(tsk->revirt_flags & REPLAY_INT3) {
	    le.eip-=1;
	}
	le.data=zreg(tsk,ecx);

	le.nrxuser = tsk->revirt_tsk->nrxuser;
	le.syscall_id=tsk->revirt_tsk->syscall_id;
	le.flags = tsk->revirt_flags;
	le.flags |= REPLAY_TSK_SETSTATE(tsk->state);


	if(tsk->revirt_flags & REPLAY_IN_COPY_USER) {
	  le.edi = revirt_edi;
	  printk("revirt_log_schedule called in __copy_user_atom. edi==0x%x\n", 
	  		revirt_edi);
	 }
	else
	  le.edi = 0;

	/*
	 * If we're blocked, find out where.  Code copied from
	 * arch/i386/kernel/process.c:get_wchan().  Note that the
	 * switch_to() macro has already stored the proper esp in the
	 * task struct, so the fact that tsk == current is OK.
	 *
	 * FIXME are we sure that tsk->state != TASK_RUNNING implies
	 * that revirt_log_schedule was called from __switch_to?
	 *
	 * OK, for now we look anytime process is in umode.  Will this break
	 * anything?
	 */
	le.wchan=0;

	if(tsk->state != TASK_RUNNING) {
		le.wchan = revirt_getwchan(tsk);
	} 

	/* Implies active!=NULL */
	if(stype == STYPE_SYNCBLOCK)  {
	    active->epochs_skipped++;
	    if(!(revirt_options.replayflags & REPLAY_OPT_LOG_SYNCBLOCK))
		write_log = 0;
	}

	/* See if this might be a redundant schedule:
	    - Is an actual schedule entry
	    - Next task passed
	    - task being scheduled in is not a logging process 
	    	FIXME - multiple replay groups
	*/
	if(tsk==next) {
	    printk("revirt_log_schedule: funny, tsk(last) == next!\n");
	}

	if(stype==STYPE_SCHEDULE) {
	    if(next && !(next->revirt_flags&REPLAY_LOG)) {
		clear_pids=0;
		write_log=0;
		sched_tmp=le;
		sched_tmp_tsk=tsk;
	    } else {
		tsk->revirt_flags &= ~(REPLAY_FIRST_AFTER_SIG);
	    }
	} else {
	    tsk->revirt_flags &= ~(REPLAY_FIRST_AFTER_SIG);
	}

    	if(clear_pids) { 
    	    clear_active_pids(active);
    	    if(le.active.i >1) {
    		clear_branchdata_group(tsk);
    	    } else {
    		clear_branchdata(tsk);
    	    }
    	}

	if(write_log) {
	    if(tsk->revirt_tsk->group) {
		tsk->revirt_tsk->group->zt_ctl.last_sched_id 
			= tsk->revirt_tsk->group->log_id;
	    }
	    revirt_write_log((char *)&le, ENTRY_SCHED, sizeof(le), tsk->revirt_tsk->group);
	}
}

EXPORT_SYMBOL_NOVERS(revirt_log_schedule);

#undef last_sched
#undef first_sched

void revirt_log_fork(struct task_struct *parent,struct task_struct * child) {
        struct rlog_entry_fork le;

        //le.type= ENTRY_FORK;
        le.pid= parent->pid;
        le.child=child->pid;

        revirt_write_log((char *)&le, ENTRY_FORK, sizeof(le), NULL);
}

void revirt_log_rdtsc(int hi, int lo, unsigned long eip) {
	struct rlog_rdtsc le;

	le.hi=hi;
	le.lo=lo;
	le.eip=eip;
	//le.syscall_id = current->revirt_tsk->syscall_id;

	revirt_write_log((char *)&le, ENTRY_RDTSC, sizeof(le), NULL);
}

void revirt_log_in(u32 byte, u16 port, u32 eip) {
	struct rlog_entry_in le;

	le.byte = byte;
	le.port=port;
	le.eip=eip;
	Rprintk("revirt_log_in: @0x%x in 0x%x == 0x%x\n",
			eip, port, byte, 0, 0);
	revirt_write_log((char *)&le, ENTRY_IN, sizeof(le), NULL);
}

void revirt_log_generic(int type, int len, char * data) {
    struct rlog_entry_generic ge;
    int i;

    if(len > REPLAY_GENERIC_SIZE) {
      printk("revirt_log_generic: WARNING, len %d > max size %d! truncating\n",
      	len, REPLAY_GENERIC_SIZE);
      len=REPLAY_GENERIC_SIZE;
    }

    ge.type=type;
    for(i=0; i<len; i++) 
       ge.data[i]=data[i];
#ifdef REPLAY_SYSCALL_LOGS_BRANCHCOUNT
    copy_branchdata(current, &ge.ctr);
#endif
    revirt_write_log((char *)&ge, ENTRY_GENERIC, sizeof(ge), NULL);
}

EXPORT_SYMBOL_NOVERS(revirt_log_generic);
#if 0
void revirt_log_dma(struct task_struct * tsk, void * dest, void *src,
    int size, int seq) {
    struct rlog_dma_entry le;
    struct iovec v[2];

    le.pid=tsk->pid;
    le.dest=dest;
    le.size=size;
    le.eip=0x0;
    le.seq = seq;
    v[0].iov_base=&le;
    v[0].iov_len=sizeof(le);
    v[1].iov_base=src;
    v[1].iov_len=size;
    revirt_writev_log(ENTRY_DMA, v, 2, NULL);
}

EXPORT_SYMBOL_NOVERS(revirt_log_dma);
#endif

EXPORT_SYMBOL_NOVERS(revirt_log_rdtsc);
EXPORT_SYMBOL_NOVERS(revirt_log_in);

void revirt_log_exit(void) {
        struct rlog_exit_entry ee;
	struct task_struct *tsk=current;
	struct revirt_group_struct * rgroup=NULL;
	struct revirt_active_pid_st * active;

	if(tsk->revirt_tsk && ((rgroup=tsk->revirt_tsk->group))) {
		active= &tsk->revirt_tsk->group->active;
		update_active_pids(tsk, active);
		ee.active = *active;
		clear_active_pids(active);
	} else {
	    active=NULL;
	    ee.active.i=-1;
	}

	ee.pid=tsk->pid;

	revirt_write_log((char *)&ee, ENTRY_EXIT, sizeof(ee), rgroup);

	clear_branchdata_group(tsk);
}

asmlinkage void revirt_log_sysret(struct pt_regs  * regs) {
	struct rlog_entry_sysret se;
	struct task_struct * tsk=current;

	if(!tsk->revirt_tsk) {
		printk("revirt_log_sysret: pid %d has no revirt_tsk!\n",
				tsk->pid);
	}


	se.pid = tsk->pid;
	// sigreturn clobbers orig_eax
	se.syscall = tsk->revirt_tsk->current_syscall;  
	se.syscall_id=tsk->revirt_tsk->syscall_id;
#ifdef REPLAY_WHOLE_PTREGS
	se.regs = *regs;
#else
	se.regs.eax = regs->eax;
#endif
#ifdef REPLAY_SYSCALL_LOGS_BRANCHCOUNT
	copy_branchdata(tsk, &se.ctr);
#endif
	se.nrxuser=tsk->revirt_tsk->nrxuser;

	tsk->revirt_tsk->nrxuser=0;

	revirt_write_log((char *)&se, ENTRY_SYSRET, sizeof(se), tsk->revirt_tsk->group);
}

void revirt_log_sigframe(struct __kernel_sigframe * frame) {
	struct rlog_entry_sigframe e;

	e.pid = current->pid;
#ifdef REPLAY_WHOLE_SIGFRAME
	e.frame = *frame;
#else
	e.frame.sig = frame->sig;
	e.frame.trapno = frame->sc.trapno;
#endif

	revirt_write_log((char *)&e, ENTRY_SIGFRAME, sizeof(e), NULL);
} 

void revirt_log_signal(struct siginfo * sig) {
	struct task_struct * tsk=current;
	struct rlog_entry_signal le_r;
	int i, size;
	struct rlog_entry_signal * le;

	le = &le_r;
	size = sizeof(le_r);

	le->pid = tsk->pid;

	le->pending_signal = tsk->pending.signal;
	le->blocked = tsk->blocked;

	/*
	 * We need to do this because force_siginfo may change SIG_IGN
	 * to SIG_DFL.
	 * 
	 * Note that there is no 0 signal; so all signals are shifted
	 * down one -- so sigaddset will change the i+1 back to i for
	 * action[0].
	 */

	if(tsk->sig) {
	    for(i=0; i<_NSIG_WORDS; i++) 
	    	le->dfl.sig[i] = 0;
	    for(i=0; i<_NSIG; i++) {
	    	if(tsk->sig->action[i].sa.sa_handler == SIG_DFL) {
			sigaddset(&le->dfl, i+1);
		}
	    }
	} else {
		printk("rlog: Strange, log_signals called for zombie...\n");
		printk("rlog: pid %d, sig %d.  Not logging.\n",
			tsk->pid, sig->si_signo);
		return;
	}

	tsk->revirt_flags |= REPLAY_FIRST_AFTER_SIG;

#ifdef REPLAY_WHOLE_SIGINFO
	le->info = *sig;
#else
	le->si_signo = sig->si_signo;
	le->si_errno = sig->si_errno;
	le->si_code = sig->si_code;
	le->_pad[0] = sig->_sifields._pad[0];
	le->_pad[1] = sig->_sifields._pad[1];
#endif
	revirt_write_log((char *)le, ENTRY_SIGNAL, size, NULL);
}

void revirt_log_rt_sigframe(struct __kernel_rt_sigframe * frame) {
	struct rlog_entry_rt_sigframe e;

	e.pid = current->pid;
#ifdef REPLAY_WHOLE_SIGFRAME
	e.frame = *frame;
#else
	e.frame.sig = frame->sig;
	e.frame.trapno = frame->uc.uc_mcontext.trapno;
#endif
	revirt_write_log((char *)&e, ENTRY_RT_SIGFRAME, sizeof(e), NULL);
}
