/*
 * perfctr-p4.c - counter and control info for p4
 *
 * Copyright (C) 2002-2003 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <revirt/perfctr.h>
#if defined(CONFIG_MPENTIUM4)

/**************************  P4 CODE ********************************/
 /*
  * Needed MSRs
  */
#define MSR_IQ_COUNTER0			0x30c
#define MSR_IQ_CCCR0			0x36c
#define MSR_CRU_ESCR2			0x3cc

#define MSR_MS_COUNTER0			0x304
#define	MSR_MS_CCCR0			0x364
#define MSR_TBPU_ESCR0			0x3c2

#define MSR_IQ_COUNTER1			0x30d
#define	MSR_IQ_CCCR1			0x36d
#define MSR_CRU_ESCR0			0x3b8

#define	MSR_IQ_COUNTER2			0x30e
#define	MSR_IQ_CCCR2			0x36e
#define MSR_CRU_ESCR3			0x3cd

#define MSR_IQ_COUNTER4			0x310
#define MSR_IQ_CCCR4			0x370
#define MSR_RAT_ESCR0			0x3bc

#define MSR_IQ_COUNTER3			0x30f
#define MSR_IQ_CCCR3			0x36f
#define MSR_CRU_ESCR1			0x3b9

#define MSR_BPU_COUNTER0		0x300
#define MSR_BPU_CCCR0			0x360
#define MSR_FSB_ESCR0			0x3a2

 /*
  * Pentium 4 bitfields
  */
#define ESCR_OS				(1<<3)
#define ESCR_USR			(1<<2)
#define ESCR_EVSEL(x)			(((x)&((1<<6)-1))<<25)
#define ESCR_EVMASK(x)			(((x)&((1<<16)-1))<<9)

#define CCCR_EN				(1<<12)
#define CCCR_RESERVED			(3<<16)
#define CCCR_OVF_PMI			(1<<26)
#define CCCR_OVF			(1<<31)
#define CCCR_ESCR_SEL(x)		(((x)&7)<<13)
#define CCCR_CMP			(1<<18)
#define CCCR_EDGE			(1<<24)
#define CCCR_THRESHOLD(x)		(((x)&0xf)<<20)




/*
 * global_power_events
 * (i.e., cycles not halted)
 */
#if 0
#define EMON_GLOBAL_POWER_EVENTS_COUNTER	MSR_BPU_COUNTER0
#define EMON_GLOBAL_POWER_EVENTS_CCCR_MSR	MSR_BPU_CCCR0
#define EMON_GLOBAL_POWER_EVENTS_CCCR_VAL	\
		( CCCR_EN | CCCR_RESERVED | CCCR_ESCR_SEL(6) )
#define EMON_GLOBAL_POWER_EVENTS_ESCR_MSR	MSR_FSB_ESCR0
#define EMON_GLOBAL_POWER_EVENTS_ESCR_VAL	\
		( ESCR_USR | ESCR_EVSEL(5) | ESCR_EVMASK(1) )
#endif


struct revirt_pfctr_s revirt_pfctr = {
	//.count=6,
	.count=1,
	.overflow= {
#ifdef USE_OLD_BC
# warning "USE_OLD_BC hasn't been tested since the code overhaul"
		.control_msr=MSR_IQ_CCCR0,
		.counter_index=1,
#else
		.control_msr=MSR_MS_CCCR0,
		.counter_index=0,
#endif
		.enable_bit= CCCR_OVF_PMI,
		.clear_mask= ~(CCCR_OVF),
	},
#ifdef USE_OLD_BC
	.subtract_exceptions=1,
#else
	.subtract_exceptions=0,
#endif
	.list = {
	   /* 0
	    * For retired-branch_type 
 	    * ESCR event mask:
 	    * 	0: unconditional
 	    * 	1: conditional
	    * 	2: call
	    * 	3: return
	    * 	4: indirect
	    * Notes: Bit 0 isn't documented in table A-1 (or A-2), but in
	    * table A-6.
	    * 
	    * Bit 4 counts things already counted in bits 0-3; this 
	    * double-counting confuses zr when it does a single-step and 
	    * sees "two" branches.
	    */
		{.counter_msr=MSR_MS_COUNTER0,
		.arch={.p4={
			.cccr_msr=MSR_MS_CCCR0,
			.cccr_val=CCCR_EN | CCCR_RESERVED | CCCR_ESCR_SEL(2),
			.escr_msr=MSR_TBPU_ESCR0,
			.escr_val=ESCR_USR | ESCR_EVSEL(4) | ESCR_EVMASK(15)}}},
	   /* 1
	    * branch_retired -- counts all branches, pred or mispred.
	    *
	    * Notes: We count all branches, taken or not; shouldn't hurt.
	    *
	    * This counter also counts iret and rsm instructions *into* 
	    * ring 3; subtract interrupt/exception/smi counts to make
	    * this repeatable.
	    */ 
		{ .counter_msr=MSR_IQ_COUNTER0,
		.arch={.p4={
			.cccr_msr=MSR_IQ_CCCR0,
			.cccr_val= CCCR_EN | CCCR_RESERVED | CCCR_ESCR_SEL(5),
			.escr_msr=MSR_CRU_ESCR2,
			.escr_val=ESCR_USR | ESCR_EVSEL(6) | ESCR_EVMASK(15)}}},
	   /* 2 
 	    * For instr_retired
 	    * ESCR event mask:
 	    * 	0: nonbogus, nontagged
 	    * 	1: nonbogus, tagged
 	    * 	2: bogus, nontagged
 	    * 	3: bogus, tagged
 	    * For instructions: 0x3
 	    */
		{.counter_msr=MSR_IQ_COUNTER1,
		.arch={.p4={
			.cccr_msr=MSR_IQ_CCCR1,
			.cccr_val=CCCR_EN | CCCR_RESERVED | CCCR_ESCR_SEL(4),
			.escr_msr=MSR_CRU_ESCR0,
			.escr_val=ESCR_USR | ESCR_EVSEL(2) | ESCR_EVMASK(3) }}},
	   /* 3 
 	    * front_end_event (loads & stores)
 	    * Requires two "counters" to be set up, only one of which actually 
	    * counts... uop_type, chooses which of ^ to count.  Doesn't count, 
	    * but the associated counter should be enabled anyway, according 
	    * to ia32-3 15.9.6.8 "Counter usage guideline".
	    * 
	    * Paired with #5, "uop_type".
 	    */
		{.counter_msr=MSR_IQ_COUNTER2,
		.arch={.p4={
			.cccr_msr=MSR_IQ_CCCR2,
			.cccr_val=CCCR_EN | CCCR_RESERVED | CCCR_ESCR_SEL(5),
			.escr_msr=MSR_CRU_ESCR3,
			.escr_val=ESCR_USR | ESCR_EVSEL(8) | ESCR_EVMASK(1)}}},
	   /* 4 
   	    * uops_retired
   	    */
		{.counter_msr=MSR_IQ_COUNTER3,
		.arch={.p4={
			.cccr_msr=MSR_IQ_CCCR3,
			.cccr_val=CCCR_EN | CCCR_RESERVED | CCCR_ESCR_SEL(4),
			.escr_msr=MSR_CRU_ESCR1,
			.escr_val=ESCR_USR | ESCR_EVSEL(1) | ESCR_EVMASK(1)}}},
	   /* 5
	    * uop_type -- see #3
	    */
		{.counter_msr=MSR_IQ_COUNTER4,
		.arch={.p4={.cccr_msr=MSR_IQ_CCCR4,
			.cccr_val=CCCR_EN | CCCR_RESERVED | CCCR_ESCR_SEL(2),
			.escr_msr=MSR_RAT_ESCR0,
			.escr_val=ESCR_USR | ESCR_EVSEL(2) | ESCR_EVMASK(4)}}},
	},
};
#else
# warning "perfctr-p4.c compiled, but CONFIG_MPENTIUM4 not defined!"
#endif
