/*
 * perfctr-p6.c Counters and MSRs for p6
 *
 * Copyright (C) 2002 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <revirt/perfctr.h>

/************************** P6 CODE ************************************
 * See IA-32-3 15.11 and A.2
 ***********************************************************************/


#define P6_EV_INST_RETIRED		0xc0
#define P6_EV_BR_TAKEN_RETIRED		0xc9
#define P6_EV_HW_INT_RX			0xc8

#define P6_EVTSEL(x)			((x)&0xff)
#define P6_UNITMASK(x)			(((x)&0xff)<<8)
#define P6_USRMODE	 		(1<<16)
#define P6_OSMODE			(1<<17)
#define P6_INT				(1<<20)
#define P6_EN				(1<<22)

#define MSR_P6_EVTSEL			0x186
#define MSR_P6_EVTSEL0			0x186
#define MSR_P6_EVTSEL1			0x187

#define MSR_P6_CTR			0xc1
#define MSR_P6_CTR0			0xc1
#define MSR_P6_CTR1			0xc2

#warning "P6 counters have never been used successfully.  Good luck."

struct revirt_pfctr_s revirt_pfctr = {
	.count=2,
	.overflow={
		.control_msr=MSR_P6_EVTSEL0,
		.enable_bit=P6_INT,
		.clear_mask=0,
		.counter_index=0,
	},
	.subtract_exceptions=1,
	.list = {
		/* 0
		 * BR_TAKEN_RETIRED
		 */
		{ .counter_msr=MSR_P6_CTR0,
		  .arch={.p6={.evtsel_msr=MSR_P6_EVTSEL0,
			  .evtsel_val=P6_EVTSEL(0xc9) | P6_EN | P6_USRMODE }}},
		/* 1
		 * HW_INT_RX -- delivered interrupts
		 */
		{ .counter_msr=MSR_P6_CTR1,
		  .arch={.p6={.evtsel_msr=MSR_P6_EVTSEL1,
			  .evtsel_val=P6_EVTSEL(0xc8) | P6_EN | P6_USRMODE }}}
	}
};
