/*
 * proc_rlog.c - Functions for /proc/rlogd 
 *
 * Copyright (C) 2002 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/types.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/poll.h>
#include <linux/module.h>

#include <asm/uaccess.h>
#include <asm/io.h>

/*
 * Needed by kdemon() for procfs
 */
#include <linux/stat.h>
#include <linux/proc_fs.h>

#include <revirt/rlog.h>

#define REVIRT_DEBUG_LOG_SIZE 500

struct revirt_debug_struct {
    const char * string;
    unsigned long arg[5];
} revirt_debug_log[REVIRT_DEBUG_LOG_SIZE];

int revirt_dlog_write_index=0;
int revirt_dlog_read_index=0;

spinlock_t revirt_dlog_lock = SPIN_LOCK_UNLOCKED;

//wait_queue_head_t revirt_debug_log_read_wait;

int revirt_debug_log_printk(char type, const char * string, unsigned long a0, 
	unsigned long a1, unsigned long a2, unsigned long a3, 
	unsigned long a4) {
    int flags;

    /* For now, 'Q' will be for 'quiet' printks -- i.e., only sent to 
     * the debug log, for whatever reason.  No extra action needs to be done
     * here, however. */

    if(revirt_options.replayflags & REPLAY_OPT_DEBUGLOG) {
	spin_lock_irqsave(&revirt_dlog_lock, flags);
	revirt_debug_log[revirt_dlog_write_index].string=string;
	revirt_debug_log[revirt_dlog_write_index].arg[0]=a0;
	revirt_debug_log[revirt_dlog_write_index].arg[1]=a1;
	revirt_debug_log[revirt_dlog_write_index].arg[2]=a2;
	revirt_debug_log[revirt_dlog_write_index].arg[3]=a3;
	revirt_debug_log[revirt_dlog_write_index].arg[4]=a4;

	revirt_dlog_write_index = 
		(revirt_dlog_write_index+1) % REVIRT_DEBUG_LOG_SIZE;

	if(revirt_dlog_write_index == revirt_dlog_read_index) {
	    revirt_dlog_read_index =
		(revirt_dlog_read_index+1) % REVIRT_DEBUG_LOG_SIZE;
	}
   
	spin_unlock_irqrestore(&revirt_dlog_lock, flags);
    }

    if(revirt_options.replayflags & REPLAY_OPT_REPLAY_DEBUG
    	&& type=='R') {
	printk(string, a0, a1, a2, a3, a4);
    }
    if(revirt_options.replayflags & REPLAY_OPT_ZTRACE_DEBUG
    	&& type=='Z') {
	printk(string, a0, a1, a2, a3, a4);
    }
    if(revirt_options.replayflags & REPLAY_OPT_CHECKPOINT_DEBUG
    	&& type=='C') {
	printk(string, a0, a1, a2, a3, a4);
    }

    return 0;
}

EXPORT_SYMBOL_NOVERS(revirt_debug_log_printk);

int revirt_dlog_open(void) {
    //revirt_dlog_read_index= revirt_dlog_write_index;
    //printk("revirt_dlog_open\n");
    return 0;
}

int revirt_dlog_release(void) {
	return 0;
}

int revirt_dlog_read(char * d, size_t l) {
    char buf[160];
    int flags;
    int total=0;
    int in_buffer=0;
    int do_write=0;


#if 1
    printk("revirt_dlog_read: write index %d, read_index %d, want %d bytes.\n",
    		revirt_dlog_write_index, revirt_dlog_read_index, l);
#endif

    while(1) {
	/* print one line, and find out how long it is... */
	spin_lock_irqsave(&revirt_dlog_lock, flags);

	/* If there are no more lines to print, exit... */
        if(revirt_dlog_write_index == revirt_dlog_read_index)  {
	    in_buffer=0;
	} else {
	    in_buffer=snprintf(buf, 160, 
		revirt_debug_log[revirt_dlog_read_index].string,
		revirt_debug_log[revirt_dlog_read_index].arg[0],
		revirt_debug_log[revirt_dlog_read_index].arg[1],
		revirt_debug_log[revirt_dlog_read_index].arg[2],
		revirt_debug_log[revirt_dlog_read_index].arg[3],
		revirt_debug_log[revirt_dlog_read_index].arg[4]);
#if 0
	    printk("revirt_dlog_read: snprintf returned %d\n", in_buffer);
#endif
	    if(total+in_buffer < l) {
	    	do_write=1;
		revirt_dlog_read_index=
			(revirt_dlog_read_index + 1)
			% REVIRT_DEBUG_LOG_SIZE;
#if 0
		printk("revirt_dlog_read: revirt_dlog_read_index moved to %d\n",
				revirt_dlog_read_index);
#endif
	    } else {
	    	do_write=0;
	    }

	}

	spin_unlock_irqrestore(&revirt_dlog_lock, flags);

	if(in_buffer && do_write) {
	    if((do_write=copy_to_user(d+total, buf, in_buffer)!=0)) {
#if 0
	    	printk("copy_to_user (%x + %d, %x, %d) == %d\n",
			d, total, buf, in_buffer);
#endif
	    	if(!total) 
		    total=-EFAULT;
		goto out;
	    }
	    total += in_buffer;
#if 0
	    printk("total now %d\n", total);
#endif
	} else 
	    goto out;

	if(current->need_resched)
		schedule();
    }
    out:
     return total;
}

unsigned int revirt_dlog_poll(void) {
    int flags;
    int error;

    spin_lock_irqsave(&revirt_dlog_lock, flags);
    error = (revirt_dlog_write_index - revirt_dlog_read_index) % REVIRT_DEBUG_LOG_SIZE;
    spin_unlock_irqrestore(&revirt_dlog_lock, flags);
    return error;
}

static int proc_revirt_dlog_open(struct inode * inode, struct file * file)
{
	return revirt_dlog_open();
}

static int proc_revirt_dlog_release(struct inode * inode, struct file * file)
{
	(void) revirt_dlog_release();
	return 0;
}

static ssize_t proc_revirt_dlog_read(struct file * file, char * buf,
			 size_t count, loff_t *ppos)
{
	return revirt_dlog_read(buf,count);
}

static unsigned int proc_revirt_dlog_poll(struct file *file, poll_table * wait)
{
	int n, ret=0;
	//poll_wait(file, &revirt_debug_log_read_wait, wait);
	if ((n=revirt_dlog_poll())) {
		ret= POLLIN | POLLRDNORM;
	}
	//printk("do_revirtlog(POLL) returns %d\n", n);
	return ret;
}

struct file_operations proc_revirt_dlog_operations = {
	read:		proc_revirt_dlog_read,
	poll:		proc_revirt_dlog_poll,
	open:		proc_revirt_dlog_open,
	release:	proc_revirt_dlog_release,
};

void  proc_revirt_dlog_init(void) {
        struct proc_dir_entry *entry;
	int i;

        /*
         * Create /proc/revirt_dlog
         */
        entry=create_proc_entry("revirt_dlog", S_IRUSR, &proc_root);
        if(entry)
                entry->proc_fops = &proc_revirt_dlog_operations;

	/* No spinlocks, in init... */
	for(i=0; i<REVIRT_DEBUG_LOG_SIZE; i++) {
	    revirt_debug_log[i].string=NULL;
	}
}
