/*
 * replay.c - ReVirt replay functionality
 *
 * Copyright (C) 2002 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/linkage.h>
#include <linux/sched.h>
#include <asm/current.h>
#include <linux/smp_lock.h>
#include <asm/semaphore.h>
#include <linux/tqueue.h>
#include <linux/slab.h>
#include <asm/perfctr.h>
#include <linux/interrupt.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/processor.h>
#include <asm/debugreg.h>
#include <asm/user.h>
#include <revirt/rlog.h>
#include <revirt/ztrace.h>
#include <revirt/perfctr.h>
#include <asm/unistd.h>

#include <asm/fixmap.h>
#include <asm/apic.h>
#include <asm/mtrr.h>
#include <asm/siginfo.h>

/* needed for SYS_SOCKETPAIR */
#include <linux/net.h>

/* Moved to rlog.h */
//#define	EFLAGS_MASK (0x40DD5)
//#define REPLAY_DEBUG

#define Dprintk(x...) \
	do { if(revirt_options.replayflags & REPLAY_OPT_REPLAY_DEBUG) 	\
		printk(x);				\
	} while(0)

#include <linux/module.h>

//extern struct ztrace_rununtil_struct zt_rununtil;

void copy_branchdata(struct task_struct *, struct revirt_ctrdata *);

void check_branchcount(struct revirt_ctrdata * lctr, struct pt_regs * regs,
		struct rlog_header rh) {
    int diff=0, pause=0;
    struct task_struct * tsk=current;
    struct revirt_ctrdata rctr;
    struct ztrace_ctl_struct * ztc;

    if(!tsk->revirt_tsk) {
	    printk("check_branchcount: pid %d has no revirt_tsk!\n",
			    tsk->pid);
	    return;
    }

    if(!(ztc=get_ztc(tsk)))
	printk("check_branchcount: no rgroup, not checking ztc...\n");

    copy_branchdata(tsk, &rctr);

    if(ADJ_BC(*lctr) != ADJ_BC_R(rctr)) {
        printk("check_branchcount: maincount differs! %llx %llx (%llx) \n",
	    	ADJ_BC_R(rctr), ADJ_BC((*lctr)), 
		ADJ_BC((*lctr)) - ADJ_BC_R(rctr));
       if(ztc) {
	if((ADJ_BC_R(rctr) == 0 )
		&& ((ztc->flags == ZTRACE_BLOCK_STOP) || (ztc->flags == ZTRACE_BKPT))
		&& (ztc->brk.eip == zreg(tsk,eip))
		&& (ztc->brk.count == ADJ_BC(*lctr))) {
		/*
		 * I think this shouldn't happen...
		 */
	   printk("At requested <eip,bc>, wanted 0 bc, assuming wierd signal...\n");
	} else  if(ADJ_BC(*lctr) == 0
		&& ((ztc->flags == ZTRACE_BLOCK_STOP) || (ztc->flags == ZTRACE_BKPT))
		&& (ztc->brk.eip == zreg(tsk,eip))) {
		/* 
		 * This happens when we deliver a signal 'late' -- i.e., during logging
		 * it happened before the sysret, now it happens after (at 
		 * ztrace_tsk_enters_syscall).
		 */
	   printk("Right eip, wanted 0, got %llx, assuming missed something...\n",
	   		ADJ_BC(*lctr));

	} else {
	    printk("ADJ_BC_R(rctr): %llx cf 0\n",
	   	ADJ_BC_R(rctr));
	    printk("ztc->flags == %x, cf %x or %x\n", ztc->flags,
	   			ZTRACE_BLOCK_STOP, ZTRACE_BKPT);
	    printk("eip: %x, cf %x\n", (unsigned int)zreg(tsk, eip), ztc->brk.eip);
	    printk("ADJ_BC(*lctr): %llx, cf %llx\n", 
	    		ADJ_BC(*lctr), ztc->brk.count);
	    diff=1;
	    pause=1;
	}
       }
    }
    if(diff) {
#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
	int i;
	char buf[512], *c;
#endif

#ifdef RECORD_ALL_PFCTRS
	printk("revirt_syscall: #%lu sc %d @ 0x%x\n" \
		" [%llx %x %x %x] [%llx %x %x %x]\n" \
		" [%llx %llx %llx %llx] [%llx %llx %llx %llx]\n",
		rh.id,
		current->revirt_tsk->current_syscall,
		(regs)?(unsigned int)regs->eip:0x0,
	    	rctr.hw_branch,
		rctr.hw_intr,
		rctr.ex_noerror,
		rctr.ex_error,
		lctr->hw_branch,
		lctr->hw_intr,
		lctr->ex_noerror,
		lctr->ex_error,
		rctr.counter[0],
		rctr.counter[1],
		rctr.counter[2],
		rctr.counter[3],
		lctr->counter[0],
		lctr->counter[1],
		lctr->counter[2],
		lctr->counter[3]);
#else
	printk("revirt_syscall: #%lu sc %d @ 0x%x [%llx] [%llx]\n",
		rh.id,
		current->revirt_tsk->current_syscall,
		(regs)?((unsigned int)regs->eip):0x0,
	    	rctr.hw_branch,
		lctr->hw_branch);
#endif
#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
 	 c=buf;
	 *c=0;
	 for(i=0; i<REVIRT_NR_EX; i++) {
		 if(rctr.exceptions[i]>0) {
			 c+=sprintf(c, "[%d] %x ", i, rctr.exceptions[i]);
		 }
	 }
	 if(c>buf) {
		 printk("%s\n", buf);
	 }
#endif
   	 if(diff && (revirt_options.replayflags & REPLAY_OPT_WARNING_HALTS))
	    	ztrace_debug_pause();
    }
}

int revirt_replay_sigframe(struct __kernel_sigframe *user_frame, 
				struct __kernel_sigframe *frame){
	struct rlog_entry_sigframe * e;
	struct rlog_header rh;
	int pause;

	/*
	 * Read the log
	 */
	revirt_read_log(&rh);
	if(rh.type != ENTRY_SIGFRAME) {
		printk("revirt_replay_sigframe: expecting type %x sig %d, got %x (#%lu)\n",
				ENTRY_SIGFRAME, frame->sig,
				rh.type, rh.id);
		ztrace_debug_pause();
	}
	e=(struct rlog_entry_sigframe *)revirt_scall_buf;

	/*
	 * Check for any deviation
	 */
	pause=0;
#define LDIFF(x,y) \
	if(e->frame.x != frame->x) { 			 \
	  printk("revirt_replay_sigframe: %s differs: %lx %lx\n", \
	  	y, frame->x, e->frame.x);	  	 \
  	  pause = 1;					 \
	}
#define DIFF(x,y) \
	if(e->frame.x != frame->x) { 			 \
	  printk("revirt_replay_sigframe: %s differs: %x %x\n", \
	  	y, (unsigned int)frame->x, (unsigned int)e->frame.x);	\
  	  pause = 1;					 \
	}
	DIFF(sig, "sig");
#ifdef REPLAY_WHOLE_SIGFRAME
	DIFF(pretcode, "pretcode");
	LDIFF(sc.eax, "eax");
	LDIFF(sc.ebx, "ebx");
	LDIFF(sc.ecx, "ecx");
	LDIFF(sc.edx, "edx");

	LDIFF(sc.edi, "edi");
	LDIFF(sc.esi, "esi");
	LDIFF(sc.ebp, "ebp");
	LDIFF(sc.esp, "esp");
	LDIFF(sc.eip, "eip");
	DIFF(sc.cs, "cs");
	DIFF(sc.ss, "ss");
	DIFF(sc.gs, "gs");
	DIFF(sc.fs, "fs");
	DIFF(sc.es, "es");
	DIFF(sc.ds, "ds");

	/*
	 * only the bits masked below are restored by restore_sigcontext()
	 * in arch/i386/kernel/signal.c.
	 * I suppose we'd really prefer to check the actual flags
	 * restored during log and replay, but this will have to do for now...
	 */
	if((e->frame.sc.eflags & REVIRT_EFLAGS_MASK) 
		!= (frame->sc.eflags & REVIRT_EFLAGS_MASK)) {
	    printk("revirt_replay_sigframe: restored eflags differ: 0x%lx 0x%lx\n",
			frame->sc.eflags & REVIRT_EFLAGS_MASK,
	    		e->frame.sc.eflags & REVIRT_EFLAGS_MASK);
	    
	    pause=1;
	}

	int i;

	for(i=0; i<8; i++) {
	 if(e->frame.retcode[i] != frame->retcode[i]) {
	   printk("revirt_replay_sigframe: retcode[%d] differs: %x %x\n",
	   		i, frame->retcode[i], e->frame.retcode[i]);
	   pause=1;
	 }
	}
#endif /* REPLAY_WHOLE_SIGFRAME */

#undef DIFF
#undef LDIFF
     	if(pause) {
     	  printk("   pausing...\n");
	  ztrace_debug_pause();
     	}

	/*
	 * Finally, copy the frame.
	 */
	//return __copy_to_user(user_frame, &e->frame, sizeof(e->frame));
#ifdef REPLAY_WHOLE_SIGFRAME
	return __copy_to_user(user_frame, &e->frame, sizeof(e->frame));
#else
	frame->sc.trapno = e->frame.trapno;
	return __copy_to_user(user_frame, frame, sizeof(*frame));
#endif
}

int revirt_replay_rt_sigframe(struct __kernel_rt_sigframe *user_frame, 
			struct __kernel_rt_sigframe *frame){
	struct rlog_entry_rt_sigframe * e;
	struct rlog_header rh;
	int pause;

	/*
	 * Read the log
	 */
	revirt_read_log(&rh);
	if(rh.type != ENTRY_RT_SIGFRAME) {
		printk("revirt_replay_rt_sigframe: expecting logtype %x sig %d, got %x(%lu)\n",
				ENTRY_RT_SIGFRAME, frame->sig,
				rh.type, rh.id);
		ztrace_debug_pause();
	}
	e=(struct rlog_entry_rt_sigframe *)revirt_scall_buf;
	
#define DIFF(x, y) \
	if(e->frame.x != frame->x) { 			 \
	  printk("revirt_replay_rt_sigframe: %s differs: %x %x\n", \
	  	y, (unsigned int)frame->x, (unsigned int)e->frame.x);	 \
  	  pause = 1;					 \
	}

	/*
	 * Check for any deviation
	 */
	Dprintk("revirt_replay_rt_sigframe: handling signal %x\n", frame->info.si_signo);

	pause=0;
	DIFF(sig, "sig");
#ifdef REPLAY_WHOLE_SIGFRAME
	DIFF(pretcode, "pretcode");
	DIFF(pinfo, "pinfo");
	DIFF(puc, "puc");

	DIFF(info.si_signo, "info.si_signo");
	DIFF(info.si_errno, "info.si_errno");
	DIFF(info.si_code, "info.si_code");
	DIFF(info.si_errno, "info.si_errno");

	// This is probably OK, as long as we clear the structure first
	// i.e., I suspect the differences are just random stack differences,
	// not true changes.  Nevertheless, we need it to be the same both
	// times.
	{
	 int i;
	 for(i=0;i < SI_PAD_SIZE; i++) {
	  if(e->frame.info._sifields._pad[i] != frame->info._sifields._pad[i]) {
	   printk("revirt_replay_rt_sigframe: frame.info._sifields._pad[%d] differs: %x %x\n",
	   		i, 
			frame->info._sifields._pad[i],
	 		e->frame.info._sifields._pad[i]);
	   pause=1;
	  }
	 }
	}

	DIFF(uc.uc_flags, "uc.uc_flags");
	DIFF(uc.uc_link, "uc.uc_link");
	DIFF(uc.uc_stack.ss_sp, "uc.uc_stack.ss_sp");
	DIFF(uc.uc_stack.ss_flags, "uc.uc_stack.ss_flags");
	DIFF(uc.uc_stack.ss_size, "uc.uc_stack.ss_size");
	
	DIFF(uc.uc_mcontext.gs, "gs");
	DIFF(uc.uc_mcontext.fs, "fs");
	DIFF(uc.uc_mcontext.es, "es");
	DIFF(uc.uc_mcontext.ds, "ds");
	DIFF(uc.uc_mcontext.__gsh, "__gsh");
	DIFF(uc.uc_mcontext.__fsh, "__fsh");
	DIFF(uc.uc_mcontext.__esh, "__esh");
	DIFF(uc.uc_mcontext.__dsh, "__dsh");
	DIFF(uc.uc_mcontext.edi, "edi");
	DIFF(uc.uc_mcontext.esi, "esi");
	DIFF(uc.uc_mcontext.ebp, "ebp");
	DIFF(uc.uc_mcontext.esp, "esp");
	DIFF(uc.uc_mcontext.ebx, "ebx");
	DIFF(uc.uc_mcontext.ecx, "ecx");
	DIFF(uc.uc_mcontext.edx, "edx");
	DIFF(uc.uc_mcontext.eax, "eax");
	// We know that trapno changes
	DIFF(uc.uc_mcontext.err, "err");
	DIFF(uc.uc_mcontext.eip, "eip");
	DIFF(uc.uc_mcontext.cs, "cs");
	DIFF(uc.uc_mcontext.__csh, "__csh");
	// eflags checked below
	DIFF(uc.uc_mcontext.esp_at_signal, "esp_at_signal");
	DIFF(uc.uc_mcontext.ss, "ss");
	DIFF(uc.uc_mcontext.__ssh, "__ssh");
	DIFF(uc.uc_mcontext.fpstate, "* fpstate");
	DIFF(uc.uc_mcontext.oldmask, "oldmask");
	DIFF(uc.uc_mcontext.cr2, "cr2");
	
	/*
	 * only the bits masked below are restored by restore_sigcontext.
	 * I suppose we'd really prefer to check the actual flags
	 * restored during log and replay, but this will have to do for now...
	 */
	if((e->frame.uc.uc_mcontext.eflags & REVIRT_EFLAGS_MASK) 
		!= (frame->uc.uc_mcontext.eflags & REVIRT_EFLAGS_MASK)) {
	    printk("revirt_replay_rt_sigframe: restored eflags differ: 0x%lx 0x%lx\n",
			frame->uc.uc_mcontext.eflags & REVIRT_EFLAGS_MASK,
	    		e->frame.uc.uc_mcontext.eflags & REVIRT_EFLAGS_MASK);
	    pause=1;
	}
	{ int i;
	 for(i=0; i<8; i++) {
	  if(e->frame.retcode[i] != frame->retcode[i]) {
	    printk("revirt_replay_rt_sigframe: retcode[%d] differs: %x %x\n",
	   		i, 
			frame->retcode[i], 
			e->frame.retcode[i]);
	    pause=1;
	  }
	 }
	 for(i=0; i<_NSIG_WORDS; i++) {
	  if(e->frame.uc.uc_sigmask.sig[i] != frame->uc.uc_sigmask.sig[i]) {
	   printk("revirt_replay_rt_sigframe: uc.uc_sigmask.sig[%d] differs: %lx %lx\n",
	   		i, 
			frame->uc.uc_sigmask.sig[i], 
			e->frame.uc.uc_sigmask.sig[i]);
	    pause=1;
	  }
	 }
	}

	DIFF(fpstate.cw, "fpstate.cw");
	DIFF(fpstate.sw, "fpstate.sw");
	DIFF(fpstate.tag, "fpstate.tag");
	DIFF(fpstate.ipoff, "fpstate.ipoff");
	DIFF(fpstate.cssel, "fpstate.cssel");
	DIFF(fpstate.dataoff, "fpstate.dataoff");
	DIFF(fpstate.datasel, "fpstate.datasel");
	DIFF(fpstate.status, "fpstate.status");
	DIFF(fpstate.magic, "fpstate.magic");
	DIFF(fpstate.mxcsr, "fpstate.mxcsr");
	DIFF(fpstate.reserved, "fpstate.reserved");

	{ int i, j;
	 for(i=0; i<8; i++) {
		for(j=0; j<4; j++) {
			DIFF(fpstate._st[i].significand[j], 
					"fpstate._sti.significand");
		}
	 	DIFF(fpstate._st[i].exponent, "fpstate.sti[].exponent");
	 }

	 for(i=0; i<6; i++) {
		DIFF(fpstate._fxsr_env[i], "fpstate._fxsr_env[]");
	 }
	 for(i=0; i<8; i++) {
		for(j=0; j<4; j++)
			DIFF(fpstate._fxsr_st[i].significand[j],
					"fpstate.fxsr_st[].significand[]");
		DIFF(fpstate._fxsr_st[i].exponent, 
				"fpstate._fxsr_st[].exponent");
		for(j=0; j<3; j++) 
			DIFF(fpstate._fxsr_st[i].padding[j],
				"fpstate._fxsr_st[].padding[]");
	 }
	 for(i=0; i<8; i++) {
		for(j=0; j<4; j++)
			DIFF(fpstate._xmm[i].element[j],
				"fpstate._xmm[].element[]");
	 }
	 for(i=0; i<56; i++) {
		DIFF(fpstate.padding[i], "fpstate._padding[]");
	 }
	}
//#else
# if 0
# define DIFF(x, y) \
	if(e->x != frame->x) { 				 \
	  printk("revirt_replay_rt_sigframe: %s differs: %x %x\n", \
	  	y, (unsigned int)frame->x, (unsigned int)e->x);	 \
  	  pause = 1;					 \
	}
# endif
#endif

#undef DIFF

     	if(pause) {
     	  printk("   pausing...\n");
	  ztrace_debug_pause();
     	}
	/*
	 * Finally, copy the frame
	 */
#ifdef REPLAY_WHOLE_SIGFRAME
	return __copy_to_user(user_frame, &e->frame, sizeof(e->frame));
#else
	frame->uc.uc_mcontext.trapno = e->frame.trapno;
	return __copy_to_user(user_frame, frame, sizeof(*frame));
#endif
}

/*
 * Move this to another file
 */


#if 0

#define REPLAY_EXCLUDE_SOME

#ifndef REPLAY_ALL_SYSCALLS
# define REPLAY_UNLOGGED \
	case __NR_sigprocmask: \
	case __NR_mmap: \
	case __NR_rt_sigreturn: \
	case __NR_getpid: \
	case __NR_munmap: \
	case __NR_lseek: \
	case __NR_sigaltstack: \
	case __NR_write: 	\
	case __NR_getguestpgd: \
	case __NR_switchguest: \
	case __NR_rt_sigsuspend: \
	case __NR__llseek:
#else
# define REPLAY_UNLOGGED
#endif

#define REPLAY_LOGGED \
	case __NR_read: \
	case __NR_socketcall: \
	case __NR_gettimeofday:
#endif

int revirt_read_ahead(struct rlog_header *);
int revirt_read_next_sysret(struct rlog_header *, struct rlog_entry_sysret *, s64 *,int, s64);

/*
 * FIXME -- this assumes that any syscall that returns an error (-1 thru -124) had no 
 * effect and can be simply skipped; it also assumes as a corrolary that if it returns 
 * an error, there will be no logs before the sysret.
 * 
 * The short write handling assumes that there are no side effects to a short write
 * other than... a short write. =)  (E.g., no signals generated.)
 */

#if 0
// All these had to be moved to revirt_tsk to deal with blockings...
static unsigned long syserror_save;
static int syserror_write_cleanup;
static int syserror_replay_unlogged;
#endif

void handle_syserror(struct pt_regs * regs, struct rlog_header * rh, 
	struct rlog_entry_sysret * le) {
	struct task_struct * tsk=current;
	struct revirt_tsk_s * rtsk=tsk->revirt_tsk;
	//struct rlog_header rh;
	//struct rlog_entry_sysret le_r, *le=&le_r;

	if(rh->type == ENTRY_SYSRET) {
	    if(le->syscall_id != rtsk->syscall_id) {
	    	printk("handle_syserror: next sysret id %lld(#%lu %d), skipping %lld (%d) [pause]\n",
			le->syscall_id, rh->id, le->syscall, rtsk->syscall_id,
			revirt_current_syscall);
		ztrace_debug_pause();
	    } else if(le->pid != tsk->pid) {
	    	/*
		 * This shouldn't happen anymore
		 */
	    	printk("handle_syserror: pid %d, got sysret for pid %d [pause]\n",
			tsk->pid, le->pid);
		ztrace_debug_pause();
	    } else if(le->syscall != revirt_current_syscall) {
	    	printk("handle_syserror: got syscall %d, expecting %d\n",
			revirt_current_syscall, le->syscall);
	    } else if(le->syscall == __NR_rt_sigreturn
	    		|| le->syscall == __NR_sigreturn) {
	    	Dprintk("handle_syserror: ignoring 'error' for sys_sigreturn\n");
	    } else if((le->syscall == __NR_waitpid 
	    			|| le->syscall == __NR_wait4)
	    	      && le->regs.eax == -ERESTARTSYS) {
		Rprintk("handle_syserror: replaying waitX (%d) with -ERESTARTSYS\n",
			le->syscall, 0, 0, 0, 0);

	    } else if(((int)le->regs.eax < 0) 
	    			&& (((int)le->regs.eax > -125) 
					|| (le->regs.eax == -ERESTARTSYS))) {
		/* 
		 * -ERESTARTSYS  is -512 normally ; it's only returned if there's a
		 * signal (presumably unblocked), and it's handled in handle_signal
		 * -- either swizzled to -EAGAIN, or causes the syscall to restart.
		 * 
		 * Replaying it should cause it to happen again properly... I think...
		 */
	    	Dprintk("handle_syserror: #%lu syscall %d returned %d, not re-exec'ing\n",
			rh->id, le->syscall, (int)le->regs.eax);
		regs->orig_eax=0;
	    } else if((le->syscall == __NR_write) && (le->regs.eax != regs->edx)) {
	    	Dprintk("handle_syserror: #%lu write wanted %ld, got %ld, adjusting edx\n",
			rh->id, regs->edx, le->regs.eax);
		rtsk->syserror_save = regs->edx;
		rtsk->syserror_write_cleanup=1;
		regs->edx = le->regs.eax;
	    }
	} 
}

void cleanup_syserror(struct pt_regs * regs) {
	struct task_struct * tsk=current;
	struct revirt_tsk_s * rtsk;

	if(!(rtsk=tsk->revirt_tsk))
		return;
	if(rtsk->syserror_write_cleanup) {
	    if(rtsk->current_syscall == __NR_write) {
		regs->edx = rtsk->syserror_save;
	    }
	    rtsk->syserror_write_cleanup=0;
	}
}


void __revirt_replay_dma(void) {
    struct rlog_header rh;
    struct rlog_dma_entry * de;
    revirt_read_log(&rh);
    if(rh.type != ENTRY_DMA) {
	printk("ztrace_dma: expecting type %x, got %x!\n",
		ENTRY_DMA, rh.type);
	ztrace_debug_pause();
	return;
    }
    de = (struct rlog_dma_entry *)revirt_scall_buf;
    Rprintk("revirt_replay_dma: copying %d bytes to %x\n", 
	    de->size, de->dest, 0, 0, 0);
    copy_to_user((void *)de->dest, revirt_scall_buf+sizeof(*de), de->size);
}

void revirt_replay_dma(struct pt_regs * regs) {
    struct rlog_header rh;
    struct rlog_dma_entry *de=(struct rlog_dma_entry *)revirt_scall_buf;
    int seq=0;
    while(1) {
	revirt_read_ahead(&rh);
	if(rh.type == ENTRY_DMA) {
	    if(de->seq == seq)  {
		__revirt_replay_dma();
		seq++;
	    } else {
	    	Rprintk(" log seq %d, != %d, ending DMA replay\n", 
			de->seq, seq, 0, 0, 0);
		break;
	    }
	} else {
	    break;
	}
    }
    Rprintk("Found %d DMA entries\n", seq, 0, 0, 0, 0);
} 

EXPORT_SYMBOL_NOVERS(revirt_replay_dma);

/*
 * Called before a syscall either during logging or replay; turns on
 * XUSER and/or DATASQUASH if appropriate
 */
asmlinkage void revirt_presys(struct pt_regs regs) {
	struct revirt_ctrdata cnt;
	struct task_struct * tsk = current;
	struct revirt_tsk_s * rtsk;
	struct rlog_header *rh;
	struct rlog_entry_sysret *le;
	s64 * cache_syscall_id;
	struct ztrace_ctl_struct * ztc;

	if(!(rtsk=tsk->revirt_tsk)) {
		printk("revirt_presys: pid %d has no revirt_tsk!\n",
				tsk->pid);
		return;
	}

	if(!(ztc=get_ztc(tsk)))
	    printk("revirt_presys: pid %d has rtsk but no rgroup!\n", tsk->pid);

	/*
	 * Note that revirt_current_syscall should be kept 'up-to-date',
	 * but there's a time between entry.S and here that it's out of sync.
	 */
	rtsk->current_syscall = revirt_current_syscall;

	/* FIXME -- make this a normal swctr, like interrupts &c */
	rtsk->syscall_id++;
	rtsk->nrxuser=0;

      	copy_branchdata(tsk, &cnt);
	if(tsk->revirt_flags & REPLAY_REPLAY) {
	    Rprintk("revirt_presys: pid %d @ 0x%lx calls syscall %d (%lld) [%llx]\n",
	    		tsk->pid, regs.eip, rtsk->current_syscall, 
			rtsk->syscall_id, ADJ_BC(cnt));
	    Rprintk("  (0x%lx %lx %lx %lx %lx)\n",
			regs.ebx, regs.ecx, regs.edx, regs.esi, regs.edi);
	}
	/*
	 * Do some self-checks to make sure replay is working right
	 */
	if(ztc && tsk->pid == ztc->pid) {
	 if((ztc->flags & ZTRACE_BKPT) && (ADJ_BC(cnt) > ztc->brk.count)) {
		struct rlog_header rh;
		revirt_read_ahead(&rh);
		printk("revirt_presys: count %llx, zr wanted %llx next log #%lu\n",
				ADJ_BC(cnt), ztc->brk.count, rh.id);
		printk("revirt_presys: #%lu, type %d\n", rh.id, rh.type);
		if(revirt_options.replayflags & REPLAY_OPT_WARNING_HALTS) {
			ztrace_debug_pause();
		}
	 }
	 if((ztc->flags & ZTRACE_PERF) && (cnt.hw_branch > 0x100 + ztc->perferr)) {
		printk("revirt_presys: ZT_PERF, hw_count %llx\n",
				cnt.hw_branch - ztc->perferr );
		if(revirt_options.replayflags & REPLAY_OPT_WARNING_HALTS) {
			ztrace_debug_pause();
		}
	 }
	}
	/*
	 * Make sure this is a legal syscall for our VM
	 * 
	 * FIXME for UML we need a flag so that we can stop certain calls after it
	 * finishes bootstrapping.
	 */
	if((revirt_options.replayflags & REPLAY_OPT_RESTRICT_SYSCALLS) 
		/*&& tsk->umlinux.umlTraced*/) {
		if(!FD_ISSET(rtsk->current_syscall, &revirt_options.allowed_syscalls)) {
			printk("revirt_presys: pid %d @%lx makes forbidden syscall %d(0x%lx 0x%lx 0x%lx 0x%lx 0x%lx)\n",
				tsk->pid, regs.eip, rtsk->current_syscall,
				regs.ebx, regs.ecx, regs.edx, regs.esi, regs.edi);
			if(revirt_options.replayflags & REPLAY_OPT_WARNING_HALTS) {
				ztrace_debug_pause();
			}
		}
	}

	/*
	 * Log or replay appropriate system calls
	 */
	rtsk->syserror_replay_unlogged=0;
	rh=&rtsk->sysret_cache.rh;
	le=&rtsk->sysret_cache.le;
	cache_syscall_id=&rtsk->sysret_cache.syscall_id;

	if(tsk->revirt_flags & REPLAY_REPLAY) {
	    /*
	     * We don't want to have to search through the list every time, 
	     * so we keep a per-process cache of the last sysret read.  We 
	     * only need to read it if 
	     *    (1) the cache is empty (rh->type == -1), 
	     * or (2) the cache contains a SYSRET entry whose ID is 'out of 
	     * date'.  Notably, if it contains an EXIT entry, it won't search 
	     * through again.
	     *
	     * It's possible to get the following setup:
	     *   Schedule log sid == N
	     *   Sysret log sid == N
	     * In that case, it's possible that the cache_syscall_id will 
	     * equal N but the rh header will be ENTRY_SCHED.  If so, we want 
	     * to read the actual sysret entry, if available.  
	     * (read_next_sysret will only stop if the log it finds is 
	     * strictly greater than the current sid.)
	     */
	    Rprintk("revirt_presys: pid %d sid %d cache:type %x id %d\n",
	    		tsk->pid, rtsk->syscall_id, rh->type, 
			*cache_syscall_id, 0);
	    if(rh->type == (__u8)-1
	       || ((rh->type != ENTRY_EXIT
	         	&& rh->type != ENTRY_END)
	       		&& (*cache_syscall_id < rtsk->syscall_id))
		|| (rh->type != ENTRY_SYSRET
	       		&& (*cache_syscall_id == rtsk->syscall_id))
			) {
		revirt_read_next_sysret(rh, le, cache_syscall_id, 
				tsk->pid, rtsk->syscall_id);
		Rprintk("next_sysret returns type %x id %d\n",
				rh->type, *cache_syscall_id, 0, 0, 0);
	    } 
	}

	if(!FD_ISSET(rtsk->current_syscall, 
			&revirt_options.unlogged_syscalls)) {

	    /*
	     * Special cases for some system calls
	     */
	    if(rtsk->current_syscall == __NR_read) {
	    	/*
		 * Some fd's we want / need to log, and some we don't.  
		 * We set these in unlogged_fds, and don't log them.
		 * Eventually, we want to admit the possibility of an unlogged
		 * fd returning an error (handle_syserror), but we can look
		 * at that later...
		 */
	     	if(FD_ISSET(regs.ebx, &revirt_options.unlogged_fds)) {
	    	    Dprintk("revirt_presys: not logging fd %ld\n", regs.ebx);
	   	    goto nolog;
	    	} else {
		    /* One of the logged fds */
	    	    if(current->revirt_flags & REPLAY_REPLAY) {
		    	Dprintk("revirt_presys: not re-exec'ing logged fd %ld\n",
				regs.ebx);
	    		regs.orig_eax=0;		    
	    	    }
	    	    goto log;
	       	}
	    }

	    if(rtsk->current_syscall == __NR_poll) {
		/* replay_only (no rexec) */
		    if(tsk->revirt_flags & REPLAY_REPLAY) {
		    	    Dprintk("special-casing sys_poll() on replay\n");
		    	    regs.orig_eax=0;
		    }
	    }

    	} else {
	    /* "Unlogged" syscall */
	    if((rh->type != ENTRY_SYSRET)) {
	    	Rprintk("revirt_presys: pid %d unlogged syscall, rh->type %x, nolog.\n",
			tsk->pid, rh->type, 0, 0, 0);
	    	goto nolog;
	    }
	    if((le->syscall_id != rtsk->syscall_id)) {
	    	Rprintk("revirt_presys: pid %d unlogged syscall id %ld, cache %ld\n", 
			tsk->pid,
			(unsigned int)le->syscall_id, 
			(unsigned int)rtsk->syscall_id, 0, 0);

	    	goto nolog;
	    }

    	    Rprintk("revirt_presys: pid %d unlogged syscall %d id %ld was logged, replaying\n",
			tsk->pid, le->syscall, le->syscall_id, 0, 0);
	    rtsk->syserror_replay_unlogged=1;
	    goto log;
	}

	    log:
		if(tsk->revirt_flags & REPLAY_LOG) 
	    	    tsk->revirt_flags |= REPLAY_XUSER;
	    	else if(tsk->revirt_flags & REPLAY_REPLAY) {
	    	    /* 
	   	    * If there was any strangeness in the return (syscall
	   	    * failed, for instance), compensate for it.  If handle_syserror
	   	    * didn't cancel the syscall, set the flags.
	   	    *
	   	    * There's some problem with exiting, and I suspect it might
	   	    * have to do with trying to read off the end of the file.
	   	    */
    		    if(!ztc || !(ztc->flags &ZTRACE_WAITEXIT))
    			handle_syserror(&regs, rh, le);

    		    if(regs.orig_eax) {
    			tsk->revirt_flags |= REPLAY_DATASQUASH;	
    			tsk->revirt_flags |= REPLAY_SYSCALL;
    		    }
		}
		else {
			printk("revirt_presys: Huh? [pause]\n");
			ztrace_debug_pause();
		}
		goto out;
    	nolog:
	/* Is this really necessary here? */
	if(tsk->revirt_flags & REPLAY_REPLAY)
#if 0
		if((ztc->flags & ZTRACE_XUSER_STOP) && 
		       (rtsk->nrxuser==ztc->xstop.nrxuser)) 
#endif
		       		ztrace_tsk_xuser_schedule(tsk);
	out:
		return;
}

asmlinkage void revirt_replay_syscall(struct pt_regs * regs);
void revirt_log_sysret(struct pt_regs *);

asmlinkage void revirt_postsys(struct pt_regs regs) {
	struct pt_regs * pregs = &regs;
	struct task_struct * tsk=current;
	struct revirt_tsk_s * rtsk;

	tsk->revirt_flags &= ~(REPLAY_DATASQUASH|REPLAY_XUSER|REPLAY_SYSCALL);
	if(!(rtsk=tsk->revirt_tsk)) {
		printk("revirt_postsys: pid %d has no revirt_tsk!\n",
				tsk->pid);
		return;
	}

	/*
	 * This was done in revirt_presys, but in certain cases ([lr]execve)
	 * we have a postsys w/o a presys.  Sync it up.
	 */
	rtsk->current_syscall = revirt_current_syscall;

	/*
	 * Figure out whether to log the syscall or not.  Most syscalls we always log;
	 * some syscalls we only log if there's an error.  Others are special-cased:
	 * write() needs to be logged if it returns an error OR if it's a short write.
	 *
	 * We have to decide this here because we don't know whether it succeeded or
	 * not until it's finished; but OTOH, in order to get XUSERS we need to 
	 */
	if(FD_ISSET(rtsk->current_syscall, 
			&revirt_options.unlogged_syscalls)) {
	   if(tsk->revirt_flags & REPLAY_LOG) {
	    if(revirt_current_syscall == 119) {
		goto nolog;
	    } else if((int)regs.eax < 0 
	    		&& ((int)regs.eax > -125 || (int)regs.eax == -512) ) {
		if(revirt_current_syscall != __NR_waitpid)
		    printk("revirt_postsys: unlogged syscall %d returned %d, logging...\n",
			revirt_current_syscall, (int) regs.eax);
		goto log;
	    } else if(revirt_current_syscall==__NR_write 
	    		&& regs.eax != regs.edx ) {
		Rprintk("revirt_postsys: write returned %d, wanted %d. logging...\n",
			(int)regs.eax, (int)regs.edx, 0, 0, 0);
		goto log;
	    	
	    }
	   } else if(tsk->revirt_flags & REPLAY_REPLAY) {
		if(rtsk->syserror_replay_unlogged) {
		    Rprintk("revirt_postsys: syserror_replay_unlogged set, replaying\n", 0, 0, 0, 0, 0);
		    goto log;
		}
	   }
	    goto nolog;
	} 
	    log:
		if(tsk->revirt_flags & REPLAY_LOG)
		    revirt_log_sysret(pregs);
		else if(tsk->revirt_flags & REPLAY_REPLAY) {
		    cleanup_syserror(pregs);
		    revirt_replay_syscall(pregs);
		    if(revirt_options.replayflags & REPLAY_OPT_LOGSYS_ON_REPLAY)
			    revirt_log_sysret(pregs);
		}
		else {
			printk("revirt_postsys: Huh? [pause]\n");
			ztrace_debug_pause();
		}

	nolog:

    	/*
      	* This needs to happen after replaying the syscall, because the
	* copy_to_user's are squashed & replayed...
	*/
	if(revirt_current_syscall == __NR_socketcall
	    && regs.ebx == SYS_SOCKETPAIR
	    && regs.eax == 0) {
	    /* 
	    * UML uses socketpair rather than pipe; we want the returned
	    * fd's to be unlogged.  This is for proper running, not just
	    * efficiency: if the socketpairs are logged, the read()s will
	    * be suppressed but writes() won't: so when the buffers
	    * fill up, annoying things happen.
	    *
	    * eax: __NR_socketcall
	    * ebx: SYS_SOCKETPAIR
	    * ecx: unsigned int * args[6]
		* socketpair(
		    args[0] int d,
		    args[1] int type,
		    args[2] int protocol,
		    args[3] int sv[2]);
	    */
	    unsigned int a[6];
	    int fds[2];
	    if(copy_from_user(a, (char *)regs.ecx, sizeof(int)*6)) {
		    printk("revirt_postsys: fault trying to get socketpair!\n");
	    }  else {
	     if(copy_from_user(fds, (char *)a[3], sizeof(int)*2)) {
		    printk("revirt_postsys: fault trying to get socketpair!\n");
	     } else {
		    printk("revirt_postsys: socketpair fds {%d,%d}\n",
		    	    fds[0], fds[1]);
		    if(fds[0] > 0 && fds[0] < 255) {
			FD_SET(fds[0],&revirt_options.unlogged_fds);
		    } else {
			ztrace_debug_pause();
		    }
		    if(fds[1] > 0 && fds[1] < 255) {
			FD_SET(fds[1],&revirt_options.unlogged_fds);
		    } else
			ztrace_debug_pause();
	     }
	    }
	}

	if(revirt_current_syscall == __NR_open
		&& regs.eax > 0) {
	    Dprintk("revirt_postsys: %s opened on fd %d\n", (char *)regs.ebx, regs.eax);
    	    /* FIXME -- make this more generic */

	    /* FIXME -- this is a security risk... */
	    char * p=(char *)regs.ebx, *dev="/dev", *hda="hda", *raw="raw",
	    	*cowfile="cowfile";
	    int i, j;

	    for(i=0; i<8; i++) {
	    	if(p[i]!=cowfile[i])
		    break;
	    }

	    if(i==8) 
	    	goto unlogged;

	    /* Make sure the pathname is /dev */
	    for(i=0; i<4; i++) {
	    	if(p[i]!=dev[i])
		  goto not_unlogged;
	    }

	    /* See if it contains "hda" or "raw" */
	    for(i=4; i < 256 && p[i]; i++) {
	    	for(j=0; j<3; j++) {
		    if(p[i+j]!=hda[j])
		    	break;
		}
		if(j==3)
		    goto unlogged;

	    	for(j=0; j<3; j++) {
		    if(p[i+j]!=raw[j])
		    	break;
		}
		if(j==3)
		    goto unlogged;
	    }
	    goto not_unlogged;

	   unlogged: 
		printk("revirt_postsys: %s opened 'unlogged' on fd %d\n",
			p, regs.eax);
		FD_SET(regs.eax, &revirt_options.unlogged_fds);

	    not_unlogged:

	}

	if(revirt_current_syscall == __NR_close
	  	&& regs.eax == 0
		&& FD_ISSET(regs.ebx, &revirt_options.unlogged_fds)) {
		printk("revirt_postsys: closing unlogged fd %ld, clearing flag.\n",
				regs.ebx);
		FD_CLR(regs.ebx, &revirt_options.unlogged_fds);
	}
	    
	/* FIXME -- add open() as well, probably... */
	/* FIXME -- need to track close() for security reasons */
	/* FIXME -- for general procs, this needs to be per-process... */
}

void ztrace_tsk_rexec_done(void);

void revirt_replay_rdtsc(int *hi, int *lo, unsigned long eip) {
	struct rlog_header rh;
	struct rlog_rdtsc * le;

	revirt_read_log(&rh);
	if(rh.type != ENTRY_RDTSC) {
	    printk("revirt_replay_rdtsc: entry %d (#%lu), wanted entry %d\n",
	    		rh.type, rh.id, ENTRY_RDTSC);
	    ztrace_debug_pause();
	    return;
	} 
	le= (struct rlog_rdtsc *) revirt_scall_buf;
	if(le->eip != eip) {
	    printk("revirt_replay_rdtsc: WARNING, got eip %x, wanted %x\n",
	    		eip, le->eip);
	    ztrace_debug_pause();
	}
	*hi = le->hi;
	*lo = le->lo;
}

u32 revirt_replay_in(u16 port, u32 eip) {
    struct rlog_header rh;
    struct rlog_entry_in *le;

    revirt_read_log(&rh);
    if(rh.type != ENTRY_IN) {
	    printk("revirt_replay_in: entry %d (#%lu), wanted entry %d\n",
	    		rh.type, rh.id, ENTRY_IN);
	    ztrace_debug_pause();
	    return 0;
    }
    le= (struct rlog_entry_in *) revirt_scall_buf;
	if(le->eip != eip) {
	    printk("revirt_replay_in: WARNING, got eip %x, wanted %x\n",
	    		eip, le->eip);
	    ztrace_debug_pause();
	}
	if(le->port != port) {
	    printk("revirt_replay_in: WARNING, got port %x, wanted %x\n",
	    		port, le->port);
	    ztrace_debug_pause();
	}
    Rprintk("revirt_replay_in: @0x%x in 0x%x == 0x%x\n",
    		le->eip, le->port, le->byte, 0, 0);
    return le->byte;
}

asmlinkage int revirt_replay_generic(int type, int len, char * data) {
    struct rlog_header rh;
    struct rlog_entry_generic *ge; 
    int i;

    if(len > REPLAY_GENERIC_SIZE) {
      printk("revirt_replay_generic: WARNING, len %d > max size %d! truncating\n",
      	len, REPLAY_GENERIC_SIZE);
      len=REPLAY_GENERIC_SIZE;
    }

    revirt_read_log(&rh);
    if(rh.type != ENTRY_GENERIC) {
    	printk("revirt_replay_generic: type %d (#%lu), wanted type %d\n",
		rh.type, rh.id, ENTRY_GENERIC);
	ztrace_debug_pause();
	return 0;
    }
    ge = (struct rlog_entry_generic *)revirt_scall_buf;
    if(ge->type != type) {
    	printk("revirt_replay_genric: generic type %d, wanted %d! (#%lu)\n", 
		ge->type, type, rh.id);
	ztrace_debug_pause();
    }
#ifdef REPLAY_SYSCALL_LOGS_BRANCHCOUNT
    check_branchcount(&ge->ctr, NULL, rh);
#endif
    Rprintk("revirt_replay_generic: type %d, len %d\n", type, len, 0, 0, 0);
    for(i=0; i<len; i++) {
    	data[i] = ge->data[i];
    }
    return len;
}


EXPORT_SYMBOL_NOVERS(revirt_replay_in);
EXPORT_SYMBOL_NOVERS(revirt_replay_generic);

asmlinkage void revirt_replay_syscall(struct pt_regs * regs) {
	int ret = -ENOSYS, r=0;
	unsigned int t;
	struct rlog_header rh;
	struct rlog_put_user_entry * pe;
	struct rlog_copy_user_entry * ce;
	struct rlog_entry_sysret * se;
	struct rlog_clear_user_entry *cle;
	struct task_struct * tsk=current;
	int check_eax=1;

	Dprintk("revirt_replay_syscall: rexec returned %ld.\n", regs->eax);
	/*
	 * Handle_syserror may have set this to 0 to skip it; set it back so that
	 * signals work right.
	 */
	if(revirt_current_syscall && !regs->orig_eax) {
	    check_eax=0;
	    regs->orig_eax = revirt_current_syscall;
	}

	/*
	 * See if we missed a wanted block...
	 */
	ztrace_tsk_rexec_done();

	if(!tsk->revirt_tsk) {
		printk("revirt_replay_syscall: pid %d has no revirt_tsk!\n",
				tsk->pid);
		return;
	}
	tsk->revirt_tsk->nrxuser=0;

	Dprintk("revirt_replay_syscall: replaying\n");
	while(1) {
	    revirt_read_log(&rh);

	    switch(rh.type) {
	    	case ENTRY_PUT_USER:
		    pe=(struct rlog_put_user_entry *)revirt_scall_buf;

		    if(pe->pid != tsk->pid) {
		    	printk("revirt_replay_syscall %d (%lu):pid %d, expecting %d\n",
				tsk->revirt_tsk->current_syscall, 
				rh.id, tsk->pid, pe->pid);
			printk("  pausing...\n");
			ztrace_debug_pause();
		    }
		    /*
		     * Ok, this is wierd.  put_user() is called with 
		     * any size integer, without being told its size, and
		     * is expected to figure it out with compiler techniques.
		     * It does so using the sizeof(*(ptr)) compiler operator.
		     * As a result, __put_user_size is always "called" with
		     * a constant: 1, 2, or 4 (for this architecture).  This
		     * is replaced with switch(2) { } (or 1 or 4), and the 
		     * optimizer optimizes away all other cases, including
		     * the default.
		     *
		     * If ever anyone calls put_user with a pointer not of
		     * size 1,2, or 4, the default: case is instantiated, 
		     * which calls __put_user_bad.  Since there is no
		     * __put_user_bad, this causes the link to fail.
		     *
		     * But it will do the same thing if we call __put_user_size
		     * with a variable, such as pe->ssize, since the compiler
		     * doesn't know what values pe->ssize may take.  So
		     * we call __put_user_size with constants.
		     */
		   ztrace_tsk_xuser_schedule(tsk);
		    Dprintk("revirt_replay_user: pid %d(%lu)__put_user_size(%llx, %lx, %d)\n",
		    		tsk->pid, rh.id, pe->xx, pe->pptr, pe->ssize);
		    switch(pe->ssize) {
		      case 1:
		    	__put_user_size(pe->xx, pe->pptr, 1, r);
			break;
		      case 2:
		    	__put_user_size(pe->xx, pe->pptr, 2, r);
			break;
		      case 4:
		    	__put_user_size(pe->xx, pe->pptr, 4, r);
			break;
		      case 8:
		    	__put_user_size(pe->xx, pe->pptr, 8, r);
			break;
		      default:
		      	printk("revirt_replay_user: bad put_user size: %d [pause]\n", 
					pe->ssize);
			ztrace_debug_pause();
		    }
		    tsk->revirt_tsk->nrxuser++;
		    if(r!=pe->rretval) {
		    	printk("revirt_replay_syscall %d:putuser ret %d,expecting %d\n", 
				tsk->revirt_tsk->current_syscall, r, pe->rretval);
		    }
		    break;


		case ENTRY_COPY_USER:
		  do {
		    char * d;
		    int i, l;

		    ce=(struct rlog_copy_user_entry *)revirt_scall_buf;
		    if(ce->pid != tsk->pid) {
		    	printk("revirt_replay_syscall %d: pid %d, expecting %d(%lu)\n",
				tsk->revirt_tsk->current_syscall, 
				tsk->pid, 
				ce->pid, rh.id);
			printk("  pausing...\n");
			ztrace_debug_pause();
		    }
		    if(ce->syscall != tsk->revirt_tsk->current_syscall) {
		    	printk("revirt_replay_syscall:syscall %d, expecting %d(%lu) [pause]\n", 
				tsk->revirt_tsk->current_syscall, ce->syscall, rh.id);
			ztrace_debug_pause();
		    }

		    ztrace_tsk_xuser_schedule(tsk);

		    l=ce->ssize;
		    t=ce->tto;
		    i=0;
		    if((d=kmalloc(l, GFP_KERNEL))==NULL) {
		    	printk("revirt_replay copy_user: cannot allocate buffer.\n");
			BUG();
		    }
		    memcpy(d, revirt_scall_buf+sizeof(*ce), l);
#if 0
		    /*
		     * We changed the semantics of ztrace_tsk_xuser_schedule to do a
		     * lot of the checking there instead of here; that kind of messes
		     * with the logic down below.  This is a low-probability case, so
		     * we'll fix this if it comes up.
		     */
		    /*
		     *  tsk->revirt_tsk->nrxuser is incremented at the beginning
		     * of a copy_user, but at the end of this, as it needs
		     * to be to interrupt between xusers.  But when 
		     * interrupting an xuser, the logged nrxuser is one
		     * more, so we compensate.
		     */
		    while(ztc->flags & ZTRACE_XUSER_STOP && 
		    	   (ztc->xstop.nrxuser-(ztc->xstop.edi!=0))
			   			==tsk->revirt_tsk->nrxuser) {
			int delta;
			printk("copy_user: edi=0x%x\n", ztc->xstop.edi);
			if(!ztc->xstop.edi) {
				ztrace_tsk_xuser_schedule(tsk);
				continue;
			}
			delta=ztc->xstop.edi - t;
			if(delta > l) {
			    printk("rp_copyuser: delta %d, l %d\n", 
			    		delta, l);
			    printk("  pausing...\n");
			    ztrace_debug_pause();
			    BUG();
			}
			r=delta;
			printk("revirt_replay_syscall: pid %d __copy_user_atom(0x%x,0x%lx,%d)\n",
					tsk->pid, t, (long unsigned int)d+i, r);
			__copy_user_atom(t,d+i,r);
			if(r) {
		    	 printk("rp_syscall:__copy_user_atom failed at %d\n",
					r);
			 printk("   syscall:%d syscall_id:%lld\n", ce->syscall,
					ce->syscall_id);
			 //BUG();
		    	}
			t+=delta;
			i+=delta;
			l-=delta;
			ztrace_tsk_xuser_schedule(tsk);
		    }
#endif
		    r=l;
		    Dprintk("revirt_replay_syscall: pid %d(%lu) __copy_user_atom(0x%x,0x%lx,%d)\n",
					tsk->pid, rh.id, 
					t, (long unsigned int)d+i, r);
		    __copy_user_atom(t, d+i, r);
		    tsk->revirt_tsk->nrxuser++;
		    kfree(d);
		    if(r) {
		    	printk("revirt_replay_syscall:__copy_user_atom failed at %d\n",
					r);
			printk("   syscall:%d(%lu) syscall_id:%lld\n", ce->syscall,
					rh.id, ce->syscall_id);
			//BUG();
		    }
		  } while(0);
		    break;
		case ENTRY_CLEAR_USER:
		    cle=(struct rlog_clear_user_entry *)revirt_scall_buf;
		    if(cle->pid != tsk->pid) {
		    	printk("revirt_replay_syscall %d: pid %d, expecting %d [pause]\n", 
				tsk->revirt_tsk->current_syscall, tsk->pid, 
				cle->pid);
			ztrace_debug_pause();
		    }
		    if(cle->syscall != tsk->revirt_tsk->current_syscall) {
		    	printk("revirt_replay_syscall:syscall %d, expecting %d [pause]\n", 
				tsk->revirt_tsk->current_syscall, cle->syscall);
			ztrace_debug_pause();
		    }
#if 0
		    if((ztc->flags & ZTRACE_XUSER_STOP) && 
		       (tsk->revirt_tsk->nrxuser==ztc->xstop.nrxuser)) 
#endif
		       		ztrace_tsk_xuser_schedule(tsk);
		    r=__clear_user((void *)cle->to, cle->n);
		    tsk->revirt_tsk->nrxuser++;
		    if(r) {
		    	printk("revirt_replay_syscall:__clear_user failed at %d\n",
					r);
			printk("   syscall:%d \n", cle->syscall);
			//BUG();
		    }
		    break;
		case ENTRY_SYSRET:
		    se=(struct rlog_entry_sysret *)revirt_scall_buf;
		    if(se->pid != tsk->pid) {
		    	printk("revirt_replay_syscall sys %d id %lu: pid %d, expecting %d [pause]\n", 
				tsk->revirt_tsk->current_syscall, 
				rh.id,
				tsk->pid, 
				se->pid);
			ztrace_debug_pause();
		    }
		    if(se->syscall != tsk->revirt_tsk->current_syscall) {
		    	printk("revirt_replay_syscall:syscall %d, expecting %d(%lu) [pause]\n", 
				tsk->revirt_tsk->current_syscall, se->syscall,
				rh.id);
			ztrace_debug_pause();
		    }
		    if(se->syscall_id != tsk->revirt_tsk->syscall_id) {
		    	printk("revirt_replay_syscall:syscall_id %lld, expecting %lld (%lu) [pause]\n", 
				tsk->revirt_tsk->syscall_id, se->syscall_id,
				rh.id);
			ztrace_debug_pause();
		    }
		    ztrace_tsk_xuser_schedule(tsk);

		    /*
		     *  Check for syscalls with changed values.... we really need
		     * to make this a bitfield of logged syscalls...
		     */
		   switch(se->syscall) {
		   	case __NR_read:
			case __NR_socketcall:
			case __NR_ioctl:
			case __NR_poll:
				break;
			case __NR_nanosleep:
			    /* If nanosleep is woken up early for any reason,
			       it returns -EINTR w/o checking for a signal; 
			       but zR purposely wakes it up early to get
			       idle time compression (no sense waiting around
			       for no reason); so here we check for that case,
			        and ignore it. */
			    if(se->regs.eax != regs->eax
			    	&& regs->eax == -EINTR
				&& !signal_pending(current)) 
				break;
			/* FALLTHRU */
			default:
		   { 
		    int pause=0;
#ifdef REPLAY_WHOLE_PTREGS
#define DIFF(x,y) \
	if(se->regs.x != regs->x) { 			 \
	  printk("revirt_replay_syscall: rex'd %s differs: %x %x\n", \
	  	y, (unsigned int)regs->x, (unsigned int)se->regs.x);	  \
  	  pause = 1;					 \
	}
		    DIFF(ebx, "ebx");
		    DIFF(ecx, "ecx");
		    DIFF(edx, "edx");
		    DIFF(esi, "esi");
		    DIFF(edi, "edi");
		    DIFF(eip, "eip");
		    DIFF(esp, "esp");
		    DIFF(ebp, "ebp");
		    /*
		     * Only the bits from REVIRT_EFLAGS_MASK are actually
		     * accessible and affect flow in ring 3,
		     * so we only check those.
		     */
		     if((se->regs.eflags & REVIRT_EFLAGS_MASK)
			!= (regs->eflags & REVIRT_EFLAGS_MASK)) {
			printk("revirt_replay_syscall: rex'd eflags differs: %lx %lx\n",
				regs->eflags & REVIRT_EFLAGS_MASK,
				se->regs.eflags & REVIRT_EFLAGS_MASK);
			pause=1;
		    }
#undef DIFF
#endif
		    if(se->regs.eax != regs->eax) {
		    	if(!check_eax) {
#if 0
		    	if(regs->orig_eax == 0 
			       && ( ((int)se->regs.eax < 0 && (int)se->regs.eax > -125)
				    || ((int)se->regs.eax == -512)) ){
#endif
			 Dprintk("revirt_replay_syscall: tolerating skipped exec\n");
			} else {
		    	 printk("revirt_replay_syscall: rex'd eax differs: %lx %lx\n",
					regs->eax, se->regs.eax);
			 pause=1;
			}
		    }

#ifdef REPLAY_SYSCALL_LOGS_BRANCHCOUNT
		    check_branchcount(&se->ctr, regs, rh);
#endif
		    if(pause){
		    	printk("revirt_replay_syscall: syscall %d(%lu) should probably be logged [pause]\n",
				se->syscall, rh.id);
			ztrace_debug_pause();
		    }
		    } // default:
		   }// switch(syscall)


		   regs->eax = se->regs.eax;

		   ret = regs->eax;
		   Dprintk("revirt_replay_sycall: pid %d ret %d @ 0x%lx\n", 
		    		tsk->pid, ret, regs->eip);
# ifdef REPLAY_SYSCALL_LOGS_BRANCHCOUNT
#  ifdef REPLAY_RECORD_ALL_PFCTRS
		   Dprintk("revirt_replay_syscall: logged count: [%llx %x %x %x]\n",
		    		se->ctr.hw_branch,
		    		se->ctr.hw_intr,
		    		se->ctr.ex_noerror,
		    		se->ctr.ex_error);
#  else
		   Dprintk("revirt_replay_syscall: logged count: [%llx]\n",
		    		se->ctr.hw_branch);
#  endif
# endif
		    tsk->revirt_tsk->nrxuser++;
		    goto out;
		default:
		    printk("Bad syscall entry: %d (%lu) [pause]\n", 
				    rh.type, rh.id);
		    if(rh.type == ENTRY_RT_SIGFRAME) {
		    	struct rlog_entry_rt_sigframe * sys;
		    	sys=(struct rlog_entry_rt_sigframe *)revirt_scall_buf;
#ifdef REPLAY_WHOLE_SIGFRAME 
	printk(" sig: %d eip:%lx\n", sys->frame.sig, sys->frame.uc.uc_mcontext.eip); 
#else
	printk(" sig: %d\n", sys->frame.sig);
#endif
		    }
		    ztrace_debug_pause();
	    }

	    /* Play nice */
	    if(current->need_resched)
	    	schedule();
	    /*
	     * Check for whether this is appropriate are done in the function...
	     */
	    ztrace_tsk_xuser_schedule(tsk);

	}
    out:
	return;
}

asmlinkage int revirt_writev_log(int, const struct iovec *, int, struct revirt_group_struct *);

unsigned long revirt_copyuser(unsigned long to, unsigned long from, 
	unsigned long size) {
    struct rlog_copy_user_entry le;
    struct task_struct * tsk = current;

    if(!tsk->revirt_tsk && !(tsk->revirt_flags & REPLAY_ZOMBIE)) {
	    printk("revirt_copyuser: pid %d has no revirt_tsk!\n",
			    tsk->pid);
	    return -EFAULT;
    }

    /*
     * There are times when the same function may write either to userland
     * or to kernel; e.g., low-level file reads.  We need to make sure to
     * log & replay only those which go into userland.
     *
     * FIXME: we need to use a #define for this to make it portable
     */
    if(to > 0xc0000000) {
    	int flag;
	flag = tsk->revirt_flags & REPLAY_XUSER;
	tsk->revirt_flags &= ~REPLAY_XUSER;
	__copy_user_atom(to, from, size);
	tsk->revirt_flags |= flag;
	return size;
    }

    if(tsk->revirt_flags & REPLAY_LOG) {
	if(tsk->revirt_flags & (REPLAY_XUSER)) {
	    /* log both metadata and data */
            struct iovec v[2]; 
            le.tto = to;
            le.ssize = size;
            le.pid = tsk->pid;
            le.syscall_id = tsk->revirt_tsk ? tsk->revirt_tsk->syscall_id:0;
	    le.syscall = revirt_current_syscall;
            v[0].iov_base = &le;
            v[0].iov_len = sizeof(le);
            v[1].iov_base = (void *)from;
            v[1].iov_len = size;
            revirt_writev_log(ENTRY_COPY_USER, v, 2, NULL);
	}
	__copy_user_atom(to, from, size);
	if(size==0) {
	    /* FIXME -- by putting this here, if we ever go back to a multi-process
	     * model, schedules between xusers for unlogged syscalls will fail
	     * silently. */
	    if(tsk->revirt_flags & REPLAY_XUSER && tsk->revirt_tsk)
		tsk->revirt_tsk->nrxuser++;
	} else 
		printk("revirt_copyuser: pid %d size returned %lu\n", tsk->pid, size);
	return size;
    }
    if(tsk->revirt_flags & REPLAY_REPLAY) {
	if(tsk->revirt_flags & REPLAY_DATASQUASH) {
	    if(to < 0xc0000000) {
		Dprintk("squashing copy_user(0x%lx,0x%lx,%ld)\n",
			to, from, size);
		return 0;
	    } else {
	    	printk("WARNING: calling copy_user with kernel destination\n");
		printk("   copy_user(0x%lx,0x%lx,%ld)\n", to, from, size);
		ztrace_debug_pause();
		BUG();
	    }
	}
	else {
		__copy_user_atom(to, from, size);
		if(size==0) {
		    if(tsk->revirt_flags & REPLAY_XUSER && tsk->revirt_tsk)
			tsk->revirt_tsk->nrxuser++;
		} else 
			printk("revirt_copyuser: pid %d size returned %lu\n", tsk->pid, size);
		return size;
	}
    }
    printk("revirt_replay_copy_user: huh?\n");
    ztrace_debug_pause();
    return -1;
}

void revirt_log_exit(void);

void revirt_exit(struct task_struct * tsk) {
	static int recursive=0;
	struct ztrace_ctl_struct * ztc;


	if(tsk->revirt_flags & REPLAY_ZOMBIE) {
	    if(recursive)
	    	return;
	    printk("revirt_exit: task %d REPLAY_ZOMBIE already set!\n",
	    		tsk->pid);
	    recursive=1;
	    return;
	}
	
	recursive=0;

	tsk->revirt_flags |= REPLAY_ZOMBIE;

	tsk->revirt_flags &= ~(REPLAY_XUSER|REPLAY_DATASQUASH);

	ztc=get_ztc(tsk);

        if (tsk->revirt_flags&REPLAY_LOG)  {
		/*
		 * No more important schedule logs...
		 */
		tsk->revirt_flags |= REPLAY_SUPPRESS_SCHED;
                revirt_log_exit();
        } else if(tsk->revirt_flags & REPLAY_REPLAY) {
	    if(ztc) {
	      if(tsk->revirt_tsk->sched) {
                if(ztc->flags & ZTRACE_WAITEXIT) {
                        Dprintk("revirt_exit: pid %d waking up scheduler\n",
                                tsk->pid);
                } else {
                        printk("revirt_exit: pid %d exiting unexpectedly?\n",
                                tsk->pid);
                }
                /* FIXME: do we need to worry about locks here? */
                tsk->revirt_tsk->sched->revirt_flags &= ~REPLAY_WAIT;
		ztc->yield = ZTRACE_WAITEXIT;
                ztc->flags = ZTRACE_WAITEXIT;
	      } else {
		    printk("revirt_exit: tsk %d exiting, no scheduler\n",
				    tsk->pid);
	      }
	    } else {
		    printk("revirt_exit: tsk %d exiting, no revirt_tsk [pause]\n", 
		    		tsk->pid);
		    ztrace_debug_pause();
	    }
	    tsk->revirt_flags=0;
        }

	Rprintk("revirt_exit: pid %d\n", tsk->pid, 0, 0, 0, 0);
	put_rtsk(tsk);
}
