/*
 * ztrace.c - Interface for zR user-level schedule replayer
 *
 * Copyright (C) 2002-2003 George Dunlap
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <linux/module.h>
#include <linux/sched.h>
#include <asm/uaccess.h>
#include <asm/user.h>

#include <asm/perfctr.h>
#include <linux/slab.h>

#include <asm/hardirq.h> // for in_interrupt()

#include <revirt/rlog.h>
#include <revirt/ztrace.h>
#include <revirt/perfctr.h>

/* for ZTRACE_SMI */
#include <asm/io.h>

//#define P4DEBUG

/*
 * FIXME -- take out syscalls 5 and 6 (open and close) 
 * Probably also some more
 */
struct revirt_options_st revirt_options = {
	//.logflags= REPLAY_DEFAULT_LOGFLAGS,
	.replayflags=REPLAY_DEFAULT_REPLAYFLAGS,
	.xreplay = {
		.packetrate=5, 
		.packetsize=1000 },
	.network_fd=UML_NETWORK_FD, 	/* network_fd */
	.unlogged_fds={}, 		/* unlogged_fds */
	.allowed_syscalls = {
	 .fds_bits={
		/* FIXME -- we need to take out open, close, ptrace,
			setitimer, rt_sigaction, sigaltstack,  */
	  /* 1 exit 
	     3 read
	     4 write
	     5 open
	     6 close
	     19 lseek
	     20 getpid
	     26 ptrace */
	  0x418007a, 
	  /* 55  fcntl */
	  0x800000, 
	  /* 78 gettimeofday
	     90 mmap
	     91 munmap */
	  0xc004000, 
	  /* 104 setitimer
	     126 sigprocmask */
	  0x40000100, 
	  /* 140 _llseek*/
	  0x1000, 
	  /* 168 poll
	     173 rt_sigreturn
	     174 rt_sigaction
	     179 rt_sigsuspend
	     186 sigaltstack */
	  0x4086100, 
	  0x0,
	  /* 233 ztrace  FIXME -- need to limit this functionality */ 
	  /* 235 getguestpgd 
	     236 switchguest */
	  0x1a00 } } ,
	  .unlogged_syscalls = {
#ifdef REPLAY_ALL_SYSCALLS
	    .fds_bits = { 0, 0, 0, 0, 0, 0, 0, 
		/* 224-255 
	    	    Always unlogged:
			231 lexecve
			232 rexecve
		 */
		0x00000180 }
#else
	    .fds_bits = {
	    	/* 0-31 
		
			4 write
			6 close
			7 waitpid
			19 lseek
			20 getpid
			26 ptrace
		*/
		0x041800d0,
		/* 32-63 
		    	37 kill
		 */
		0x00000020, 
		/* 64-95 
			90 mmap [26]
			91 munmap [27]
		*/
		0x0c000000,
		/* 96-127 
			104 setitimer
			114 wait4
			119 sigreturn
			125 mprotect
	    		126 sigprocmask
		*/
		0x60840100,
		/* 128-159 
			140 _llseek
		*/
		0x00001000,
		/* 160-191 
			173 rt_sigreturn (13)
			174 rt_sigaction (14)
			175 rt_sigprocmask (15)
			179 rt_sigsuspend (19)
			186 sigaltstack (26)
		*/
		0x0408e000,
		/* 192-223 
			192 mmap2
		 */
		0x1,
		/* 224-255 
			#231 lexecve
			#232 rexecve
		 */
		0x00000000,
		/* 256-287 
			274 lexecve
			275 rexecve
		*/
		0x000c0000
	    }
#endif
	  },
	 .use_unlogged_fds = {
	    .fds_bits={
	    	/* 0-31  
		 *  6 open
		 */
		0x00000040,
		/* 32-63 */
		0x00000000, 
		/* 64-95 */
		0x00000000,
		/* 96-127 */
		0x00000000,
		/* 128-159 */
		0x00000000,
		/* 160-191 */
		0x00000000,
		/* 192-223 
		 *   221 fcntl 
		 */
		0x20000000,
		/* 224-255 */
		0x00000000,
		/* 256-287  */
		0x00000000
	    }
	 },
	 .logbuf= {
		 .size=(1 << 21),	/* This must be a power of 2...*/
		 .mask=((1 << 21)-1) }, /* for this to work */
 };

EXPORT_SYMBOL_NOVERS(revirt_options);



//int zt_flags=0;
//struct ztrace_rununtil_struct zt_rununtil;
//struct ztrace_xuser_stop zt_xstop;
//int zt_yield=0;
//int zt_wchan=0;
//struct task_struct * zt_tsk=NULL;
//s64 zt_perferr=0;
//int zt_pid=0;
//int zt_rhid=0;

spinlock_t ztrace_yieldto_lock = SPIN_LOCK_UNLOCKED;

//#define ZTRACE_DEBUG

/****************************************************************
 *
 * Call-backs from important kernel places...
 *
 ****************************************************************/
/* 
 * Called from arch/i386/kernel/traps.c:do_debug(), for single-stepping and 
 * hardware breakpoints.
 */
void ztrace_tsk_debug(void) {
	struct task_struct *tsk=current;
	struct ztrace_ctl_struct * ztc;
	
	if(!(ztc=get_ztc(tsk))) 
	    return;

	if(tsk->pid != ztc->pid)
		return;

	if((tsk->revirt_flags & REPLAY_SSTEP) 
		|| (ztc->flags & (ZTRACE_BKPT|ZTRACE_PERF))) {
    	    ztc->yield=ZTRACE_SSTEP;
    	    tsk->revirt_flags &= ~REPLAY_SSTEP;
	    if((tsk->revirt_flags & REPLAY_USER_TRAP) == 0) {
	      zreg(tsk,eflags) &= ~TRAP_FLAG;
	    }
    	    ztrace_yield_to(tsk->revirt_tsk->sched);
	} else {
	    printk("ztrace_task_sstep: REPLAY_SSTEP and ZTRACE_BKPT unset!\n");
	    printk("   pausing...\n");
	    ztrace_debug_pause();
	}
}

/*
 * Called from arch/i386/kernel/apic.c:smp_perfcntr_overflow()
 * when we get a performance counter overflow interrupt
 */

/*
 * FIXME For multiple logging / replay, this needs to be muxed somehow...
 */
struct ztrace_ctl_struct * perf_ztc = NULL;

void ztrace_tsk_perf(void) {
    struct task_struct *tsk=current;

    perfcntr_disable_overflow();

    if(!perf_ztc)
    	return;

    /*
     * FIXME -- p4-specific debug code
     */
#ifdef P4DEBUG
    do {
    	int l,h;
#define CCCR_OVF (1<<31)
	rdmsr(revirt_pfctr.overflow.control_msr, l, h);
	if(!(l & CCCR_OVF)) {
		printk("ztrace_tsk_perf: CCCR_OVF not set! %x\n", l);
	}

    } while(0);
#endif

    /*
     * Clear LVTPC mask, and overflow flag in counter control MSR (P4)
     */
    perfctr_clear_ovf();

    /*
     * Sanity checks
     */
    if(!(perf_ztc->flags & ZTRACE_PERF)) {
    	printk("ztrace_tsk_perf: ZTRACE_PERF unset!\n");
	goto out;
    }

    if(!perf_ztc->tsk) {
    	printk("ztrace_tsk_perf: got ovf, but perf_ztc->tsk == NULL!\n");
	goto out;
    }

    if(!perf_ztc->tsk->revirt_tsk) {
	    printk("zt_tsk_perf: perf_ztc->tsk (pid %d) has no revirt_tsk\n",
			    tsk->pid);
	    goto out;
    }

    /*
     * This needs to be atomic, for if we get swapped out, we may never
     * get run again.  need_resched should be safe before this section,
     * because the scheduler can't run, but we'll protect it here just to
     * be safe.
     *
     * If the process is waiting on a lock (lock_page, for example), this will
     * cause deadlock.  (It gets the lock and is "woken up", but can't progress due
     * to REPLAY_WAIT; then zR tries to lock the same page.)  So if we're blocked,
     * we set the TRAP flag and have ztrace_tsk_debug take care of it.
     *
     * No, even that won't work; if the process was on its way to being blocked
     * when the perf interrupt happens, you have the same problem.  Maybe we just need
     * make everything happen in do_debug...
     *
     * No, alas!  The trap flag is sometimes squashed.  *sigh* OK, we'll add a flag, and
     * check for it before we check for signals on the return path...
     */
    if(perf_ztc->tsk->revirt_flags & REPLAY_REPLAY)
	perf_ztc->tsk->revirt_flags |= REPLAY_DELAY_PERF;
    else
	Zprintk("ztrace_tsk_perf: pid %d not replaying, got perf\n",
		perf_ztc->tsk->pid, 0, 0, 0, 0);

    if(perf_ztc->tsk->state != TASK_RUNNING) {
    	printk("ztrace_tsk_perf: warning, pid %d blocked\n", perf_ztc->tsk->pid);
    }

 //   restore_flags(flags);
  out:

    if(revirt_options.replayflags & REPLAY_OPT_ZTRACE_DEBUG) {
	printk("ztrace_tsk_perf exiting\n");
	do {
	    struct revirt_ctrdata ctr;
	    copy_branchdata(tsk, &ctr);
	    printk("pid %d, count %llx\n", tsk->pid, ADJ_BC(ctr));
	} while(0);
    }

    return;
}

asmlinkage void ztrace_delay_perf(void) {
	struct task_struct * tsk=current;

	if(perf_ztc 
		&& perf_ztc->tsk 
		&& (perf_ztc->tsk->revirt_flags & REPLAY_REPLAY)
		&& perf_ztc->tsk->revirt_tsk) {

	    Zprintk("ztrace_delay_perf: pid %d yielding to pid %d due to DELAY_PERF\n",
	    		tsk->pid, perf_ztc->tsk->revirt_tsk->sched->pid,
			0, 0, 0);

	    perf_ztc->yield=ZTRACE_PERF;
	    /* FIXME -- why was this set, and do we need to work around it? */
	    //perf_ztc->flags=0;

	    tsk->revirt_flags &= ~REPLAY_DELAY_PERF;

    	    ztrace_yield_to(perf_ztc->tsk->revirt_tsk->sched);
	}
}

/*
 * Called when a replayed task blocks.  We'd like to be able to check
 * to make sure we're in the same wchan as before, but that means calling
 * this function in switch_to, which causes all kinds of problems.  For one,
 * at that point, the next process has already been scheduled; for two,
 * it's tough to call schedule again from there.
 *
 * As it is, we don't call schedule, since we'll return there anyway.  We
 * just block this process and wake up its scheduler.
 *
 * I suppose that means it might be called from __switch_to... we'd just have
 * to wait until the next process was scheduled out again before zr
 * was run.
 * 
 * *sigh* The problem is, if the system is not busy, the next process is
 * the idle process, which just sleeps for a timeslice... extremely 
 * inefficient.   I don't remember why we didn't call this from schedule(),
 * I think we'll have to move it back there...
 */ 
void ztrace_tsk_blocks(struct task_struct * tsk) {
    unsigned long flags;
    struct ztrace_ctl_struct *ztc;

    if(!(ztc=get_ztc(tsk)))
    	return;

    if(ztc->flags != ZTRACE_EPOCH_STOP && tsk->pid != ztc->pid)
    	return;

    if(ztc->flags == ZTRACE_EPOCH_STOP 
    	&& tsk->revirt_flags & REPLAY_SYNC_BLOCK) {
	    struct revirt_group_struct * rgroup = tsk->revirt_tsk->group;

	    update_active_pids(tsk, &rgroup->active);
	    rgroup->active.epochs_skipped++;

	    if(ztc->yield == ZTRACE_RESTORE_CHKPT) {
	    	Zprintk("ztrace_tsk_blocks: pid %d yielding to zR due to RESTORE_CHKPT after %d epochs\n",
			tsk->pid, rgroup->active.epochs_skipped, 0, 0, 0);
		ztc->pid=tsk->pid;
		goto yield;
	    } else if(rgroup->active.epochs_skipped == ztc->active.epochs_skipped) {
		Zprintk("ztrace_tsk_blocks: pid %d yielding after %d epochs to pid %d due to SYNC_BLOCK\n",
	    		tsk->pid, 
			ztc->active.epochs_skipped,
			tsk->revirt_tsk->sched->pid, 0, 0);
		ztc->yield=ZTRACE_EPOCH;
		ztc->pid=tsk->pid;
		goto yield;
	    }else if(rgroup->active.epochs_skipped > ztc->active.epochs_skipped) {
	    	printk("ztrace_tsk_blocks: active.epochs skipped %d, wanted %d!\n",
				rgroup->active.epochs_skipped,
				ztc->active.epochs_skipped);
		ztrace_debug_pause();
	    } else {
		Zprintk("ztrace_tsk_blocks: pid %d epochs_skipped %d, want %d\n", 
	    		tsk->pid, rgroup->active.epochs_skipped, 
			ztc->active.epochs_skipped, 0, 0);
	    }
    } else if(ztc->flags & ZTRACE_BLOCK_STOP 
    		&& !(tsk->revirt_flags & REPLAY_NO_YIELD_ON_BLOCK) ) {
	/* 
	 * OK, here's the deal.  We need to call getwchan() so that we can
	 * tell if we're blocking in the right place.  But we can't call
	 * getwchan() until after switch_to, because otherwise tsk->thread.esp
	 * doesn't make any sense.  But we can't call ztrace_tsk_blocks after
	 * switch_to, because then the idle thread gets scheduled too often
	 * and we replay reeeeeaaaly slllloooooowwwwlly.  
	 *
	 * So, we yield to zR at every block, and let zR (in ztrace_block) 
	 * check the wchan...
	 */

            ztc->yield=ZTRACE_BLOCK;
	    goto yield;
    } else if(ztc->flags & ZTRACE_BKPT 
    		&& (zreg(tsk, eip) == ztc->brk.eip)
		&& (ztc->brk.flags & REPLAY_SYSCALL)
    		&& !(tsk->revirt_flags & REPLAY_NO_YIELD_ON_BLOCK) ) {
	    /*
	     * There's sometimes a race condition between the time a process puts
	     * itself on the waitqueue and the time it actually calls schedule; if this
	     * happens, it will be logged as merely a task switch, rather than a block.
	     * But during replay, this race condition won't happen, so the task will
	     * block rather than bkpt.
	     *
	     * I'm pretty sure there's no way for a task to switch in the middle of a
	     * system call right now: so if we block, and the flags say REPLAY_SYSCALL,
	     * then yield.
	     */
            ztc->yield=ZTRACE_BLOCK;
	    goto yield;
    } 
    return;

    yield:
	{
	    struct task_struct * p = tsk;

            spin_lock_irqsave(&ztrace_yieldto_lock, flags);
	    tsk->revirt_tsk->sched->revirt_flags &= ~REPLAY_WAIT;
	    revirt_debug_log_printk('Q', "ztrace_block [yield]: waking pid %d\n",
	    	tsk->revirt_tsk->sched->pid, 0, 0, 0, 0);
	    do {
		p->revirt_flags |= REPLAY_WAIT;
		revirt_debug_log_printk('Q', "ztrace_block [yield]: WAITing pid %d\n",
	    	p->pid, 0, 0, 0, 0);
		if(p->revirt_tsk)
			p=p->revirt_tsk->next;
		else
			break;
	    } while(p!=tsk);
            spin_unlock_irqrestore(&ztrace_yieldto_lock, flags);
	}
}


/*
 * ztrace_tsk_enters_syscall -- called from entry.S just after entering
 * the kernel.
 * 
 * Stop if:
 *  - we've been asked to stop at the next system call entry (SYSENTER_STOP)
 *  or
 *  - We're doing breakpoint replay and the return eip equals the stop eip.
 *
 *  NOTE that nrsyscalls and current_syscall are now set in revirt_presys,
 *   which is AFTER ztrace_tsk_enters_syscall() is called; they can't be
 *   trusted at this point.
 *
 *  Well, nrsyscalls won't have been incremented yet, which is exactly what we want;
 * current_syscall we can deal with as we need to.
 */
asmlinkage void ztrace_tsk_enters_syscall(void) {
	struct task_struct *tsk= current;
	struct ztrace_ctl_struct * ztc;

	if(!(ztc=get_ztc(tsk)))
		return;

	if(tsk->pid != ztc->pid)
		return;

	if(ztc->flags & ZTRACE_BKPT) {
		if(!tsk->thread.debugreg[7]) {
			printk("tsk_enters_syscall: pid %d BKPT, but !debugreg!\n",
					tsk->pid);
			printk("[%lx %lx %lx %lx %lx %lx %lx %lx] [pause]\n",
				tsk->thread.debugreg[0],
				tsk->thread.debugreg[1],
				tsk->thread.debugreg[2],
				tsk->thread.debugreg[3],
				tsk->thread.debugreg[4],
				tsk->thread.debugreg[5],
				tsk->thread.debugreg[6],
				tsk->thread.debugreg[7]);
			ztrace_debug_pause();
		}
	}

	if(ztc->flags & ZTRACE_SYSENTER_STOP) {

		Zprintk("ztrace: pid %d yielding due to SYSENTER_STOP\n",
				tsk->pid, 0, 0, 0, 0);

		//ztc->flags &= ~ZTRACE_SYSENTER_STOP;
		ztc->yield=ZTRACE_SYSCALL_ENTER;
		ztrace_yield_to(tsk->revirt_tsk->sched);
	} 
}

/*
 * Called from entry.S, after returning from the syscall, but before
 * signal delivery / rescheduling stuff.
 *
 * I think ZTRACE_BLOCK wants to stop here because it *expected* to be blocked
 * in the middle of the last system call; for some reason it wasn't.  Same
 * thing with ztrace_tsk_rexec_done(), which happens before this on the same
 * codepath.  Not sure why I have it both places... =(
 */

void ztrace_rununtil_set_bkpt(struct task_struct * tsk, unsigned int eip);

asmlinkage void ztrace_tsk_exits_syscall(void) {
	struct task_struct *tsk= current;
	struct ztrace_ctl_struct * ztc;

	if(!(ztc=get_ztc(tsk)))
		return;

	if(tsk->pid != ztc->pid)
		return;

	if(ztc->flags & ZTRACE_BKPT) {
		if(!tsk->thread.debugreg[7]) {
			printk("tsk_exits_syscall: pid %d BKPT, but !debugreg!\n",
					tsk->pid);
			printk("[%lx %lx %lx %lx %lx %lx %lx %lx] [pause]\n",
				tsk->thread.debugreg[0],
				tsk->thread.debugreg[1],
				tsk->thread.debugreg[2],
				tsk->thread.debugreg[3],
				tsk->thread.debugreg[4],
				tsk->thread.debugreg[5],
				tsk->thread.debugreg[6],
				tsk->thread.debugreg[7]);
			printk("Restoring breakpoint values from ztc->brk...\n");
			ztrace_rununtil_set_bkpt(tsk, ztc->brk.eip);
			printk("[%lx %lx %lx %lx %lx %lx %lx %lx] [pause]\n",
				tsk->thread.debugreg[0],
				tsk->thread.debugreg[1],
				tsk->thread.debugreg[2],
				tsk->thread.debugreg[3],
				tsk->thread.debugreg[4],
				tsk->thread.debugreg[5],
				tsk->thread.debugreg[6],
				tsk->thread.debugreg[7]);
			ztrace_debug_pause();
		}
	}

	if(ztc->flags & ZTRACE_XUSER_STOP) {
		printk("ztrace: exiting syscall, ZTRACE_XUSER_STOP set!\n");
		printk("ztrace: ztc->xstop.nrexuser=%lx, tsk->revirt_tsk->nrxuser=%lx\n",
				ztc->xstop.nrxuser, tsk->revirt_tsk->nrxuser);
		ztrace_debug_pause();
	}
	
	if(ztc->flags & ZTRACE_SYSEXIT_STOP) {

	    Zprintk("ztrace: pid %d yielding at sysexit due to SYSEXIT_STOP\n",
	    		tsk->pid, 0, 0, 0, 0);

	    //tsk->revirt_flags |= REPLAY_AT_EXIT;
	    //ztc->flags &= ~ZTRACE_SYSEXIT_STOP;
	    ztc->yield=ZTRACE_SYSCALL_EXIT;
	    ztrace_yield_to(tsk->revirt_tsk->sched);
	} 
	else if(tsk->revirt_flags & REPLAY_SSTEP ) {

	    Zprintk("ztrace: pid %d yielding at sysexit due to REPLAY_SSTEP\n",
	    		tsk->pid, 0, 0, 0, 0);

	    tsk->revirt_flags &= ~REPLAY_SSTEP;
	    ztc->yield=ZTRACE_SYSCALL_EXIT;
	    ztrace_yield_to(tsk->revirt_tsk->sched);
	} else if(ztc->flags & ZTRACE_BKPT && zreg(tsk,eip)==ztc->brk.eip) {
		Zprintk("ztrace: pid %d yielding after syscall @%x due to BKPT\n",
			 tsk->pid, ztc->brk.eip, 0, 0, 0);

	    ztc->yield=ZTRACE_SYSCALL_EXIT;
		ztrace_yield_to(tsk->revirt_tsk->sched);
	} else if (ztc->flags & ZTRACE_BLOCK_STOP 
			&& zreg(tsk,eip)==ztc->brk.eip
			&& ztc->brk.flags & REPLAY_SYSCALL) {
	    /* 
	     * I'm pretty sure we only want to block here if REPLAY_SYSCALL is set,
	     * which means that there should be no way that SIGBLOCK could be set...
	     */
	    if(ztc->brk.flags & REPLAY_SIGBLOCK) {

	     Zprintk("sysexit: pid %d NOT yielding: BLOCK, but SIGBLOCK\n",
	     		tsk->pid, 0, 0, 0, 0);

	    } else {

	     Zprintk("ztrace: pid %d yielding after syscall due to BLOCK\n",
			tsk->pid, 0, 0, 0, 0);

	     ztc->yield=ZTRACE_SYSCALL_EXIT;
	     ztc->brk.wchan=0;
	     ztrace_yield_to(tsk->revirt_tsk->sched);
	    }
	} else if(ztc->flags == ZTRACE_RESTORE_CHKPT) {
	    Zprintk("ztrace_tsk_exits_syscall: pid %d yielding after syscall due to RESTORE_CHKPT\n",
	    		tsk->pid, 0, 0, 0, 0);
	    ztc->yield = ZTRACE_RESTORE_CHKPT;
	    ztrace_yield_to(tsk->revirt_tsk->sched);
	}
}


/*
 * Called from revirt/replay.c:revirt_replay_sycall().
 * This is called during replay, after the system call is re-executed,
 * before the xusers &c are replayed.
 *
 * See note for ztrace_tsk_syscall_exit for ZTRACE_BLOCK.
 * 
 * We need this here so that we don't get wierd results from checking the
 * count on return from a syscall.  (I.e., "got N, wanted 0").
 */
void ztrace_tsk_rexec_done(void) {
	struct task_struct * tsk = current;
	struct ztrace_ctl_struct * ztc;

	if(!(ztc=get_ztc(tsk)))
		return;

	if(tsk->pid != ztc->pid)
		return;

	/*
	 * If we wanted a syscall to block and it didn't, pause here so that we take
	 * the sysret logs out in the right order.
	 *
	 * There are other places to block than a syscall!  So we only want to yield
	 * here if we were supposed to block in the syscall.
	 */
	if((ztc->flags  & ZTRACE_BLOCK_STOP  )
                        && (zreg(tsk,eip)==ztc->brk.eip )
			&& ztc->brk.flags & REPLAY_SYSCALL) {
	    if(ztc->brk.flags & REPLAY_SIGBLOCK) {
	      /* FIXME -- actually, this shouldn't happen if REPLAY_SYSCALL is set... */
	      Zprintk("sysexit: pid %d NOT yielding: BLOCK, but SIGBLOCK\n", 
	      		tsk->pid, 0, 0, 0, 0);
	    } else {
              Zprintk("ztrace: pid %d yielding to pid %d after rexec due to BLOCK eip:0x%lx\n",
                        tsk->pid, tsk->revirt_tsk->sched->pid, zreg(tsk,eip),
			0, 0);
             ztc->yield=ZTRACE_XSTOP;
             ztc->brk.wchan=0;
             ztrace_yield_to(tsk->revirt_tsk->sched);
	    }
        } 
	/*
	 * OK....I think this is what we want.  In the case of a wchan, we stop
	 * where we blocked before, and block here if we miss it.  In the case of
	 * a normal xuser, we want to stop after the re-execution.
	 *
	 * I think the reason we stopped before re-execution before was to deal with
	 * a process blocking for a signal.  Now we deal with that via wchans; and the
	 * problem we're having now is sys_write() (etc) failing during replay because
	 * the signal was delivered too early.
	 *
	 * In the case where the signal really does interrupt the system call, I think
	 * we can deal with that using the syscall fail system we're going to end
	 * up setting up.
	 */
	else if ((ztc->flags & ZTRACE_BKPT_IN_KERNEL) 
			&& (zreg(tsk,eip)==ztc->brk.eip)) {
		Zprintk("ztrace: pid %d yielding after rexec due to BKPT_IN_KERNEL\n",
				tsk->pid, 0, 0, 0, 0);
		ztc->yield=ZTRACE_XSTOP;
		ztrace_yield_to(tsk->revirt_tsk->sched);
	}
}

/*
 * Called from revirt/replay.c:revirt_replay_sycall()
 * From varioius places: after / during put_user and copyuser replays.
 */
void ztrace_tsk_xuser_schedule(struct task_struct * tsk) {
	//struct task_struct * tsk = current;
	struct ztrace_ctl_struct * ztc;

	if(!(ztc=get_ztc(tsk)))
		return;

	if(tsk->pid != ztc->pid)
		return;

	if(ztc->flags & ZTRACE_XUSER_STOP) {
	    if(tsk->revirt_tsk->nrxuser==ztc->xstop.nrxuser) {
		if(ztc->xstop.edi != 0) {
			printk("ztrace_tsk_xuser_schedule: pid %d (#%lu) WARNING: edi == 0x%x! [pause]\n",
				tsk->pid, ztc->rhid, ztc->xstop.edi);
			ztrace_debug_pause();
		}
		Zprintk("ztrace_tsk_xuser_schedule: pid %d yielding due to XUSER_STOP nrxuser %lu @ 0x%x\n",
			tsk->pid, tsk->revirt_tsk->nrxuser, 
			(unsigned int)zreg(tsk,eip), 0, 0);
		ztc->yield = ZTRACE_XSTOP;
		ztrace_yield_to(tsk->revirt_tsk->sched);
	    } else {
	    	Zprintk("ztrace_tsk_xuser_schedule: pid %d xuser %lu, want %lu\n",
			tsk->pid, tsk->revirt_tsk->nrxuser, ztc->xstop.nrxuser,
			0, 0);
	    }
	}
}

/*
 * Called from arch/i386/kernel/signal.c:do_signal, after handle_signal.
 */
void ztrace_tsk_gets_signal(void) {
	struct task_struct * tsk = current;
	struct revirt_ctrdata ctr;
	struct ztrace_ctl_struct * ztc;

	if(!(ztc=get_ztc(tsk)))
		return;

	if(!(tsk->revirt_flags & REPLAY_REPLAY))
		return;

	if(tsk->pid != ztc->pid)
		return;

	/*
	 * After delivering a signal, we need to let it be handled so that
	 * zr has accurate information about what the process is going to do
	 * next.
	 *
	 * If we're single-stepping and receive a segv, the trap doesn't 
	 * happen, but the flag gets cleared; and if we're
	 * at a breakpoint and we've just received a SEGV, instruction
	 * breakpoints are temporarily disabled; so we need to check for
	 * those conditions and yield anyway.
	 */
        if(revirt_options.replayflags & REPLAY_OPT_ZTRACE_DEBUG) {
	    copy_branchdata(tsk, &ctr);
	    printk("ztrace_tsk_gets_signal: @ 0x%x %llx [%llx]\n", 
			(unsigned int)zreg(tsk,eip), ADJ_BC_R(ctr), ctr.hw_branch);
	}

	if(ztc->flags & ZTRACE_SIGNAL_STEP) {
            ztc->yield=ZTRACE_SIGNALQUEUE;

	    Zprintk("ztrace_tsk_gets_signal: yielding due to SIGNALQUEUE\n",
	    		0, 0, 0, 0, 0);
	    //ztc->flags= 0;
	    if(tsk->revirt_tsk->sched)
	     	ztrace_yield_to(tsk->revirt_tsk->sched);
	} else if(tsk->revirt_flags & REPLAY_SSTEP) {
            ztc->yield=ZTRACE_SIGNALQUEUE;
            tsk->revirt_flags &= ~REPLAY_SSTEP;

	    Zprintk("ztrace_tsk_gets_signal: yielding due to SSTEP\n",
	    	0, 0, 0, 0, 0);

            ztrace_yield_to(tsk->revirt_tsk->sched);
	} else if(ztc->flags & ZTRACE_BKPT
                        && zreg(tsk,eip)==ztc->brk.eip) {
		/* 
		 * We need to be able to deliver a signal before the process
		 * continues executing; but at this point, do_signal has
		 * already been called.  So we clear the RF flag so that
		 * the breakpoint succeeds.
		 */
		if(zreg(tsk,eflags) & X86_EFLAGS_RF) {
			zreg(tsk,eflags) &= ~X86_EFLAGS_RF;

			Zprintk("ztrace_tsk_gets_signal: clearing RF\n",
				0, 0, 0, 0, 0);

		}
        }
#if 0 // REPLAY_LOG_SIGNAL_AT_DELIVERY
	else if(ztc->flags & ZTRACE_BKPT
			&& ztc->brk.count==0) {
	    printk("ztrace_tsk_gets_signal: 0 count, yielding\n");
            ztc->yield=ZTRACE_SIGNALQUEUE;
	    if(tsk->revirt_tsk->sched)
	     	ztrace_yield_to(tsk->revirt_tsk->sched);
	}
#endif
}

/********************************************************************
 *
 * Functions called from sys_ztrace
 *
 ********************************************************************/
int ztrace_set_opt(struct task_struct * tsk, unsigned long addr, 
			unsigned long data) {
	u32 tmask;
	int err=0;
	fd_set fd_tmp;

	printk("ztrace_set_opt: %lx %lx\n", addr, data);
	if(addr & REPLAY_OPT_VALUE) {
	    printk("ztrace_set_opt: using value\n");
	    switch(addr) {
	    	case REPLAY_OPT_PACKETRATE:
		    if(data <= REPLAY_PBUF_MAX_SIZE)	
			revirt_options.xreplay.packetrate=data;
		    else
		    	err=-EINVAL;
		    break;
		case REPLAY_OPT_PACKETSIZE:
		    revirt_options.xreplay.packetsize=data;
		    break;
		case REPLAY_OPT_NETWORKFD:
		    /* FIXME -- check to make sure we're not ptraced 
		     * (i.e., only lilo can switch these, not guest kernel) */
		    revirt_options.network_fd=data;
		    break;
		case REPLAY_OPT_UNLOGGEDFDS:
		    if(copy_from_user(&fd_tmp, (fd_set *)data, sizeof(fd_set))) 
			err=-EFAULT;
		    else
		    	revirt_options.unlogged_fds=fd_tmp;
		    break;
		case REPLAY_OPT_LOGBUFSIZE:
		    /*
		     * revirt_alloc_logbuf will resize this if necessary
		     */
		    revirt_options.logbuf.size=data;
		    revirt_options.logbuf.mask=data-1;
		    printk("logbufsize: size %d mask %x\n", 
				    revirt_options.logbuf.size,
				    revirt_options.logbuf.mask);
		    revirt_alloc_log_buf(1);
		    err=0;
		    break;
		case REPLAY_OPT_LOG_SYSCALL:
		    printk("ztrace_set_opt: clearing bit for syscall %d\n", (int)data);
		    FD_CLR(data, &revirt_options.unlogged_syscalls);
		    err=0;
		    break;
		case REPLAY_OPT_UNLOG_SYSCALL:
		    printk("ztrace_set_opt: setting bit for syscall %d\n", (int)data);
		    FD_SET(data, &revirt_options.unlogged_syscalls);
		    err=0;
		    break;
		default:
		    err=-EINVAL;
	    }
	} else {
	    printk("ztrace_set_opt: using optmask\n");
    	    tmask = revirt_options.replayflags;
    	    tmask &= ~(REPLAY_OPT_REPLAYMASK & addr);
    	    tmask |= (REPLAY_OPT_REPLAYMASK & data);
    	    printk("ztrace: setting replayflags to %x\n", tmask);
    	    revirt_options.replayflags = tmask;
	}
	return err;
}

/************************
 * Data control functions
 ************************/
int ztrace_get_reg(struct task_struct *tsk,unsigned long addr,
		   unsigned long* val)
{
	long * p;
	unsigned long lval;

	if(addr > EAX)
		return -EINVAL;

	p=&(zreg(tsk,ebx));
	lval = *(p+addr);

	put_user(lval, val);
	return 0;
}

int ztrace_put_reg(struct task_struct *tsk,unsigned long addr,
			unsigned long val) 
{
	long * p;

	if(addr < FRAME_SIZE )  {
		/*
		 * Well, Linux changed the setup of the ptregs struct,
		 * but couldn't change asm/ptrace.h for backwards
		 * compatability; so they map the old junk onto the new junk.
		 * This is basically a copy of 
		 * arch/i386/kernel/ptrace.c:getreg.
		 */
		switch(addr) {
			case FS:
				tsk->thread.fs = val;
				break;
			case GS:
				tsk->thread.gs = val;
				break;
			case DS:	
			case ES:	
			case CS:	
			case SS:	
				val &= 0xffff;
				/* FALLTHRU */
			default:
				/*
				 * Don't let zr change trap flag, etc.
				 */
				if(addr==EFL) {
				    val = (REVIRT_EFLAGS_MASK & val)
					| ((~REVIRT_EFLAGS_MASK) 
					       & zreg(tsk,eflags));
				}
				if(addr > GS)
					addr -= 2;
				p=&(zreg(tsk,ebx));
				*(p+addr) = val;
		}
		return 0;
	} else 
		return -EINVAL;	
}

static int ztrace_getstate(struct task_struct *tsk,
			   struct ztrace_state_struct *state)
{
	struct revirt_tsk_s * rtsk;
	struct revirt_group_struct * rgroup;

	if((rtsk=tsk->revirt_tsk)) {
	    rgroup=rtsk->group;
	} else
	    rgroup=NULL;

	/*
	 * sys_ztrace calls this with a tmp variable, and does copy_user
	 */
	int i;
	state->regs = *(((struct pt_regs *) tsk->thread.esp0)-1);
	state->revirt_flags = tsk->revirt_flags;
	state->revirt_flags |= REPLAY_TSK_SETSTATE(tsk->state);
	/*
	 * Copy next 16 bytes of code, and previous 16 bytes of code
	 *
	 * Put 'canaries' in so we can see if the reads failed (probably
	 * due to invalid addresses).
	 */
	for (i=0;i<4;i++) {
		if(ztrace_read_long(tsk,state->regs.eip+i*4,
				 (unsigned long*) (state->code+i*4))) {
			*(unsigned long *)(state->code+i*4) = 0xdeadbeef;
		}
	}
	for (i=0;i<4;i++) {
		if(ztrace_read_long(tsk,state->regs.eip-16+i*4,
				 (unsigned long*) (state->pcode+i*4))) {
			*(unsigned long *)(state->pcode+i*4) = 0xdeadbeef;
		}
	}

	for(i=0; i<8; i++)
		state->debugreg[i]=tsk->thread.debugreg[i];
	state->cr2 = tsk->thread.cr2;
	state->trap_no = tsk->thread.trap_no;
	state->error_code = tsk->thread.error_code;
	copy_branchdata(tsk, &state->ctr);

	if(rtsk) {
	    state->nrxuser = rtsk->nrxuser;
	    state->syscall_id=rtsk->syscall_id;
	    state->sysret_cache.rh = rtsk->sysret_cache.rh;
	    state->sysret_cache.syscall_id = rtsk->sysret_cache.syscall_id;
	    if(rgroup) {
		state->zt_ctl = rgroup->zt_ctl;
		state->active = rgroup->active;
		state->replay_pos = rgroup->replay_pos;
		state->log_pos = rgroup->log_pos;
		state->log_id = rgroup->log_id;
	    } else {
		state->active.i=0;
		state->active.epochs_skipped=-1;
		state->replay_pos=0;
		state->log_pos=0;
		state->log_id=0;
	    }
	} else { 
	    state->syscall_id=0;
	    state->nrxuser=0;
	    state->sysret_cache.rh.type = 0;
	    state->sysret_cache.syscall_id = 0;
	}

	state->wchan = revirt_getwchan(tsk);
	//state->zt_flags = ztc->flags;
	//state->zt_brk = zt_rununtil;
	//state->zt_perferr = zt_perferr;
	//state->zt_pid=zt_pid;
	//state->zt_rhid=zt_rhid;
	return 0;
}

static int ztrace_peek(struct task_struct * tsk, unsigned long addr, 
				unsigned long * p) {
	unsigned long tmp;
	int copied;

	copied = access_process_vm(tsk, addr, &tmp, sizeof(tmp), 0);
	if(copied != sizeof(tmp))
		return -EIO;
	return put_user(tmp, p);
}


/*********************************************
 * Functions which block & let traced pid run
 *********************************************/

int send_signal(int, struct siginfo *, struct sigpending *);

static int ztrace_signalqueue(struct task_struct * tsk, char * data) {
	siginfo_t si;
	struct rlog_entry_signal le;
//	char * p;
	int i, ret;

	if(copy_from_user(&le, data, sizeof(le))) {
		return -EFAULT;
	}

	if(tsk->sig == NULL) {
		return -EFAULT;
	}

	spin_lock_irq(&tsk->sigmask_lock);

	for(i=0; i<_NSIG; i++) {
	    if(sigismember(&le.dfl, i+1) && 
	    		(tsk->sig->action[i].sa.sa_handler == SIG_IGN)) {
//		printk("ztrace: pid %d sig %d SIG_IGN to SIG_DFL\n", 
//				current->pid, i+1);
		tsk->sig->action[i].sa.sa_handler = SIG_DFL;
	    }
	}

	Zprintk("ztrace: changing pid %d sigmask 0x%lx %lx to 0x%lx %lx\n",
		tsk->pid, tsk->blocked.sig[0], tsk->blocked.sig[1],
		le.blocked.sig[0], le.blocked.sig[1]);

	tsk->blocked = le.blocked;

	flush_signals(tsk);

#ifdef REPLAY_WHOLE_SIGINFO
	si = le.info;
#else
	memset(&si, 0, sizeof(si));
	si.si_signo = le.si_signo;
	si.si_errno = le.si_errno;
	si.si_code = le.si_code;
	si._sifields._pad[0] = le._pad[0];
	si._sifields._pad[1] = le._pad[1];
#endif

	Zprintk("ztrace: placing signal %d in queue\n", si.si_signo,
		0, 0, 0, 0);

	if((ret=send_signal(si.si_signo, &si, &tsk->pending)))
		return ret;
	if(!sigismember(&tsk->blocked, si.si_signo) 
				&& tsk->state & TASK_INTERRUPTIBLE) 
		wake_up_process(tsk);

	recalc_sigpending(tsk);

	spin_unlock_irq(&tsk->sigmask_lock);
	tsk->revirt_flags |= REPLAY_FIRST_AFTER_SIG;

       return 0;
}

static int ztrace_sstep(struct task_struct *tsk, struct ztrace_ctl_struct * ztc, 
	int waitfl)
{
	s64 count;
	struct revirt_ctrdata ctr;

	/* Take count for a sanity check... */
	copy_branchdata(tsk, &ctr);
	count = ADJ_BC(ctr);

	/* Set SSTEP flag & trap flag, and yield */
	tsk->revirt_flags |= REPLAY_SSTEP;
	zreg(tsk,eflags) |= TRAP_FLAG;

	ztrace_yield_to(tsk);

	if((tsk->revirt_flags & REPLAY_USER_TRAP) == 0) {
	  zreg(tsk,eflags) &= ~TRAP_FLAG;
	}
	tsk->revirt_flags &= ~REPLAY_SSTEP;

	/* Sanity check */
	copy_branchdata(tsk, &ctr);
	if(ADJ_BC(ctr) - count > 2) {
		printk("ztrace: sstep more than one branch? c: %llx [pause]\n",
			count);
		ztrace_debug_pause();
	}

	if(ztc)
	    return ztc->yield;
	else
	    return 0;
}

/*
 * At this point, tsk should be about to execute an int 80.  We want
 * tsk to run until in ztrace_tsk_enters_syscall.
 */
int ztrace_syscall_enter(struct task_struct *tsk, struct ztrace_ctl_struct * ztc) {

	//ztc->flags &= ZTRACE_ACTMASK;
	ztc->flags=ZTRACE_SYSENTER_STOP;
	ztrace_yield_to(tsk);
	ztc->flags=0;

	return ztc->yield;
}

/*
 * SYSCALL_EXIT will stop after the syscall is reexecuted, before 
 * xusers are replayed.
 */
int ztrace_syscall_exit(struct task_struct *tsk, struct ztrace_ctl_struct * ztc) {

	//zt_flags &= ZTRACE_ACTMASK;
	//
	ztc->flags=ZTRACE_SYSEXIT_STOP;
	ztrace_yield_to(tsk);
	ztc->flags=0;

	return ztc->yield;
}


/*
 * Helper functions for ztrace_block and _rununtil
 */

static void ztrace_setbits_dr(struct task_struct *tsk, int reg, unsigned val) {
	if (reg<0 || reg>7 || reg==4 || reg==5)
		return ;
	tsk->thread.debugreg[reg] |= val;  
}
static void ztrace_clearbits_dr(struct task_struct *tsk, int reg, unsigned val) {
	if (reg<0 || reg>7 || reg==4 || reg==5)
		return ;
	tsk->thread.debugreg[reg] &= (~val);
}

void ztrace_rununtil_set_bkpt(struct task_struct * tsk, unsigned int eip) {
    Zprintk("ztrace_rununtil_set_bkpt: pid %d eip 0x%x\n", tsk->pid,
    		eip, 0, 0, 0);
  ztrace_setbits_dr(tsk, 7, ZTRC_DR7_L0 | ZTRC_DR7_LE |
		    ZTRC_DR7_RW0_EX | ZTRC_DR7_LEN0_0 );
  ztrace_clearbits_dr(tsk,0,0xffffffff);
  ztrace_setbits_dr(tsk, 0, eip); 
}

void ztrace_clear_bkpt(struct task_struct * tsk) {
    Zprintk("ztrace_clear_bkpt: pid %d\n", tsk->pid, 0, 0, 0, 0);
  ztrace_clearbits_dr(tsk, 7, ZTRC_DR7_L0 | ZTRC_DR7_LE |
		    ZTRC_DR7_RW0_EX | ZTRC_DR7_LEN0_0 );
    Zprintk("ztrace_clear_bkpt: tsk->thread.debugreg[7] 0x%x\n",
    		tsk->thread.debugreg[7],0,0,0,0);
}

int ztrace_sync_ztc(struct ztrace_ctl_struct * ztc) {
    struct task_struct *tsk = ztc->tsk;
    struct revirt_tsk_s * rtsk;
    int counter=revirt_pfctr.overflow.counter_index;

    if(tsk)
    	rtsk= tsk->revirt_tsk;
    else
    	rtsk=NULL;

    if(ztc->flags & (ZTRACE_BKPT|ZTRACE_BKPT_IN_KERNEL)) {
	    if(rtsk) {
	    	Zprintk("ztrace_tsk_sync: pid %d bkpt @ 0x%x\n",
				tsk->pid, ztc->brk.eip, 0, 0, 0);
		ztrace_rununtil_set_bkpt(tsk, ztc->brk.eip);
	    } else {
	    	printk("ztrace_sync_ztc: ZTRACE_BKPT, but no rtsk!\n");
	    }
    } else {
	ztrace_clear_bkpt(tsk);
    }

    if(ztc->flags & ZTRACE_PERF) {
	if(rtsk) {
	    perf_ztc = ztc;
#if 0
	    Zprintk("ztrace_sync_ztc: enabling perfoverflow. pid %d val %lld err %lld\n",
		tsk->pid, tsk->revirt_tsk->perfctr[counter].ctr, ztc->perferr);
#else
	    Zprintk("ztrace_sync_ztc: enabling perfoverflow. pid %d val %d err %ld\n",
		tsk->pid, (long)tsk->revirt_tsk->perfctr[counter].ctr, ztc->perferr,
		0, 0);
#endif
	    /*
	    * FIXME -- enable per task?
	    */
	    perfcntr_enable_overflow();
       	} else {
    	    printk("ztrace_sync_ztc: ZTRACE_PERF, but no rtsk!\n");
     	}
    } else {
	perfcntr_disable_overflow();
    }

    return 0;
}

/*
 * This needs careful examination... anything may change between in the yield, except:
 *   + data / location of ztc
 */
static int ztrace_ctl(struct task_struct *tsk, 
			   struct ztrace_ctl_struct * ztc,
			   struct ztrace_ctl_struct * ztc_user) {
			   //struct ztrace_breturn *bret) {
    int ret=0;
    //int copy_breturn=0;
    unsigned int local_wchan=0;

    if(copy_from_user(ztc, ztc_user, sizeof(*ztc)))  {
    	ret=-EFAULT;
	goto out;
    }

    ztc->tsk = tsk;
    //ztc->bret = bret;

  again:  
    tsk->revirt_flags &= ~(REPLAY_FIRST_AFTER_SIG);

    switch(ztc->flags) {
	case ZTRACE_CHECKPOINT_WAIT:
	    Zprintk("ztrace_ctl: next schedule is a final checkpoint, waiting for roll-back...\n", 0, 0, 0, 0, 0);
	    break;
    	case ZTRACE_EPOCH_STOP:
	    Zprintk("ztrace_ctl: waiting for SYNC_BLOCK, target pid %d\n", 
	    		tsk->pid, 0, 0, 0, 0);
	    break;

	case ZTRACE_BKPT_IN_KERNEL:
	    Zprintk("ztrace_ctl: funny, BKPT_IN_KERNEL, no BKPT!\n", 0, 0, 0, 
	    		0, 0);
	    /* FALLTHRU */
	case ZTRACE_BKPT:
	    /* FALLTHRU */
	case ZTRACE_BKPT|ZTRACE_BKPT_IN_KERNEL:
	    /* FIXME -- do this someplace else */
	    tsk->revirt_flags &= ~REPLAY_SSTEP;

	    /*
	     * For decode & break "single-step" over string instructions
	     */
	    if(ztc->brk.flags & REPLAY_SSTEP) {
	       tsk->revirt_flags |= REPLAY_SSTEP;
	    }

	    break;

	case ZTRACE_PERF:
	    if(tsk->revirt_tsk) {
		int counter=revirt_pfctr.overflow.counter_index;
		/*
		 * zR stores the "total branches taken since last record" in the
		 * hardware counters; we need to maintain this illusion.
		 */
		ztc->perfctr_save=tsk->revirt_tsk->perfctr[counter].ctr;

		/*
		* ztc->perferr should only be read by the replayed process, which should
		* be blocked right now; no need to put a lock around these two.
		*/
		tsk->revirt_tsk->perfctr[counter].ctr = -ztc->brk.count;
    
    		ztc->perferr = ztc->brk.count + ztc->perfctr_save;
#if 0
    		Zprintk("ztrace_sync_ztc: PERF Want %lld branches; setting counter to %llx, ztc->perferr to %llx (perfctr_save %llx)\n",
       		    ztc->brk.count, 
		    tsk->revirt_tsk->perfctr[counter].ctr, 
		    ztc->perferr, 
		    ztc->perfctr_save, 0);
#else
    		Zprintk("ztrace_sync_ztc: PERF Want %d branches; setting counter to %x, ztc->perferr to %x (perfctr_save %x)\n",
       		    ztc->brk.count, 
		    tsk->revirt_tsk->perfctr[counter].ctr, 
		    ztc->perferr, 
		    ztc->perfctr_save, 0);
#endif
	    }
	    break;
    }

    ztrace_sync_ztc(ztc);

    ztc->can_checkpoint=1;

    /*
     * FIXME we need to find a better way to do this...  But for now, this should work;
     * we just need to get RUNUNTIL and friends to have an empty active() list...
     *
     * Nothing after this yield should depend on anything local beforehand, as we expect
     * to take a checkpoint somewhere in here...
     */
    ztrace_yield_to_group(&ztc->active);

    /*
     * This is necessary if we do checkpoint/replay stuff
     */
    tsk = ztc->tsk;

    ztc->can_checkpoint=0;

    if(!tsk) {
    	Zprintk("ztrace_ctl: ztc->tsk null after return!\n", 0, 0, 0, 0, 0);
    } else {
	Zprintk("ztrace_ctl: ztc.flags %x yield %d pid @%x\n", 
		ztc->flags, ztc->yield, zreg(tsk, eip), 0, 0);
    }

    switch(ztc->flags) {
    	case ZTRACE_EPOCH_STOP:
	    if(tsk) {
		ret=ztc->yield;
	    } else {
	    	printk("ztrace_ctl: tsk diappeared [ZTRACE_EPOCH_STOP]\n");
		ret=-EINTR;
		goto out;
	    }
	    break;

	case ZTRACE_BKPT_IN_KERNEL:
	    /* FALLTHRU */
	case ZTRACE_BKPT:
	    /* FALLTHRU */
	case (ZTRACE_BKPT|ZTRACE_BKPT_IN_KERNEL):
	    if(tsk) {
		if(ztc->yield == ZTRACE_RESTORE_CHKPT) {
		    printk("WARNING: restoring in unexpected place (BKPT)\n");
		}

		if(ztc->brk.flags & REPLAY_SSTEP) {
		    tsk->revirt_flags &= ~REPLAY_SSTEP;
		}

		ztrace_clear_bkpt(tsk);
		//copy_breturn=1;
		local_wchan=revirt_getwchan(tsk);
		ret=ztc->yield;
	    } else {
	    	printk("ztrace_ctl: tsk diappeared [ZTRACE_BKPT]\n");
		ret=-EINTR;
		goto out;
	    }
	    break;

	case ZTRACE_BLOCK_STOP:
	    if(tsk) {
		if(ztc->yield == ZTRACE_BLOCK) {
		    local_wchan = revirt_getwchan(tsk);
    
    		    if(zreg(tsk, eip) != ztc->brk.eip
    			|| ztc->brk.wchan != local_wchan) {
       	
			Zprintk("ztrace_block: pid %d blocked @%x:%x, want %x:%x\n",
    			    tsk->pid, (unsigned int)zreg(tsk,eip), local_wchan,
    			    ztc->brk.eip, ztc->brk.wchan);
       			goto again;
		    }
    		    Zprintk("ztrace_block: pid %d @0x%x:0x%x, returning\n", 
       			    tsk->pid, (unsigned int)zreg(tsk, eip), local_wchan,
			    0, 0);
    
    		} else {
    		    Zprintk("ztrace_block: ztc->yield %x, returning\n",
    			    ztc->yield, 0, 0, 0, 0);
    		}

		if(ztc->yield == ZTRACE_RESTORE_CHKPT) {
		    printk("WARNING: restoring in unexpected place (BLOCK)\n");
		}

	    //copy_breturn=1;
    		ret=ztc->yield;
	    } else {
	    	printk("ztrace_ctl: tsk diappeared [ZTRACE_BLOCK_STOP]\n");
		ret=-EINTR;
		goto out;
	    }
	    break;

	case ZTRACE_PERF:
	    if(tsk) {
		if(ztc->yield == ZTRACE_RESTORE_CHKPT) {
		    printk("WARNING: restoring in unexpected place (PERF)\n");
		}
		perf_ztc=NULL;
		tsk->revirt_tsk->perfctr[revirt_pfctr.overflow.counter_index].ctr 
		    += ztc->brk.count + ztc->perfctr_save;
		ztc->perferr=0;
		//copy_breturn=1;
		ret=ztc->yield;
	    } else {
	    	printk("ztrace_ctl: tsk diappeared [ZTRACE_PERF]\n");
		ret=-EINTR;
		goto out;
	    }
	    break;
	case ZTRACE_RESTORE_CHKPT:
	    ret=ztc->yield;
	    break;

	default:
		if(ztc->yield == ZTRACE_RESTORE_CHKPT) {
		    printk("WARNING: restoring in unexpected place (%x)\n",
		    		ztc->flags);
		}
	    ret=ztc->yield;
    }

    if(ztc->bret) {
      if(tsk) {
	struct ztrace_breturn br;
	copy_branchdata(tsk, &br.ctr);
	br.eip= zreg(tsk,eip);
	br.wchan = local_wchan;
	br.epochs_skipped=tsk->revirt_tsk->group->active.epochs_skipped;
#if 0
    	Zprintk("Copying breturn: eip 0x%x wchan 0x%x count %llx\n epochs_skipped %d\n",
			br.eip, br.wchan, br.ctr.hw_branch, br.epochs_skipped,
			0);
#else
    	Zprintk("Copying breturn: eip 0x%x wchan 0x%x count %x\n epochs_skipped %d\n",
			br.eip, br.wchan, br.ctr.hw_branch, br.epochs_skipped,
			0);
#endif
	copy_to_user(ztc->bret,&br,sizeof(br));
      } else {
      	printk("ztrace_ctl: ztc->bret set, but tsk==NULL!\n");
      }
    }

    /* Random debugging stuff */
    if(ztc->flags & ZTRACE_BKPT && ztc->yield == ZTRACE_BLOCK) {
    	Zprintk("ztrace_rununtil: pid %d (#%u) blocked @0x%x:%x rununtil_flags %x\n",
	    	tsk->pid, ztc->rhid, (unsigned int)zreg(tsk, eip), 
		local_wchan, ztc->brk.flags);
    }

    out:
	ztc->flags = 0;
    	return ret;

}



DECLARE_WAIT_QUEUE_HEAD(ztrace_debug);

void ztrace_debug_pause(void) {
	if(signal_pending(current)) {
		printk("ztrace_debug_pause: current->pending.signal=={%lx %lx} [zt_yield]\n",
				current->pending.signal.sig[0], 
				current->pending.signal.sig[1]);
		if(current->pending.signal.sig[0] & 0x100) {
			printk("SIGKILL pending!  Calling do_exit...\n");
			do_exit(9);
		} else {
		    ztrace_yield_to(NULL);
		}
	} else if(in_interrupt()){
		printk("ztrace_debug_pause: Called from interrupt!!!\n");
		ztrace_yield_to(NULL);
	} else {
	    interruptible_sleep_on(&ztrace_debug);
	}
}

EXPORT_SYMBOL_NOVERS(ztrace_debug_pause);

int ztrace_debug_wake(void) {
	wake_up(&ztrace_debug);
	return 0;
}

//struct waitevent_ctl_struct waitevent_ctl = { .cmd=WAITEVENT_NONE, .pid=0 }
struct waitevent_ctl_struct waitevent_ctl = { .marker=0 };
int waitevent_enabled=0;
int waitevent_is_waiting=0;
DECLARE_MUTEX_LOCKED(waitevent_sem);
DECLARE_MUTEX_LOCKED(waitevent_ctl_sem);

int ztrace_waitevent(int cmd, int val) {
    int err=0;
    struct task_struct * tsk;

    switch(cmd) {
    	case WAITEVENT_ENABLE:
	    printk("ztrace_waitevent: setting waitevent_enabled to %d\n", val);
    	    waitevent_enabled=val;
    	    break;
	case WAITEVENT_SETMARK:
	    printk("ztrace_waitevent: setting marker to %d\n", val);
	    waitevent_ctl.marker=val;
	    break;
	case WAITEVENT_CONT:
	    if(!waitevent_is_waiting) {
	    	printk("ztrace_waitevent: WARNING, calling up() when no one is waiting!\n");
	    } 
	    up(&waitevent_sem);
	    break;
	case WAITEVENT_WAIT:
	    err=down_interruptible(&waitevent_ctl_sem);
	    break;
	case WAITEVENT_IS_WAITING:
	    err=waitevent_is_waiting;
	    break;
	default:
	    err=-EINVAL;
    }
    return err;
}


int ztrace_marker(int mark) {
    struct rlog_marker_entry me;

    do_gettimeofday(&me.tv);

    if(current->revirt_flags & REPLAY_REPLAY) {
	    struct rlog_header rh;
	    revirt_read_log(&rh);
	    if(rh.type != ENTRY_MARKER) {
		    printk("ztrace_marker: expecting type %d, got %d! [pause]\n",
		    	    ENTRY_MARKER, rh.type);
		    ztrace_debug_pause();
	    }
    } 
    if(current->revirt_flags & REPLAY_LOG
       || (revirt_options.replayflags & REPLAY_OPT_ALWAYS_WRITE_MARKER)) {
    	    //me.type = ENTRY_MARKER;
    	    me.mark=mark;
    	    revirt_write_log((char *)&me, ENTRY_MARKER, sizeof(me), NULL);
    }
    printk("ztrace_marker: pid %d, mark %d, %d.%06d sec\n",
    		    current->pid, mark, 
    		    (int)me.tv.tv_sec, 
    		    (int)me.tv.tv_usec);

    if(waitevent_enabled) {
    	if(waitevent_ctl.marker==0
		|| waitevent_ctl.marker==mark) {
	    int flags;
	    waitevent_is_waiting=1;
	    up(&waitevent_ctl_sem);
	    /* This can't be _interruptible, since UML gets signals during logging
	     * pretty regularly...*/
	    down(&waitevent_sem);
	    waitevent_is_waiting=0;
	}
    }
    return 0;
}

EXPORT_SYMBOL_NOVERS(ztrace_marker);

/*************************
 * Random testing stuff...
 *************************/
#if 0
int ztrace_copyuser(struct task_struct * tsk, unsigned int dest, 
			unsigned int len) {
    char * buf;
    unsigned int i;

    buf=kmalloc(len, GFP_KERNEL);
    if(!buf)
       return -ENOMEM;
    for(i=0; i<len; i++)
    	buf[i]=(char) i;
//    printk("ztrace_copyuser: copy_to_user(0x%x, 0x%x, 0x%x)\n",
//    		dest, (unsigned int)buf, len);
    if(copy_to_user((void *)dest, buf, len))
    	return -EFAULT;
    kfree(buf);
    return 0;
}

int ztrace_xstoptest(struct task_struct * tsk, unsigned int dest, 
		unsigned int len) {
	char * d = (char *)dest;
	int r=0;

	if(current->revirt_flags & REPLAY_LOG) {
	    r|=put_user('L', d++);
	    schedule();
	    r|=put_user('o',d++) ;
	    schedule();
	    r|=put_user('g',d++);
	    schedule();
	    r|=copy_to_user(d, "ging", 4);
	    schedule();
	    d+=4;
	    r|=copy_to_user(d, " Stuff\0", 7);
	    schedule();
	    d+=7;
	} else {
	    r|=put_user('R',d++);
	    r|=put_user('e',d++);
	    r|=put_user('p',d++);
	    r|=copy_to_user(d, "layi", 4);
	    d+=4;
	    r|=copy_to_user(d, "ng!\0\0\0\0", 7);
	    d+=7;
	}
	if(r)
	    return -EFAULT;
	else
	    return 0;
}
#endif


int ztrace_smi(void) {
#if 0
	unsigned int lo, hi;
	unsigned int ctr1, ctr2, null;
	unsigned int ctr3, ctr4;
	/*
	 *  Set up the counter to count kernel stuff
	 */
	perfctr_enable_counters();

	/*
	 * Generate an SMI -- platform specific
	 */
	asm("mov $0x884, %dx ; in %dx, %al");
	//e=inb(0x884);
	
	/*
	 * Generate an SMI - platform generic
	 */
	err=(0x2) << 8;
	printk("apic_write(APIC_ICR (0x%x), 0x%x)\n", 
			APIC_ICR, err);
	apic_write(APIC_ICR, err);
	printk("returns\n");
	err=((0x2) << 8) | (1 << 18);
	printk("apic_write(APIC_ICR (0x%x), 0x%x)\n", 
			APIC_ICR, err);
	apic_write(APIC_ICR, err);
	printk("returns\n");

	perfctr_enable_counters();
#endif
	return 0;
}

void show_stack(unsigned long *);

int ztrace_printstack(struct task_struct * tsk) {
	unsigned long * esp;
	esp=(unsigned long *)tsk->thread.esp;
	printk("Stack for pid %d @ 0x%lx:\n", tsk->pid, (unsigned long)esp);
	show_stack(esp);
	return 0;
}

int ztrace_clearbranch(struct task_struct * tsk, int all) {
    if(all) {
    	clear_branchdata_group(tsk);
#if 0
	do {
	    Zprintk("ztrace_clearbranch: clearing pid %d\n", p->pid);
	    clear_branchdata(p);
	    if(p->revirt_tsk)
		    p=p->revirt_tsk->next;
	} while(p!=tsk);
#endif
    } else {
	Zprintk("ztrace_clearbranch: clearing pid %d\n", tsk->pid, 0, 0, 0, 0);
	clear_branchdata(tsk);
    }
      /* 
       * FIXME do we need to keep the other stuff during replay?
       */
    if(tsk->revirt_tsk && tsk->revirt_tsk->group) {
	Zprintk("ztrace_clearbranch: clearing active pids\n", 0, 0, 0, 0, 0);
	clear_active_pids(&tsk->revirt_tsk->group->active);
    }

    return 0;
}

int ztrace_resume(struct task_struct *tsk) {
    int err;
    struct task_struct *p;

    if(!tsk->revirt_tsk || !tsk->revirt_tsk->group) {
	printk("tsk->pid %d tsk->revirt_tsk %x tsk->revirt_tsk->group %x\n",
		tsk->pid, 
		(unsigned int)tsk->revirt_tsk, 
		(unsigned int)tsk->revirt_tsk->group);
	err=-EPERM;
	goto out;
    }
    
    p=tsk;
    do {
	printk("ztrace_resume: resuming pid %d\n", p->pid);
	if((p->revirt_flags & REPLAY_USER_TRAP) == 0) {
	  zreg(p,eflags) &= ~TRAP_FLAG;
	}
	ztrace_clear_bkpt(p);
	p->revirt_flags &= ~REPLAY_WAIT;
	p->revirt_flags &= ~REPLAY_SSTEP;

	p->revirt_tsk->sched=NULL;

	p=p->revirt_tsk->next;
    } while(p != tsk);
    err=0;
  out:
    return err;
}


int ztrace_pause(struct task_struct *tsk) {
    int err;
    struct task_struct *p;


    if(!tsk->revirt_tsk || !tsk->revirt_tsk->group) {
	printk("tsk->pid %d tsk->revirt_tsk %x tsk->revirt_tsk->group %x\n",
		tsk->pid, 
		(unsigned int)tsk->revirt_tsk, 
		(unsigned int)tsk->revirt_tsk->group);
	err=-EPERM;
	goto out;
    }

    p=tsk;
    do {
	Zprintk("ztrace_pause: clearing revirt flags for pid %d\n", p->pid,
		0,0,0,0);

	if((p->revirt_flags & REPLAY_USER_TRAP) == 0) {
	    Zprintk(" ztrace_pause: clearing trap flag for pid %d\n", p->pid,
	    	0,0,0,0);
	    zreg(p,eflags) &= ~TRAP_FLAG;
	}

	Zprintk(" ztrace_pause: clearing breakpoitn for pid %d\n", p->pid,
		0,0,0,0);
	ztrace_clear_bkpt(p);

	if(p->revirt_flags & REPLAY_REPLAY)
	    p->revirt_flags &= ~REPLAY_WAIT|REPLAY_SSTEP;

	/* Sometimes logging & normal need REPLAY_WAIT for sections that
	 * are suppress_sched() enforced */
	p->revirt_flags &= ~(REPLAY_LOG|REPLAY_REPLAY);

	p=p->revirt_tsk->next;
    } while (p != tsk);

    /* Make this more general... */
    perfcntr_disable_overflow();

    err=0;
   out:
    return err;
}

int ztrace_attach(struct task_struct * tsk, struct revirt_active_pid_st * activep) {
    int err;
    struct task_struct * sched=current, *p;
    struct revirt_active_pid_st active;
    struct ztrace_ctl_struct * ztc;

    err=-EPERM;

    if(tsk==sched) {
    	printk("tsk == sched (%d %d)\n", tsk->pid, sched->pid);
    	goto out;
    }
    if(!tsk->revirt_tsk || !tsk->revirt_tsk->group)  {
    	printk("tsk->pid %d tsk->revirt_tsk %x; tsk->revirt_tsk->group %x\n",
		tsk->pid, 
		(unsigned int)tsk->revirt_tsk, 
		(unsigned int)(tsk->revirt_tsk)?tsk->revirt_tsk->group:NULL);
 	goto out;
    }

    if(!(tsk->revirt_flags & (REPLAY_REPLAY|REPLAY_LOG))) {
    	printk("tsk->revirt_flags %x\n", tsk->revirt_flags);
    	goto out;
    }

    ztc=get_ztc(tsk);

    /* First, attach to everyone who's running */
    p=tsk;
    do {
	if(p->revirt_tsk->sched) {
	    printk("ztrace_attach: WARNING, re-attaching to pid %d!\n", tsk->pid);
	} else {
	    printk("ztrace_attach: pid %d had no scheduler, pointing to %d\n",
	    		p->pid, sched->pid);
	}
	p->revirt_tsk->sched = sched;
	p=p->revirt_tsk->next;
	if(!p || !p->revirt_tsk) {
	    printk("WARNING, chain broken!  Returning -EPERM...\n");
	    err=-EPERM;
	    goto out;
	}
    } while(p!=tsk);
    tsk->revirt_tsk->group->sched=sched;

    /* Then, if we're logging, wait for the go-ahead to start replaying */
    if(tsk->revirt_flags & REPLAY_LOG) {
    	printk("ztrace_attach: pid %d attaching to logged process %d, blocking 'till first roll-back\n",
			sched->pid, tsk->pid);
	ztrace_yield_to(NULL);
	if(ztc->yield != ZTRACE_RESTORE_CHKPT) {
	    Zprintk("ztc->yield not ZTRACE_RESTORE_CHKPT, assuming failure...\n",
	    		0,0,0,0,0);
	    err=-EPERM;
	    goto out;
	}
    }

    /* Then find everyone who's running.  We should've attached to everyone already. */
    active.i=0;
    p=tsk;
    do {
    	if(active.i > REVIRT_MAX_ACTIVE_PIDS) {
    	    printk("active.i %d\n", active.i);
    	    return -EPERM;
	}
	printk("ztrace_attach: found pid %d\n", p->pid);
    	active.pid[active.i]=p->pid;
	/* Should no longe rhappen... */
	if(!p->revirt_tsk->sched) {
	    printk("ztrace_attach: WARNING, pid %d had no scheduler, pointing to %d\n",
	    		p->pid, sched->pid);
	    p->revirt_tsk->sched = sched;
	}
       	active.i++;
	p=p->revirt_tsk->next;
	if(!p || !p->revirt_tsk) {
	    printk("WARNING, chain broken!  Returning -EPERM...\n");
	    err=-EPERM;
	    goto out;
	}
    } while(p!=tsk);

    /* Now tell zR who's running, and return. */
    if(copy_to_user(activep, &active, sizeof(active))) {
	err=-EFAULT;
	goto out;
    }
    err=0;
   
  out:
    return err;
}

int ztrace_dma(struct task_struct * tsk) {
	tsk->revirt_flags |= REPLAY_DMA;
#if 0
	revirt_read_log(&rh);
	if(rh.type != ENTRY_DMA) {
		printk("ztrace_dma: expecting type %x, got %x!\n",
			ENTRY_DMA, rh.type);
		ztrace_debug_pause();
		return 0;
	}
	de = (struct rlog_dma_entry *)revirt_scall_buf;
	printk("ztrace_dma: copying %d bytes to %x\n", de->size, de->dest);
	access_process_vm(tsk, de->dest, revirt_scall_buf+sizeof(*de), 
				de->size, 1);
#endif			
	return 0;

}

asmlinkage int sys_ztrace(int command,int pid,unsigned long addr, 
			unsigned long data, unsigned int len)
{
	int err;
	struct task_struct *tsk=NULL;
	int waitfl;
	unsigned long val;
	struct ztrace_state_struct state;
	struct ztrace_ctl_struct * ztc=NULL;

	switch(command) {
		case ZTRACE_MARKER:
		case ZTRACE_WAITEVENT:
		case ZTRACE_SMI:
		    tsk=NULL; 
		    break;
		default:
		    err= -ESRCH;
		    tsk= find_task_by_pid(pid);
		    if (!tsk) 
			goto out;
		    ztc=get_ztc(tsk);
	}

	err= 0;
	waitfl= 0;
	/*
	 * FIXME -- we need to setup checks for some actions to make sure they
	 * don't trace themselves.
	 *
	 * I suppose we should also eventually do checks to make sure 
	 * only root can call these. =)
	 */
	switch(command) {
	  case ZTRACE_SET_OPT:
	  	err=ztrace_set_opt(tsk, addr, data);
		break;
	  /* Data control stuff */
	  case ZTRACE_GETREG:
	  	err= ztrace_get_reg(tsk,addr,&val);
	  	if (!err)
	  		err= put_user(val,(unsigned long*) data);
	  	break;
	  case ZTRACE_PUTREG:
		err=ztrace_put_reg(tsk, addr, data);
	  	break;
	  case ZTRACE_ATTACH:
	  	err=ztrace_attach(tsk, (struct revirt_active_pid_st *)addr);
#if 0
	  	if(tsk!=current && tsk->revirt_tsk && tsk->revirt_tsk->group) {
		    tsk->revirt_tsk->sched
		    	= tsk->revirt_tsk->group->sched
			= current;
		} else
		    err=-EPERM;
#endif
	  	break;
	  case ZTRACE_RESUME:
		if(addr)
		    err=ztrace_pause(tsk);
		else
		    err= ztrace_resume(tsk);
		break;

	  case ZTRACE_DETACH:
	  	if(data && tsk->revirt_tsk)
		    tsk->revirt_tsk->sched=NULL;

		if((tsk->revirt_flags & REPLAY_USER_TRAP) == 0) {
		  zreg(tsk,eflags) &= ~TRAP_FLAG;
		}
		/* FIXME: do we need to worry about the lock here? */
		tsk->revirt_flags &= ~REPLAY_WAIT;
		tsk->revirt_flags &= ~REPLAY_SSTEP;
		break;
	  case ZTRACE_GETSTATE:
	  	err= ztrace_getstate(tsk,&state);
	  	if (!err)
			copy_to_user((void *)data,&state,sizeof(state));
	  	break;
	  case ZTRACE_PEEK:
	  	err = ztrace_peek(tsk, addr, (unsigned long *)data);
		break;

	  /* Flow control */
	  case ZTRACE_SSTEP_WAIT:
	  	waitfl=1;
		/* FALLTHRU */
	  case ZTRACE_SSTEP:
	  	ztc->pid=tsk->pid;
	  	err=ztrace_sstep(tsk,ztc,waitfl);
	  	break;
	  case ZTRACE_SYSCALL_ENTER:
	  	ztc->pid=tsk->pid;
	  	err=ztrace_syscall_enter(tsk,ztc);
	  	break;
	  case ZTRACE_SYSCALL_EXIT:
	  	ztc->pid=tsk->pid;
	  	err=ztrace_syscall_exit(tsk,ztc);
	  	break;
	  case ZTRACE_CTL:
	  	if(!ztc) {
	    	    err=-EPERM;
		    break;
		}
		err=ztrace_ctl(tsk,
				ztc,
				(struct ztrace_ctl_struct *)addr);
				//(struct ztrace_breturn *)data);
		break;
#if 0
	  case ZTRACE_RUNUNTIL:
	  	ztc->pid=tsk->pid;
	  	err= ztrace_rununtil(tsk,
				(struct ztrace_rununtil_struct *)addr,
				(struct ztrace_breturn *)data);
	  	break;
	  case ZTRACE_BLOCK:
	  	ztc->pid=tsk->pid;
	  	err = ztrace_block(tsk,
				(struct ztrace_rununtil_struct *)addr,
				(struct ztrace_breturn *)data);
		break;
	  case ZTRACE_XSTOP:
	  	ztc->pid=tsk->pid;
	  	err=ztrace_xstop(tsk,(struct ztrace_xuser_stop *)addr);
		break;
	  case ZTRACE_EXIT:
	  	err=ztrace_exit(tsk,
				(struct ztrace_rununtil_struct *)addr);
		break;
	  case ZTRACE_PERFOVF:
	  	ztc->pid=tsk->pid;
	  	err=ztrace_perf(tsk, 
				(struct ztrace_rununtil_struct *)addr, 
				(struct ztrace_breturn *)data);
		break;
	  case ZTRACE_EPOCH:
	  	ztc->pid=tsk->pid;
		err=ztrace_epoch(tsk, (struct ztrace_rununtil_struct *)addr);
		break;
#endif
	  /* Process state changes */
	  case ZTRACE_WAKEUP:
		if(tsk)
		    wake_up_process(tsk);
		break;
	  case ZTRACE_SIGNALQUEUE:
	  	ztc->pid=tsk->pid;
	  	err = ztrace_signalqueue(tsk, (void *)data);
		break;
	  case ZTRACE_CLEARBRANCH:
	  	ztrace_clearbranch(tsk, addr);
		err=0;
		break;
	  case ZTRACE_DMA:
	  	err=ztrace_dma(tsk);
		break;

	  /* Debugging help */
	  case ZTRACE_DEBUGWAKE:
	  	err=ztrace_debug_wake();
		break;
	  case ZTRACE_MARKER:
	  	err=ztrace_marker(pid);
		break;
	  case ZTRACE_PRINTSTACK:
		err=ztrace_printstack(tsk);
		break;
	  /* Random testing stuff */
	  case ZTRACE_SMI:
		err=ztrace_smi();
		break;
	  case ZTRACE_WAITEVENT:
	  	err=ztrace_waitevent((int)addr, (int)data);
		break;
	  default:
	  	err= -ENOSYS;
	}

out:
	return err;
}


EXPORT_SYMBOL_NOVERS(sys_ztrace);

