#ifndef __CHECKPOINT_H__
#define __CHECKPOINT_H__

#define SPECDATA_FILE_EXT ".spd"
#define SPECBIT_FILE_EXT  ".spb"
#define SPEC_NUM_BYTES_EXT 4
#define NUM_SPEC_BLOCKS 4000

#include "myutil.h"

// STK I moved the structure defs back to the c file 
// (There were there on purpose)

typedef struct data_file_struct *data_file_t;
typedef struct undo_redo_struct *undo_redo_t;
typedef struct checkpoint_struct *checkpoint_t;
typedef struct sparse_file_struct *sparse_file_t;
typedef struct list_file_struct *list_file_t;


// checkpoint object
checkpoint_t cp_new(char *baseName, unsigned long long id);
checkpoint_t cp_open(char *baseName, unsigned long long id);
unsigned long cp_delete(checkpoint_t cp, int unlinkUndoRedoFiles); // return ref count
void cp_increment_ref_count(checkpoint_t cp);
unsigned long cp_get_ref_count(checkpoint_t cp);
undo_redo_t cp_get_undo(checkpoint_t cp);
undo_redo_t cp_get_redo(checkpoint_t cp);
checkpoint_t cp_get_next(checkpoint_t cp);
checkpoint_t cp_get_prev(checkpoint_t cp);
void cp_set_next(checkpoint_t cp, checkpoint_t next);
void cp_set_prev(checkpoint_t cp, checkpoint_t prev);
unsigned long long cp_get_id(checkpoint_t cp);
// list operations
void cp_insert(checkpoint_t prevCp, checkpoint_t newCp);
void cp_remove(checkpoint_t cp);

// undo redo file object
void ur_new(char *undoFileName, char *redoFileName, char *blockDataFileName,
	    undo_redo_t *undo, undo_redo_t *redo);
void ur_open(char *undoFileName, char *redoFileName, char *blockDataFileName,
	     undo_redo_t *undo, undo_redo_t *redo);
void ur_close(undo_redo_t ur);
void ur_close_unlink(undo_redo_t ur);
void ur_reset(undo_redo_t *ur, char *dataFileName);
void ur_reset_time(undo_redo_t ur);

unsigned long ur_get_num_rec(undo_redo_t ur);
void ur_get_rec(undo_redo_t ur, unsigned long idx, loff_t *offset, 
		unsigned long long *id, void *data);
void ur_get_rec_meta(undo_redo_t ur,unsigned long idx, loff_t *offset,
		     unsigned long long *id); 

int ur_is_block_present(undo_redo_t ur, loff_t offset);
void ur_add_block(undo_redo_t ur, loff_t offset, unsigned long long id, void *data);
void ur_update_time(undo_redo_t ur, loff_t offset, unsigned long long id);
unsigned long long ur_get_time(undo_redo_t ur, loff_t offset);

// implies efficient search, may allocate some memory for the search. this memory
// can be freed after you are done searching.  Returns the index of the
// block that holds the data.  This function assumes that the block IS in the
// list
unsigned long ur_find_block(undo_redo_t ur, loff_t offset);
void ur_get_rec_sorted(undo_redo_t ur, unsigned long idx, loff_t *offset, 
		       unsigned long long *id, void *data);
void ur_get_rec_meta_sorted(undo_redo_t ur,unsigned long idx, loff_t *offset,
			    unsigned long long *id); 
void ur_free_find_data(undo_redo_t ur);

void ur_invalidate(undo_redo_t ur, unsigned long idx);
void ur_remove_invalid(undo_redo_t ur);
//void ur_destructive_move(undo_redo_t *dst, undo_redo_t *src);

// this is for the sparse file.  It does not really belong here since we will
// have a single sparse file for each cow-disk, not for each checkpoint.
// that is ok though
sparse_file_t sf_new(char *baseName);
void sf_delete(sparse_file_t sf, struct file *diskFile);

int sf_get_fd(sparse_file_t sf);
int sf_is_block_present(sparse_file_t sf, loff_t offset);
void sf_add_blocks(sparse_file_t sf, loff_t offset, unsigned long numBytes,
		   struct file *diskFile);
void sf_flush_block(sparse_file_t sf, loff_t offset, int diskFd);
void sf_flush_all_blocks(sparse_file_t sf, struct file *diskFile);
void sf_set_file_pos(sparse_file_t sf, loff_t offset);


data_file_t df_new(char *fileName, unsigned long recordSize);
data_file_t df_open(char *fileName, unsigned long recordSize);
void df_inc_object_ref_count(data_file_t df);
void df_close(data_file_t df);
void df_close_unlink(data_file_t df);

int df_is_rec_present(data_file_t df, loff_t key);
void df_add_rec(data_file_t df, loff_t key, void *data);
void df_get_rec(data_file_t df, loff_t key, void *data);
void df_get_key_by_idx(data_file_t df, unsigned long idx, loff_t *keyOut);
void df_get_rec_by_idx(data_file_t df, unsigned long idx, void *data, loff_t *keyOut);
unsigned long df_find_rec(data_file_t df, loff_t key);
unsigned long df_get_num_rec(data_file_t df);

void df_invalidate(data_file_t df, loff_t key);
void df_remove_invalid(data_file_t df);
void df_inc_rec_ref_count(data_file_t df, loff_t key);


list_file_t lst_new(char *fileName, unsigned long recordSize, int useMemBitmap);
void lst_close(list_file_t lst);
void lst_close_unlink(list_file_t lst);

int lst_is_rec_present(list_file_t lst, loff_t key);
void lst_add_rec(list_file_t lst, loff_t key, void *data);
void lst_get_rec_by_idx(list_file_t lst, unsigned long idx, 
			void *data, loff_t *keyOut);
unsigned long lst_get_num_rec(list_file_t lst);


#endif
