#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/unistd.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/param.h>
#include <linux/fs.h>
#include <linux/slab.h>
#include <linux/file.h>

#include <asm/uaccess.h>

#include <asm/signal.h>
#include <linux/mman.h>

#include <asm/system.h>
#include <asm/segment.h>

#include <asm/semaphore.h>
#include <asm/pgalloc.h>

#include "myutil.h"

static void bring_in_page(struct mm_struct *mm, struct vm_area_struct *vma,
			  unsigned long start, unsigned long write) {
  struct task_struct *tsk = current;

  switch (handle_mm_fault(mm, vma, start, write)) {
  case 1:
    tsk->min_flt++;
    break;
  case 2:
    tsk->maj_flt++;
    break;
  case 0:
    //if (i) return i;
    //return -EFAULT;
    return;
  default:
    //if (i) return i;
    //return -ENOMEM;      
    return;
  }
}

void protect_mm_range(struct mm_struct *mm, unsigned long start,
		      unsigned long len) {
  struct vm_area_struct *vma;
  pgd_t *pgd;
  pmd_t *pmd;
  pte_t *pte;
  unsigned long address;
  unsigned long restartAddress = 0;
  struct page *ptepage;

  down_read(&mm->mmap_sem);

  for(address = start; address < (start+len); address += PAGE_SIZE) {
    vma = find_vma(mm, address);
    if((vma->vm_flags & VM_WRITE) && (vma->vm_flags & VM_SHARED) && vma->vm_file) {
    start_loop:
      pgd = pgd_offset(mm,address);
      
      if(pgd_none(*pgd)) goto restart_loop;
      if(pgd_bad(*pgd)) goto restart_loop;
      
      pmd = pmd_offset(pgd, address);
      
      if(pmd_none(*pmd)) goto restart_loop;
      if(pmd_bad(*pmd)) goto restart_loop;
      
      pte = pte_offset(pmd, address);
      
      if(pte_none(*pte)) goto restart_loop;
      if(!pte_present(*pte)) {
	printk("We have a page that is swapped out\n");
	assert(0);
      }
      
      ptepage = pte_page(*pte);
      if((!VALID_PAGE(ptepage)) || PageReserved(ptepage)) {
	printk("I am not sure what this is, but we did not wp\n");
	assert(0);
      }
      
      ptep_set_wrprotect(pte);
      if(mm == current->mm) {
	flush_tlb_page(vma,address);
      }
    }
  }

  up_read(&mm->mmap_sem);
  return;

 restart_loop:
  up_read(&mm->mmap_sem);

  if(address == restartAddress) {
    printk("could not make address 0x%08lx present, bailing\n",address);
    return;
  }
  restartAddress = address;

  down_write(&mm->mmap_sem);
  
  bring_in_page(mm, vma, address, 1);

  up_write(&mm->mmap_sem);
  down_read(&mm->mmap_sem);
  
  goto start_loop;
}

void protect_mm(struct mm_struct *mm) {
  struct vm_area_struct *vma;
  pgd_t *pgd;
  pmd_t *pmd;
  pte_t *pte;
  unsigned long address;
  unsigned long restartAddress = 0;
  struct page *ptepage;

  down_read(&mm->mmap_sem);

  for(vma = mm->mmap; vma ; vma = vma->vm_next) {
    if((vma->vm_flags & VM_WRITE) && (vma->vm_flags & VM_SHARED) && vma->vm_file) {
      for(address = vma->vm_start; address < vma->vm_end; address += PAGE_SIZE) {
      start_loop:
	pgd = pgd_offset(mm,address);
	
	if(pgd_none(*pgd)) goto restart_loop;
	if(pgd_bad(*pgd)) goto restart_loop;
	
	pmd = pmd_offset(pgd, address);
	
	if(pmd_none(*pmd)) goto restart_loop;
	if(pmd_bad(*pmd)) goto restart_loop;
	
	pte = pte_offset(pmd, address);
	
	if(pte_none(*pte)) goto restart_loop;
	if(!pte_present(*pte)) {
	  printk("We have a page that is swapped out\n");
	  assert(0);
	}
	
	ptepage = pte_page(*pte);
	if((!VALID_PAGE(ptepage)) || PageReserved(ptepage)) {
	  printk("I am not sure what this is, but we did not wp\n");
	  assert(0);
	}
	
	ptep_set_wrprotect(pte);
	if(mm == current->mm) {
	  flush_tlb_page(vma,address);
	}
      }
    }
  }

  up_read(&mm->mmap_sem);
  return;

 restart_loop:
  up_read(&mm->mmap_sem);

  if(address == restartAddress) {
    printk("could not make address 0x%08lx present, bailing\n",address);
    return;
  }
  restartAddress = address;

  down_write(&mm->mmap_sem);
  
  bring_in_page(mm, vma, address, 1);

  up_write(&mm->mmap_sem);
  down_read(&mm->mmap_sem);
  
  goto start_loop;
}

/*
// STK
// sorry about the ugly gotos, too much time reading the kernel I guess :)
// this function will make the page present if it is not already
static void *getKernelVirtualAddress(void *guestVirtualAddr) {
  pgd_t *pgd;
  pmd_t *pmd;
  pte_t *pte;
  struct mm_struct *mm;
  struct page *ptepage;
  unsigned long address = (unsigned long) guestVirtualAddr;
  unsigned long alreadyFault = 0;

  // now that we know it is present, walk the page table to get the
  // kernel virtual address of the page
  mm = current->mm;

 restart_get_kva:
  pgd = pgd_offset(mm,address);
  
  if(pgd_none(*pgd)) goto make_present_label;
  if(pgd_bad(*pgd)) goto make_present_label;
  
  pmd = pmd_offset(pgd, address);
  
  if(pmd_none(*pmd)) goto make_present_label;
  if(pmd_bad(*pmd)) goto make_present_label;
  
  pte = pte_offset(pmd, address);
  
  if(pte_none(*pte)) goto make_present_label;
  if(!pte_present(*pte)) {
    goto make_present_label;
  }
  
  ptepage = pte_page(*pte);
  if((!VALID_PAGE(ptepage)) || PageReserved(ptepage)) {
    goto make_present_label;
  }
  
  return ptepage->virtual;
  
 make_present_label:
  assert(!alreadyFault);
  alreadyFault = 1;
  address &= PAGE_MASK;
  down_write(&current->mm->mmap_sem);
  make_pages_present(address, address + PAGE_SIZE);
  up_write(&current->mm->mmap_sem);
  
  goto restart_get_kva;

}
*/




