#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/unistd.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/param.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <linux/slab.h>

#include <linux/ptrace.h>
#include <asm/ptrace.h>
#include <asm/signal.h>
#include <asm/io.h>
#include <linux/mman.h>
#include <linux/mm.h>
#include <linux/sched.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/desc.h>

#include <linux/vmalloc.h>
#include <asm/semaphore.h>

#include <asm/pgalloc.h>

#include <linux/file.h>
#include <asm/mmu_context.h>

#include <revirt/checkpoint.h>

#include "myutil.h"
#include "mem_check.h"
#include "uml.h"
#include "cow_mem.h"
#include "checkpoint.h"

extern int expand_files(struct files_struct *files, int nr);

//#define MEM_CHECK_DEBUG
//#define DISABLE_COW_MEM
//#define MEM_CHECK_DEBUG
//#define FULL_MEM_VERIFY

#define OMIT_SHARED_FILE_WRITE 1
#define INCLUDE_SHARED_FILE_WRITE 0

#define MAX_COPIED_PAGES (1024 * 1024)

#ifdef FULL_MEM_VERIFY
#warning you have full memory verification on, this will be slow
#endif

struct mem_check_struct {
  struct list_head link;
  struct gen_check_struct *check;
  unsigned long long guestSyscallCount;
  bitmap_t undo_bitmap;
  bitmap_t redo_bitmap;
  list_file_t undo_redo_data;
  struct file *vm_file;
  unsigned long checksum;
#ifdef FULL_MEM_VERIFY
  fileUtil_t fullMem;
#endif
};

static struct mem_check_struct *currentCheckpoint = NULL;
static struct list_head mem_check_list = LIST_HEAD_INIT(mem_check_list);

extern struct mm_struct *stk_copy_mm(struct mm_struct *oldmm);
//extern long sys_dup2(unsigned int oldfd,unsigned int newfd);
extern long do_mprotect(struct mm_struct *mm, unsigned long start, size_t len,
			unsigned long prot);

static fileUtil_t getTmpFile(char *dirName);
static void writeProtectGuest(struct task_struct *kernTsk);

#ifdef MEM_CHECK_DEBUG
static unsigned long calcChecksum(struct file *vm_file) {
  unsigned long checksum = 0;
  unsigned long *data;
  loff_t pos = 0;
  unsigned long idx;
  int ret;

  data = util_alloc_small(PAGE_SIZE);
  while((ret = util_kernel_read(vm_file,
				data,
				PAGE_SIZE,
				pos)) == PAGE_SIZE) {
    for(idx = 0; idx < (PAGE_SIZE / sizeof(checksum)); idx++) {
      checksum += data[idx];
    }

    pos += PAGE_SIZE;
  }

  printk("util_kernel_read ret = %d\n",ret);

  util_free_small(data);

  return checksum;
}

#endif

#ifdef FULL_MEM_VERIFY
static void myCopyFile(fileUtil_t dst, struct file *vm_file) {
  unsigned long *data;
  loff_t pos = 0;
  int ret;

  data = util_alloc_small(PAGE_SIZE);
  while((ret = util_kernel_read(vm_file,
				data,
				PAGE_SIZE,
				pos)) == PAGE_SIZE) {
    ret = util_write(dst, data, PAGE_SIZE,pos);
    assert(ret == PAGE_SIZE);

    pos += PAGE_SIZE;
  }

  util_free_small(data);
}

static void verifyMemPage(fileUtil_t fullMem, unsigned char *data, loff_t offset,
			  char *msg) {
  unsigned char *pgdata = util_alloc_small(PAGE_SIZE);
  int ret;
  unsigned long idx, i;
  unsigned long *ptr;

  if(fullMem) {
    ret = util_read(fullMem, pgdata, PAGE_SIZE, offset);
    assert(ret == PAGE_SIZE);

    for(idx = 0; idx < PAGE_SIZE; idx++) {
      if(pgdata[idx] != data[idx]) {
	printk("************* WARNING %s: page 0x%08llx 0x%02x != 0x%02x\n",
	       msg, offset+idx, pgdata[idx], data[idx]);
	for(i = idx - 52; i < (idx + 24); i += 4) {
	  ptr = (unsigned long *) (data + i);
	  if(i == idx) {
	    printk("*");
	  } else {
	    printk(" ");
	  }
	  printk("    0x%08lx\n",*ptr);
	}
      }
    }
  }
  
  util_free_small(pgdata);
}

static void  validateMem(struct mem_check_struct *curr,
			 struct mem_check_struct *prev) {
  void *data;
  loff_t pos;
  int ret;

  // make sure that any data that no modifications were found
  // actually were not modified
  if((curr->fullMem == NULL) || (prev->fullMem == NULL)) {
    return;
  }

  data = util_alloc_small(PAGE_SIZE);
  pos = 0;
  while((ret = util_read(curr->fullMem, data, PAGE_SIZE, pos)) == PAGE_SIZE) {
    if(!util_is_bit_set(prev->undo_bitmap, pos / PAGE_SIZE)) {
      verifyMemPage(prev->fullMem, data, pos, "vaildate mem");
    }
    pos += PAGE_SIZE;
  }
  util_free_small(data);
}

#endif
 
static void restoreVmaData(struct task_struct *tsk, struct gen_check_struct *check) {
  loff_t pos;
  void *data;
  unsigned long addr;
  unsigned long bytes;
  struct kern_vma_data *kvma;
  unsigned long idx;
  int ret;
  int wpFail;

  data = util_alloc_small(PAGE_SIZE);

  pos = 0;
  for(idx = 0; idx < check->vma_array_len; idx++) {		
    int newProt;
    kvma = check->vma_array + idx;

    wpFail = 0;
    /*
    if((kvma->flags & VM_WRITE) == 0) {
      if(tsk == current) {
	ret = do_mprotect(tsk->mm, 
			  (unsigned long) kvma->startAddr,
			  kvma->len,
			  PROT_READ | PROT_WRITE);
      } else {
	ret = -1;
      }
      if(ret != 0) {
	wpFail = 1;
	printk("failed to wp vma (ret == %d) starting at 0x%08lx for proc %s (%d)\n",
	       ret, (unsigned long) kvma->startAddr, tsk->comm, tsk->pid);
      }
    }
    */

    for(bytes = 0; bytes < kvma->len; bytes += PAGE_SIZE) {	
      // read the data from the file
      ret = util_read(check->vma_data_file,
		      data,			     
		      PAGE_SIZE,
		      pos);
      assert(ret == PAGE_SIZE);
      pos += ret;
      
      // copy the data
      addr = (unsigned long) kvma->startAddr;
      addr += bytes;
      //if(!wpFail) {
      //if(tsk == current) {
      //  ret = copy_to_user((void *) addr, data, PAGE_SIZE);
      //  assert(ret == 0);
      //} else {
      ret = access_process_vm(tsk, addr, data, PAGE_SIZE, 1);
      assert(ret == PAGE_SIZE);
      //}
      //}
    }

    if((kvma->flags & VM_WRITE) == 0) {
      if(!wpFail) {
	newProt = 0;
	if(kvma->flags & VM_READ) newProt |= PROT_READ;
	if(kvma->flags & VM_EXEC) newProt |= PROT_EXEC;
	ret = do_mprotect(tsk->mm,
			  (unsigned long) kvma->startAddr,
			  kvma->len,
			  newProt);
	assert(ret == 0);
      }
    }
  }

  util_free_small(data);
}

static void restoreCowMem(struct task_struct *kernTsk, 
			  struct gen_check_struct *check) {
  struct mem_check_struct *memcheck,*mc;
  unsigned long long sysCount = kernTsk->uml->guestSyscallCount;
  unsigned long long targetSysCount;
  void *data;
  loff_t key;
  bitmap_t ur_bm = NULL;
  char *copiedPages;
  int ret;
  unsigned long idx, numRec;
  unsigned long restoredPages = 0;

#ifdef DISABLE_COW_MEM
  return;
#endif

  memcheck = NULL;

  list_for_each_entry(mc, &mem_check_list, link) {
    if(mc->check == check) {
      memcheck = mc;
      break;
    }
  }

  assert(memcheck);
  assert(memcheck == mc);
  targetSysCount = memcheck->guestSyscallCount;

  //if(sysCount == targetSysCount) return;

  copiedPages = util_alloc_large(MAX_COPIED_PAGES);
  memset(copiedPages, 0, MAX_COPIED_PAGES);

  data = util_alloc_small(PAGE_SIZE);

  while(&memcheck->link != &mem_check_list) {
    if(sysCount > targetSysCount) {
      if(sysCount < memcheck->guestSyscallCount) break;

      ur_bm = memcheck->undo_bitmap;
    }

    if(sysCount < targetSysCount) {
      if(sysCount > memcheck->guestSyscallCount) break;

      ur_bm = memcheck->redo_bitmap;
    }

    if(sysCount == targetSysCount) {
      ur_bm = NULL;
    }

    numRec = lst_get_num_rec(memcheck->undo_redo_data);
    for(idx = 0; idx < numRec; idx++) {
      lst_get_rec_by_idx(memcheck->undo_redo_data, idx, NULL, &key);
      
      if(!copiedPages[key]) {
	if((ur_bm == NULL) || util_is_bit_set(ur_bm,key)) {
	  // XXX if we have multiple files, keep track per vm_file, a big mess
	  copiedPages[key] = 1;
	  restoredPages++;	  
	  
	  lst_get_rec_by_idx(memcheck->undo_redo_data, idx, data, NULL);
	  ret = util_kernel_write(memcheck->vm_file, data, PAGE_SIZE, 
				  key * PAGE_SIZE);
	  assert(ret == PAGE_SIZE);
	}
      }
    }

    if(sysCount > targetSysCount) {
      memcheck = list_entry(memcheck->link.next, struct mem_check_struct, link);
    } else if (sysCount < targetSysCount) {
      memcheck = list_entry(memcheck->link.prev, struct mem_check_struct, link);
    } else {
      assert(sysCount == targetSysCount);
      break;
    }
  }

  currentCheckpoint = mc;
  if(currentCheckpoint->link.next == &mem_check_list) {
    writeProtectGuest(kernTsk);
  }

  util_free_small(data);
  util_free_large(copiedPages);

#ifdef MEM_CHECK_DEBUG
  if(mc->checksum && mc->vm_file) {
    unsigned long checksum = calcChecksum(mc->vm_file);

    printk("verifying checksum\n");
    if(checksum != mc->checksum) {
      printk("*********** WARNING ************\n");
      printk("checksum != mc->checksum when going from %llu to %llu\n",
	     sysCount, targetSysCount);
    }
  }
#endif

}

static void restoreKernVma(struct task_struct *kernTsk, 
			   struct uml_check_struct *chk) {
  // re-install the kernel mm into the kernel process
  struct mm_struct *old_mm;
  struct mm_struct *kern_mm = stk_copy_mm(chk->uml_kern_check.mm);
  
  assert(kern_mm != NULL);
  
  old_mm = kernTsk->mm;
  kernTsk->mm = kern_mm;
  kernTsk->active_mm = kern_mm;
  if(kernTsk == current) {
    activate_mm(old_mm, kern_mm);
  }
  
  mmput(old_mm);

  restoreVmaData(kernTsk, &chk->uml_kern_check);

  restoreCowMem(kernTsk, &chk->uml_kern_check);
}

static void restoreZrVma(struct task_struct *kernTsk, 
			 struct uml_check_struct *chk) {
  struct mm_struct *zr_mm = stk_copy_mm(chk->zr_check.mm);

  // switch to the zr process mm struct    
  mmput(revirt_get_sched(kernTsk)->mm);
  revirt_get_sched(kernTsk)->mm = zr_mm;
  revirt_get_sched(kernTsk)->active_mm = zr_mm;
  *(USER_REGS(revirt_get_sched(kernTsk))) = chk->zr_check.regs;

  restoreVmaData(revirt_get_sched(kernTsk), &chk->zr_check);
}

#define INIT_VMA_ARRAY_SIZE 256

static void init_gen_check(char *dirName, struct gen_check_struct *check,
			   struct mm_struct *mm) {
  // set up the data file
  assert(check->vma_data_file == NULL);
  check->vma_data_file = getTmpFile(dirName);
  
  // now store the main kernel process mm struct
  check->mm = stk_copy_mm(mm);

  // store the locals to the check struct
  assert(check->vma_array == NULL);
  check->vma_array = util_alloc_large(sizeof(struct kern_vma_data) * 
				      INIT_VMA_ARRAY_SIZE);
  check->vma_array_len = 0;
  check->vma_array_alloc = INIT_VMA_ARRAY_SIZE;
}

static void add_vma_data(struct gen_check_struct *check, void *startAddr,
			 unsigned long len, unsigned long flags,
			 struct task_struct *tsk) {
  void *data = util_alloc_small(len);
  void *tmpArray;
  unsigned long vmaIdx = check->vma_array_len;
  int ret;

  if(vmaIdx >= check->vma_array_alloc) {
    tmpArray = util_alloc_large(sizeof(struct kern_vma_data) *
				check->vma_array_alloc * 2);
    memcpy(tmpArray, check->vma_array, check->vma_array_alloc *
	   sizeof(struct kern_vma_data));
    check->vma_array_alloc *= 2;
    util_free_large(check->vma_array);
    check->vma_array = tmpArray;
  }

  check->vma_array[vmaIdx].startAddr = startAddr;
  check->vma_array[vmaIdx].len = len;
  check->vma_array[vmaIdx].flags = flags;

  if(tsk == current) {
    ret = copy_from_user(data, startAddr, len);
    assert(ret == 0);
  } else {
    ret = access_process_vm(tsk, (unsigned long) startAddr, data, len, 0);
    assert(ret == len);
  }

  ret = util_append(check->vma_data_file,
		    data,
		    len);  
  assert(ret == len);


  check->vma_array_len++;
  util_free_small(data);
}


static void storeVma(struct task_struct *tsk, char *dirName,
		     struct gen_check_struct *check, int omitSharedFile) {
  unsigned long idx;
  //struct mm_struct *mm = current->mm;
  struct vm_area_struct *vma;
  struct kern_vma_data *kern_vma_array = NULL;
  unsigned long kern_vma_array_len = 0;
  struct mm_struct *mm = tsk->mm;
  unsigned long addr;
  unsigned long bytes;
  struct file **fileArray = NULL;

  init_gen_check(dirName,check,mm);

  // first loop and get the number of vma structs
  down_read(&mm->mmap_sem);

  kern_vma_array_len = 0;
  for(vma = mm->mmap; vma ; vma = vma->vm_next) {
    kern_vma_array_len++;
  }
  assert(kern_vma_array_len > 0);

  // alloc the array
  assert(kern_vma_array == NULL);
  kern_vma_array = util_alloc_small(sizeof(struct kern_vma_data) * 
				    kern_vma_array_len);
  fileArray = util_alloc_small(sizeof(struct file *) * 
			       kern_vma_array_len);

  // now set up the copy
  idx = 0;
  for(vma = mm->mmap; vma ; vma = vma->vm_next) {
    kern_vma_array[idx].startAddr = (void *) vma->vm_start;
    kern_vma_array[idx].len = vma->vm_end - vma->vm_start;
    kern_vma_array[idx].flags = vma->vm_flags;
    fileArray[idx] = vma->vm_file;
    idx++;
  }
  up_read(&mm->mmap_sem);

  // because of locking, do the actual copying
  for(idx = 0; idx < kern_vma_array_len; idx++) {
    for(bytes = 0; bytes < kern_vma_array[idx].len; bytes += PAGE_SIZE) {
      addr = (unsigned long) kern_vma_array[idx].startAddr;
      addr += bytes;
      if(!(omitSharedFile && fileArray[idx] && 
	   (kern_vma_array[idx].flags & VM_WRITE) &&
	   (kern_vma_array[idx].flags & VM_SHARED))) {
	add_vma_data(check, (void *) addr, PAGE_SIZE, kern_vma_array[idx].flags,tsk);
      }
    }
  }

  util_free_small(fileArray);
  util_free_small(kern_vma_array);
}

static struct mem_check_struct *alloc_mem_check_struct(char *dirName,
						       struct gen_check_struct *check,
						       unsigned long long sysCount) {
  struct mem_check_struct *memcheck;
  char *fileName;

  memcheck = util_alloc_small(sizeof(struct mem_check_struct));
  memset(memcheck, 0, sizeof(struct mem_check_struct));
  memcheck->check = check;
  memcheck->guestSyscallCount = sysCount;

  fileName = util_alloc_small(strlen(dirName) + 40);

  //sprintf(fileName,"%scow-mem.%llu.undo.bit",dirName,sysCount);
  //memcheck->undo_bitmap = util_new_bitmap(fileName);
  //util_unlink(fileName);
  memcheck->undo_bitmap = util_new_mem_bitmap();

  //sprintf(fileName,"%scow-mem.%llu.redo.bit",dirName,sysCount);
  //memcheck->redo_bitmap = util_new_bitmap(fileName);
  //util_unlink(fileName);
  memcheck->redo_bitmap = util_new_mem_bitmap();

  sprintf(fileName,"%scow-mem.%llu.data",dirName,sysCount);
  memcheck->undo_redo_data = lst_new(fileName, PAGE_SIZE, 1);

  util_free_small(fileName);

  return memcheck;
}

static void free_mem_check_struct (struct mem_check_struct *memcheck) {
  util_delete_bitmap(memcheck->undo_bitmap);
  util_delete_bitmap(memcheck->redo_bitmap);
  lst_close_unlink(memcheck->undo_redo_data);

#ifdef FULL_MEM_VERIFY
  if(memcheck->fullMem) {
    util_close(memcheck->fullMem);
  }
#endif

  if(memcheck->vm_file != NULL) {
    fput(memcheck->vm_file);
  }
  
  //memset(memcheck,0,sizeof(struct mem_check_struct));

  util_free_small(memcheck);
}

#define CREATE_UNDO 1
#define CREATE_REDO 0
static void createUndoRedo(struct mem_check_struct *memcheck, int createUndo) {
  struct mem_check_struct *mc;
  unsigned long numRec, idx;
  int ret;
  loff_t key;
  unsigned long pgoff;
  void *data;

  if(createUndo == CREATE_UNDO) {
    mc = list_entry(memcheck->link.next, struct mem_check_struct, link);
  } else {
    mc = list_entry(memcheck->link.prev, struct mem_check_struct, link);
  }  

  if(mc->vm_file) {
    // create a redo record for the new checkpoint
    data = util_alloc_small(PAGE_SIZE);
    numRec = lst_get_num_rec(mc->undo_redo_data);
    for(idx = 0; idx < numRec; idx++) {
      lst_get_rec_by_idx(mc->undo_redo_data, idx, NULL, &key);
      
      if(((createUndo == CREATE_REDO) && util_is_bit_set(mc->undo_bitmap, key)) ||
	 ((createUndo == CREATE_UNDO) && util_is_bit_set(mc->redo_bitmap, key))) {
	if(createUndo == CREATE_REDO) {
	  util_set_bit(memcheck->redo_bitmap, key);
	} else {
	  util_set_bit(memcheck->undo_bitmap, key);
	}

	if(memcheck->vm_file == NULL) {
	  get_file(mc->vm_file);
	  memcheck->vm_file = mc->vm_file;
	}

	if(!lst_is_rec_present(memcheck->undo_redo_data, key)) {
	  pgoff = (unsigned long) key;
	  ret = util_kernel_read(memcheck->vm_file, data, PAGE_SIZE, 
				 pgoff * PAGE_SIZE);
	  assert(ret == PAGE_SIZE);	
	  lst_add_rec(memcheck->undo_redo_data, key, data);

#ifdef FULL_MEM_VERIFY
	  verifyMemPage(memcheck->fullMem, data, pgoff * PAGE_SIZE,"create redo");
#endif

	}
      }
    }
    util_free_small(data);
  }
}

static void storeKernVma(struct task_struct *kernTsk, 
			 struct uml_check_struct *chk, char *dirName) {
  struct mem_check_struct *memcheck,*mc;
  int linkAdded;

  assert(kernTsk->uml);
  assert(kernTsk->uml->isKernel);

#ifdef DISABLE_COW_MEM
  storeVma(kernTsk, dirName, &chk->uml_kern_check,
  	   INCLUDE_SHARED_FILE_WRITE);
#else  
  storeVma(kernTsk, dirName, &chk->uml_kern_check,
	   OMIT_SHARED_FILE_WRITE);
#endif

  memcheck = alloc_mem_check_struct(dirName,&chk->uml_kern_check,
				    kernTsk->uml->guestSyscallCount);

  linkAdded = 0;
  list_for_each_entry(mc, &mem_check_list, link) {
    if(mc->guestSyscallCount >= memcheck->guestSyscallCount) {
      assert(mc->guestSyscallCount != memcheck->guestSyscallCount);
      list_add_tail(&memcheck->link, &mc->link);
      linkAdded = 1;
      break;
    }
  }
  
  if(!linkAdded) {
    list_add_tail(&memcheck->link, &mem_check_list);
  }

  currentCheckpoint = memcheck;

  if(memcheck->link.prev != &mem_check_list) {
    createUndoRedo(memcheck, CREATE_REDO);
  }

  if(memcheck->link.next == &mem_check_list) {
    writeProtectGuest(kernTsk);
  } else {
    createUndoRedo(memcheck, CREATE_UNDO);
  }

#ifdef MEM_CHECK_DEBUG
  if(currentCheckpoint->vm_file) {
    printk("calculating checksum\n");
    currentCheckpoint->checksum = calcChecksum(currentCheckpoint->vm_file);
  }
#endif			  

#ifdef FULL_MEM_VERIFY
  if(currentCheckpoint->vm_file) {
    if(currentCheckpoint->fullMem) {
      util_close(currentCheckpoint->fullMem);
    }
    currentCheckpoint->fullMem = getTmpFile(dirName);
    printk("copying memory to file\n");
    myCopyFile(currentCheckpoint->fullMem, currentCheckpoint->vm_file);

    if(currentCheckpoint->link.prev != &mem_check_list) {
      struct mem_check_struct *prev = list_entry(currentCheckpoint->link.prev,
						 struct mem_check_struct,
						 link);
      validateMem(currentCheckpoint, prev);
    }

  }
#endif
    
}


static void storeZrVma(struct task_struct *kernTsk, struct uml_check_struct *chk, 
		       char *dirName) {
  assert(kernTsk->uml);
  assert(kernTsk->uml->isKernel);

  storeVma(revirt_get_sched(kernTsk), dirName, &chk->zr_check,
	   INCLUDE_SHARED_FILE_WRITE);

  chk->zr_check.regs = *(USER_REGS(revirt_get_sched(kernTsk)));
}

static void my_put_unused_fd(unsigned fd, struct task_struct *tsk) {
  struct files_struct *files = tsk->files;

  write_lock(&files->file_lock);
  __put_unused_fd(files, fd);
  write_unlock(&files->file_lock);
}

static int my_get_unused_fd(struct task_struct *tsk) {
  struct files_struct * files = tsk->files;
  int fd, error;

  error = -EMFILE;
  write_lock(&files->file_lock);

 repeat:
  fd = find_next_zero_bit(files->open_fds,
			  files->max_fdset,
			  files->next_fd);

  /*                                                                             
   * N.B. For clone tasks sharing a files structure, this test                   
   * will limit the total number of files that can be opened.                    
   */
  if (fd >= tsk->rlim[RLIMIT_NOFILE].rlim_cur)
    goto out;

  /* Do we need to expand the fdset array? */
  if (fd >= files->max_fdset) {
    error = expand_fdset(files, fd);
    if (!error) {
      error = -EMFILE;
      goto repeat;
    }
    goto out;
  }

  /*                                                                             
   * Check whether we need to expand the fd array.                               
   */
  if (fd >= files->max_fds) {
    error = expand_fd_array(files, fd);
    if (!error) {
      error = -EMFILE;
      goto repeat;
    }
    goto out;
  }
  
  FD_SET(fd, files->open_fds);
  FD_CLR(fd, files->close_on_exec);
  files->next_fd = fd + 1;
#if 1
  /* Sanity check */
  if (files->fd[fd] != NULL) {
    printk(KERN_WARNING "get_unused_fd: slot %d not NULL!\n", fd);
    files->fd[fd] = NULL;
  }
#endif
  error = fd;

 out:
  write_unlock(&files->file_lock);
  return error;
}

static void my_fd_install(unsigned int fd, struct file *file, 
			  struct task_struct *tsk) {
  struct files_struct *files = tsk->files;

  write_lock(&files->file_lock);
  if (files->fd[fd])
    BUG();
  files->fd[fd] = file;
  write_unlock(&files->file_lock);
}

static int my_sys_open(char *tmp, int flags, int mode, struct task_struct *tsk) {
  int fd, error;

#if BITS_PER_LONG != 32
  flags |= O_LARGEFILE;
#endif

  fd = my_get_unused_fd(tsk);
  if (fd >= 0) {
    struct file *f = filp_open(tmp, flags, mode);
    error = PTR_ERR(f);
    if (IS_ERR(f))
      goto out_error;
    my_fd_install(fd, f, tsk);
  }
 out:
  return fd;

 out_error:
  my_put_unused_fd(fd, tsk);
  fd = error;
  goto out;
}
  
static void my_sys_close(int fd, struct task_struct *tsk) {
  struct file * filp;
  struct files_struct *files = tsk->files;

  write_lock(&files->file_lock);
  if (fd >= files->max_fds)
    goto out_unlock;
  filp = files->fd[fd];
  if (!filp)
    goto out_unlock;
  files->fd[fd] = NULL;
  FD_CLR(fd, files->close_on_exec);
  __put_unused_fd(files, fd);
  write_unlock(&files->file_lock);
  filp_close(filp, files);
  return;

 out_unlock:
  write_unlock(&files->file_lock);
}

static long my_sys_dup2(unsigned int oldfd, unsigned int newfd, 
			struct task_struct *tsk)
{
  int err = -EBADF;
  struct file * file, *tofree;
  struct files_struct * files = tsk->files;

  write_lock(&files->file_lock);
  if (!(file = fcheck_files(files, oldfd)))
    goto out_unlock;
  err = newfd;
  if (newfd == oldfd)
    goto out_unlock;
  err = -EBADF;
  if (newfd >= tsk->rlim[RLIMIT_NOFILE].rlim_cur)
    goto out_unlock;
  get_file(file);                 /* We are now finished with oldfd */

  err = expand_files(files, newfd);
  if (err < 0)
    goto out_fput;

  /* To avoid races with open() and dup(), we will mark the fd as                
   * in-use in the open-file bitmap throughout the entire dup2()                 
   * process.  This is quite safe: do_close() uses the fd array                  
   * entry, not the bitmap, to decide what work needs to be                      
   * done.  --sct */
  /* Doesn't work. open() might be there first. --AV */

  /* Yes. It's a race. In user space. Nothing sane to do */
  err = -EBUSY;
  tofree = files->fd[newfd];
  if (!tofree && FD_ISSET(newfd, files->open_fds))
    goto out_fput;

  files->fd[newfd] = file;
  FD_SET(newfd, files->open_fds);
  FD_CLR(newfd, files->close_on_exec);
  write_unlock(&files->file_lock);

  if (tofree)
    filp_close(tofree, files);
  err = newfd;
 out:
  return err;
 out_unlock:
  write_unlock(&files->file_lock);
  goto out;

 out_fput:
  write_unlock(&files->file_lock);
  fput(file);
  goto out;
}

static void restoreProcMm(struct task_struct *kernTsk, struct uml_check_struct *chk, 
			  struct task_struct *userTsk) {
  // first, free up the current proc_mm's
  struct files_struct *files = kernTsk->files;
  unsigned long fd;
  struct file *file;
  unsigned long idx;
  int foundProcMm = 1;
  int newFd;
  //mm_segment_t oldFs;
  struct mm_struct *prev_mm, *tmp_mm;
  unsigned long proc_mm_array_len;
  struct proc_mm_struct *proc_mm_array;
  unsigned long proc_mm_user_mm_fd;
  int oldFd;

  // set up local variables
  proc_mm_array_len = chk->proc_mm_array_len;
  proc_mm_array = chk->proc_mm_array;
  proc_mm_user_mm_fd = chk->proc_mm_user_mm_fd;
	
  // close up all proc mm file descriptors
  while(foundProcMm) {
    foundProcMm = 0;
    read_lock(&files->file_lock);
    for(fd = 0; fd < files->max_fds; fd++) {
      if(FD_ISSET(fd,files->open_fds)) {
	file = files->fd[fd];
	if(isProcMm(file)) {
	  foundProcMm = 1;
	  read_unlock(&files->file_lock);
	  my_sys_close(fd, kernTsk);
	  break;
	}
      }
    }
    if(!foundProcMm) {
      read_unlock(&files->file_lock);
    }
  }

  // now open up the old mm's and set them up
  // XXX FIXME I am not sure about locking here		
  for(idx = 0; idx < proc_mm_array_len; idx++) {
    //oldFs = get_fs();
    //set_fs(get_ds());
    newFd = my_sys_open("/proc/mm",O_RDWR,0666,kernTsk);
    //set_fs(oldFs);
    assert(newFd >= 0);
    // check if we need to tweak the fd
    if(newFd != proc_mm_array[idx].fd) {
      oldFd = newFd;
      newFd = my_sys_dup2(newFd, proc_mm_array[idx].fd, kernTsk);
      assert(newFd == proc_mm_array[idx].fd);
      my_sys_close(oldFd, kernTsk);
    }

    file = files->fd[proc_mm_array[idx].fd];
    mmput((struct mm_struct *) file->private_data);
    file->private_data = stk_copy_mm(proc_mm_array[idx].mm);

    if(proc_mm_array[idx].fd == proc_mm_user_mm_fd) {
      // this fd is the current mm, make sure to
      // increment so it does not get freed twice.
      // this is because switch mm of new ptrace
      // increments the mm_users, so we have to
      // becauase it is current for the user proc
      prev_mm = userTsk->mm;
      tmp_mm = (struct mm_struct *) file->private_data;
      atomic_inc(&tmp_mm->mm_users);
      userTsk->mm = tmp_mm;
      userTsk->active_mm = tmp_mm;
      mmput(prev_mm);
    }

  } 

}

unsigned long pageCount = 0;
void resetPageCount(void) {
  pageCount = 0;
}


void modifiedPage(struct task_struct *tsk, void *addr) {
  struct vm_area_struct *vma;
  unsigned long pgoff = 0;
  unsigned long address = ((unsigned long) addr);
  struct file *file = NULL;
  int ret;
  void *data;

#ifdef DISABLE_COW_MEM
  assert(0);
#endif

  if(!currentCheckpoint) return;

  assert(tsk->uml);

  vma = find_vma(tsk->mm, (unsigned long) addr);

  if(vma == NULL) return;

  if((vma->vm_file == NULL) || !(vma->vm_flags & VM_SHARED) ||
     !(vma->vm_flags & VM_WRITE)) {
    return;
  }

  file = vma->vm_file;
  pgoff = vma->vm_pgoff;
  pgoff += (address - vma->vm_start) / PAGE_SIZE;
      
  if(file && currentCheckpoint) {
    if(currentCheckpoint->vm_file == NULL) {
      get_file(file);
      currentCheckpoint->vm_file = file;
    } else {
      assert(currentCheckpoint->vm_file == file);
    }

    if(!util_is_bit_set(currentCheckpoint->undo_bitmap, pgoff)) {
      util_set_bit(currentCheckpoint->undo_bitmap, pgoff);

      if(!lst_is_rec_present(currentCheckpoint->undo_redo_data,pgoff)) {
	data = util_alloc_small(PAGE_SIZE);
	ret = util_kernel_read(currentCheckpoint->vm_file, data, 
			       PAGE_SIZE, pgoff*PAGE_SIZE);
	assert(ret == PAGE_SIZE);
	lst_add_rec(currentCheckpoint->undo_redo_data, pgoff, data);

#ifdef FULL_MEM_VERIFY
	verifyMemPage(currentCheckpoint->fullMem, data, pgoff*PAGE_SIZE, "update undo");
#endif
	util_free_small(data);
	pageCount++;
      } else {
#ifdef FULL_MEM_VERIFY
	data = util_alloc_small(PAGE_SIZE);
	ret = util_kernel_read(currentCheckpoint->vm_file, data, 
			       PAGE_SIZE, pgoff*PAGE_SIZE);
	assert(ret == PAGE_SIZE);
	verifyMemPage(currentCheckpoint->fullMem, data, pgoff*PAGE_SIZE, "already there undo");
	util_free_small(data);
#endif
      }
    }      
  }
}

void handleProcMm(struct mm_struct *mm, void *addr, unsigned long len,
		  unsigned long flags) {
  unsigned long address;
  unsigned long idx;

#ifdef DISABLE_COW_MEM
  return;
#endif

  if(currentCheckpoint != NULL) {
    // only protect if we do not have any checkpoints beyond where we are
    if((currentCheckpoint->link.next == &mem_check_list) && (flags & PROT_WRITE)) {
      if(mm == current->mm) {
	// assume sharing, copy it if there is not already a copy
	address = (unsigned long) addr;
	for(idx = 0; idx < len; idx += PAGE_SIZE) {
	  modifiedPage(current, (void *) (address + idx));
	}
      } else {
	// fault it in
	protect_mm_range(mm, (unsigned long) addr, len);		  
      }
    }
  }    
}

static void storeProcMm(struct task_struct *kernTsk, struct uml_check_struct *chk, 
			struct task_struct *userTsk) {
  struct files_struct *files = kernTsk->files;
  unsigned long fd;
  struct file *file;
  unsigned long idx;

  unsigned long proc_mm_array_len;
  struct proc_mm_struct *proc_mm_array = NULL;
  unsigned long proc_mm_user_mm_fd;


  read_lock(&files->file_lock);
	
  // first, count the proc_mm's and find the current user procMm
  assert(proc_mm_array == NULL);
  proc_mm_array_len = 0;
  proc_mm_user_mm_fd = (unsigned long) -1;
  for(fd = 0; fd < files->max_fds; fd++) {
    if(FD_ISSET(fd,files->open_fds)) {
      file = files->fd[fd];
      if(isProcMm(file)) {
	//printk("we have a proc_mm_fops\n");
	proc_mm_array_len++;
	if(userTsk->mm == 
	   (struct mm_struct *) file->private_data) {
	  assert(proc_mm_user_mm_fd == (unsigned long) -1);
	  proc_mm_user_mm_fd = fd;
					
	}
      }
    }
  }

  if(proc_mm_array_len > 0) {
    assert(proc_mm_user_mm_fd != (unsigned long) -1);
    
    // allocate the data
    proc_mm_array = (struct proc_mm_struct *) kmalloc(sizeof(struct proc_mm_struct) * proc_mm_array_len, GFP_KERNEL);
    assert(proc_mm_array != NULL);
    
    // now copy the mm data
    idx = 0;
    for(fd = 0; fd < files->max_fds; fd++) {
      if(FD_ISSET(fd,files->open_fds)) {
	file = files->fd[fd];
	if(isProcMm(file)) {
	  proc_mm_array[idx].fd = fd;
	  proc_mm_array[idx].mm = stk_copy_mm((struct mm_struct *) file->private_data);
	  idx++;
	}
      }
    }
    assert(idx == proc_mm_array_len);   
  }
  
  read_unlock(&files->file_lock);

  // set up local back up
  chk->proc_mm_array_len = proc_mm_array_len;
  chk->proc_mm_array = proc_mm_array;
  chk->proc_mm_user_mm_fd = proc_mm_user_mm_fd;
}

static void storeRegs(struct task_struct *kernTsk, struct uml_check_struct *chk, 
		      struct task_struct *userTsk) {
  chk->uml_kern_check.regs = *(USER_REGS(kernTsk));
  chk->user_regs = *(USER_REGS(userTsk));
}

static void restoreRegs(struct task_struct *kernTsk, struct uml_check_struct *chk, 
			struct task_struct *userTsk) {
  *(USER_REGS(kernTsk)) = chk->uml_kern_check.regs;
  *(USER_REGS(userTsk)) = chk->user_regs;
}

static fileUtil_t getTmpFile(char *dirName) {
  fileUtil_t file;
  unsigned long count;
  unsigned long rnd;
  char *fileName;
  unsigned long len;

  len = strlen(dirName);
  len += 15; // the size of the number and null char

  fileName = util_alloc_small(len);

  count = 0;
  do {
    assert(count < 100000);
    rdtscl(rnd);
    sprintf(fileName,"%s%lu",dirName,rnd);
    file = util_open(fileName, O_RDWR | O_LARGEFILE | O_EXCL | O_CREAT | O_TRUNC,
		     0600);
    count++;
  } while(file == NULL);

  util_unlink(fileName);

  util_free_small(fileName);

  return file;
}

static void writeProtectGuest(struct task_struct *kernTsk) {
  struct files_struct *files = kernTsk->files;
  struct file *file;
  int fd;
  extern int debugPrint;

#ifdef DISABLE_COW_MEM
  return;
#endif

  // write protect all of the current pages
  resetPageCount();
  debugPrint = 1;
  protect_mm(kernTsk->mm);
  debugPrint = 0;

  read_lock(&files->file_lock);  
  for(fd = 0; fd < files->max_fds; fd++) {
    if(FD_ISSET(fd, files->open_fds)) {
      file = files->fd[fd];
      if(isProcMm(file)) {
	protect_mm((struct mm_struct *) file->private_data);
      }
    }
  }
  read_unlock(&files->file_lock);

}

struct uml_check_struct *checkpointProcMm(struct task_struct *kernTsk, 
					  char *dirName) {
  struct uml_check_struct *chk = NULL;
	
  chk = (struct uml_check_struct *) kmalloc(sizeof(struct uml_check_struct), 
					    GFP_KERNEL);
  assert(chk != NULL);
  memset(chk,0,sizeof(struct uml_check_struct));

  storeKernVma(kernTsk, chk, dirName);
  storeProcMm(kernTsk, chk, kernTsk->uml->userList->tsk);
  storeRegs(kernTsk, chk, kernTsk->uml->userList->tsk);

  // now checkpoint revirt and zr
  chk->revirt_check_state = revirt_take_checkpoint(kernTsk);
  //assert(current->uml->chk->revirt_check_state != NULL);
  if(chk->revirt_check_state != NULL) {
    if(revirt_get_sched(kernTsk) != NULL) {
      storeZrVma(kernTsk, chk, dirName);
    }
  } else {
    printk("revirt_take_checkpoint returned null\n");
  }

  return chk;
}

int rollbackProcMm(struct task_struct *kernTsk, struct uml_check_struct *chk) {
  int ret;

  //printk("rolling back kernel vma\n");
  restoreKernVma(kernTsk, chk);
  //printk("rolling back proc mm\n");
  restoreProcMm(kernTsk, chk, kernTsk->uml->userList->tsk);
  //printk("done rolling back\n");
  restoreRegs(kernTsk, chk, kernTsk->uml->userList->tsk);

  if(chk->revirt_check_state != NULL) {
    if(chk->activeTsk == NULL) {
      ret = revirt_restore_checkpoint(kernTsk, kernTsk, chk->revirt_check_state);
      assert(ret == 0);
    } else {
      ret = revirt_restore_checkpoint(kernTsk, chk->activeTsk, 
				      chk->revirt_check_state);
      assert(ret == 0);
    }

    if(revirt_get_sched(kernTsk) != NULL) {
      restoreZrVma(kernTsk, chk);
    }
  }
  
  return 0;
}

void freeGenCheckData(struct gen_check_struct *check) {
  struct mem_check_struct *memcheck,*prev,*next;
  void *data;
  loff_t key;
  unsigned long idx, numRec;

  if(check->mm != NULL) mmput(check->mm);
  
  if(check->vma_array != NULL) {
    util_free_large(check->vma_array);
    check->vma_array = NULL;
  }
  
  if(check->vma_data_file != NULL) util_close(check->vma_data_file);
  check->vma_data_file = NULL;

  check->vma_array_len = 0;

  list_for_each_entry(memcheck, &mem_check_list, link) {
    if(memcheck->check == check) {
      // move any redo and undo data out to the previous and next checkpoints
      prev = NULL;
      next = NULL;
      if(memcheck->link.prev != &mem_check_list) {
	prev = list_entry(memcheck->link.prev, struct mem_check_struct, link);
      }
      if(memcheck->link.next != &mem_check_list) {
	next = list_entry(memcheck->link.next, struct mem_check_struct, link);
      }

      if((prev || next) && memcheck->vm_file) {
	if(prev && !prev->vm_file) {
	  get_file(memcheck->vm_file);
	  prev->vm_file = memcheck->vm_file;
	}
	if(next && !next->vm_file) {
	  get_file(memcheck->vm_file);
	  next->vm_file = memcheck->vm_file;
	}
	data = util_alloc_small(PAGE_SIZE);
	numRec = lst_get_num_rec(memcheck->undo_redo_data);
	for(idx = 0; idx < numRec; idx++) {
	  lst_get_rec_by_idx(memcheck->undo_redo_data, idx, data, &key);

	  if(prev && !util_is_bit_set(prev->undo_bitmap, key) &&
	     util_is_bit_set(memcheck->undo_bitmap, key)) {
	    util_set_bit(prev->undo_bitmap,key);
	    if(!lst_is_rec_present(prev->undo_redo_data, key)) {
	      lst_add_rec(prev->undo_redo_data, key, data);

#ifdef FULL_MEM_VERIFY
	      verifyMemPage(memcheck->fullMem, data, key*PAGE_SIZE, 
			    "merge prev curr");
	      verifyMemPage(prev->fullMem, data, key*PAGE_SIZE,"merge prev");
	      if(util_is_bit_set(memcheck->redo_bitmap, key)) {
		printk("***********prev undo redo out of sync\n");
	      }
#endif
	    }
	  }

	  if(next && !util_is_bit_set(next->redo_bitmap, key) &&
	     util_is_bit_set(memcheck->redo_bitmap, key)) {
	    util_set_bit(next->redo_bitmap,key);
	    if(!lst_is_rec_present(next->undo_redo_data, key)) {
	      lst_add_rec(next->undo_redo_data, key, data);

#ifdef FULL_MEM_VERIFY
	      verifyMemPage(memcheck->fullMem, data, key*PAGE_SIZE,
			    "merge next curr");
	      verifyMemPage(next->fullMem, data, key*PAGE_SIZE,"merge next");
	      if(util_is_bit_set(memcheck->undo_bitmap, key)) {
		printk("***********next undo redo out of sync\n");
	      }
#endif

	    }
	  }
	}
	
	util_free_small(data);
      }

      if(memcheck == currentCheckpoint) {
	if(memcheck->link.prev == &mem_check_list) {
	  currentCheckpoint = NULL;
	} else {
	  currentCheckpoint = list_entry(memcheck->link.prev, 
					 struct mem_check_struct,
					 link);
	  if(currentCheckpoint->link.next == &mem_check_list) {
	    // we are at the tail now
	    writeProtectGuest(current);
	  }
	}
      }
      list_del(&memcheck->link);
      free_mem_check_struct(memcheck);
      break;
    }
  }
}

void cowGuestSyscallCb(unsigned long long guestSyscallCount) {
  struct mem_check_struct *memcheck;

#ifdef DISABLE_COW_MEM
  return;
#endif

  memcheck = NULL;
  if(currentCheckpoint != NULL) {
    if(currentCheckpoint->link.next != &mem_check_list) {
      memcheck = list_entry(currentCheckpoint->link.next,
			    struct mem_check_struct,
			    link);
    }
  } else if(mem_check_list.next != &mem_check_list) {
    memcheck = list_entry(mem_check_list.next, struct mem_check_struct, link);
  }
  if(memcheck && (memcheck->guestSyscallCount == guestSyscallCount)) {
    currentCheckpoint = memcheck;

    // if we are at the end of the list, we must write protect the guest
    // incase we go past where we were before	
    if(currentCheckpoint->link.next == &mem_check_list) {
      writeProtectGuest(current);
    }
  }

#ifdef MEM_CHECK_DEBUG
  if(currentCheckpoint != NULL) {
    assert(currentCheckpoint->guestSyscallCount <= guestSyscallCount);

    if(currentCheckpoint->link.next != &mem_check_list) {
      memcheck = list_entry(currentCheckpoint->link.next, 
			    struct mem_check_struct, link);
      assert(memcheck->guestSyscallCount > guestSyscallCount);
    }

    if(currentCheckpoint->link.prev != &mem_check_list) {
      memcheck = list_entry(currentCheckpoint->link.prev,
			    struct mem_check_struct, link);
      assert(memcheck->guestSyscallCount < guestSyscallCount);
    }
  }
#endif

}  
