#ifndef __MYUTIL_H__
#define __MYUTIL_H__

// for file IO
typedef struct fileUtilStruct *fileUtil_t;

fileUtil_t util_open(char *fileName, unsigned long flags, unsigned long mode);
int util_read(fileUtil_t f, void *buf, unsigned long numBytes, loff_t offset);
int util_write(fileUtil_t f, void *buf, unsigned long numBytes, loff_t offset);
int util_append(fileUtil_t f, void *buf, unsigned long numBytes);
void util_close(fileUtil_t f);
void util_unlink(char *fileName);
void util_truncate(fileUtil_t f, loff_t len);

#define UTIL_SEEK_SET 0
#define UTIL_SEEK_CUR 1
#define UTIL_SEEK_END 2
#define UTIL_MAX_PATH 1024
loff_t util_seek(fileUtil_t f, loff_t offset, int origin);

int util_kernel_read(struct file *file, void *buf, unsigned long numBytes, 
		     loff_t offset);
int util_kernel_write(struct file *file, void *buf, unsigned long numBytes, 
		      loff_t offset);
extern loff_t kernel_llseek(struct file *file, loff_t offset, int origin);
extern asmlinkage long sys_unlink(const char * pathname);
extern int kernel_rename(const char *oldname, const char *newname);
extern int isProcMm(struct file *);

// for memory allocation
void *util_alloc_small(unsigned long size);
void util_free_small(void *mem);
void *util_alloc_large(unsigned long size);
void util_free_large(void *mem);

// for a large bitmap that is backed by a file.  All bits are set to 0 by
// default and only return 1 if the have been explicitly set
typedef struct bitmapStruct *bitmap_t;

bitmap_t util_new_bitmap(char *fileName);
bitmap_t util_new_mem_bitmap(void);
bitmap_t util_open_bitmap(char *fileName);
void util_delete_bitmap(bitmap_t bitmap);
int util_is_bit_set(bitmap_t bitmap, loff_t offset);
void util_set_bit(bitmap_t bitmap, loff_t offset);
void util_clear_bit(bitmap_t bitmap, loff_t offset);
//void util_clear_bit(bitmap_t bitmap, loff_t offset);

struct sort_element {
  loff_t key;
  unsigned long idx;
};
void util_heap_sort(struct sort_element *a, int array_size);
struct sort_element *util_search(struct sort_element *a, int array_size, loff_t key);

static int __assert_reentrant=0;

#define assert(val)  do { 		\
	if(__assert_reentrant) 		\
		while(1); 		\
	if(!(val)) { 			\
		__assert_reentrant=1;	\
		printk("assert error: file %s, line %d\n",__FILE__, __LINE__); \
		BUG(); 			\
	}} while(0)

//#define assert(val) 

#endif
