#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/unistd.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/param.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <linux/slab.h>

#include <linux/interrupt.h>

#include <linux/ptrace.h>
#include <asm/ptrace.h>
#include <asm/signal.h>
#include <asm/io.h>
#include <linux/mman.h>
#include <linux/mm.h>
#include <linux/sched.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/desc.h>

#include <linux/pci.h>

#include <linux/vmalloc.h>
#include <asm/semaphore.h>

#include <asm/pgalloc.h>

#include <linux/file.h>
#include <linux/poll.h>

#include <asm/irq.h>

#include <revirt/checkpoint.h>

#include "uml.h"
//#include "cow_disk.h"
#include "myutil.h"
#include "mem_check.h"
#include "cow_mem.h"
#include "dbraw.h"

static void dummy_func(void) {
  // dont do anything, we will just use the address to get an idea of 
  // where the beginning of the file is
}

// for making this a device driver
static int major;
static char name[] = "uml";
static struct file_operations op;

unsigned long platoDbgPrint = 0;

#define UML_COW_DISK_BASENAME "/checkpoint/cow-hda"
#define UML_CHECK_DATA_BASE_DIR "/checkpoint/"
#define TUNTAP_FILENAME "/dev/net/tun"

#define GDB_CHECKPOINT_INVERVAL_SEC 10

#define GDB_CHECK_OFF 0
#define GDB_CHECK_PERIODIC 1

// this is for revirt search
#define HANDLE_RW_CMD
static fileUtil_t rwFile = NULL;

//static struct uml_check_struct *umlCheckArray[MAX_UML_CHECK];
//static unsigned long checkTimeRef[] = {0,1,16,81,256,625,1296,2401,4096,6561};

static int uml_open(struct inode *inode, struct file *file);
static int uml_release(struct inode *i, struct file *f);
static ssize_t uml_write(struct file *f, const char *buf, size_t len, loff_t *p);
static ssize_t uml_read(struct file *f, char *buf, size_t len, loff_t *p);
static int uml_ioctl(struct inode *inode, struct file *f,
		     unsigned int cmd, unsigned long arg);
static unsigned int uml_poll(struct file *filp, struct poll_table_struct *wait);

typedef int (*syscall_callback_sig)(struct pt_regs *regs);
//typedef int (*signal_callback_sig)(struct pt_regs *regs, int signr, int preframe);
typedef void (*do_exit_sig)(void);
typedef void (*do_fork_sig)(struct task_struct *newProc);
typedef void (*modifiedPage_sig)(struct task_struct *, void *addr);
typedef void (*procMmCb_sig)(struct mm_struct *, void *, unsigned long,unsigned long);

extern void setDoExitCb(do_exit_sig cb);
extern void setDoForkCb(do_fork_sig cb);
extern void setSyscallCb(syscall_callback_sig sysCb, 
			 syscall_callback_sig retCb,
			 syscall_callback_sig systCb, 
			 syscall_callback_sig systExitCb);


//extern void setSignalCb(signal_callback_sig sigCb);

typedef void (*inOutCbSig)(struct pt_regs *regs, unsigned short op,
                           unsigned short port,int UNUSED);
extern int registerInOutCb(inOutCbSig cb);

extern void setModifiedPageCb(modifiedPage_sig mpCb);
extern void setProcMmCb(procMmCb_sig pCb);

static cowDiskCpCb cowDiskCb = NULL;
static int cowDiskMinor = -1;
void registerCowDiskCpCb(cowDiskCpCb cb, int minor) {
  cowDiskCb = cb;
  cowDiskMinor = minor;
}


static int takeCheckpoint = 0;
static int rollbackCheckpoint = 0;
static int restoreCheck = 0;
static int inCheckpoint = 0;
static unsigned long checkpointId = INVALID_CHECK_ID;
static unsigned long checkpointIdOut = INVALID_CHECK_ID;
static struct uml_ioc_rollback_struct savedRbs;
static int checkpointIdOutReady = 0;
static struct semaphore read_sem;
static struct semaphore write_sem;
static struct semaphore uml_sem;
static struct semaphore var_lock;

static long long numDiskBytes = 0;
static int diskFd = -1;

static int umlBlocked = 0;
static int pendingCmd = 0;
static int cmdErr=0;
static int rollbackCheckpointReady = 0;
static int readBlocked = 0;
static int writeBlocked = 0;
static int pauseUml = 0;

static gdbExitCallback gdbExitCb = NULL;

// this stuff is for guest interposition
static guest_callback_sig guestSyscallCb = NULL;
static guest_callback_sig guestRetCb = NULL;
static guest_callback_sig guestTrapCb = NULL;

extern loff_t kernel_llseek(struct file *file, loff_t offset, int origin);
extern int util_kernel_read(struct file *file, void *buf, unsigned long numBytes, 
			    loff_t offset);
extern int util_kernel_write(struct file *file, void *buf, unsigned long numBytes, 
			     loff_t offset);

static int umlSystraceCb(struct pt_regs *regs);
static int umlSystraceExitCb(struct pt_regs *regs);

// for the current checkinfo data
static int currCheckinfoSet = 0;
static struct uml_ioc_checkinfo_struct currCheckinfo;
static struct task_struct *umlTsk = NULL;

// debugging
int debugPrint = 0;

// for forwarding hardware interrupts
struct uml_irq_struct {
  int active;
  int irqCount;
  int irq;
  struct task_struct *tsk;
};

#define NUM_UML_IRQ 64
static struct uml_irq_struct umlIrqList[NUM_UML_IRQ];

//static int irqCount = 0;
//static struct task_struct *umlIrqTsk = NULL;
//static int umlIrq = -1;

struct guest_dma_struct {
  void *dmaVaddr;
  unsigned long dmaGuestVaddr;
  unsigned long dmaGuestSize;
  dma_addr_t dmaAddr;
  struct task_struct *tsk;
  struct pci_dev *dev;
  int realAlloc;
};
#define NUM_GUEST_DMA 200

//static struct guest_dma_struct guestDma[NUM_GUEST_DMA];

struct guest_iomap_struct {
  unsigned long phys_addr;
  unsigned long vaddr;
  unsigned long size;
  struct task_struct *tsk;
  int realAlloc;
};
#define NUM_GUEST_IOMAP 10
//static struct guest_iomap_struct guestIomap[NUM_GUEST_IOMAP];


struct guest_ioport_struct {
  unsigned long start;
  unsigned long end;
  struct task_struct *tsk;
};
#define NUM_GUEST_IOPORT 20
//static struct guest_ioport_struct guestIoport[NUM_GUEST_IOPORT];

struct guest_devperm_struct {
  struct guest_dma_struct guestDma[NUM_GUEST_DMA];
  struct guest_iomap_struct guestIomap[NUM_GUEST_IOMAP];
  struct guest_ioport_struct guestIoport[NUM_GUEST_IOPORT];
};

static struct guest_devperm_struct *guestDev = NULL;

static void *checkpointGuestDev(void);
static void restoreGuestDevCheckpoint(void *cp);
static void deleteGuestDevCheckpoint(void *cp);
static void free_guest_dma_vaddr(void *dmaVaddr);
static void free_guest_dma(struct guest_dma_struct *dma);
static void free_guest_iomap(struct guest_iomap_struct *iom);
static void free_guest_iomap_vaddr(unsigned long iomapVaddr);
static void cleanup_guestdev(struct guest_devperm_struct *gdp, 
			     struct task_struct *tsk);
static void register_guest_ioport(unsigned long start,
				  unsigned long end);
static void register_guest_iomap(unsigned long physAddr,
				 unsigned long vaddr,
				 unsigned long size,
				 int realAlloc);
static void register_guest_dma(struct pci_dev *dev, void *dmaVaddr, 
			       unsigned long dmaGuestVaddr, 
			       unsigned long dmaGuestSize, 
			       dma_addr_t dmaAddr,
			       int realAlloc);


static int getUserCheckInfo = 0;

void setUserCheckInfo(int isUser) {
  getUserCheckInfo = isUser;
}

void setTrapFlag(struct task_struct *tsk) {
//  printk("setting trap flag\n");
  assert(tsk->uml != NULL);
  //assert(tsk->uml->isKernel);
  tsk->uml->tfRefCount++;
  USER_REGS(tsk)->eflags |= EFLAGS_TF;
  tsk->revirt_flags |= REPLAY_USER_TRAP;
}

void clearTrapFlag(struct task_struct *tsk) {
  assert(tsk->uml->tfRefCount > 0);
  tsk->uml->tfRefCount--;
//  printk("decrementing tf ref count\n");

  if(tsk->uml->tfRefCount == 0) {
//    printk("clearning trap flag\n");
    tsk->revirt_flags &= ~REPLAY_USER_TRAP;
    if(!(tsk->revirt_flags & REPLAY_SSTEP)) {
      USER_REGS(tsk)->eflags &= ~EFLAGS_TF;
    }
  }
}

int isTrapFlagSet(struct task_struct *tsk) {
  return tsk->uml->tfRefCount != 0;
}

void syncTrapFlag(struct task_struct *tsk) {
  if(isTrapFlagSet(tsk) || (tsk->revirt_flags & REPLAY_SSTEP)) {
    USER_REGS(tsk)->eflags |= EFLAGS_TF;
  } else {
    USER_REGS(tsk)->eflags &= ~EFLAGS_TF;
  }
}

void setCurrCheckinfo(struct task_struct *kernTsk, struct task_struct *userTsk) {
  int ret;

  assert(currCheckinfoSet == 0);
  assert(kernTsk->uml && kernTsk->uml->isKernel);

  currCheckinfoSet = 1;
  
  currCheckinfo.guestSyscallCount = kernTsk->uml->guestSyscallCount;
  currCheckinfo.checkId = INVALID_CHECK_ID;
  if(userTsk != NULL) {
    ret = revirt_get_tuple(&currCheckinfo.revirtId, userTsk);
    if(ret != 0) {
      // revirt is not running
      memset(&currCheckinfo.revirtId, 0, sizeof(currCheckinfo.revirtId));
    }
    currCheckinfo.eip = USER_REGS(userTsk)->eip;
  } else {
    ret = revirt_get_tuple(&currCheckinfo.revirtId, kernTsk);
    if(ret != 0) {
      // revirt is not running
      memset(&currCheckinfo.revirtId, 0, sizeof(currCheckinfo.revirtId));
    }
    currCheckinfo.eip = USER_REGS(kernTsk)->eip;
  }

  assert(umlTsk == NULL);
  umlTsk = kernTsk;
}

void clearCurrCheckinfo(void) {
  assert(currCheckinfoSet == 1);
  currCheckinfoSet = 0;
  assert(umlTsk != NULL);
  umlTsk = NULL;
}
  
#if 0
static int cmp_lt_revirt_tuple(struct revirt_tuple *left,
			       struct revirt_tuple *right) {
  
  if(left->rhid < right->rhid) {
    return 1;
  } else if(left->rhid > right->rhid) {
    return 0;
  }

  // the rhid's are equal
  return left->bc < right->bc;
}

static int cmp_eq_revirt_tuple(struct revirt_tuple *left,
			       struct revirt_tuple *right) {
  return (left->bc == right->bc) && (left->rhid == right->rhid);
}
int atOrPastGdbRevbreakId(unsigned long guestEip) {
  struct revirt_tuple rt;
  int ret;
  ret = revirt_get_tuple(&rt, current);
  assert(ret == 0);
  
  if(cmp_lt_revirt_tuple(&rt, &current->uml->gdb_revbreak_id)) {
    return 0;
  }

  if(!cmp_eq_revirt_tuple(&rt, &current->uml->gdb_revbreak_id)) {
    // this means we are past for sure
    return 1;
  }

  return guestEip >= current->uml->gdb_revbreak_eip;
}

// it is ok if we say it is not beyond when we actually are, but if this
// returns true, it has to be true for sure
int beyondCheckpoint(struct uml_check_struct *chk) {
  struct revirt_tuple rt;
  int ret;
  ret = revirt_get_tuple(&rt, current);
  assert(ret == 0);
  
  return !cmp_lt_revirt_tuple(&rt, &current->uml->gdb_revbreak_id) &&
    !cmp_eq_revirt_tuple(&rt, &current->uml->gdb_revbreak_id);
}
#endif

static struct uml_struct *alloc_uml_struct(void) {
  struct uml_struct *uml = util_alloc_small(sizeof(struct uml_struct));
  memset(uml,0,sizeof(struct uml_struct));
  uml->checkArray = util_alloc_small(sizeof(struct uml_check_struct *) * 
				     MAX_UML_CHECK);
  memset(uml->checkArray, 0, sizeof(struct uml_check_struct *) * MAX_UML_CHECK);
  return uml;
}

static void free_uml_check_struct(struct uml_check_struct *chk) {
  unsigned long idx;
  int ret;
  
  freeGenCheckData(&chk->uml_kern_check);

  // clean up any revirt state
  if(chk->revirt_check_state != NULL) {
    freeGenCheckData(&chk->zr_check);
    ret = revirt_free_checkpoint(chk->revirt_check_state);
    assert(ret == 0);
    chk->revirt_check_state = NULL;
  }
  // now clean up the user memory
  for(idx = 0; idx < chk->proc_mm_array_len; idx++) {
    mmput(chk->proc_mm_array[idx].mm);
  }

  //deleteCheckpoint(current->uml->cowDisk, chk->cowDiskId);

  deleteGuestDevCheckpoint(chk->guestDevCp);

  if(cowDiskCb != NULL) {
    cowDiskCb(COW_DISK_DELETE_CP, chk->cowDiskCpId, cowDiskMinor);
  }

  kfree(chk->proc_mm_array);
  chk->next = NULL;
  chk->prev = NULL;
  kfree(chk);
}

void deleteUmlCheckpoint(struct task_struct *kernTsk, unsigned long idx) {
  struct task_struct *tsk = kernTsk;
  assert(idx < MAX_UML_CHECK);
  assert(tsk->uml);
  assert(tsk->uml->isKernel);
  assert(tsk->uml->checkArray[idx] != NULL);

  free_uml_check_struct(tsk->uml->checkArray[idx]);
  tsk->uml->checkArray[idx] = NULL;
}

static void free_uml_struct(struct uml_struct *uml, struct file *file) {
  unsigned long idx;
  struct gdb_mod_mem *node;

  if(uml != NULL) {
    // now free up any checkpoints
    for(idx = 0; idx < MAX_UML_CHECK; idx++) {
      if(uml->checkArray[idx] != NULL) {
	assert(uml->isKernel);
	free_uml_check_struct(uml->checkArray[idx]);
	uml->checkArray[idx] = NULL;
      }
    }
    
    util_free_small(uml->checkArray);
    uml->checkArray = NULL;

    /*
    if((file != NULL) && (uml->isKernel)) {
      closeCowDisk(uml->cowDisk, file);
    }
    */

    while(uml->gdb_mem_list != NULL) {
      node = uml->gdb_mem_list;
      uml->gdb_mem_list = uml->gdb_mem_list->next;
      util_free_small(node);
    }
    
    util_free_small(uml);
  }

}

static void addToUserList(struct task_struct *kern, struct task_struct *user) {
  struct uml_list *newNode = (struct uml_list *) kmalloc(sizeof(struct uml_list),
							 GFP_KERNEL);
  assert(newNode != NULL);

  // put at the head of the list
  newNode->tsk = user;
  newNode->next = kern->uml->userList;
  kern->uml->userList = newNode;
}

static void deleteFromUserList(struct task_struct *kern, struct task_struct *user) {
  struct uml_list **node = &kern->uml->userList, *deleteme;

  while(*node != NULL) {
    if((*node)->tsk == user) {
      deleteme=*node;
      *node = (*node)->next;
      kfree(deleteme);
      return;
    }
    node = &((*node)->next);
  }

  assert(0);
}

static void cleanupKernel(void) {
  //struct file *file;

  if(current->uml->startedExit) {
    printk("recursive exit, spinning in a while loop, use SysRq to reset\n");
    while(1)
      ;

    return;
  }

  current->uml->startedExit = 1;

  // we are assuming that all of the children exit before the kernel proc
  printk("number of disk bytes = %llu\n",numDiskBytes);
  if(current->uml->userList != NULL) {
#if 0
    printk("there are still children processes; just quit without cleaning up\n");
    return;
#else
    struct uml_list * list = current->uml->userList, *p, *q;
    struct task_struct * tsk;
    printk("there are still children processes; making them orphans...\n");
    current->uml->userList == NULL;
    for(p=list; p; ) {
	tsk=p->tsk;
	printk(" freeing uml struct for pid %d\n", tsk->pid);
	kfree(tsk->uml);
	tsk->uml=NULL;
	q=p;
	p=p->next;
	kfree(q);
     /* deleteFromUserList(kerneltsk, tsk)
        kfree(tsk->uml)
	current->uml=NULL */
    }
    
#endif
  }
  // close up the cow disks
  //file = fget(diskFd);
  //if(file == NULL) {
  //printk("warning, tring to exit and cannot open the disk!!!!!\n");
  //return;
  //}
  free_uml_struct(current->uml, NULL);
  //fput(file);
  diskFd = -1;
  
  current->uml = NULL;
}

static void umlDoExitCb(void) {
  unsigned long idx;
  if(current->uml != NULL) {    
    if(current->uml->isKernel) {
      if(gdbExitCb != NULL) {
	gdbExitCb();
      }
      cleanupKernel();

      if(rwFile != NULL) {
	util_close(rwFile);
	rwFile = NULL;
      }

    } else {
      deleteFromUserList(current->uml->kernelTsk, current);
      free_uml_struct(current->uml, NULL);
      current->uml = NULL;
    }

    printk("cleaning up the guest device allocations\n");
    cleanup_guestdev(guestDev, current);
  }

  for(idx = 0; idx < NUM_UML_IRQ; idx++) {
    if((current == umlIrqList[idx].tsk) && umlIrqList[idx].active) {
      printk("uml tsk exiting, freeing irq\n");
      free_irq(umlIrqList[idx].irq,NULL);
      umlIrqList[idx].active = 0;
    }
  }
}

static void umlDoForkCb(struct task_struct *tsk) {
  if(current->uml != NULL) {
    // we have to do a little fixup because in do_fork the new
    // process struct data is set by doing a *p = *current
    if(current->uml->isKernel) {
      tsk->uml = alloc_uml_struct();
      tsk->uml->isUser = 1;
      tsk->uml->kernelTsk = current;
      addToUserList(current, tsk);
      if(current->mm != tsk->mm) {
	//printk("***** kernel fork, not using CLONE_VM, (parent=%d, child=%d)\n",
	//     current->pid, tsk->pid);
      } else {
	//printk("kernel fork, using CLONE_VM, (parent=%d, child=%d)\n",
	//     current->pid, tsk->pid);
      }
    } else {
      assert(current->uml->isUser);
      //printk("************* user fork (parent=%d, child=%d)***************\n",
      //     current->pid, tsk->pid);
      tsk->uml = NULL;
    }
  }
}

static int isAddrModSysCall(unsigned long syscallNo) {
  return (syscallNo == __NR_mmap) ||
    (syscallNo == __NR_mmap2) ||
    (syscallNo == __NR_munmap) ||
    (syscallNo == __NR_mprotect);
}

static void handleStkTrack(struct pt_regs *regs) {
  int fd;
  mm_segment_t old_fs;

  if(revirt_options.replayflags & REPLAY_OPT_STK_NOCOW) {
    // open a fd
    old_fs = get_fs();
    set_fs(get_ds());
    fd = sys_open(UML_COW_DISK_BASENAME ".spd", O_LARGEFILE | O_RDWR | O_CREAT |O_TRUNC, 0600);
    printk("umlSysCallCb: OPT_STK_NOWCOW set, placeholder fd %d\n", fd);
    set_fs(old_fs);
    assert(fd >= 0);
    assert(current->uml == NULL);
  } else {
    assert(current->uml == NULL);
    current->uml = alloc_uml_struct();
    current->uml->isKernel = 1;
    current->uml->kernelTsk = current;
    //current->uml->cowDisk = newCowDisk(UML_COW_DISK_BASENAME);
    //printk("tracking uml at pid = %d\n",current->pid);
    numDiskBytes = 0;
  }
}

static void handleLlseek(struct pt_regs *regs) {
  // This is how we know it is a hard disk, use this info later
  // to implement kernel copy on write hard disks for uml
  //printk("seeking on file %lu\n",regs->ebx);
  if(diskFd == -1) {
    diskFd = (int) regs->ebx;
    printk("setting disk fd to %d\n",diskFd);
  } else {
    if(diskFd != regs->ebx) {
      printk("seeking on a second disk, diskFd = %d, ebx = %d, numDiskBytes = %llu changing fd\n",
	     diskFd, (int) regs->ebx, numDiskBytes);
      diskFd = regs->ebx;
    }
  }
}

static void handleWrite(struct pt_regs *regs) {
  /*
  loff_t pos;
  unsigned long bytes;
  struct file *file;
  sparse_file_t sf;

  if(regs->ebx == diskFd) {
    if((regs->edx % COW_BLOCK_SIZE) != 0) {
      //printk("disk block size != 512, is %d, not writing to log\n",(int) regs->edx);	    
    } else {	    
      assert(diskFd >= 0);
      
      if(!(revirt_options.replayflags & REPLAY_OPT_STK_NOWRITE)) {  
	file = fget(diskFd); assert(file != NULL);	  
	pos = kernel_llseek(file, 0, UTIL_SEEK_CUR);
	
	// set up the speculative file
	sf = getCowSf(current->uml->cowDisk);
	sf_set_file_pos(sf, pos);
	current->uml->hasSavedFd = 1;
	current->uml->savedFd = regs->ebx;
	regs->ebx = sf_get_fd(sf);
	sf_add_blocks(sf, pos, regs->edx, file);
	fput(file);
	
	// copy cow data and let the sparse file know what
	// blocks are about to be written
	numDiskBytes += regs->edx;
	for(bytes = 0; bytes < regs->edx; bytes += COW_BLOCK_SIZE) {
	  //printk("COWDISKWRITE %lld\n",pos + bytes);
	  writeCowDisk(current->uml->cowDisk, diskFd, pos + bytes);
	}
      }
    }
  }
  */
}

static void handleRead(struct pt_regs *regs) {
  /*
  loff_t pos;
  unsigned long bytes;
  struct file *file;
  sparse_file_t sf;

  if(regs->ebx == diskFd) {	 
    if(!(revirt_options.replayflags & REPLAY_OPT_STK_NOWRITE)) {  
      file = fget(diskFd); assert(file != NULL);	  
      pos = kernel_llseek(file, 0, UTIL_SEEK_CUR);
      fput(file);
      
      sf = getCowSf(current->uml->cowDisk);
      for(bytes = 0; bytes < regs->edx; bytes += COW_BLOCK_SIZE) {
	//printk("COWDISKREAD %lld\n",pos + bytes);
	if(sf_is_block_present(sf, pos + bytes)) {
	  // this is not the common case
	  sf_flush_block(sf, pos + bytes, diskFd);
	}
      }
    }
  }
  */
}


static int umlSysCallCb(struct pt_regs *regs) {
  char *fileName;

  if(regs->orig_eax == __NR_stk_track) {
    handleStkTrack(regs);
  }

  if(current->uml) {
    if(current->uml->isKernel) {
      if(regs->orig_eax == __NR__llseek) {

	handleLlseek(regs);

      } else if(regs->orig_eax == __NR_write) {
	
	handleWrite(regs);
	
      } else if(regs->orig_eax == __NR_read) {
	
	handleRead(regs);
	
      } else if(regs->orig_eax == __NR_close) {
	if(regs->ebx == diskFd) {
	  printk("about to close the disk file descriptor, fd = %d blocks = %llu\n", 
		 diskFd, numDiskBytes);
	  diskFd = -1;
	}
      } else if(regs->orig_eax == __NR_execve) {
	printk("*****the kernel is about to exec, it must be rebooting\n");
	cleanupKernel();
      } else if(regs->orig_eax == __NR_open) {
	fileName = getname((char *) regs->ebx);
	if(!IS_ERR(fileName)) {
	  if(strcmp(fileName,TUNTAP_FILENAME) == 0) {
	    printk("opening tuntap device... ecx = 0x%08lx\n",regs->ecx);
	  }
	  putname(fileName);
	}
      } 
#if 0
      else if((regs->orig_eax == __NR_wait4) ||
		(regs->orig_eax == __NR_waitpid)) {
	if(current->uml->userList != NULL) {
	  syncTrapFlag(current->uml->userList->tsk);
	}
      }
#endif
    }
  }
  return 0;
}

#define GDB_STATE_ORIG 0
#define GDB_STATE_NEW 1

static void restoreGdbState(struct task_struct *kernTsk, int stateType) {
  struct gdb_mod_mem *gdb_mem;
  int ret;

  // before we take this checkpoint, make sure that the gdb data is restored
  gdb_mem = kernTsk->uml->gdb_mem_list;
  while(gdb_mem != NULL) {
    if(stateType == GDB_STATE_ORIG) {
      ret = access_process_vm(kernTsk,(unsigned long) gdb_mem->addr, 
			      &gdb_mem->origVal, sizeof(gdb_mem->origVal),1);
      assert(ret == sizeof(gdb_mem->origVal));
    } else if (stateType == GDB_STATE_NEW){
      ret = access_process_vm(kernTsk, (unsigned long) gdb_mem->addr, 
			      &gdb_mem->newVal, sizeof(gdb_mem->newVal),1);
      assert(ret == sizeof(gdb_mem->newVal));
    } else {
      ret = -1;
      assert(0);
    }
    gdb_mem = gdb_mem->next;
  }

  /*
  if(stateType == GDB_STATE_ORIG) {
    if(current->revirt_flags & REPLAY_SSTEP) {
      USER_REGS(current)->eflags |= EFLAGS_TF;
    } else {
      USER_REGS(current)->eflags &= ~EFLAGS_TF;
    }
  } else if(stateType == GDB_STATE_NEW) {
    syncTrapFlag();
  }
  */
}

static unsigned long userListLen(struct uml_list *uml) {
  unsigned long count = 0;

  while(uml != NULL) {
    uml = uml->next;
    count++;
  }

  return count;
}

/* GWD */
asmlinkage int revirt_write_log(char * data, int type, int len, struct revirt_group_struct *);
int checkpoint_marker(int mark) {
    struct rlog_marker_entry me;

    /* 
     * Don't put a marker of this type in during logging, until we
     * fix revirt_read_log() to skip this type entirely.
     */
    if(!(current->revirt_flags & REPLAY_LOG)) {
	do_gettimeofday(&me.tv);
	//me.type = ENTRY_CHKPT_MARKER;
	me.mark = mark;
	revirt_write_log((char *)&me, ENTRY_CHKPT_MARKER, sizeof(me), NULL);
    }
    return 0;
}


struct uml_check_struct *checkpoint(struct task_struct *kernTsk) {
  struct uml_check_struct *chk = NULL;

  if(kernTsk->uml == NULL) {
    printk("cannot take check, uml == NULL\n"); 
    return NULL;
  }
  if(!kernTsk->uml->isKernel) {
    printk("not uml kernel, cannot take check\n");
    return NULL;
  }
  
  if(userListLen(kernTsk->uml->userList) != 1) {
    printk("cannot take check, user list len = %lu\n",
	   userListLen(kernTsk->uml->userList));
    return NULL;
  }


  if(revirt_options.replayflags & REPLAY_OPT_MARK_CHECKPOINT) {
    checkpoint_marker(1);
  }

  restoreGdbState(kernTsk, GDB_STATE_ORIG);

  if((chk = checkpointProcMm(kernTsk, UML_CHECK_DATA_BASE_DIR)) != NULL) {
    // checkpoint the guest dev info
    chk->guestDevCp = checkpointGuestDev();
    
    // checkpoint the cow disk
    if(cowDiskCb != NULL) {
      chk->cowDiskCpId = cowDiskCb(COW_DISK_TAKE_CP, 0,cowDiskMinor);
      
    }

    // store the current guest syscall count
    revirt_get_tuple(&chk->revirt_id, kernTsk);
    revirt_get_tuple(&chk->user_revirt_id,kernTsk->uml->userList->tsk);      
  }

  restoreGdbState(kernTsk, GDB_STATE_NEW);

  chk->guestSyscallCount = kernTsk->uml->guestSyscallCount;

  if(revirt_options.replayflags & REPLAY_OPT_MARK_CHECKPOINT) {
    checkpoint_marker(2);
  }
 
  
  return chk;
}

int restoreCheckpoint(struct task_struct *kernTsk, struct uml_check_struct *chk) {
  struct timeval tvStart, tvStop;
  long tvSec, tvUSec;

  do_gettimeofday(&tvStart);
  if(kernTsk->uml == NULL) {
    printk("cannot rollback, uml == NULL\n"); 
    return -1;
  }
  if(!kernTsk->uml->isKernel) {
    printk("not uml kernel, cannot rollback\n");
    return -1;
  }
  
  if(userListLen(kernTsk->uml->userList) != 1) {
    printk("cannot rollback, user list len = %lu\n",
	   userListLen(kernTsk->uml->userList));
    return -1;
  }

  if(rollbackProcMm(kernTsk, chk) != 0) {
    printk("error, memory not rolled back\n");
    return -1;
  }

  // restore the guest device state
  restoreGuestDevCheckpoint(chk->guestDevCp);

  do_gettimeofday(&tvStop);

  tvSec = tvStop.tv_sec - tvStart.tv_sec;
  tvUSec = tvStop.tv_usec - tvStart.tv_usec;

  while(tvUSec < 0) {
    tvUSec += 1000000;
    tvSec--;
  }

  while(tvUSec >= 1000000) {
    tvUSec -= 1000000;
    tvSec++;
  }

  //printk("time to restore memory checkpoint = %lu.%06lu\n",tvSec,tvUSec);

  do_gettimeofday(&tvStart);

  if(cowDiskCb != NULL) {
    cowDiskCb(COW_DISK_RESTORE_CP, chk->cowDiskCpId, cowDiskMinor);
  }

  do_gettimeofday(&tvStop);

  tvSec = tvStop.tv_sec - tvStart.tv_sec;
  tvUSec = tvStop.tv_usec - tvStart.tv_usec;

  while(tvUSec < 0) {
    tvUSec += 1000000;
    tvSec--;
  }

  while(tvUSec >= 1000000) {
    tvUSec -= 1000000;
    tvSec++;
  }

  //printk("time to restore disk checkpoint = %lu.%06lu\n",tvSec,tvUSec);

  restoreGdbState(kernTsk, GDB_STATE_NEW);
  kernTsk->uml->guestSyscallCount = chk->guestSyscallCount;

  return 0;
}

/*
static int rollback(struct uml_check_struct *chk, int restore) {
  if(restore) {
    restoreCheckpoint(chk);
  }
  free_uml_check_struct(chk);
  return 0;
}
*/

// these are from the sys_wait4 system call and the user mode includes
#define WIFSTOPPED(status)    (((status) & 0xff) == 0x7f)
#define WSTOPSIG(status)      (((status) & 0xff00) >> 8)

static int isGuestSyscall(struct pt_regs *regs) {
  if(current->uml) {
    if(current->uml->isKernel) {
      if((regs->orig_eax == __NR_wait4) || (regs->orig_eax == __NR_waitpid)) {
	// make sure that the guest user process just made a 
	// system call, not a signal
	if(current->uml->userList != NULL) {
	  if(current->uml->userList->tsk != NULL) {
	    if(regs->eax == current->uml->userList->tsk->pid) {
	     if(regs->edx & WUNTRACED) {
	      int status;
	      int ret;
	      unsigned char opcode;
	      ret = copy_from_user(&status, (void *) regs->ecx, sizeof(status));
	      if(ret) {
		printk("isGuestSyscall: pid %d: copy_from_user(&status,0x%x)== %d\n",
			current->pid,
	      		(unsigned int)regs->ecx, ret);
		printk("isGuestSyscall: %lu(0x%lx 0x%lx 0x%lx 0x%lx) == %d\n",
			regs->orig_eax,
			regs->ebx,
			regs->ecx,
			regs->edx,
			regs->esi,
			(int) regs->eax);
		return 0;
	      }
	      //assert(ret == 0);
	      if(WIFSTOPPED(status) && (WSTOPSIG(status) == SIGTRAP)) {
		// make sure this is from an int80, not an int3
		ret = copy_from_user(&opcode, (void *) (regs->eip - 1), 1);
		if(ret) {
		    printk("isGuestSyscall: pid %d: copy_from_user(&opcode, 0x%x)==%d\n",
		    	current->pid,
		    	(unsigned int)regs->eip-1, ret);
		    return 0;
		    //assert(ret == 0);
		}
		if(opcode != 0xcc) {
		  return 1;
		}
	      }
	     }
	    }
	  }
	}
      }
    }
  }
  return 0;
}

static void handleCmd(struct pt_regs *regs) {
  // this is for userland checkpoints through the device interface
  down(&var_lock);

  cmdErr=0;

  while(takeCheckpoint || rollbackCheckpoint || pauseUml) {
    if(takeCheckpoint) {
      unsigned long idx;
      unsigned long cpId = INVALID_CHECK_ID;
      unsigned long tmpId = INVALID_CHECK_ID;
      int matchingCheckpoint = 0;
      for(idx = 0; idx < MAX_UML_CHECK; idx++) {
	if((current->uml->checkArray[idx] == NULL) && (tmpId == INVALID_CHECK_ID)) {
	  tmpId = idx;
	} else if(current->uml->checkArray[idx] != NULL) {
	  if(current->uml->checkArray[idx]->guestSyscallCount ==
	     current->uml->guestSyscallCount) {
	    matchingCheckpoint = 1;
	  }
	}
      }
      
      if((tmpId != INVALID_CHECK_ID) && !matchingCheckpoint) {
	current->uml->checkArray[tmpId] = checkpoint(current);
	if(current->uml->checkArray[tmpId] != NULL) {
	  cpId = tmpId;
	}
      }

      checkpointIdOut = cpId;
      if(matchingCheckpoint) {
	printk("already checkpoint at guestSyscallCount = %llu\n",
	       current->uml->guestSyscallCount);
      } else if (cpId == INVALID_CHECK_ID) {
	printk("checkpoint not taken\n");
      }

      takeCheckpoint = 0;
      inCheckpoint = 1;
      checkpointIdOutReady = 1;
      if(readBlocked) {
	readBlocked = 0;
	up(&read_sem);
      }
    }
    if(rollbackCheckpoint) {
      unsigned long idx;
      
      // find the specific data for this checkpoint
      idx = checkpointId;
      if(idx >= MAX_UML_CHECK) {
      	printk("idx %lu > MAX_UML_CHECK, returning -EINVAL\n", idx);
	cmdErr=-EINVAL;
	goto rollback_err;
      }
      if((current->uml->checkArray[idx] == NULL)) {
      	unsigned long i;
	printk("idx %lu null in checkArray, return -EINVAL\n", idx);

	for(i=0; i<MAX_UML_CHECK; i++) {
	    printk(" id %lu %s\n", i, 
	    	(current->uml->checkArray[i]==NULL)?"invalid":"valid");
	}
	cmdErr=-EINVAL;
	goto rollback_err;
      }
      
      if(restoreCheck) {
	printk("restoring checkpoint %lu\n",checkpointId);
	restoreCheckpoint(current, current->uml->checkArray[idx]);
	current->uml->gdb_revbreak_syscall_count = 
	  savedRbs.gdb_revbreak_syscall_count;
	current->uml->gdb_step_after_syscall =
	  savedRbs.gdb_step_after_syscall;

	/*
	current->uml->gdb_revbreak_id = savedRbs.gdb_revbreak_id;
	current->uml->gdb_revbreak_eip = savedRbs.gdb_revbreak_eip;
	printk("rhid = %lu, bc = %lld, eip = 0x%08lx\n",
	       current->uml->gdb_revbreak_id.rhid,
	       current->uml->gdb_revbreak_id.bc);
	       current->uml->gdb_revbreak_eip);
	printk("gdb_revbreak_syscall_count = %llu, gdb_step_after_syscall = %lu\n",
	       current->uml->gdb_revbreak_syscall_count,
	       current->uml->gdb_step_after_syscall);
	*/
      } else {
	printk("deleting checkpoint %lu\n", checkpointId);
	free_uml_check_struct(current->uml->checkArray[idx]);	
	current->uml->checkArray[idx] = 0;
      }

     rollback_err: 
      rollbackCheckpoint = 0;
      inCheckpoint = 0;
      rollbackCheckpointReady = 1;
      if(writeBlocked) {
	writeBlocked = 0;
	up(&write_sem);
      }
    }

    if(pauseUml) {
      // just clear it
      pauseUml = 0;
    }

    /* 
     * If the command completed successfully,
     * block & wait to be restarted
     */
    if(cmdErr == 0) {
	umlBlocked = 1;
    // let user mode code get checkpoint info when the vm is blocked
	setCurrCheckinfo(current, NULL);

	up(&var_lock);
	
	/* And... wait for return */
	down(&uml_sem);
    
    // reaquire the var lock before we hit the while loop
	down(&var_lock);
	clearCurrCheckinfo();
    }
    
  }
  // out_up:
  up(&var_lock);
}

/*
static void handleSparseFileWrite(struct pt_regs *regs) {
  struct file *diskFile, *specFile;
  loff_t pos, diskPos;

  if(revirt_options.replayflags & REPLAY_OPT_STK_NOWRITE) {  
    return;
  }

  assert(regs->orig_eax == __NR_write);
  assert(regs->ebx == sf_get_fd(getCowSf(current->uml->cowDisk)));
  regs->ebx = current->uml->savedFd;
  current->uml->hasSavedFd = 0;
  
  // lets make sure that the actual number of bytes written is correct
  if(regs->eax != regs->edx) {
    printk("only wrote %lu bytes\n",regs->eax);
    assert(0);
  }
  
  // now update the file pointer for the disk
  diskFile = fget(diskFd); assert(diskFile != NULL);
  specFile = fget(sf_get_fd(getCowSf(current->uml->cowDisk))); 
  assert(specFile != NULL);
  
  pos = kernel_llseek(specFile, 0, UTIL_SEEK_CUR);
  diskPos = kernel_llseek(diskFile, 0, UTIL_SEEK_CUR);
  assert((diskPos + regs->eax) == pos);
  kernel_llseek(diskFile, pos, UTIL_SEEK_SET);
  
  fput(diskFile);
  fput(specFile);
}
*/

static int umlSysCallRetCb(struct pt_regs *regs) {
  struct task_struct * tsk=current;
  struct task_struct *gdbTsk;
  int saved_flags;

  if(isGuestSyscall(regs)) {
    if(current->uml->guestSyscallStarted) {
      current->uml->guestSyscallStarted = 0;

      current->uml->guestSyscallCount++;

      if(isTrapFlagSet(current) && !(regs->eflags & EFLAGS_TF)) {
	printk("umlSysCallRetCb: resetting the trap flag\n");
	regs->eflags |= EFLAGS_TF;
      }

      if(current->uml->gdb_step_after_syscall && 
	 (current->uml->guestSyscallCount == 
	  current->uml->gdb_revbreak_syscall_count) ) {
	printk("umlSysCallRetCb: revbreak syscall (%llu) count reached, stepping\n",
	       current->uml->guestSyscallCount);
	if(getUserCheckInfo) {
	  assert(userListLen(current->uml->userList) == 1);
	  gdbTsk = current->uml->userList->tsk;
	} else {
	  gdbTsk = current;
	}
	if(!gdbTsk->uml->gdb_step_set) {
	  gdbTsk->uml->gdb_step_set = 1;
	  setTrapFlag(gdbTsk);
	}
      }

      cowGuestSyscallCb(current->uml->guestSyscallCount);

      revirt_suppress_sched(tsk, saved_flags);
      handleCmd(regs);
      revirt_restore_sched(tsk, saved_flags);
    }
    
  } else if(isAddrModSysCall(regs->orig_eax)) {
    if(current->uml && current->uml->isKernel) {
      if(debugPrint) {
	printk("guest kernel addr modification\n");
      }
    }
  }

  return 0;
}

static int isValidPort(unsigned short port) {
  unsigned long idx;
  for(idx = 0; idx < NUM_GUEST_IOPORT; idx++) {
    if((port >= guestDev->guestIoport[idx].start) && 
       (port <= guestDev->guestIoport[idx].end)) {
      return 1;
    }
  }

  return 0;
}

#define IS_IN(op) ( (((op) & 0xff) == 0xec) || (((op) & 0xff) == 0xed) || ((op) == 0xed66) )
#define IS_OUT(op) ( (((op) & 0xff) == 0xee) || (((op) & 0xff) == 0xef) || ((op) == 0xef66) )

static void inOutCb(struct pt_regs *regs, unsigned short op, 
		    unsigned short port, int UNUSED) {
  unsigned long val = 0;
  // emulate the in/out inst
  if(!isValidPort(port)) {    
    printk("failed in/out at port 0x%04x, eip = 0x%08lx\n",port,regs->eip);
  } else {
    // check the opcodes
    if(IS_IN(op)) {
      if(current->revirt_flags & REPLAY_REPLAY) {
	val = revirt_replay_in(port, regs->eip);
      } else {
	if((op & 0xff) == 0xec) {
	  val = inb(port);
	} else if(op == 0xed66) {
	  val = inw(port);
	} else if((op & 0xff) == 0xed) {
	  val = inl(port);
	} else {
	  assert(0);
	}
      }
      regs->eax = val;
	
      if(current->revirt_flags & REPLAY_LOG) {
	revirt_log_in(val, port, regs->eip);
      }
    } else {
      assert(IS_OUT(op));
      if(!(current->revirt_flags & REPLAY_REPLAY)) {
	if((op & 0xff) == 0xee) {
	  outb(regs->eax & 0xff, port);
	} else if(op == 0xef66) {
	  outw(regs->eax & 0xffff, port);
	} else if((op & 0xff) == 0xef) {
	  outl(regs->eax, port);
	} else {
	  assert(0);
	}
      }
    }
  }

  if((op & 0xff) == 0x66) {
    regs->eip += 2;
  } else {
    regs->eip += 1;
  }    

}

static void uml_irq_bh(void) {

}

static void uml_irq_callback(int irq, void *dev_id, struct pt_regs *regs) {
  assert((irq >= 0) &&  (irq < NUM_UML_IRQ));
  if(umlIrqList[irq].active) {
    umlIrqList[irq].irqCount++;
    disable_irq(irq);
    send_sig(SIGIO, umlIrqList[irq].tsk,1);
  }
}

/**
 * linux module stuff
 */
int init_module(void) 
{
  unsigned long idx;

  sema_init(&read_sem,0);
  sema_init(&uml_sem,0);
  sema_init(&write_sem,0);
  sema_init(&var_lock,1);
  

  guestDev = (struct guest_devperm_struct *) 
    kmalloc(sizeof(struct guest_devperm_struct), GFP_KERNEL);
  assert(guestDev != NULL);

  memset(guestDev, 0, sizeof(struct guest_devperm_struct));
  
  /* Initialize guestIoport with appropriate serial port ranges... */
  idx=0;
#ifdef CONFIG_SERIAL_UML_RESERVE_0
  guestDev->guestIoport[idx].start=0x3f8;
  guestDev->guestIoport[idx].end=0x3ff;
  idx++;
#endif
#ifdef CONFIG_SERIAL_UML_RESERVE_1
  guestDev->guestIoport[idx].start=0x2f8;
  guestDev->guestIoport[idx].end=0x2ff;
  idx++;
#endif
  
  /**
   * regiter for the callbacks
   */
  setSyscallCb(umlSysCallCb, umlSysCallRetCb, umlSystraceCb, umlSystraceExitCb);
  setDoForkCb(umlDoForkCb);
  setDoExitCb(umlDoExitCb);
  setModifiedPageCb(modifiedPage);
  setProcMmCb(handleProcMm);
  registerInOutCb(inOutCb);

  /**
   * set up the device driver functions
   */
  memset(&op, 0, sizeof(op));
  
  op.open = &uml_open;
  op.release = &uml_release;
  op.write = &uml_write;
  op.read = &uml_read;
  op.ioctl = &uml_ioctl;
  op.poll = &uml_poll;


  major = register_chrdev(0,name,&op);
  if(major < 0)
    return -1;

  printk("uml module loaded, dummy_func is at 0x%08lX, __NR_stk_track %d\n",
	 (unsigned long) dummy_func, __NR_stk_track); 

  init_bh(UML_BH, uml_irq_bh);

  return 0;
}

void cleanup_module(void)
{
  unsigned long idx;
  /**
   * make sure you nullify the callbacks, the kernel will (probably) crash
   * if you do not.
   */
  setSyscallCb(NULL,NULL, NULL, NULL);
  setDoForkCb(NULL);
  setDoExitCb(NULL);  
  setModifiedPageCb(NULL);
  setProcMmCb(NULL);
  registerInOutCb(NULL);   

  for(idx = 0; idx < NUM_UML_IRQ; idx++) {
    if(umlIrqList[idx].active) {
      free_irq(umlIrqList[idx].irq,NULL);
    }
  }

  if(rwFile != NULL) {
    util_close(rwFile);
    rwFile = NULL;
  }

  cleanup_guestdev(guestDev, NULL);
  kfree(guestDev);

  // unregister the device
  remove_bh(UML_BH);
  unregister_chrdev(major,name);
}

/**
 * these are the device driver functions, they are used so that user mode applications
 * can read the system call and other data.
 */
static unsigned int uml_poll(struct file *filp, struct poll_table_struct *wait) {
  unsigned int ret = 0;
  struct uml_irq_struct *umlIrq = (struct uml_irq_struct *) filp->private_data;

  if(umlIrq != NULL) {
    if(umlIrq->irqCount > 0) {
      umlIrq->irqCount--;
      ret = POLLIN | POLLRDNORM;
    }
  }

  return ret;
}

static int uml_open(struct inode *inode, struct file *file) {
  file->f_pos = 0;

  MOD_INC_USE_COUNT;

  return 0;
}

static int uml_release(struct inode *i, struct file *f) {
  MOD_DEC_USE_COUNT;

  if(!MOD_IN_USE) {
    // XXX FREE IRQ HERE???
  }

  return 0;
}

static ssize_t uml_write(struct file *f, const char *buf, size_t len, loff_t *p) {
  struct uml_ioc_rollback_struct rbs;

  down(&var_lock);
  if(pendingCmd) {
    printk("***Error: there is still a pending command!\n");
    up(&var_lock);
    return -EACCES;
  }

  if(len != sizeof(struct uml_ioc_rollback_struct)) {
    pendingCmd = 0;
    up(&var_lock);
    return -EINVAL;
  }

  pendingCmd = 1;
  cmdErr=0;

  printk("uml_write: rolling back checkpoint\n");
  rollbackCheckpoint = 1;
  rollbackCheckpointReady = 0;
  copy_from_user(&rbs, (void *) buf, sizeof(rbs));
  restoreCheck = rbs.restore;
  checkpointId = rbs.checkId;
  savedRbs = rbs;

  if(umlBlocked) {
    umlBlocked = 0;
    up(&uml_sem);
  }  

  while(!rollbackCheckpointReady) {    
    writeBlocked = 1;
    up(&var_lock);
    if(down_interruptible(&write_sem) == -EINTR) {
      writeBlocked = 0;
      pendingCmd = 0;
      rollbackCheckpoint = 0;
      return -EINTR;
    }
    down(&var_lock);
    assert(!writeBlocked);
  }

  pendingCmd = 0;

  up(&var_lock);

  if(cmdErr<0) {
    return cmdErr;
  }
  return sizeof(rbs);
}


/*
#define US_PER_S 1000000

static long long timeDiffUs(struct timeval *tv0, struct timeval *tv1) {
  long long d;

  d = ((long long) tv0->tv_sec) * US_PER_S;
  d += tv0->tv_usec;
  d -= ((long long) tv1->tv_sec) * US_PER_S;
  d -= tv1->tv_usec;

  return d;
}

static int down_timeout(struct semaphore *sem, struct timeval *tv) {
  struct timeval currTime;
  struct timeval newTime;
  long long d;

  do_gettimeofday(&currTime);
  currTime.tv_usec += tv->tv_usec;
  while(currTime.tv_usec >= US_PER_S) {
    currTime.tv_sec++;
    currTime.tv_usec -= US_PER_S;
  }
  currTime.tv_sec += tv->tv_sec;

  do {
    if(down_trylock(sem) == 0) return 0;
    schedule();
    do_gettimeofday(&newTime);
    d = timeDiffUs(&currTime, &newTime);
  } while(d <= 0);

  return -EINTR;
}

static void timerCallback(unsigned long data) {
  printk("timer callback\n");
  down(&var_lock);
  if(pendingCmd && readBlocked) {
    printk("uml_read timeout, cancel take checkpoint cmd\n");
    *((int *) data) = 1;
    readBlocked = 0;
    takeCheckpoint = 0;
    up(&read_sem);
  }
  up(&var_lock);
}
*/


static ssize_t uml_read(struct file *f, char *buf, size_t len, loff_t *p) {
  struct uml_ioc_checkpoint_struct cps;

  down(&var_lock);
  if(pendingCmd) {
    printk("***Error: there is still a pending command!\n");
    up(&var_lock);
    return -EACCES;
  }

  pendingCmd = 1;

  takeCheckpoint = 1;
  checkpointIdOutReady = 0;
  
  if(len != sizeof(struct uml_ioc_checkpoint_struct)) {
    pendingCmd = 0;
    takeCheckpoint = 0;
    up(&var_lock);
    return -EINVAL;
  }

  if(umlBlocked) {
    umlBlocked = 0;
    up(&uml_sem);
  }  

  while(!checkpointIdOutReady) {
    readBlocked = 1;
    up(&var_lock);
    //if(down_timeout(&read_sem,&timeout) == -EINTR) {
    if(down_interruptible(&read_sem) == -EINTR) {
    //if(down_timeout(&readBlocked, &timeout) == -EINTR) {
      readBlocked = 0;
      pendingCmd = 0;
      takeCheckpoint = 0;
      return -EINTR;
    }
    down(&var_lock);
    assert(!readBlocked);
  }

  cps.checkId = checkpointIdOut;
  copy_to_user(buf, &cps, sizeof(cps));
  pendingCmd = 0;

  up(&var_lock);

  return sizeof(cps);

  //BUG();
  //return 0;
}

static int uml_reset_cow_disk(struct uml_ioc_reset_cow_disk * user) {
  /*
    struct uml_ioc_reset_cow_disk *uml, uml_;
    int ret;
    struct file *diskFile;
    cow_disk_t cow;
    unsigned long long *idList;
    unsigned long idx;

    uml = &uml_;

    if(copy_from_user(uml, (void *)user, sizeof(*uml))) {
	ret=-EFAULT; 
	goto out;
    }

    if(uml->numId <= 0) {
	printk("uml->numId %lu!\n", uml->numId);
	ret=-EINVAL;
	goto out;
    }

    idList = kmalloc(sizeof(unsigned long long) * uml->numId, GFP_KERNEL);
    if(!idList) {
	ret=-ENOMEM;
	goto out;
    }

    ret=copy_from_user(idList, uml->idList, 
			 sizeof(unsigned long long) * uml->numId);
    if(ret) {
	ret=-EFAULT;
	goto out_free_idlist;
    }

    diskFile = filp_open(uml->umlHda, O_LARGEFILE | O_RDWR, 0600);

    if(!diskFile) {
	ret=-EBADF;
	goto out_free_idlist;
    }

    // load the disk checkpoints back up
    cow = newCowDisk(uml->cowBaseName);
    if(!cow) {
	goto out_close_diskfile;
    }

    for(idx = 0; idx < uml->numId; idx++) {
      printk("appending checkpoint %llu\n",idList[idx]);
        appendCowCheckpoint(cow, idList[idx]);
    }

    // now set the checkpoint state and go to the checkpoint
    printk("setting current id to %llu\n",uml->startId);
    if((ret=setCowCurrId(cow, uml->startId))<0) {
	printk("setCowCurrId returned %d!\n", ret);
	goto out_close_cowdisk;
    }
    printk("moving to checkpoint %llu\n",uml->targetId);
    if((ret=goToCheckpoint(cow, uml->targetId, diskFile))<0) {
	printk("goToCheckpoint returned %d!\n", ret);
	goto out_close_cowdisk;
    }

    printk("cleaning up after reset\n");
    out_close_cowdisk:
	closeCowDisk(cow, diskFile);
    out_close_diskfile:
	filp_close(diskFile, current->files);
    out_free_idlist:
	if(idList) {
	    kfree(idList);
	    idList=NULL;
	}
    out: 
	return ret;
*/

  return 0;
}

static int uml_ioctl_pause(int sec) {
  int timeout;
  int bail = 0;
  int count;
    bail = 0;
    count = 0;
    timeout = jiffies + ((int) sec) * HZ;
    down(&var_lock);
    pauseUml = 1;
    while(!umlBlocked) {
      if((timeout < jiffies) && (sec != 0)) {
	bail = 1;
	pauseUml = 0;
      }
      up(&var_lock);
      if(bail) {
	return -EINTR;
      }
      count++;
      schedule();
      down(&var_lock);
    }
    up(&var_lock);
    return 0;
}

static void handlePciCmd(struct uml_pci_cmd_struct *pciCmd) {
   struct pci_dev *dev = NULL;
   void *dmaVaddr;
   unsigned long dmaGuestVaddr, dmaGuestSize;
   int realAlloc;

   if(current->revirt_flags & REPLAY_REPLAY) {
     /* Most of the ioctl's we dont' want to replay... but this one
      * we need to. */
     switch(pciCmd->cmd_in) {
     case UML_PCI_CMD_ALLOC_DMA_BUFFER: case UML_PCI_CMD_FREE_DMA_BUFFER:
       break;
     default:
       return;
     }
   }
  
   if((pciCmd->vendor_in != 0) || (pciCmd->device_in != 0)) {
     dev = pci_find_device(pciCmd->vendor_in, 
			   pciCmd->device_in, NULL);
     assert(dev != NULL);
   }
   
  switch(pciCmd->cmd_in) {
  case UML_PCI_CMD_ENABLE_IRQ:
    // XXX FIXME make sure the guest has permission to do this
    enable_irq(pciCmd->irq_in);
    break;
  case UML_PCI_CMD_ENABLE_DEVICE:
    pci_enable_device(dev);
    break;
  case UML_PCI_CMD_SET_MASTER:
    pci_set_master(dev);
    break;
  case UML_PCI_CMD_SET_DMA_MASK:
    pci_set_dma_mask(dev, pciCmd->dma_mask_in);
    break;
  case UML_PCI_CMD_ALLOC_DMA_BUFFER:
    if(current->revirt_flags & REPLAY_REPLAY) {
      realAlloc = 0;
      revirt_replay_generic(0, sizeof(dmaVaddr), (char *) &dmaVaddr);
      revirt_replay_generic(0, sizeof(pciCmd->phys_buff_out), 
			    (char *) &pciCmd->phys_buff_out);
    } else {
      realAlloc = 1;
      dmaVaddr = pci_alloc_consistent(dev, pciCmd->dma_buff_size_in, 
				      (dma_addr_t *) &pciCmd->phys_buff_out);
      if(current->revirt_flags & REPLAY_LOG) {
	revirt_log_generic(0, sizeof(dmaVaddr), (char *) &dmaVaddr);
	revirt_log_generic(0, sizeof(pciCmd->phys_buff_out), 
			   (char *) &pciCmd->phys_buff_out);
      }
      memset(dmaVaddr, 0, pciCmd->dma_buff_size_in);
    }
    dmaGuestVaddr = pciCmd->guest_vaddr_in;
    dmaGuestSize = pciCmd->dma_buff_size_in;
    register_guest_dma(dev, dmaVaddr, dmaGuestVaddr, dmaGuestSize, 
		       (dma_addr_t) pciCmd->phys_buff_out, realAlloc);
    pciCmd->host_vaddr_out = (unsigned long) dmaVaddr;
    break;

  case UML_PCI_CMD_FREE_DMA_BUFFER:    
    free_guest_dma_vaddr(__va(pciCmd->phys_buff_in));
    break;
  default:
    printk("invalid pci cmd");
    assert(0);
  }
}

// XXX FIXME, add tsk check to both these
static int isGuestDmaMem(unsigned long vaddr) {
  unsigned long idx;
  for(idx = 0; idx < NUM_GUEST_DMA; idx++) {
    if(guestDev->guestDma[idx].dmaVaddr != NULL) {
      if((vaddr >= ((unsigned long) guestDev->guestDma[idx].dmaVaddr)) &&
	 (vaddr < 
	  (((unsigned long) guestDev->guestDma[idx].dmaVaddr) + 
	   guestDev->guestDma[idx].dmaGuestSize))) {
	return 1;
      }
    }
  }

  return 0;
}


static int isGuestIomapMem(unsigned long vaddr) {
  unsigned long idx;

  for(idx = 0; idx < NUM_GUEST_IOMAP; idx++) {
    if(guestDev->guestIomap[idx].vaddr != 0) {
      if((vaddr >= guestDev->guestIomap[idx].vaddr) &&
	 (vaddr < (guestDev->guestIomap[idx].vaddr + 
		   guestDev->guestIomap[idx].size))) {
	return 1;
      }
    }
  }

  return 0;
}

static void handleIomapCmd(struct uml_iomap_cmd_struct *iomapCmd) {
  int realAlloc;
  if(iomapCmd->map) {
    if(current->revirt_flags & REPLAY_REPLAY) {
      realAlloc = 0;
      revirt_replay_generic(0, sizeof(iomapCmd->vaddr_out),
			    (char *) &iomapCmd->vaddr_out);
    } else {
      realAlloc = 1;
      iomapCmd->vaddr_out = (unsigned long) ioremap(iomapCmd->phys_addr_in, 
						    iomapCmd->size_in);
    }

    if(current->revirt_flags & REPLAY_LOG) {
      revirt_log_generic(0, sizeof(iomapCmd->vaddr_out), 
			 (char *) &iomapCmd->vaddr_out);
    }
    printk("mapped phys addr 0x%08lx (size = %lu) to vaddr 0x%08lx\n", 
	   iomapCmd->phys_addr_in, iomapCmd->size_in, iomapCmd->vaddr_out);
    // find a free iomap slot to store the data in
    register_guest_iomap(iomapCmd->phys_addr_in,
			 iomapCmd->vaddr_out,
			 iomapCmd->size_in,
			 realAlloc);
  } else {
    printk("unmapping vaddr 0x%08lx\n",iomapCmd->vaddr_in);
    free_guest_iomap_vaddr(iomapCmd->vaddr_in);
  }
}

static int isValidGuestDev(unsigned short vendor, unsigned short device) {
  //XXX FIXME

  // this is for the audio device
  if((vendor == 0x8086) && (device == 0x24c5)) {
    return 1;
  }
  
  // usb devices
  if((vendor == 0x08086) && ((device == 0x24c2) ||
			     (device == 0x24c4) ||
			     (device == 0x24c7) ||
			     (device == 0x24cd))) {
    return 1;
  }


  return 0;
}

static void handleGetPciDev(struct uml_get_pci_dev_struct *getPciDev) {
  struct pci_dev *dev;
  unsigned long res;

  getPciDev->irq_out = 0;

  dev = (struct pci_dev *) getPciDev->pci_dev_iter_inout;
  while((dev = pci_find_device(PCI_ANY_ID, PCI_ANY_ID, dev)) != NULL) {
    if(isValidGuestDev(dev->vendor, dev->device)) {
      printk("found valid guest device vend/dev 0x%x/0x%x\n",dev->vendor,dev->device);
      break;
    }
  }

  if(dev != NULL) {
    getPciDev->irq_out = dev->irq;
    
    for(res=0;res<DEVICE_COUNT_RESOURCE;res++) {
      getPciDev->resource[res].start = dev->resource[res].start;
      getPciDev->resource[res].end = dev->resource[res].end;
      getPciDev->resource[res].flags = dev->resource[res].flags;
    }
    getPciDev->pci_dev_iter_inout = dev;
    getPciDev->vendor_out = dev->vendor;
    getPciDev->device_out = dev->device;
    getPciDev->class_out = dev->class;


    // add in any iosports we allow in and outs to
    // FIXME check for ioports
    for(res = 0; res < DEVICE_COUNT_RESOURCE; res++) {
      if(dev->resource[res].start && (dev->resource[res].start <= 0xffff)) {
	register_guest_ioport(dev->resource[res].start,
			      dev->resource[res].end);
      }
    }
  }
}

static void handleSyncDma(struct uml_sync_dma_struct *syncDma) {
  /*
  int ret;
  unsigned long idx;
  struct pci_dev *dev = pci_find_device(syncDma->vendor,
					syncDma->device,
					NULL);
  assert(dev != NULL);

  // find the guestDma entry
  for(idx = 0; idx < NUM_GUEST_DMA; idx++) {
    if(syncDma->dmaAddr == guestDma[idx].dmaAddr) {
      assert(syncDma->size == guestDma[idx].dmaGuestSize);
      if(syncDma->direction == PCI_DMA_FROMDEVICE) {
	ret = copy_to_user((void *) guestDma[idx].dmaGuestVaddr,
			   guestDma[idx].dmaVaddr, 
			   syncDma->size);
	assert(ret == 0);
      } else if(syncDma->direction == PCI_DMA_TODEVICE) {
	ret = copy_from_user(guestDma[idx].dmaVaddr,
			     (void *) guestDma[idx].dmaGuestVaddr,
			     syncDma->size);
	assert(ret == 0);
	if(ret != 0) {
	  printk("could not copy data from guest vaddr 0x%08lx\n",
		 guestDma[idx].dmaGuestVaddr);
	}
      } else {
	assert(0);
      }
      
      break;
    }
  }

  assert(idx < NUM_GUEST_DMA);
  //we dont actually need this since we are emulating the loads and stores
  */
}


static void handlePciConfig(struct uml_pci_config_struct *pciConfig) {
  struct pci_dev *dev = pci_find_device(pciConfig->vendor_in,
					pciConfig->device_in,
					NULL);
  // XXX FIXME, what if we have two of the same device???
  assert(dev != NULL);
  switch(pciConfig->cmd_in) {
  case PCI_WRITE_CONFIG_WORD:
    pciConfig->ret_out = pci_write_config_word(dev, pciConfig->where_in, 
					       (u16) pciConfig->val_in);
    break;
  default:
    assert(0);
  }
}

/*
static void walkUsbV(unsigned long flptr) {
  unsigned long element;
  unsigned long *elptr;
  unsigned long link;
  
  unsigned long count = 0;

  elptr = __va(flptr & ~0xf);
  link = flptr;
  element = *elptr;

  while(1) {
    if(count++ > 10) {
      printk("infinite loop\n");
      break;
    }
    if(link & 0x2) {
      assert(0);
    } else {
      unsigned long st;
      printk("TD(0x%08lx) ",element);
      st = elptr[1];
      st &= (0xff << 16);
      st >>= 16;
      printk(" status = 0x%02lx ", st);
    }

    if(element & 0x1) {
      printk ("terminate\n");
      break;
    } else {      
      elptr = __va(element & ~0xf);
      link = element;
      element = *elptr;
      printk("-> ");
    }
  }
}

static void walkUsbH(unsigned long flptr) {
  unsigned long element;
  unsigned long *elptr;
  unsigned long link;
  
  unsigned long count = 0;
  unsigned long qh_element = 0;

  if(flptr & 0x1) {
    printk("frame pointer invalid\n");
    return;
  }

  elptr = __va(flptr & ~0xf);
  link = flptr;
  element = *elptr;
  if(link & 0x2) {
    qh_element = elptr[1];
  }

  while(1) {
    if(count++ > 10) {
      printk("infinite loop\n");
      break;
    }
    if(link & 0x2) {
      printk("QH(0x%08lx,0x%08lx) ",element, qh_element);
      if(!(qh_element & 0x1)) {
	walkUsbV(qh_element & ~0xf);
      }
    } else {
      printk("TD(0x%08lx) ",element);
    }

    if(element & 0x1) {
      printk ("terminate\n");
      break;
    } else {      
      elptr = __va(element & ~0xf);
      link = element;
      element = *elptr;
      if(link & 0x2) {
	qh_element = elptr[1];
      }
      printk("\n-> ");
    }
  }
}
*/

static void handleDebugCmd(struct uml_debug_struct *debugCmd) {
  struct pci_dev *dev;

  if(debugCmd->cmd == DEBUG_PRINT_CMD) {
    debugPrint = debugCmd->debugPrint;
  } else if(debugCmd->cmd == DEBUG_WALK_FRAMELIST) {
    unsigned short base = 0xff40;
    unsigned long framebase;
    unsigned long *flptr;
    unsigned short status;

    dev = pci_find_device(0x8086, 0x24c7, NULL);
    assert(dev != NULL);

    framebase = inl(base + 8) & ~0xf;
    flptr = __va(framebase);
    printk("framebase = 0x%08lx, flptr = %p, 0x%08lx\n",framebase, flptr, *flptr);
    status = inw(base + 2);
    printk("status = 0x%04x\n",status);
    //walkUsbH(*flptr);
    
  } else if (debugCmd->cmd == DEBUG_PRINT_TD) {
    unsigned long *addr = (unsigned long *) debugCmd->addr;
    unsigned long len,idx;
    unsigned char *buf;

    printk("td at %p\n",addr);
    printk("status = 0x%08lx\n",addr[1]);
    printk("info   = 0x%08lx\n",addr[2]);
    printk("buffer = 0x%08lx\n",addr[3]);

    len = addr[2] >> 21;
    if(len == 0x7ff) len = 0;
    buf = __va(addr[3]);
    for(idx = 0; idx <= len; idx++) {
      printk("0x%02x ", buf[idx]);
    }

    printk("\n");

  } else {
    assert(0);
  }
}

static void handleRwCmd(struct uml_rw_struct *rw) {
#ifndef HANDLE_RW_CMD
  return;
#else
  int ret;
  unsigned char *data;

  if(rwFile == NULL) {
    rwFile = util_open("/tmp/rwlog.dat",O_RDWR | O_LARGEFILE | 
		       O_CREAT | O_TRUNC, 0600);
    assert(rwFile != NULL);
  } 

  util_append(rwFile, &rw->pid, sizeof(rw->pid));
  util_append(rwFile, rw->comm, sizeof(rw->comm));
  util_append(rwFile, &current->uml->guestSyscallCount, 
	      sizeof(current->uml->guestSyscallCount));
  util_append(rwFile, &rw->dataLen, sizeof(rw->dataLen));

  data = util_alloc_large(rw->dataLen);

  ret = copy_from_user(data, rw->buf, rw->dataLen);
  assert(ret == 0);
  util_append(rwFile, data, rw->dataLen);

  util_free_large(data);
#endif
}


static int uml_ioctl(struct inode *inode, struct file *f,
			 unsigned int cmd, unsigned long arg) {
  //struct uml_ioc_rollback_struct rbs;
  struct uml_ioc_checkinfo_struct cis;
  int ret;
  struct uml_check_struct *chk;
  int err=0;
  struct uml_pci_cmd_struct pciCmd;
  struct uml_iomap_cmd_struct iomapCmd;
  struct uml_get_pci_dev_struct getPciDev;
  struct uml_pci_config_struct pciConfig;
  struct uml_sync_dma_struct syncDma;
  struct uml_debug_struct debugCmd;
  struct uml_rw_struct rw;

  if(current->revirt_flags & REPLAY_REPLAY) {
    /* Most of the ioctl's we dont' want to replay... but this one
     * we need to. */
    switch(cmd) {
    case UML_IOC_GET_PCI_DEV:
    case UML_IOC_IOMAP_CMD:
    case UML_IOC_PCI_CMD:
      break;
    default:
      return 0;
    }
  }

  switch(cmd) {
  case UML_IOC_RW_CMD:
    ret = copy_from_user(&rw, (void *) arg, sizeof(rw));
    assert(ret == 0);
    handleRwCmd(&rw);
    break;
  case UML_IOC_DEBUG_CMD:
    ret = copy_from_user(&debugCmd, (void *) arg, sizeof(debugCmd));
    assert(ret == 0);
    handleDebugCmd(&debugCmd);
    break;
  case UML_IOC_SYNC_DMA:
    ret = copy_from_user(&syncDma, (void *) arg, sizeof(syncDma));
    assert(ret == 0);
    handleSyncDma(&syncDma);
    break;
  case UML_IOC_PCI_CONFIG:
    ret = copy_from_user(&pciConfig, (void *) arg, sizeof(pciConfig));
    assert(ret == 0);
    handlePciConfig(&pciConfig);
    ret = copy_to_user((void *) arg, &pciConfig, sizeof(pciConfig));
    assert(ret == 0);
    break;
  case UML_IOC_GET_PCI_DEV:
    ret = copy_from_user(&getPciDev, (void *) arg, sizeof(getPciDev));
    assert(ret == 0);
    handleGetPciDev(&getPciDev);
    ret = copy_to_user((void *) arg, &getPciDev, sizeof(getPciDev));
    assert(ret == 0);
    break;
  case UML_IOC_IOMAP_CMD:
    ret = copy_from_user(&iomapCmd, (void *) arg,sizeof(struct uml_iomap_cmd_struct));
    assert(ret == 0);
    handleIomapCmd(&iomapCmd);
    ret = copy_to_user((void *) arg, &iomapCmd, sizeof(struct uml_iomap_cmd_struct));
    assert(ret == 0);
    break;
  case UML_IOC_PCI_CMD:
    ret = copy_from_user(&pciCmd, (void *) arg, 
			 sizeof(struct uml_pci_cmd_struct));
    assert(ret == 0);
    handlePciCmd(&pciCmd);
    ret = copy_to_user((void *) arg, &pciCmd, sizeof(struct uml_pci_cmd_struct));
    assert(ret == 0);
    break;
  case UML_IOC_REQUEST_IRQ:
    //register the interrupt handler
    if(!(current->revirt_flags & REPLAY_REPLAY)) {
      // XXX FIXME, support sharing irq's at somepoint if the guest driver does
      int irq = (int) arg;
      assert((irq > 0) && (irq < NUM_UML_IRQ));
      if(umlIrqList[irq].active) {
	printk("already have irq regiesterd\n");
	err = -1;
      } else {
	if(request_irq(irq,uml_irq_callback,0,name,NULL) != 0) {
	  printk("Did not get the irq %d\n",(int) arg);
	  err = -1; // ???
	} else {
	  umlIrqList[irq].active = 1;
	  umlIrqList[irq].irqCount = 0;
	  umlIrqList[irq].irq = irq;
	  umlIrqList[irq].tsk = current;
	  f->private_data = (void *) &(umlIrqList[irq]);

	  printk("registered the interrupt (%d)\n", irq);
	}
      }
    }
    break;
  case UML_IOC_PAUSE:
    err = uml_ioctl_pause((int)arg);
    break;
  case UML_IOC_GET_CHECKINFO:
    down(&var_lock);
    ret = copy_from_user(&cis, (void *) arg, sizeof(cis));
    assert(ret == 0);
    if(!currCheckinfoSet) {
      printk("this command can only be used when gdb has the vm stopped\n");
      up(&var_lock);
      return -EACCES;
    }

    // find the checkpoint
    if(cis.checkId == INVALID_CHECK_ID) {
      // for the current
      cis = currCheckinfo;
    } else {
      // find the checkpoint
      assert(cis.checkId < MAX_UML_CHECK);
      assert(umlTsk->uml && umlTsk->uml->isKernel);
      chk = umlTsk->uml->checkArray[cis.checkId];
      if(chk == NULL) {
	memset(&cis, 0, sizeof(cis));
	cis.checkId = INVALID_CHECK_ID;
      } else {
	cis.guestSyscallCount = chk->guestSyscallCount;
	if(getUserCheckInfo) {
	  cis.revirtId = chk->user_revirt_id;
	  cis.eip = chk->user_regs.eip;
	} else {
	  cis.revirtId = chk->revirt_id;
	  cis.eip = chk->uml_kern_check.regs.eip;
	}
      }
    }
    ret = copy_to_user((void *) arg, &cis, sizeof(cis));
    assert(ret == 0);
    up(&var_lock);
    break;
#if 0
  case UML_IOC_CHECKPOINT:
    BUG();
    break;
  case UML_IOC_ROLLBACK:
    /*
    printk("uml_ioctl: rolling back checkpoint\n");
    rollbackCheckpoint = 1;
    copy_from_user(&rbs, (void *) arg, sizeof(rbs));
    restoreCheck = rbs.restore;
    checkpointId = rbs.checkId;
    */
    BUG();
    break;
#endif
  case UML_IOC_SET_GDB_STATE:
    BUG();
    /*
    down(&var_lock);
    if(pendingCmd) {
      printk("***Error: there is still a pending command!\n");
      up(&var_lock);
      return -EACCES;
    }
    
    if((arg != GDB_CHECK_OFF) && (arg != GDB_CHECK_PERIODIC)) {
      printk("illega gdb state from ioctl\n");
      up(&var_lock);
      return -EACCES;
    }
    pendingCmd = 1;    

    gdb_state_waiting = 1;
    gdb_state = arg;

    up(&var_lock);
    */
    break;
  case UML_IOC_RESTART_UML:
    down(&var_lock);
    if(umlBlocked) {
      umlBlocked = 0;
      up(&uml_sem);
    }
    up(&var_lock);
    break;
  case UML_IOC_ERESET_UML:
    if(umlBlocked) {
      umlBlocked = 0;
      up(&uml_sem);
    }
    break;
  case UML_IOC_RESET_COW_DISK: 
    err=uml_reset_cow_disk((struct uml_ioc_reset_cow_disk *)arg);
    break;
#if 0
  {
    struct uml_ioc_reset_cow_disk *uml = kmalloc(sizeof(*uml), GFP_KERNEL);
    int ret;
    struct file *diskFile;
    cow_disk_t cow;
    unsigned long long *idList;
    unsigned long idx;

    assert(uml != NULL);

    ret = copy_from_user(uml, (void *) arg, sizeof(*uml));
    assert(ret == 0);
    assert(uml->numId > 0);
    idList = kmalloc(sizeof(unsigned long long) * uml->numId, GFP_KERNEL);
    assert(idList != NULL);
    ret = copy_from_user(idList, uml->idList, 
			 sizeof(unsigned long long) * uml->numId);
    assert(ret == 0);

    diskFile = filp_open(uml->umlHda, O_LARGEFILE | O_RDWR, 0600);
    assert(diskFile != NULL);

    // load the disk checkpoints back up
    cow = newCowDisk(uml->cowBaseName);
    for(idx = 0; idx < uml->numId; idx++) {
      printk("appending checkpoint %llu\n",idList[idx]);
      appendCowCheckpoint(cow, idList[idx]);
    }

    // now set the checkpoint state and go to the checkpoint
    printk("setting current id to %llu\n",uml->startId);
    if((err=setCowCurrId(cow, uml->startId))<0) {
	printk("setCowCurrId returned %d!\n", err);
	goto out;
    }
    printk("moving to checkpoint %llu\n",uml->targetId);
    if((err=goToCheckpoint(cow, uml->targetId, diskFile))<0) {
	printk("goToCheckpoint returned %d!\n", err);
	goto out;
    }

    printk("cleaning up after reset\n");
    closeCowDisk(cow, diskFile);
    filp_close(diskFile, current->files);
    kfree(uml);

    break;
  }
#endif
  default:
    err=-EINVAL;
  }
  //out:
  return err;
}


static int umlSystraceCb(struct pt_regs *regs) {
  if(current->uml && current->uml->isUser) {
    if(guestSyscallCb != NULL) {
      guestSyscallCb(current->uml->kernelTsk);
    }
    current->uml->kernelTsk->uml->guestSyscallStarted = 1;

  }

  
  return 0;
}

static int umlSystraceExitCb(struct pt_regs *regs) {
  if(current->uml && current->uml->isUser) {
    if(guestRetCb != NULL) {
      guestRetCb(current->uml->kernelTsk);
    }
  }

  return 0;
}

void registerGuestCallbacks(guest_callback_sig gsysCb,
			    guest_callback_sig gretCb,
			    guest_callback_sig gtrapCb) {
  guestSyscallCb = gsysCb;
  guestRetCb = gretCb;
  guestTrapCb = gtrapCb;
}

void setGdbExitCallback(gdbExitCallback cb) {
  gdbExitCb = cb;
}

#define MODRM_TO_R(modrm)   (((modrm) & 0x34) >> 3)
#define MODRM_TO_MOD(modrm) (((modrm) & 0xc0) >> 6)
#define MODRM_TO_RM(modrm)  ((modrm) & 0x07)

#define SIB_SCALE(sib) (((sib) & 0xc0) >> 6)
#define SIB_INDEX(sib) (((sib) & 0x38) >> 3)
#define SIB_BASE(sib)  ((sib) & 0x07)


#define BIT_8  0
#define BIT_16 1
#define BIT_32 2

static void setRegValue(unsigned long val,unsigned char r, struct pt_regs *regs,
			int numBits) {
  switch(r) {
  case 0:
    regs->eax = val;
    break;
  case 1:
    regs->ecx = val;
    break;
  case 2:
    regs->edx = val;
    break;
  case 3:
    regs->ebx = val;
    break;
  case 4:
    regs->esp = val; assert(numBits != BIT_8);
    break;
  case 5:
    regs->ebp = val; assert(numBits != BIT_8);
    break;
  case 6:
    regs->esi = val; assert(numBits != BIT_8);
    break;
  case 7:
    regs->edi = val; assert(numBits != BIT_8);
    break;
  }
}

static unsigned long getRegValue(unsigned char r, struct pt_regs *regs,
				 int numBits) {
  unsigned long ret = 0;
  switch(r) {
  case 0:
    ret = regs->eax;
    break;
  case 1:
    ret = regs->ecx;
    break;
  case 2:
    ret = regs->edx;
    break;
  case 3:
    ret = regs->ebx;
    break;
  case 4:
    ret = (numBits == BIT_8) ? regs->eax >> 8 : regs->esp;
    break;
  case 5:
    ret = (numBits == BIT_8) ? regs->ecx >> 8 : regs->ebp;
    break;
  case 6:
    ret = (numBits == BIT_8) ? regs->edx >> 8 : regs->esi;
    break;
  case 7:
    ret = (numBits == BIT_8) ? regs->ebx >> 8 : regs->edi;
    break;
  default:
    assert(0);
  }

  if(numBits == BIT_8) {
    ret = ret & 0xff;
  } else if(numBits == BIT_16) {
    ret = ret & 0xffff;
  }
  
  return ret;
}

#define IS_TWO_BYTE_ESCAPE(op) ( ((op) == 0xf) )

#define IS_PUSH(op) ( ((op) == 0xff) )

#define IS_LOAD(op)  ( ((op) == 0x8b) || ((op) == 0x8a) || ((op) == 0xb) || ((op) == 0xb6) || IS_PUSH(op) )
#define IS_STORE_IMM(op) ( (op) == 0xc7 )
#define IS_STORE(op) ( ((op) == 0x89) || ((op) == 0x88) || IS_STORE_IMM(op) )
#define IS_8BIT_LS(op) ( ((op) == 0x88) || ((op) == 0x8a) )
#define IS_OR(op) ( ((op) == 0xb) )
#define IS_ZERO_FILL_8BIT(op) ( ((op) == 0xb6) )
#define IS_ZERO_FILL(op) ( IS_ZERO_FILL_8BIT(op) )

#define IS_REP(op) ( ((op) == 0xf3) )
#define IS_STR_OP(op) ( ((op) == 0xab) || IS_MOVS(op)  )
#define IS_MOVSB(op) ( ((op) == 0xa4) )
#define IS_MOVS(op) ( IS_MOVSB(op) || ((op) == 0xa5) )

static int isValidEmuAddr(unsigned long addr) { 
  int ret = isGuestIomapMem(addr) || isGuestDmaMem(addr);

  if(!ret) {
    //printk("invalid address 0x%08lx\n",addr);
  }
  return ret;
}

static int emulateRepStr(struct pt_regs *regs, unsigned char op, int numBits, 
			 unsigned long eip) {
  int curet;
  int isRep = IS_REP(op);
  unsigned long count = 0;
  unsigned long ecx;
  unsigned long edi;
  unsigned long esi;
  unsigned long numBytes = 0;
  unsigned long val = 0;

  if(isRep) {
    curet = copy_from_user(&op, (void *) eip++, sizeof(op));
    assert(curet == 0);
  }

  if(!IS_STR_OP(op)) {
    printk("unhadled string op = 0x%02x, eip = 0x%08lx\n",op,regs->eip);
    return 0;
  }

  if(IS_MOVSB(op)) {
    numBits = BIT_8;
  }

  if(numBits == BIT_8) {
    numBytes = 1;
  } else if(numBits == BIT_16) {
    numBytes = 2;
  } else if(numBits == BIT_32) {
    numBytes = 4;
  } else {
    assert(0);
  }

  ecx = regs->ecx;
  edi = regs->edi;
  esi = regs->esi;
  while((isRep && ecx) || (!isRep && (count == 0))) {
    if(IS_MOVS(op)) {
      if(edi >= PAGE_OFFSET) {
	if(!isValidEmuAddr(edi)) return 0;
	if(esi >= PAGE_OFFSET) return 0;

	if(!(current->revirt_flags & REPLAY_REPLAY)) {
	  curet = copy_from_user((void *) edi, (void *) esi, numBytes);
	  assert(curet == 0);
	}
      } else {
	if(!isValidEmuAddr(esi)) return 0;
	
	if(current->revirt_flags & REPLAY_REPLAY) {
	  revirt_replay_generic(0,numBytes,(char *) &val);
	} else {
	  memcpy(&val,(void *) esi, numBytes);
	}
	curet = copy_to_user((void *) edi, &val, numBytes);
	assert(curet == 0);
	if(current->revirt_flags & REPLAY_LOG) {
	  revirt_log_generic(0,numBytes,(char *) &val);
	}
      }
      if((regs->eflags & X86_EFLAGS_DF) == 0) {
	esi += numBytes;
	edi += numBytes;
      } else {
	esi -= numBytes;
	edi -= numBytes;
      }
    } else if(op == 0xab) {
      if(!isValidEmuAddr(edi)) return 0;

      if(!(current->revirt_flags & REPLAY_REPLAY)) {
	memcpy((void *) edi, &regs->eax, numBytes);
      }

      if((regs->eflags & X86_EFLAGS_DF) == 0) {
	edi += numBytes;
      } else {
	edi -= numBytes;
      }
    }

    count++;
    if(isRep) ecx--;
  }

  regs->ecx = ecx;
  regs->edi = edi;
  regs->esi = esi;
  regs->eip = eip;
  return 1;
}

static int emulateLoadStore(struct pt_regs *regs) {
  unsigned long idx;
  int numBits = BIT_32;
  unsigned char op;
  unsigned char modRM;
  unsigned char sib;
  int curet;
  unsigned long eip = regs->eip;
  unsigned long addr = 0;
  unsigned long val;
  unsigned long index;
  unsigned long disp = 0;
  unsigned char firstByte = 0;

  curet = copy_from_user(&op, (void *) eip++, sizeof(op));
  if(curet != 0) return 0;

  if(op == 0x66) {
    numBits = BIT_16;
    curet = copy_from_user(&op, (void *) eip++, sizeof(op));
    assert(curet == 0);
  }

  if(IS_TWO_BYTE_ESCAPE(op)) {
    firstByte = op;
    curet = copy_from_user(&op, (void *) eip++, sizeof(op));
    assert(curet == 0);
  }

  if(IS_REP(op) || IS_STR_OP(op)) {
    return emulateRepStr(regs, op, numBits, eip);
  }

  if(IS_LOAD(op) || IS_STORE(op)) {
    if(IS_8BIT_LS(op)) numBits = BIT_8;
    // 8, 16 or 32 bit move from r/m to r
    curet = copy_from_user(&modRM, (void *) eip++, sizeof(modRM));
    assert(curet == 0);

    if(MODRM_TO_RM(modRM) == 4) {
      // sib
      curet = copy_from_user(&sib, (void *) eip++, sizeof(sib));
      assert(curet == 0);
      assert(sib != 5);
      addr = getRegValue(SIB_BASE(sib), regs, BIT_32);
      index = getRegValue(SIB_INDEX(sib), regs, BIT_32);	    
      for(idx = 0; idx < SIB_SCALE(sib); idx++) {
	index *= 2;
      }
      addr += index;
    } else if((MODRM_TO_RM(modRM) == 5) && (MODRM_TO_MOD(modRM) == 0)) {
      //disp 32
      // this style of load store is not used by the read/write macros (I hope)
      assert(0);
    } else {
      addr = getRegValue(MODRM_TO_RM(modRM),regs, BIT_32);
    }

    if(MODRM_TO_MOD(modRM) == 0x00) {
      disp = 0;
    } else if(MODRM_TO_MOD(modRM) == 0x2) {
      curet = copy_from_user(&disp, (void *) eip, 4);
      assert(curet == 0);
      eip += 4;
    } else if(MODRM_TO_MOD(modRM) == 0x1) {
      disp = 0;
      curet = copy_from_user(&disp, (void *) eip, 1);
      assert(curet == 0);
      eip += 1;
    } else {
      printk("unknown mod 0x%x, eax = 0x%08lx\n",MODRM_TO_MOD(modRM),regs->eax);
      return 0;
    }

    addr += disp;

    if(!isGuestIomapMem(addr) && !isGuestDmaMem(addr)) {
      printk("invalid addr = 0x%08lx at eip = 0x%08lx\n",addr,regs->eip);
      printk("eax = 0x%08lx\n",regs->eax);
      printk("ebx = 0x%08lx\n",regs->ebx);
      printk("ecx = 0x%08lx\n",regs->ecx);
      return 0;
    }
    if(IS_LOAD(op)) {
      if(current->revirt_flags & REPLAY_REPLAY) {
	revirt_replay_generic(0,sizeof(val), (char *) &val);
      } else {
	val = *((unsigned long *) addr);
	if(current->revirt_flags & REPLAY_LOG) {
	  revirt_log_generic(0,sizeof(val), (char *) &val);
	}
      }
      if(IS_OR(op)) {
	val |= getRegValue(MODRM_TO_R(modRM), regs, numBits);
      } else if(IS_ZERO_FILL(op)) {
	if(IS_ZERO_FILL_8BIT(op)) {
	  val &= 0xff;
	}
      }
      if(IS_PUSH(op)) {
	assert(numBits == BIT_32);
	curet = copy_to_user((void *) (regs->esp - sizeof(val)), &val, sizeof(val));
	assert(curet == 0);
      } else {
	setRegValue(val, MODRM_TO_R(modRM), regs, numBits); 
      }
      if(debugPrint) {
	printk("load address 0x%08lx = 0x%08lx\n", addr, *((unsigned long *) addr));
      }
    } else {
      assert(IS_STORE(op));
      if(IS_STORE_IMM(op)) {
	assert(numBits == BIT_32);
	curet = copy_from_user(&val, (void *) eip, 4);
	assert(curet == 0);
	eip += 4;
      } else {
	val = getRegValue(MODRM_TO_R(modRM),regs, numBits);
      }
      if(debugPrint) {
	printk("store addr 0x%08lx = 0x%08lx\n",addr, val);
      }
      if(!(current->revirt_flags & REPLAY_REPLAY)) {
	if(numBits == BIT_8) {
	  *((unsigned char *) addr) = (unsigned char) val;
	} else if(numBits == BIT_16) {
	  *((unsigned short *) addr) = (unsigned short) val;
	} else {
	  assert(numBits == BIT_32);
	  *((unsigned long *) addr) = val;
	}
      }
    }
    
    if(IS_PUSH(op)) {
      assert(numBits == BIT_32);
      regs->esp -= 4;
    }
    regs->eip = eip;
    return 1;
  } else {
    printk("unhandled opcode = 0x%x\n",op);
  }
  
  return 0;
}

int umlDeviceSigHandler(struct pt_regs *regs, int signr, int preframe) {
  int ret = signr;

  if(current->uml && current->uml->isKernel && preframe) {
    if(signr == SIGSEGV) {
      if(emulateLoadStore(regs)) {
	ret = 0;
      }
    }
  }

  if(current->uml && current->uml->isKernel && (ret == SIGSEGV)) {
    //printk("guest kernel segv at 0x%08lx\n",regs->eip);
  }  

  return ret;
}

static void free_guest_dma_vaddr(void *dmaVaddr) {
  unsigned long idx;

  for(idx = 0; idx < NUM_GUEST_DMA; idx++) {
    if((guestDev->guestDma[idx].dmaVaddr == dmaVaddr) &&
       (guestDev->guestDma[idx].tsk == current)) {
      free_guest_dma(guestDev->guestDma + idx);
    }
  }
}

static void free_guest_dma(struct guest_dma_struct *dma) {
  if(dma->dmaVaddr && dma->realAlloc) {
    pci_free_consistent(dma->dev,
			dma->dmaGuestSize,
			dma->dmaVaddr,
			dma->dmaAddr);
  }

  memset(dma, 0, sizeof(*dma));
}

static void free_guest_iomap(struct guest_iomap_struct *iom) {
  if(iom->vaddr && iom->realAlloc) {
    iounmap((void *) iom->vaddr);
  }

  memset(iom, 0, sizeof(*iom));
}

static void free_guest_iomap_vaddr(unsigned long iomapVaddr) {
  unsigned long idx;

  for(idx = 0; idx < NUM_GUEST_IOMAP; idx++) {
    if((guestDev->guestIomap[idx].vaddr == iomapVaddr) &&
       (guestDev->guestIomap[idx].tsk == current)) {
      free_guest_iomap(guestDev->guestIomap + idx);
    }
  }
}

static void free_guest_ioport(struct guest_ioport_struct *iop) {
  memset(iop, 0, sizeof(*iop));
}

static void cleanup_guestdev(struct guest_devperm_struct *gdp, 
			     struct task_struct *tsk) {
  unsigned long idx;

  for(idx = 0; idx < NUM_GUEST_DMA; idx++) {
    if((tsk == NULL) || (tsk == gdp->guestDma[idx].tsk)) {
      free_guest_dma(gdp->guestDma + idx);
    }
  }

  for(idx = 0; idx < NUM_GUEST_IOMAP; idx++) {
    if((tsk == NULL) || (tsk == gdp->guestIomap[idx].tsk)) {
      free_guest_iomap(gdp->guestIomap + idx);
    }
  }

  for(idx = 0; idx < NUM_GUEST_IOPORT; idx++) {
    if((tsk == NULL) || (tsk == gdp->guestIoport[idx].tsk)) {
      free_guest_ioport(gdp->guestIoport + idx);
    }
  }
} 

static void register_guest_ioport(unsigned long start,
				  unsigned long end) {
  unsigned long idx = 0;
  while(1) {
    if((guestDev->guestIoport[idx].start == 0) &&
       (guestDev->guestIoport[idx].end == 0)) {
      guestDev->guestIoport[idx].start = start;
      guestDev->guestIoport[idx].end = end;
      guestDev->guestIoport[idx].tsk = current;
      break;
    } else {
      idx++;
      assert(idx < NUM_GUEST_IOPORT);
    }
  }
}
  
static void register_guest_iomap(unsigned long physAddr,
				 unsigned long vaddr,
				 unsigned long size,
				 int realAlloc) {
  unsigned long idx; 

  for(idx = 0; idx < NUM_GUEST_IOMAP; idx++) {
    if(guestDev->guestIomap[idx].vaddr == 0) {
      guestDev->guestIomap[idx].vaddr = vaddr;
      guestDev->guestIomap[idx].phys_addr = physAddr;
      guestDev->guestIomap[idx].size = size;
      guestDev->guestIomap[idx].realAlloc = realAlloc;
      guestDev->guestIomap[idx].tsk = current;
      break;
    }
  }
  assert(idx < NUM_GUEST_IOMAP);
}
static void register_guest_dma(struct pci_dev *dev, void *dmaVaddr, 
			       unsigned long dmaGuestVaddr, 
			       unsigned long dmaGuestSize, 
			       dma_addr_t dmaAddr,
			       int realAlloc) {
  unsigned long idx;

  for(idx = 0; idx < NUM_GUEST_DMA; idx++) {
    if(guestDev->guestDma[idx].dmaVaddr == NULL) {
      guestDev->guestDma[idx].dmaVaddr = dmaVaddr;
      guestDev->guestDma[idx].dmaGuestVaddr = dmaGuestVaddr;
      guestDev->guestDma[idx].dmaGuestSize = dmaGuestSize;
      guestDev->guestDma[idx].dmaAddr = dmaAddr;
      guestDev->guestDma[idx].dev = dev;
      guestDev->guestDma[idx].realAlloc = realAlloc;
      guestDev->guestDma[idx].tsk = current;
      break;
    }
  }

  assert(idx < NUM_GUEST_DMA);
}

// XXX FIXME, these only work during replay for now, adding support
// for regular operation should be easy.  Just free current resources
// and re-allocate any new ones that are restored
#define DISABLE_GUEST_DEV_CP

#ifdef DISABLE_GUEST_DEV_CP
#warning "guest devices will not work with checkpointing if they are enabled"
#endif
static void *checkpointGuestDev(void) {
#ifndef DISABLE_GUEST_DEV_CP
  struct guest_devperm_struct *gdcp;
  static int warn = 1;

  if(!(current->revirt_flags & REPLAY_REPLAY)) {
    if(warn) {
      warn = 0;
      printk("WARNING: guest device checkpointing not possible when not replaying\n");
    }
    return NULL;
  }

  gdcp = kmalloc(sizeof(*gdcp), GFP_KERNEL);
  assert(gdcp != NULL);
  

  memcpy(gdcp, guestDev, sizeof(*gdcp));
  
  return gdcp;
#endif
  return NULL;
}

static void restoreGuestDevCheckpoint(void *cp) {
#ifndef DISABLE_GUEST_DEV_CP
  if(!(current->revirt_flags & REPLAY_REPLAY)) {
    return;
  }
  
  // XXX FIXME, free current resources first
  memcpy(guestDev, cp, sizeof(*guestDev));

  // XXX FIXME, reallocate any real reasources just added in, not sure about
  // how to force dma addresses though
#endif
}

static void deleteGuestDevCheckpoint(void *cp) {
#ifndef DISABLE_GUEST_DEV_CP
  if(!(current->revirt_flags & REPLAY_REPLAY)) {
    return;
  }

  // XXX FIXME, free resources for this checkpoint

  kfree(cp);
#endif
}
