#ifndef __UML_H__
#define __UML_H__

#include "../include/linux/ioctl.h"
#include "../include/revirt/rlog_min.h"
#include <asm/types.h>

#define INVALID_CHECK_ID ((unsigned long) -1)

struct uml_ioc_checkpoint_struct {
  unsigned long checkId; // from the device to the application
};


struct uml_ioc_rollback_struct {
  unsigned long checkId; // from the appilcation to the device
  unsigned long restore; // from the application to the device
  unsigned long long gdb_revbreak_syscall_count;
  unsigned long gdb_step_after_syscall;
};

struct uml_ioc_reset_cow_disk {
  char umlHda[1024];
  char cowBaseName[1024];
  unsigned long long startId;
  unsigned long long targetId;
  unsigned long long *idList;
  unsigned long numId;
};

struct uml_ioc_checkinfo_struct {
  // note, invalid check id here will give you the current info
  unsigned long checkId;                // from the app to the device
  struct revirt_tuple revirtId;         // from the device to the app
  unsigned long long guestSyscallCount; // from the device to the app
  unsigned long eip;
};

struct umlgdb_ioc_watch_struct {
  unsigned long set;   // either set or clear the watchpoint
  unsigned long addr;  // address of the data in question
  unsigned long regNo; // this can be either 1, 2, or 3
};

struct umlgdb_ioc_debug_guest_user_struct {
  unsigned long numPid;
  pid_t *pidArray;
};

#ifndef DEVICE_COUNT_RESOURCE
#define DEVICE_COUNT_RESOURCE 12
#endif

struct uml_resource {
  unsigned long start, end;
  unsigned long flags;
};

struct uml_get_pci_dev_struct {
  void * pci_dev_iter_inout;
  unsigned int irq_out;
  struct uml_resource resource[DEVICE_COUNT_RESOURCE];
  unsigned short vendor_out;
  unsigned short device_out;
  unsigned int class_out;
};

#define UML_PCI_CMD_ENABLE_DEVICE     0
#define UML_PCI_CMD_SET_MASTER        1
#define UML_PCI_CMD_ALLOC_DMA_BUFFER  2
#define UML_PCI_CMD_FREE_DMA_BUFFER   3
#define UML_PCI_CMD_SET_DMA_MASK      4
#define UML_PCI_CMD_ENABLE_IRQ        5
struct uml_pci_cmd_struct {
  unsigned long cmd_in;
  unsigned short vendor_in;
  unsigned short device_in;
  unsigned long long dma_mask_in;
  unsigned long phys_buff_in;
  unsigned long guest_vaddr_in;
  ssize_t dma_buff_size_in;
  unsigned int irq_in;
  unsigned long phys_buff_out;
  unsigned long host_vaddr_out;
};

struct uml_iomap_cmd_struct {
  unsigned long map; // 1 map, 0 free
  unsigned long phys_addr_in;
  unsigned long size_in;
  unsigned long vaddr_in;
  unsigned long vaddr_out;
};

#define PCI_WRITE_CONFIG_WORD 0

struct uml_pci_config_struct {
  unsigned short vendor_in;
  unsigned short device_in;
  unsigned long cmd_in;
  int where_in;
  unsigned long val_in;
  unsigned long val_out;
  int ret_out;
};

struct uml_sync_dma_struct {
  unsigned short vendor;
  unsigned short device;
  unsigned long dmaAddr;
  ssize_t size;
  int direction;
};

#define DEBUG_PRINT_CMD 0
#define DEBUG_WALK_FRAMELIST 1
#define DEBUG_PRINT_TD 2

struct uml_debug_struct {
  unsigned long cmd;
  unsigned long debugPrint;
  unsigned long addr;
};

struct uml_rw_struct {
	pid_t pid;
	char comm[16];
	unsigned long dataLen;
	char *buf;
};

#define MAX_UML_CHECK 50

#define UML_IOC_MAGIC 's'
#define UML_IOC_CHECKPOINT _IOR(UML_IOC_MAGIC,0,struct uml_ioc_checkpoint_struct)
#define UML_IOC_ROLLBACK   _IOW(UML_IOC_MAGIC,1,struct uml_ioc_rollback_struct)
#define UML_IOC_RESTART_UML  _IO(UML_IOC_MAGIC,2)
#define UML_IOC_RESET_COW_DISK _IOW(UML_IOC_MAGIC,3, struct uml_ioc_reset_cow_disk)
#define UML_IOC_SET_GDB_STATE _IOW(UML_IOC_MAGIC,4,unsigned long)
#define UML_IOC_GET_CHECKINFO _IOWR(UML_IOC_MAGIC,5,struct uml_ioc_checkinfo_struct)
#define UML_IOC_PAUSE         _IOW(UML_IOC_MAGIC,6,int)
#define UML_IOC_ERESET_UML  _IO(UML_IOC_MAGIC,7)
#define UML_IOC_REQUEST_IRQ _IOW(UML_IOC_MAGIC,8,int)
#define UML_IOC_PCI_CMD     _IOWR(UML_IOC_MAGIC,9, struct uml_pci_cmd_struct)
#define UML_IOC_IOMAP_CMD   _IOWR(UML_IOC_MAGIC,10,struct uml_iomap_cmd_struct)
#define UML_IOC_GET_PCI_DEV   _IOWR(UML_IOC_MAGIC,11,struct uml_get_pci_dev_struct)
#define UML_IOC_PCI_CONFIG    _IOWR(UML_IOC_MAGIC,12,struct uml_pci_config_struct)
#define UML_IOC_SYNC_DMA      _IOW(UML_IOC_MAGIC,13,struct uml_sync_dma_struct)
#define UML_IOC_DEBUG_CMD      _IOW(UML_IOC_MAGIC,14,struct uml_debug_struct)
#define UML_IOC_RW_CMD      _IOW(UML_IOC_MAGIC,15,struct uml_rw_struct)

#define UMLGDB_IOC_MAGIC 'g'
#define UMLGDB_IOC_RESET     _IO(UMLGDB_IOC_MAGIC,0)
#define UMLGDB_IOC_WATCH     _IOW(UMLGDB_IOC_MAGIC,1,struct umlgdb_ioc_watch_struct)
#define UMLGDB_IOC_CLEAR_TF_AFTER_SYSCALL _IO(UMLGDB_IOC_MAGIC,2)
#define UMLGDB_IOC_DEBUG_GUEST_USER _IOW(UMLGDB_IOC_MAGIC,3,struct umlgdb_ioc_debug_guest_user_struct)
#define UMLGDB_IOC_BREAK_NEXT     _IO(UMLGDB_IOC_MAGIC,4)


#define PLATO_IOC_MAGIC 'p'

#define PLATO_MODE_ON 1
#define PLATO_MODE_OFF 0
#define PLATO_IOC_SET_MODE _IO(PLATO_IOC_MAGIC, 0)

struct plato_mem_struct {
  unsigned long bufferAddr;
  unsigned long guestAddr;
  unsigned long len;
  unsigned long isKern;
  unsigned long ret;
};

#define PLATO_IOC_READ_MEM _IOWR(PLATO_IOC_MAGIC,1,struct plato_mem_struct)
#define PLATO_IOC_WRITE_MEM _IOWR(PLATO_IOC_MAGIC,2,struct plato_mem_struct)
#define PLATO_IOC_RESTART _IO(PLATO_IOC_MAGIC,3)
#define PLATO_IOC_SET_KERN_REGS _IOW(PLATO_IOC_MAGIC,4,struct pt_regs)

#define PLATO_IOC_CHECKPOINT     _IOR(PLATO_IOC_MAGIC,5,unsigned long)
#define PLATO_IOC_RESTORE_CHECK  _IO(PLATO_IOC_MAGIC,6)
#define PLATO_IOC_DELETE_CHECK   _IO(PLATO_IOC_MAGIC,7)
#define PLATO_IOC_SET_USER_REGS _IOW(PLATO_IOC_MAGIC,8,struct pt_regs)

#define PLATO_REPLAY_STATE_NONE 0
#define PLATO_REPLAY_STATE_LOGGING 1
#define PLATO_REPLAY_STATE_REPLAY 2
#define PLATO_IOC_QUERY_REPLAY_STATE _IOR(PLATO_IOC_MAGIC,9,unsigned long)
#define PLATO_IOC_QUERY_HOST_PID _IOR(PLATO_IOC_MAGIC,10,pid_t)
#define PLATO_IOC_QUERY_USER_MODE_TRAP _IOR(PLATO_IOC_MAGIC,11,unsigned long)
#define PLATO_IOC_QUERY_GUEST_KERN_REGS _IOR(PLATO_IOC_MAGIC,12,struct pt_regs)

#define PLATO_IOC_DBG_PRINT _IOW(PLATO_IOC_MAGIC,13,unsigned long)

struct plato_sigprocmask {
  int how;
  sigset_t *set;
  sigset_t *oldset;
  int ret;
};
#define PLATO_IOC_SIGPROCMASK _IOWR(PLATO_IOC_MAGIC, 14, struct plato_sigprocmask)


// for use in kernel mode modules, not user mode applications
#ifdef __KERNEL__
#define USER_REGS(p) (((struct pt_regs *) (p)->thread.esp0)-1) 
int atOrPastGdbRevbreakId(unsigned long guestEip);
void setCurrCheckinfo(struct task_struct *kernTsk, struct task_struct *userTsk);
void clearCurrCheckinfo(void);

typedef void (*gdbExitCallback)(void);
void setGdbExitCallback(gdbExitCallback cb);

void setTrapFlag(struct task_struct *tsk);
void clearTrapFlag(struct task_struct *tsk);
int isTrapFlagSet(struct task_struct *tsk);
void syncTrapFlag(struct task_struct *tsk);

typedef int (*guest_callback_sig)(struct task_struct *kernTsk);
void registerGuestCallbacks(guest_callback_sig gsysCb,
			    guest_callback_sig gretCb,
			    guest_callback_sig gtrapCb);
int umlDeviceSigHandler(struct pt_regs *regs, int signr, int preframe);

struct uml_check_struct *checkpoint(struct task_struct *kernTsk);
int restoreCheckpoint(struct task_struct *kernTsk, struct uml_check_struct *chk);
void deleteUmlCheckpoint(struct task_struct *kernTsk, unsigned long idx);


#define EFLAGS_TF 0x100
#endif

#endif
