#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/unistd.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/param.h>
#include <linux/fs.h>
#include <asm/uaccess.h>
#include <linux/slab.h>

#include <linux/ptrace.h>
#include <asm/ptrace.h>
#include <asm/signal.h>
#include <asm/io.h>
#include <linux/mman.h>
#include <linux/mm.h>
#include <linux/sched.h>

#include <asm/system.h>
#include <asm/segment.h>
#include <asm/desc.h>

#include <linux/vmalloc.h>
#include <asm/semaphore.h>

#include <asm/pgalloc.h>

#include <linux/file.h>
#include <revirt/checkpoint.h>
#include <asm/debugreg.h>

#include <linux/time.h>

#include "uml.h"
#include "myutil.h"

#define ETX_CHAR 3
#define BREAKPOINT_OPCODE 0xcc

// pasted in from /usr/include/sys/time.h to do timing measurements
# define timersub(a, b, result)                                               \
  do {                                                                        \
    (result)->tv_sec = (a)->tv_sec - (b)->tv_sec;                             \
    (result)->tv_usec = (a)->tv_usec - (b)->tv_usec;                          \
    if ((result)->tv_usec < 0) {                                              \
      --(result)->tv_sec;                                                     \
      (result)->tv_usec += 1000000;                                           \
    }                                                                         \
  } while (0)

// for making this a device driver
static int major;
static char name[] = "umlgdb";
static struct file_operations op;

static int umlgdb_open(struct inode *inode, struct file *file);
static int umlgdb_release(struct inode *i, struct file *f);
static ssize_t umlgdb_write(struct file *f, const char *buf, size_t len, loff_t *p);
static ssize_t umlgdb_read(struct file *f, char *buf, size_t len, loff_t *p);
static ssize_t plato_read(struct file *f, char *buf, size_t len, loff_t *p);
static int umlgdb_ioctl(struct inode *inode, struct file *f,
		     unsigned int cmd, unsigned long arg);

typedef int (*syscall_callback_sig)(struct pt_regs *regs);
typedef int (*signal_callback_sig)(struct pt_regs *regs, int signr, int preframe);

extern void setSignalCb(signal_callback_sig sigCb);
extern long do_sigprocmask(int how, old_sigset_t *set, old_sigset_t *oset, 
			   struct task_struct *tsk);

static int *guestUserDebugInitArray = NULL;
static pid_t *guestUserDebugPidArray = NULL;
static unsigned long numGuestUserDebugPid = 0;

void gdb_handle_exception(struct pt_regs *regs, int sigval, pid_t threadPid,
			  int gotSIGINT);

// io buffering
#define IO_BUFFER_SIZE 1024
struct buffer_list {
  unsigned char data[IO_BUFFER_SIZE];
  unsigned char *head;
  unsigned char *tail;
  struct semaphore list_sem;
};

static pid_t getGuestPid(struct task_struct *kernTsk);
static int write_to_buffer(struct buffer_list *list, const unsigned char *data,
                           unsigned long numBytes, int isUserData);
static int read_from_buffer(struct buffer_list *list, unsigned char *data,
                            unsigned long numBytes, int isUserData);
static struct buffer_list writeBuffer;
static struct buffer_list readBuffer;
static struct semaphore interface_lock;
static struct semaphore uml_sem;
static struct semaphore read_sem;
static struct semaphore plato_sem;
static int readBlocked = 0;
static int umlBlocked = 0;
static int debugOn = 0;
static int debugInit = 0;
static int debugSetInit = 0;
static int breakNextChance = 0;
static struct task_struct *gdbTsk = NULL;
extern unsigned long platoDbgPrint;

static int platoMode = PLATO_MODE_OFF;
static unsigned long gdbDeleteIdx = INVALID_CHECK_ID;
static unsigned long gdbRestoreIdx = INVALID_CHECK_ID;
static unsigned long gdbCheckIdx = INVALID_CHECK_ID;
static unsigned long platoUserModeTrap = 0;

typedef int (*is_int3_cb_t)(struct task_struct * tsk, 
			    struct pt_regs *regs);
extern int registerPlatoInt3Cb(is_int3_cb_t cb);

static void dummy_func(void) {
  // dont do anything, we will just use the address to get an idea of 
  // where the beginning of the file is
}

int isGuestKernelDebug(void) {
  return numGuestUserDebugPid == 0;
}

static int getGuestUserDebugIdx(pid_t pid, unsigned long *idx) {
  unsigned long i;
  for(i = 0; i < numGuestUserDebugPid; i++) {
    if(guestUserDebugPidArray[i] == pid) {
      if(idx != NULL) *idx = i;
      return 1;
    }
  }
  return 0;
}

static void setGuestUserDebugInit(pid_t pid, int init) {
  unsigned long idx;
  if(getGuestUserDebugIdx(pid, &idx)) {
    guestUserDebugInitArray[idx] = init;
  }
}

static int isGuestUserDebugInit(pid_t pid) {
  unsigned long idx;
  if(getGuestUserDebugIdx(pid, &idx)) {
    return guestUserDebugInitArray[idx];
  }

  return 0;
}

static int isGuestUserDebugPid(pid_t pid) {
  return getGuestUserDebugIdx(pid, NULL);
}

#define DBG_LEN(reg,len) (((len)-1) << (18 + (reg)*4))

#define DBG_RW_DW 0x1
#define DBG_RW(reg,mode)  ((mode) << (16 + (reg)*4))

#define DBG_L(reg) (0x1 << ((reg) * 2))

static void setWatchAddr(unsigned long dbgNo, unsigned long addr) {
  // XXX FIXME check the var align to make this bigger when possible
  unsigned long dataLen = 1;
  if(gdbTsk == NULL) {
    printk("error, gdbTsk == NULL\n");
    return;
  }

  assert((dbgNo >= 1) && (dbgNo <= 3));

  gdbTsk->thread.debugreg[dbgNo] = addr;
  
  gdbTsk->thread.debugreg[7] |= (DBG_L(dbgNo) | 
				 DBG_RW(dbgNo,DBG_RW_DW) | 
				 DBG_LEN(dbgNo,dataLen));
}

static void clearWatchAddr(unsigned long dbgNo) {
  unsigned long dataLen = 4; // this will clear the whole thing

  if(gdbTsk == NULL) {
    printk("error, gdbTsk == NULL\n");
    return;
  }

  assert((dbgNo >= 1) && (dbgNo <= 3));

  gdbTsk->thread.debugreg[dbgNo] = 0;

  gdbTsk->thread.debugreg[7] &= ~(DBG_L(dbgNo) | 
				  DBG_RW(dbgNo,DBG_RW_DW) | 
				  DBG_LEN(dbgNo,dataLen));  
}

static void platoWakeSleepingReaders(void) {
  if(readBlocked) {
    readBlocked = 0;
    up(&plato_sem);
  }
}

static void llTrapToGdb(struct pt_regs *regs, int sigval, int gotSIGINT,
			unsigned long isFromUserMode) {
  struct task_struct * tsk=current;
  pid_t threadPid;
  int saved_flags;

  if(platoDbgPrint) printk("llTrapToGdb proc %d\n", tsk->pid);
  platoUserModeTrap = isFromUserMode;
  /* We still need this here to keep the kernel process from waking up */
  revirt_suppress_sched(tsk, saved_flags);
  if(!current->uml->isKernel) {
    threadPid = getGuestPid(current->uml->kernelTsk);
    setCurrCheckinfo(current->uml->kernelTsk, current);
    gdbTsk = current->uml->kernelTsk;
  } else {
    threadPid = 0;
    setCurrCheckinfo(current, NULL);
    gdbTsk = current;
  }
  
  if(platoMode == PLATO_MODE_OFF) {
    gdb_handle_exception(regs,sigval,threadPid, gotSIGINT);
  } else {
    while(1) {
      down(&interface_lock);
      umlBlocked = 1;
      // wake up any waiting readers
      platoWakeSleepingReaders();
      up(&interface_lock);

      down(&uml_sem);
      
      // check for any waiting commands
      if(gdbCheckIdx != INVALID_CHECK_ID) {
	assert(0);
	//gdbTsk->uml->checkArray[gdbCheckIdx] = checkpoint(gdbTsk);

	gdbCheckIdx = INVALID_CHECK_ID;
	platoWakeSleepingReaders();
      } else if(gdbRestoreIdx != INVALID_CHECK_ID) {	
	assert(0);
	//restoreCheckpoint(gdbTsk, gdbTsk->uml->checkArray[gdbRestoreIdx]);

	gdbRestoreIdx = INVALID_CHECK_ID;
	platoWakeSleepingReaders();
      } else if(gdbDeleteIdx != INVALID_CHECK_ID) {
	assert(0);
	//deleteUmlCheckpoint(gdbDeleteIdx);
	
	gdbDeleteIdx = INVALID_CHECK_ID;
	platoWakeSleepingReaders();
      } else {
	break;
      }
    } 

    // we are back
    down(&interface_lock);
    gdbTsk = NULL;
    if(readBlocked) {
      readBlocked = 0;
      up(&plato_sem);
    }
    up(&interface_lock);
  }
    
  gdbTsk = NULL;
  clearCurrCheckinfo();

  syncTrapFlag(current);
  revirt_restore_sched(tsk, saved_flags);
}

static void trapToGdbUserMode(struct pt_regs *regs, int sigval, int gotSIGINT) {
  llTrapToGdb(regs, sigval, gotSIGINT, 1);
}

static void trapToGdb(struct pt_regs *regs, int sigval, int gotSIGINT) {
  llTrapToGdb(regs, sigval, gotSIGINT, 0);
}

static int isBreakPoint(struct pt_regs *regs) {
  struct gdb_mod_mem *gdb_mem;
  unsigned char c=0;
  int ret;

  assert(current->uml);

  // read in the opcode
  ret = copy_from_user(&c, (void *) (regs->eip - 1), sizeof(c));
  assert(ret == 0);

  if(platoMode == PLATO_MODE_ON) {
    // all int3's are breakpoints
    return c == BREAKPOINT_OPCODE;
  }

  gdb_mem = current->uml->gdb_mem_list;
  while(gdb_mem != NULL) {
    if(gdb_mem->addr == (void *) (regs->eip - 1)) {
      // read the opcode and check if there is a int3 instruction there
      if(c == BREAKPOINT_OPCODE) {
	return 1;
      }
    }
    gdb_mem = gdb_mem->next;
  }

  // check if it was a hardware break
  if(current->thread.debugreg[6] & (DR_TRAP1|DR_TRAP2|DR_TRAP3)) {
    //printk("hdwatch debugreg[6] = 0x%08lx\n",current->thread.debugreg[6]);
    return 1;
  }

  return 0;
}

static int umlGuestKernelSignal(struct pt_regs *regs, int signr, int preframe) {
  int ret = signr;
  int breakPoint = 0;
  int gotSIGINT = 0;
  int forceTrap = 0;

  struct timeval tvbefore;
  struct timeval tvafter; 
  struct timeval tvdifference;

  if(platoDbgPrint) printk("guest kernel signal %d, preframe = %d\n",signr, preframe);

  // figure out if this is a breakpoint
  if(!preframe) {
    // check the trap flags
    if(isTrapFlagSet(current)) {
      if((regs->eflags & EFLAGS_TF) == 0) {
	printk("warning, trap flag cleared, resetting before signal(%d) delivery\n",
	       signr);
	regs->eflags |= EFLAGS_TF;
      }
    }
  } else if(signr == SIGTRAP) {
    assert(preframe);
    if(isTrapFlagSet(current) || isBreakPoint(regs)) {
      breakPoint = 1;
    } else {
      unsigned long opcode = 0;
      copy_from_user(&opcode, (void *) (regs->eip - 2), sizeof(opcode));
      printk("warning, unknown trap to gdb, eip = 0x%08lx opcode(eip-2) = 0x%08lx\n"
	     "eflags = 0x%08lx signr %d, pid = %d, eax = %lu\n",
	     regs->eip, opcode, regs->eflags, signr, current->pid, regs->orig_eax);
      {
	unsigned long myidx;
	unsigned long stackData = 0;
	printk("stack contents:\n");
	for(myidx = 0; myidx < 256; myidx++) {
	  stackData = 0;
	  copy_from_user(&stackData, (void *) (regs->esp + (myidx * 4)), sizeof(stackData));
	  if((stackData > 0xa0000000) && (stackData < 0xa1000000)) {
	    printk("0x%08lx: 0x%08lx\n",regs->esp + (myidx * 4),stackData);
	  }
	}
      }
    }
  } else if(signr == SIGUSR1) {
    current->thread.trap_no = 0;
  }

  if((breakPoint || debugInit || breakNextChance) && debugOn && preframe) {
    if(debugInit) {
      debugInit = 0;
      forceTrap = 1;
    } 
    if(breakNextChance) {
      breakNextChance = 0;
      gotSIGINT = 1;
      forceTrap = 1;
    } 

    if(breakPoint) {
      ret = 0;
    }

    if(isTrapFlagSet(current) || isBreakPoint(regs) || forceTrap) {
      if(forceTrap) {
        // start time
        do_gettimeofday(&tvbefore);
      }
      trapToGdb(regs, signr, gotSIGINT);
      if(forceTrap) {
        // end time
        do_gettimeofday(&tvafter);
        timersub(&tvafter, &tvbefore, &tvdifference);
        printk("Time for trap: %d sec %d usec\n", tvdifference.tv_sec, tvdifference.tv_usec);
      }
    } else if(!isTrapFlagSet(current) && (regs->eflags & EFLAGS_TF)) {
      printk("warning, we got a single step that we were not expecting, clear it\n");
      regs->eflags &= ~EFLAGS_TF;
    } else {
      unsigned long opcode = 0;
      copy_from_user(&opcode, (void *) (regs->eip - 2), sizeof(opcode));
      printk("error, unknown trap to gdb, eip = 0x%08lx opcode(eip-2) = 0x%08lx\n"
	     "eflags = 0x%08lx signr %d\n",
	     regs->eip, opcode, regs->eflags, signr);
      assert(0);
    }
  }
  
  return ret;
}

static int umlGuestUserSignal(struct pt_regs *regs, int signr, int preframe) {
  int ret;
  unsigned long oldEip, newEip;

  ret = signr;

  if(isBreakPoint(regs)) {
    trapToGdbUserMode(regs, signr, 0);
    return 0;
  }

  if((platoMode == PLATO_MODE_ON) && (signr == SIGTRAP)) {
    // this is probably a bug if the code reaches here
    printk("WARNING, guest user SIGTRAP platoMode == on, did not stop\n");
  }

  if((current->thread.debugreg[6] & (DR_TRAP1|DR_TRAP2|DR_TRAP3)) ||
     isTrapFlagSet(current)) {
    if(!isTrapFlagSet(current)) {
      //correct so gdb still thinks it is a software breakpoint      
      oldEip = regs->eip;
      regs->eip++;
      newEip = regs->eip;
      trapToGdbUserMode(regs, signr, 0);
      if(regs->eip == newEip) {
	regs->eip = oldEip;
      }
    } else {
      trapToGdbUserMode(regs, signr, 0);
    }


    ret = 0;
  }

  if(isTrapFlagSet(current) && !(regs->eflags & EFLAGS_TF)) {
    printk("warning: resetting trap flag at signal for user debugging\n");
    syncTrapFlag(current);
  }

  return ret;
} 
  

static int umlSignalHandler(struct pt_regs *regs, int signr, int preframe) {
  int ret;

  ret = signr;

  if(current->uml && current->uml->isKernel) {
    ret = umlGuestKernelSignal(regs, signr, preframe);
  } else if(current->uml && current->uml->isUser) {
    if((signr == SIGTRAP) && (current->uml->syscall_trace_sigtrap)) {
      current->uml->syscall_trace_sigtrap = 0;
      // do not modify the signal, let it pass
    } else {
      if((signr == SIGTRAP) && !isGuestKernelDebug() &&
	 isGuestUserDebugPid(getGuestPid(current->uml->kernelTsk))) {
	ret = umlGuestUserSignal(regs, signr, preframe);
      } else if((signr == SIGTRAP) && 
		((current->thread.debugreg[6] & (DR_TRAP1|DR_TRAP2|DR_TRAP3)) ||
		 isTrapFlagSet(current))) {
	// we got a breakpoint on another process, kill the signal and
	// try to let it replay, I think this is where we would use the rf
	// but it seems that someone is doing it for us
	ret = 0;
      } else if((signr == SIGTRAP) && (platoMode == PLATO_MODE_ON)) {
	// all traps in plato mode go the debugger
	ret = umlGuestUserSignal(regs, signr, preframe);
      }
    }

    if(isTrapFlagSet(current)) {
      /*
      unsigned char c;
      int myret;
      myret = copy_from_user(&c, (void *) regs->eip, sizeof(c));
      // check if the next instruction is a int 0x80
      if((myret == 0) && (c == 0xcd)) {
	printk("clearing trap flag for system call during single step\n");
	regs->eflags &= ~EFLAGS_TF;
      }
      */
    }

  }


  return ret ? umlDeviceSigHandler(regs, ret, preframe) : ret;
}

static void umlGdbExitCb(void) {
  if(debugOn) {
    trapToGdb(USER_REGS(current), SIGKILL, 0);
  }
}

// this is a bigtime hack, too bad
static pid_t getGuestPid(struct task_struct *kernTsk) {
  int ret;
  pid_t pid;
  unsigned long guestCurrent = (unsigned long) (USER_REGS(kernTsk)->esp & (~0x3fff));
  
  pid = 0;
  ret = access_process_vm(kernTsk, guestCurrent + 124,
			  &pid, sizeof(pid), 0);
  assert(ret == sizeof(pid));

  return pid;
}

static int guestSyscallRetCb(struct task_struct *kernTsk) {  
  pid_t pid = getGuestPid(kernTsk);
  //unsigned long idx;
  if(pid && isGuestUserDebugPid(pid) && !isGuestUserDebugInit(pid)) {
    setGuestUserDebugInit(pid, 1);

    printk("found guest user pid = %d, trapping to gdb\n",pid);
    trapToGdb(USER_REGS(kernTsk), SIGTRAP, 0);
  } 

  if(isTrapFlagSet(current) && !(USER_REGS(current)->eflags & EFLAGS_TF)) {
    printk("warning: resetting trap flag for user debugging\n");
    syncTrapFlag(current);
  }

  return 0;   
}

// for now we are assuming any int3's from the guest are from
// plato when plato is active
static int platoInt3Cb(struct task_struct *tsk, struct pt_regs *regs) {
  return (platoMode == PLATO_MODE_ON) && (tsk->uml);
}

/**
 * linux module stuff
 */
int init_module(void) 
{
  readBuffer.head = NULL;
  readBuffer.tail = NULL;
  sema_init(&readBuffer.list_sem,1);
  writeBuffer.head = NULL;
  writeBuffer.tail = NULL;
  sema_init(&writeBuffer.list_sem,1);

  setSignalCb(umlSignalHandler);
  setGdbExitCallback(umlGdbExitCb);

  sema_init(&interface_lock,1);
  sema_init(&uml_sem,0);
  sema_init(&read_sem,0);  
  sema_init(&plato_sem,0);  

  registerGuestCallbacks(NULL,guestSyscallRetCb, NULL);
  registerPlatoInt3Cb(platoInt3Cb);

  /**
   * set up the device driver functions
   */
  memset(&op, 0, sizeof(op));
  
  op.open = &umlgdb_open;
  op.release = &umlgdb_release;
  op.write = &umlgdb_write;
  op.read = &umlgdb_read;
  op.ioctl = &umlgdb_ioctl;
  
  major = register_chrdev(0,name,&op);
  if(major < 0)
    return -1;

  printk("umlgdb module loaded, dummy_func is at 0x%08lX\n",
	 (unsigned long) dummy_func); 

  return 0;
}

void cleanup_module(void)
{
  /**
   * make sure you nullify the callbacks, the kernel will (probably) crash
   * if you do not.
   */
  setSignalCb(NULL);
  setGdbExitCallback(NULL);
  registerPlatoInt3Cb(NULL);

  if(guestUserDebugInitArray != NULL) util_free_small(guestUserDebugInitArray);
  if(guestUserDebugPidArray != NULL) util_free_small(guestUserDebugPidArray);

  // unregister the device
  unregister_chrdev(major,name);
}

/**
 * these are the device driver functions, they are used so that user mode applications
 * can read the system call and other data.
 */
static int umlgdb_open(struct inode *inode, struct file *file) {
  file->f_pos = 0;
  MOD_INC_USE_COUNT;

  down(&interface_lock);

  if(!debugOn) {
    readBuffer.head = NULL;
    readBuffer.tail = NULL;
    sema_init(&readBuffer.list_sem,1);
    writeBuffer.head = NULL;
    writeBuffer.tail = NULL;
    sema_init(&writeBuffer.list_sem,1);
    debugSetInit = 1;
  }

  debugOn = 1;

  up(&interface_lock);

  return 0;
}

static int umlgdb_release(struct inode *i, struct file *f) {
  MOD_DEC_USE_COUNT;

  if(!MOD_IN_USE) {
    debugOn = 0;
  }
  return 0;
}

static ssize_t umlgdb_write(struct file *f, const char *buf, size_t len, loff_t *p) {
  int ret;
  char tmpChar;

  if(debugSetInit) {
    debugSetInit = 0;
    debugInit = 1;
  }

  if(len == sizeof(tmpChar)) {
    if(copy_from_user(&tmpChar, buf, sizeof(tmpChar)) == 0) {
      if(tmpChar == ETX_CHAR) {
	breakNextChance = 1;
      }
    }
  }

  down(&interface_lock);
  ret = write_to_buffer(&writeBuffer, buf, len, 1);
  if(ret <= 0) {
    printk("warning, write buffer is full!!!\n");    
    return 0;
  }
  if(umlBlocked) {
    umlBlocked = 0;  
    up(&uml_sem);
  }
  up(&interface_lock);

  return ret;
}

static void plato_handle_mem(struct plato_mem_struct *platoMem, 
			     unsigned long platoCmd) {
  int ret;
  unsigned char *buf;
  struct task_struct *tsk;
  int write;

  platoMem->ret = 0;
  assert(platoMem->len > 0);
  assert(gdbTsk != NULL);
  assert(gdbTsk->uml);
  assert(gdbTsk->uml->isKernel);

  if(platoMem->isKern) {
    tsk = gdbTsk;
  } else {
    // we have to have at least one child proc...
    assert(gdbTsk->uml->userList != NULL);
    assert(gdbTsk->uml->userList->tsk != NULL);
    if(gdbTsk->uml->userList->next != NULL) {
      printk("******** WARNING ********\n");
      printk("more than one user process, using the first...\n");
      printk("This is probably not what you want.\n");
    }
    tsk = gdbTsk->uml->userList->tsk;
  }

  if(platoMem->len > 2*PAGE_SIZE) {
    buf = util_alloc_small(platoMem->len);
  } else {
    buf = util_alloc_large(platoMem->len);
  }
  assert(buf != NULL);
  write = (platoCmd == PLATO_IOC_READ_MEM) ? 0 : 1;

  /*
  printk("PLATO: mem ");
  if(write) {
    printk("write ");
  } else {
    printk("read ");
  }
  printk("at guest addr 0x%08lx, user address 0x%08lx, len = %lu\n",
	 platoMem->guestAddr, platoMem->bufferAddr, platoMem->len);
  */

  if(write) {
    ret = copy_from_user(buf, (void *) platoMem->bufferAddr, platoMem->len);
    assert(ret == 0);
  }

  ret = access_process_vm(tsk, platoMem->guestAddr, buf, platoMem->len, write);
  if(ret != platoMem->len) {
    /*
    printk("could not %s guest memory starting at 0x%08lx len = %lu\n",
	   (write) ? "write" : "read", platoMem->guestAddr,
	   platoMem->len);
    */
    if(ret == 0) {
      //printk("ret == 0, adjusting...\n");
      platoMem->ret = 1;
    } else {
      platoMem->ret = ret;
    }
  }

  if(!write && (platoMem->ret == 0)) {
    ret = copy_to_user((void *) platoMem->bufferAddr, buf, platoMem->len);
    assert(ret == 0);
  }

  if(platoMem->len > 2*PAGE_SIZE) {
    util_free_small(buf);
  } else {
    util_free_large(buf);
  }

}

static ssize_t plato_read(struct file *f, char *buf, size_t len, loff_t *p) {
  int ret;

  down(&interface_lock);
  while(gdbTsk == NULL) {
    readBlocked = 1;
    up(&interface_lock);

    if(down_interruptible(&plato_sem) == -EINTR) {
      // ??? do we need to lock something
      readBlocked = 0;
      return -EINTR;
    }

    down(&interface_lock);
  }

  if(platoUserModeTrap) {
    assert(gdbTsk->uml && gdbTsk->uml->isKernel);
    ret = copy_to_user(buf, USER_REGS(gdbTsk->uml->userList->tsk),
		       sizeof(struct pt_regs));
    if(gdbTsk->uml->userList->next != NULL) {
      printk("WARNING: there are more than one user mode procs, using the first\n");
    }
  } else {
    ret = copy_to_user(buf, USER_REGS(gdbTsk), sizeof(struct pt_regs));
  }
  up(&interface_lock);

  return (ret == 0) ? sizeof(struct pt_regs) : -EFAULT;
}
static ssize_t umlgdb_read(struct file *f, char *buf, size_t len, loff_t *p) {
  int ret;

  if(debugSetInit) {
    debugSetInit = 0;
    debugInit = 1;
  }

  if(platoMode == PLATO_MODE_ON) {
    return plato_read(f, buf, len, p);
  }

  down(&interface_lock);
  while((ret = read_from_buffer(&readBuffer, buf, len, 1)) == 0) {
    readBlocked = 1;
    up(&interface_lock);

    if(down_interruptible(&read_sem) == -EINTR) {
      // ??? do we need to lock something
      readBlocked = 0;
      return -EINTR;
    }

    down(&interface_lock);
  }

  up(&interface_lock);

  return ret;
}

extern void setUserCheckInfo(int isUser);

static void platoWaitForReply(void) {
  assert(umlBlocked);

  down(&interface_lock);
  umlBlocked = 0;
  readBlocked = 1;
  up(&interface_lock);
  
  up(&uml_sem);
  
  down(&plato_sem);
}

static void platoIocDeleteCheck(unsigned long checkIdx) {
  if(gdbTsk == NULL) {
    printk("cannot delete checkpoint when vm is running\n");
    return;
  }

  if((checkIdx >= MAX_UML_CHECK) || (gdbTsk->uml->checkArray[checkIdx] == NULL)) {
    printk("cannot delete this checkpoint\n");
    return;
  }

  deleteUmlCheckpoint(gdbTsk, checkIdx);
}

static void platoIocRestoreCheck(unsigned long checkIdx) {
  if(gdbTsk == NULL) {
    printk("cannot restore checkpoint when vm is running\n");
    return;
  }

  if((checkIdx >= MAX_UML_CHECK) || (gdbTsk->uml->checkArray[checkIdx] == NULL)) {
    printk("cannot delete this checkpoint\n");
    return;
  }

  restoreCheckpoint(gdbTsk, gdbTsk->uml->checkArray[checkIdx]);
}    

static unsigned long platoIocCheckpoint(void) {
  unsigned long idx;
  unsigned long checkIdx = INVALID_CHECK_ID;

  if(gdbTsk == NULL) {
    printk("cannot take checkpoint when vm is running\n");
    return INVALID_CHECK_ID;
  }
  
  assert(gdbTsk->uml->isKernel);

  // check to make sure there are not any other checkpoints at this
  // guest syscall
  for(idx = 0; idx < MAX_UML_CHECK; idx++) {
    if(gdbTsk->uml->checkArray[idx] != NULL) {
      if(gdbTsk->uml->checkArray[idx]->guestSyscallCount == 
	 gdbTsk->uml->guestSyscallCount) {
	printk("ERROR: already a checkpoint at this guest syscall count\n");
	return INVALID_CHECK_ID;
      }
    }
  }

  // reserve a checkpoint slot
  for(idx = 0; idx < MAX_UML_CHECK; idx++) {
    if(gdbTsk->uml->checkArray[idx] == NULL) {
      checkIdx = idx;
      gdbTsk->uml->checkArray[idx] = checkpoint(gdbTsk);
      if(platoUserModeTrap) {
	assert(gdbTsk->uml->userList->tsk != NULL);
	gdbTsk->uml->checkArray[idx]->activeTsk = gdbTsk->uml->userList->tsk;
	if(gdbTsk->uml->userList->next != NULL) {
	  printk("WARNING: there are more than one user mode procs, using the first\n");
	}
      }
      break;
    }
  }

  if(checkIdx == INVALID_CHECK_ID) {
    printk("no free checkpoints, checkpoint not taken\n");
    return checkIdx;
  }

  return checkIdx;
}

static int umlgdb_ioctl(struct inode *inode, struct file *f,
			 unsigned int cmd, unsigned long arg) {

  int ret;
  struct umlgdb_ioc_watch_struct w;
  struct umlgdb_ioc_debug_guest_user_struct dbu;
  struct plato_mem_struct platoMem;
  struct plato_sigprocmask platoSigProc;
  unsigned long replayState;

  if(cmd == UMLGDB_IOC_RESET) {
    down(&interface_lock);
    if(umlBlocked) {
      umlBlocked = 0;
      up(&uml_sem);
    }
    debugOn = 0;
    up(&interface_lock);
  } else if(cmd == PLATO_IOC_SIGPROCMASK) {
    assert(gdbTsk != NULL); assert(gdbTsk->uml && gdbTsk->uml->isKernel);
    ret = copy_from_user(&platoSigProc, (void *) arg, sizeof(platoSigProc));
    assert(ret == 0);
    platoSigProc.ret = do_sigprocmask(platoSigProc.how,
				      (old_sigset_t *) platoSigProc.set,
				      (old_sigset_t *) platoSigProc.oldset, 
				      gdbTsk);
    ret = copy_to_user((void *) arg, &platoSigProc, sizeof(platoSigProc));
    assert(ret == 0);
  } else if(cmd == PLATO_IOC_DBG_PRINT) {
    platoDbgPrint = arg;
  } else if(cmd == PLATO_IOC_QUERY_USER_MODE_TRAP) {
    assert(gdbTsk != NULL); assert(gdbTsk->uml);
    ret = copy_to_user((void *) arg, &platoUserModeTrap, sizeof(platoUserModeTrap));
    assert(ret == 0);
  } else if(cmd == PLATO_IOC_QUERY_HOST_PID) {
    assert(gdbTsk != NULL); assert(gdbTsk->uml);
    ret = copy_to_user((void *) arg, &gdbTsk->pid, sizeof(gdbTsk->pid));
    assert(ret == 0);
  } else if(cmd == PLATO_IOC_QUERY_GUEST_KERN_REGS) {
    assert(gdbTsk != NULL); assert(gdbTsk->uml); assert(gdbTsk->uml->isKernel);
    ret = copy_to_user((void *) arg, USER_REGS(gdbTsk), sizeof(struct pt_regs));
    assert(ret == 0);
  } else if(cmd == PLATO_IOC_QUERY_REPLAY_STATE) {
    assert(gdbTsk != NULL); assert(gdbTsk->uml);
    replayState = PLATO_REPLAY_STATE_NONE;
    if(gdbTsk->revirt_flags & REPLAY_REPLAY) {
      replayState = PLATO_REPLAY_STATE_REPLAY;
    } else if(gdbTsk->revirt_flags & REPLAY_LOG) {
      replayState = PLATO_REPLAY_STATE_LOGGING;
    }
    ret = copy_to_user((void *) arg, &replayState, sizeof(unsigned long));
    assert(ret == 0);
  } else if(cmd == PLATO_IOC_SET_USER_REGS) {
    assert(gdbTsk != NULL); assert(gdbTsk->uml); assert(gdbTsk->uml->isKernel);
    assert(gdbTsk->uml->userList && gdbTsk->uml->userList->tsk);
    if(gdbTsk->uml->userList->next != NULL) {
      printk("warning, more that one uml user mode process, using the first\n");
    }
    ret = copy_from_user(USER_REGS(gdbTsk->uml->userList->tsk), (void *) arg,
			 sizeof(struct pt_regs));
    assert(ret == 0);
  } else if(cmd == PLATO_IOC_SET_KERN_REGS) {
    //printk("PLATO: setting kernel registers\n");
    assert(gdbTsk != NULL); assert(gdbTsk->uml); assert(gdbTsk->uml->isKernel);
    ret = copy_from_user(USER_REGS(gdbTsk), (void *) arg, sizeof(struct pt_regs));
    assert(ret == 0);
  } else if(cmd == PLATO_IOC_SET_MODE) {
    printk("PLATO: entering plato mode %lu\n",arg);
    platoMode = arg;
    if(arg == PLATO_MODE_ON) {
      debugSetInit = 1;
    }
  } else if(cmd == PLATO_IOC_RESTART) {
    //printk("PLATO: restarting vm\n");
    if(umlBlocked) {
      platoWaitForReply();
    }
  } else if((cmd == PLATO_IOC_READ_MEM) || (cmd == PLATO_IOC_WRITE_MEM)) {    
    ret = copy_from_user(&platoMem, (void *) arg, sizeof(platoMem));
    assert(ret == 0);
    plato_handle_mem(&platoMem, cmd);
    ret = copy_to_user((void *) arg, &platoMem, sizeof(platoMem));
    assert(ret == 0);
  } else if(cmd == UMLGDB_IOC_WATCH) {
    ret = copy_from_user(&w, (void *) arg, sizeof(w));
    assert(ret == 0);
    if(w.set) {
      setWatchAddr(w.regNo, w.addr);
    } else {
      clearWatchAddr(w.regNo);
    }
  } else if(cmd== UMLGDB_IOC_CLEAR_TF_AFTER_SYSCALL) {
    assert(gdbTsk != NULL);
    assert(gdbTsk->uml != NULL);
    //assert(gdbTsk->uml->isKernel);
    if(gdbTsk->uml->gdb_step_set) {
      gdbTsk->uml->gdb_step_set = 0;
      printk("umlgdb_ioctl: clearing trap flag after syscall step\n");
      clearTrapFlag(gdbTsk);
    } else {
      printk("umlgdb_ioctl: gdb_step_set not set, tf already clear\n");
    }
  } else if (cmd == UMLGDB_IOC_DEBUG_GUEST_USER) {
    assert(gdbTsk != NULL);
    assert(gdbTsk->uml);
    ret = copy_from_user(&dbu,(void *) arg, sizeof(dbu));
    assert(ret == 0);
    assert(dbu.numPid > 0);

    guestUserDebugPidArray = (pid_t *) util_alloc_small(sizeof(pid_t) * dbu.numPid);
    ret = copy_from_user(guestUserDebugPidArray, dbu.pidArray, 
			 sizeof(pid_t) * dbu.numPid);
    assert(ret == 0);

    guestUserDebugInitArray = (int *) util_alloc_small(sizeof(int) * dbu.numPid);
    memset(guestUserDebugInitArray, 0, sizeof(int) * dbu.numPid);

    numGuestUserDebugPid = dbu.numPid;
    setUserCheckInfo(1);
  } else if(cmd == UMLGDB_IOC_BREAK_NEXT) {
    breakNextChance = 1;
  } else if(cmd == PLATO_IOC_CHECKPOINT) {
    unsigned long cpId = platoIocCheckpoint();
    ret = copy_to_user((unsigned long *) arg, &cpId, sizeof(cpId));
    if(ret != 0) {
      printk("Took checkpoint %lu, could not return value to user\n", cpId);
      return -1;
    }
  } else if(cmd == PLATO_IOC_RESTORE_CHECK) {
    platoIocRestoreCheck(arg);
  } else if(cmd == PLATO_IOC_DELETE_CHECK) {
    platoIocDeleteCheck(arg);
  } else {
    printk("umlgdb_ioctl: unknown ioctl %u\n",cmd);
  }

  return 0;
}



static int write_to_buffer(struct buffer_list *list, const unsigned char *data, 
			   unsigned long numBytes, int isUserData) {
  unsigned long idx;

  down(&list->list_sem);
  if((list->head == NULL) || (list->tail == NULL)) {
    list->head = list->data;
    list->tail = list->data;
  } 

  for(idx = 0; idx < numBytes; idx++) {
    if(!isUserData) {
      *list->head = data[idx];
    } else {
      copy_from_user(list->head, data+idx, 1);
    }

    if(++list->head >= (list->data + IO_BUFFER_SIZE)) {
      list->head = list->data;
    }

    if(list->head == list->tail) {
      printk("warning, buffer full\n");
      up(&list->list_sem);
      return idx;
    }
  }

  up(&list->list_sem);
  return idx;
}
    
static int read_from_buffer(struct buffer_list *list, unsigned char *data,
			    unsigned long numBytes, int isUserData) {
  unsigned long idx;

  down(&list->list_sem);

  if((list->head == NULL) || (list->tail == NULL)) {
    list->head = list->data;
    list->tail = list->data;
  }

  for(idx = 0; idx < numBytes; idx++) {
    if(list->head == list->tail) {
      up(&list->list_sem);
      return idx;
    }

    if(!isUserData) {
      data[idx] = *list->tail;
    } else {
      copy_to_user(data+idx,list->tail,1);
    }

    if(++list->tail >= (list->data + IO_BUFFER_SIZE)) {
      list->tail = list->data;
    }
  }

  up(&list->list_sem);
  return idx;
}

void putDebugChar(char c) {
  int ret;

  if(!debugOn) return;

  down(&interface_lock);
  ret = write_to_buffer(&readBuffer, &c, sizeof(char), 0);
  if(ret <= 0) {
    printk("putDebugChar: read buyffer is full\n");
    up(&interface_lock);
    return;
  }
  if(readBlocked) {
    readBlocked = 0;
    up(&read_sem);
  }
  up(&interface_lock);
}

int getDebugChar(void) {
  char c;
  
  down(&interface_lock);
  while((read_from_buffer(&writeBuffer, &c, sizeof(char), 0) == 0) && debugOn) {    
    umlBlocked = 1;
    up(&interface_lock);

    down(&uml_sem);
    down(&interface_lock);
  }

  up(&interface_lock);

  if(!debugOn) {
    return -1;    
  }

  return c;
}
