// this is for the kernel implementation of some algorithms
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/unistd.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/param.h>
#include <linux/fs.h>
#include <linux/slab.h>

#include <asm/uaccess.h>

#include <asm/signal.h>
#include <linux/mman.h>

#include <asm/system.h>
#include <asm/segment.h>

#include <asm/semaphore.h>

#include "myutil.h"

static void downheap(struct sort_element *a, int k, int N) {
  struct sort_element T = a[k - 1];
  int j;
  while (k <= N/2) {
    j = k + k;
    if ((j < N) && (a[j - 1].key < a[j].key)) {
      j++;
    }
    if (T.key >= a[j - 1].key) {
      break;
    } else {
      a[k - 1] = a[j - 1];
      k = j;
    }
  }
  a[k - 1] = T;
}

void util_heap_sort(struct sort_element *a, int array_size) {
  int N = array_size;
  int k;
  struct sort_element T;

  if(N <= 1) return;

  for (k = N/2; k > 0; k--) {
    downheap(a, k, N);
  }
  do {
    T = a[0];
    a[0] = a[N - 1];
    a[N - 1] = T;
    N = N - 1;
    downheap(a, 1, N);
  } while (N > 1);

  // XXX this is a parinoid check to make sure sorting works, comment it
  // out after you are sure that this is working
  for(k = 1; k < array_size; k++) {
    if(a[k].key < a[k-1].key) {
      printk("a[%d].key = %lld, a[%d].key = %lld\n",k,a[k].key,k-1,a[k-1].key);
      for(k=0;k < array_size; k++) {
	printk("a[%d].key = %lld\n",k,a[k].key);
      }
      assert(0);
    }
  }
}


// simple binary search
struct sort_element *util_search(struct sort_element *a, int array_size, loff_t key) {
  struct sort_element *left, *right;
  struct sort_element *middle;

  left = &(a[0]);
  right = &(a[array_size-1]);

  // check if it is out of bounds
  if((key < left->key) || (key > right->key)) {
    return NULL;
  }

  // now if it is one of our initial outer elements
  if(key == left->key) return left;
  if(key == right->key) return right;

  // and now the calculation
  middle = left + ((right - left) / 2);
  while(middle->key != key) {
    if(middle == left) return NULL;

    if(key > middle->key) {
      left = middle;
    } else {
      right = middle;
    }

    middle = left + ((right - left) / 2);
  }

  assert(middle->key == key);
  return middle;
}
