// this is for the kernel implementation of the file io utilities
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/unistd.h>
#include <linux/string.h>
#include <linux/types.h>
#include <linux/param.h>
#include <linux/fs.h>
#include <linux/slab.h>

#include <asm/uaccess.h>

#include <asm/signal.h>
#include <linux/mman.h>

#include <asm/system.h>
#include <asm/segment.h>

#include <asm/semaphore.h>

#include "myutil.h"

/* For revirt_{suppress,restore}_sched */
#include <revirt/rlog.h>

struct fileUtilStruct {
  struct file *fp;
};

fileUtil_t util_open(char *fileName, unsigned long flags, unsigned long mode) {
  fileUtil_t file = (fileUtil_t) util_alloc_small(sizeof(struct fileUtilStruct));
  file->fp = filp_open(fileName, flags, mode);
  assert(file->fp != NULL);
  return file;
}

struct file *util_getfp(fileUtil_t f) {
  return f->fp;
}

// we will not export explicitly, but we will let people use this if they 
// want
int util_kernel_read(struct file *file, void *buf, unsigned long numBytes, 
		     loff_t offset) {
  mm_segment_t old_fs;
  loff_t pos = offset;
  int result = -ENOSYS;
  int flags;
  struct task_struct * tsk=current;

  // this is a modified version of kernel_read from fs/exec.c
  if (!file->f_op->read)
    goto fail;

  revirt_suppress_sched(tsk, flags);
  old_fs = get_fs();
  set_fs(get_ds());
  result = file->f_op->read(file, buf, numBytes, &pos);
  set_fs(old_fs);
  revirt_restore_sched(tsk, flags);
 fail:
  return result;  
}

void util_unlink(char *fileName) {
  mm_segment_t old_fs;

  old_fs = get_fs();
  set_fs(get_ds());
  sys_unlink(fileName);
  set_fs(old_fs);
}

int util_kernel_write(struct file *file, void *buf, unsigned long numBytes, 
		      loff_t offset) {
  mm_segment_t old_fs;
  loff_t pos = offset;
  int result = -ENOSYS;
  int flags;
  struct task_struct * tsk=current;

  // this is a modified version of kernel_read from fs/exec.c
  if (!file->f_op->write)
    goto fail;

  revirt_suppress_sched(tsk, flags);

  old_fs = get_fs();
  set_fs(get_ds());
  result = file->f_op->write(file, buf, numBytes, &pos);
  set_fs(old_fs);

  revirt_restore_sched(tsk, flags);
 fail:
  return result;
}
int util_read(fileUtil_t f, void *buf, unsigned long numBytes, loff_t offset) {
  return util_kernel_read(f->fp, buf, numBytes, offset);
}

int util_write(fileUtil_t f, void *buf, unsigned long numBytes, loff_t offset) {
  return util_kernel_write(f->fp, buf, numBytes, offset);
}

loff_t util_seek(fileUtil_t f, loff_t offset, int origin) {
  return kernel_llseek(f->fp, offset, origin);
}

int util_append(fileUtil_t f, void *buf, unsigned long numBytes) {
  loff_t offset = util_seek(f, 0, UTIL_SEEK_END);
  return util_write(f, buf, numBytes, offset);
}

void util_close(fileUtil_t f) {
  filp_close(f->fp, current->files);

  util_free_small(f);
}

void util_truncate(fileUtil_t f, loff_t len) {
  int ret = do_truncate(f->fp->f_dentry, len);
  assert(ret == 0);
}
