#include <stdlib.h>
#include <stdio.h>

#include <unistd.h>
#include <asm/unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include <ctype.h>

#include <linux/autoconf.h>
#include <revirt/rlog.h>

#define BUFSIZE (1024 * 1024) 
unsigned char buf[BUFSIZE];

#define NUMTYPES 40

int main(int argc, char * argv[]) {
	int fd;
	int i, j, l;
	struct rlog_header rh;
	char * names[] = { "SCHED", "FORK", "EXIT", "TOP", "SIGNAL", 
		"EXECPARM", 0, 0, 0, 0, 0, 0, 0, 0, 0, "PUT_USER", 
		"COPY_USER", "SYSRET", "CLEAR_USER", "COPY_USER_NODATA",
		"ENTRY_SIGFRAME", "ENTRY_RT_SIGFRAME", "ENTRY_MARKER",
		"ENTRY_CHPT_MARKER" , "ENTRY_RDTSC", "ENTRY_END", "ENTRY_IN" ,
		0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
	char * stype_names[] =  STYPE_NAMES;

	unsigned records=0, bytes=0, dbytes=0;

	unsigned type_count[NUMTYPES];
	unsigned type_data[NUMTYPES];
	unsigned syscall_count[256][12];
	unsigned signal_count[32];
	unsigned pidmap[12];
	int pidmapmax=0;
	unsigned copyuser_data[256];
	unsigned putuser_data[256];
	unsigned stype_count[STYPE_MAX];
	unsigned stype_data [STYPE_MAX];
	int fd_count[256], fd_data[256];
	int log_size_buckets[16];
	int max_log_size=0;

	int put_user_cumulator=0;

	for(i=0; i<NUMTYPES; i++) {
		type_count[i]=type_data[i]=0;
	}

	for(i=0; i<256; i++) {
		copyuser_data[i]=putuser_data[i]=fd_count[i]=fd_data[i]=0;
		for(j=0; j<12; j++) {
			syscall_count[i][j]=0;
		}
	}

	for(i=0; i<32; i++) {
		signal_count[i]=0;
	}

	for(i=0; i<3; i++) {
	    stype_count[i]=stype_data[i]=0;
	}

	for(i=0; i<16; i++) {
	    log_size_buckets[i]=0;
	}

	if(argc!=2) {
		fprintf(stderr, "Usage: %s logfile\n", argv[0]);
	}
	printf("Opening file %s\n", argv[1]);
	if((fd=open(argv[1], O_RDONLY))<0) {
		perror("opening file");
		exit(1);
	}
	while((l=read(fd, &rh, sizeof(rh)))>0) {
		records++;
		bytes+=l + rh.len;
		if(!(records % 100000)) 
			fprintf(stderr, "%d records [%d bytes] read...\n", records, bytes);
#if 0
		if(type_count[rh.type & 0x7f]==0) {
		    fprintf(stderr, "First instance of log: 0x%x #%ld %d %s\n",
		    	rh.type, rh.id, rh.len, names[rh.type & 0x7F]);
		}
#endif
		type_count[rh.type & 0x7f]++;
		type_data[rh.type & 0x7f] += rh.len;

		if(rh.len > max_log_size) {
			max_log_size=rh.len;
		}
		for(i=0; i<16; i++) {
		    if(rh.len < (1 << (i+1))) {
		    	log_size_buckets[i]++;
			break;
		    }
		}

		if((l=read(fd, buf, rh.len))<rh.len) {
			if(l<0)
				perror("reading file");
			else
				fprintf(stderr, "Bad file format: wanted %d, got %d\n", rh.len, l);
			break;
		}

		if(1) {
			union {
				struct rlog_entry_signal signal;
				struct rlog_put_user_entry put_user;
				struct rlog_copy_user_entry copy_user;
				struct rlog_clear_user_entry clear_user;
				struct rlog_entry_sysret sysret;
				struct rlog_schedule_entry sched;
				struct rlog_entry_sigframe sigframe;
				struct rlog_entry_rt_sigframe rt_sigframe;
				struct rlog_marker_entry mark;
			} * le = (__typeof(le))buf;

		       switch(rh.type) {
			  case ENTRY_MARKER:
				printf("Mark %d %ld.%06d lo %d sys %d sig %d\n",
					le->mark.mark,
					le->mark.tv.tv_sec,
					le->mark.tv.tv_usec,
					bytes,
					type_count[ENTRY_SYSRET & 0x7f],
					type_count[ENTRY_SIGNAL & 0x7f]);
				break;
			  case ENTRY_SCHED:
				stype_count[le->sched.stype]++;
				stype_data[le->sched.stype] += rh.len;
				break;
			  case ENTRY_CLEAR_USER:
			  case ENTRY_FORK:
				break;
			  case ENTRY_PUT_USER:
			  	put_user_cumulator+=rh.len;
#if 0
			  	putuser_data[le->put_user.syscall] += 
						//le->put_user.ssize;
						rh.len;
#endif
			  	break;
			  case ENTRY_COPY_USER:
			  	copyuser_data[le->copy_user.syscall] += 
						//le->copy_user.ssize;
						rh.len;
				break;
			  case ENTRY_SYSRET:
				j=-1;
			  	for(i=0; i<pidmapmax ; i++) {
					if(pidmap[i]==le->sysret.pid) {
						j=i;
						break;
					}
				}
				if(j<0) {
					if(pidmapmax > 12) {
					  fprintf(stderr, "Maxed outpids...\n");
					  goto printout;
					}
					j=pidmapmax;
					pidmap[j]=le->sysret.pid;
					pidmapmax++;
				}
			  	syscall_count[le->sysret.syscall][0]++;
			  	syscall_count[le->sysret.syscall][j+1]++;
				putuser_data[le->sysret.syscall]+=put_user_cumulator;
				put_user_cumulator=0;
#ifdef REPLAY_WHOLE_PTREGS
				{
				  int fd=le->sysret.regs.ebx;
				  switch(le->sysret.syscall) {
				    int fd;
				    case __NR_read:
				      fd_read[fd]++;
				      fd_data[fd]+=le->sysret.regs.eax;
				      break;
				    case __NR_open:
				      fd_open[fd]++;
				      break;
				    case __NR_close:
				      fd_close[fd]++;
				      break;
				    case __NR_close:
				      fd_close[fd]++;
				      break;
				    case __NR_fcntl64:
				      fd_fcntl[fd]++;
				      break;
				}
#if 0
				if(le->sysret.syscall == __NR_read) {
				    int fd= le->sysret.regs.ebx;
				    fd_read[fd]++;
				    fd_data[fd]+=le->sysret.regs.eax;
				}
#endif

#endif
				break;
			  case ENTRY_SIGNAL:	
				break;
		      }
		} 
	}

    printout:
	dbytes = bytes - (records * sizeof(rh));
	printf("%d records, %d data (%d in file)\n", 
		records, dbytes, bytes);
	printf("Log types:\n");

	for(j=0; j<NUMTYPES; j++) {
	    if(names[j] && type_count[j]) {
	    	int size = type_count[j] * sizeof(rh) + type_data[j];
		printf(" %s cnt: %d data:%d %%%2.1f (%d %%%2.1f in file)\n",
		    names[j], 
		    type_count[j], 
		    type_data[j], 
		    (float)type_data[j] / (float)dbytes * 100,
		    size, 
		    (float) size / bytes * 100);
	    }
	    if(j == (ENTRY_SCHED & 0x7f)) {
		for(i=0; i<STYPE_MAX; i++) {
		    if(stype_count[i]>0) {
		     printf("\t%s cnt: %d data: %d (%d in file)\n",
	    		stype_names[i], 
			stype_count[i],
			stype_data[i],
			stype_data[i] + stype_count[i] * sizeof(rh));
		    }
		}
	    }
	}
#if 0
#define PRNT(x) printf(" %s cnt: %d data:%d %%%2.1f (%d %%%2.1f in file)\n",\
	names[x&0x7F], \
	type_count[x & 0x7F], \
	type_data[x & 0x7F], (float)type_data[x&0x7f] / (float)dbytes * 100,\
	type_count[x & 0x7F] * sizeof(rh) + type_data[x & 0x7f], \
	(float) (type_count[x & 0x7f] * sizeof(rh) + type_data[x & 0x7f]) / \
		bytes * 100);

	PRNT(ENTRY_SCHED);
	for(i=0; i<3; i++) {
	    printf("\t%s cnt: %d data: %d (%d in file)\n",
	    		stype_names[i], 
			stype_count[i],
			stype_data[i],
			stype_data[i] + stype_count[i] * sizeof(rh));
	}
	PRNT(ENTRY_FORK);
	PRNT(ENTRY_EXIT);
	PRNT(ENTRY_TOP);
	PRNT(ENTRY_SIGNAL);
	PRNT(ENTRY_EXECPARM);
	PRNT(ENTRY_PUT_USER);
	PRNT(ENTRY_COPY_USER);
	PRNT(ENTRY_SYSRET);
	PRNT(ENTRY_CLEAR_USER);
	PRNT(ENTRY_COPY_USER_NODATA);
	PRNT(ENTRY_SIGFRAME);
	PRNT(ENTRY_MARKER);
	//PRNT(ENTRY_SENDERIP);
	PRNT(ENTRY_RDTSC);
#undef PRNT
#endif
	printf("Signal count:\n");
	for(i=0; i<8; i++) {
	  printf("%2d:%6d %2d:%6d %2d:%6d %2d:%6d\n", 
	  	i, signal_count[i],
	  	i+8, signal_count[i+8],
	  	i+16, signal_count[i+16],
	  	i+24, signal_count[i+24]);
	}
	printf("Syscall: Count copy_user put_user\n");
	for(i=0; i<256; i++) {
	   printf("%3d: %10d %10d %10d\n", i, syscall_count[i][0], 
		copyuser_data[i], putuser_data[i]);
	}
       for(j=0; j<pidmapmax; j++) {
        printf("\n---Pid %d:%d:\n", j, pidmap[j]);
	for(i=0; i<256; i++) {
	   if(syscall_count[i][j+1])
		printf("%3d: %10d\n", i, syscall_count[i][j+1]);
	}
       }
#ifdef REPLAY_WHOLE_PTREGS
       printf("\n--FD statistics--\n");
       for(i=0; i<256; i++) {
	if(fd_read[i] || fd_data[i] || fd_open[i] || fd_close[i] || fd_fcntl[i]) {
	    printf("%3d: %5d %10d %5d %5d %5d\n", i, 
	    	fd_read[i], 
	    	fd_open[i], 
	    	fd_close[i], 
	    	fd_fcntl[i], 
		fd_data[i]);
	}
       }
#endif
	printf("\n--Record size statistics--\n");
	for(i=0; i<16; i++) {
		if(log_size_buckets[i]>0) {
			printf("%2d:%8d\n", i, log_size_buckets[i]);
		}
	}
	printf("Max log size: %d\n", max_log_size);
       return 0;
}
