#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

int ifd=0, ofd=0;


void usage(char * name) {
    fprintf(stderr, "Usage: %s [-i input] [-o output] [-f from] [-t to] [-l len]\n", name);
}

int open_file(char * fname, int flags) {
    int r;
    r=open(fname, flags, 0666);
    if(r<0) {
	fprintf(stderr, "Cannot open file %s: %s\n", fname, strerror(errno));
	exit(1);
    }
    return r;
}

int main(int argc, char * argv[]) {
    int c;
    
    int ifd=0, ofd=0;
    loff_t from, to, len, nlen, rlen; 
    int bsize=1024*1024*10;
    char * buf;

    from=to=len=0;

    while((c=getopt(argc, argv, "i:o:f:t:l:hb:"))>=0) {
	switch(c) {
	    case 'i':
		fprintf(stderr, "openting input file %s\n", optarg);
	    	ifd=open_file(optarg, O_RDONLY);
		fprintf(stderr, "got fd %d\n", ifd);
		break;
	    case 'o':
		fprintf(stderr, "openting output file %s\n", optarg);
	    	ofd=open_file(optarg, O_CREAT|O_TRUNC|O_WRONLY);
		fprintf(stderr, "got fd %d\n", ofd);
		break;
	    case 'f':
	    	from=atoll(optarg);
		if(to)
		    len=to-from;
		break;
	    case 't':
	    	to=atoll(optarg);
		if(from)
		    len=to-from;
		break;
	    case 'l':
	    	len=atoll(optarg);
		break;
	    case 'h':
		usage(argv[0]);
		exit(0);
		break;
	    case 'b':
	    	bsize = atoi(optarg);
		break;
	}
    }
    if(!from) {
	fprintf(stderr, "%s: no from!\n", argv[0]);
	usage(argv[0]);
	exit(1);
    }
    if(!len) {
	fprintf(stderr, "%s: need either to or len!\n", argv[0]);
	usage(argv[0]);
	exit(1);
    }

    buf=malloc(bsize);
    if(!buf) {
	fprintf(stderr, "cannot allocate %d-byte buffer\n", bsize);
	exit(1);
    }

    if(lseek64(ifd, from, SEEK_SET)<0) {
	perror("lseek64");
	exit(1);
    }

    while(len>0) {
	nlen = (bsize < len ) ? bsize : len;
	rlen=read(ifd, buf, nlen);
	if(rlen < 0) {
	    perror("read");
	    exit(1);
	}
	rlen=write(ofd, buf, rlen);
	if(rlen < 0) {
	    perror("write");
	    exit(1);
	}
	len -= rlen;
    }
    
}
