#include <stdlib.h>
#include <stdio.h>

#include <unistd.h>
#include <asm/unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include <ctype.h>

#include <linux/autoconf.h>
#include <revirt/rlog.h>
#include <revirt/flags.h>

/*
 * So asm/siginfo.h will compile...
 */
typedef long clock_t;

//#include <asm/siginfo.h>

void print_top(struct rlog_header * h /*, struct rlog_entry_top * le*/) {
	printf("%u:Top (0x%x %d %lld)\n", h->id, h->type, h->len, h->pos);
}

char * stype_names[] = STYPE_NAMES;

void print_stype(int stype) {
    if(stype < STYPE_MAX) {
	printf("%s", stype_names[stype]);
    } else {
	printf("%d (unknown)", stype);
    }
}

void print_bc(struct revirt_ctrdata * le) {
#ifdef REPLAY_RECORD_ALL_PERFCTRS
	int i;
	printf(" x:0x%llx  [ %x %x %x | "
		le->hw_branch, 
		le->hw_intr,
		le->ex_pf,
		le->ex_dev);
	for(i=0; i<REVIRT_NR_PFCTRS; i++) {
		printf("%llx ", le->counter[i]);
	}
	printf("\n");
	if(le->hw_intr != le->ex_dev) {
		printf(" ---Interrupt counts do not match---\n");
	}
#else
	printf(" x:0x%llx\n", le->hw_branch);
#endif

#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
#warning listing exceptions
	for(i=0; i<REVIRT_NR_EX; i++) {
		if(le->exceptions[i]>0)
			printf("  [%d] 0x%x\n", i, le->exceptions[i]);
	}
#endif
}

char * rflags_names[] = RFLAGS_NAMES;

void print_sched(struct rlog_header *h, struct rlog_schedule_entry * le) {
	int i;
	printf("%u:Sched (0x%x %d %lld) ", h->id, h->type, h->len, h->pos);
	printf("pid:%5d flags:0x%08x ( ",
		le->pid,
		le->flags);

	for(i=0; i<RFLAGS_MAX; i++) {
	    if(le->flags & (1 << i)) {
		printf("%s ", rflags_names[i]);
	    }
	}

	printf(")\n tsk->state: ");
	switch(REPLAY_TSK_GETSTATE(le->flags)) {
	  case 0: printf("TASK_RUNNING"); break;
	  case 1: printf("TASK_INTERUPTIBLE"); break;
	  case 2: printf("TASK_UNINTERUPTIBLE"); break;
	  case 4: printf("TASK_ZOMBIE"); break;
	  case 8: printf("TASK_STOPPED"); break;
	  default: printf("unknown");
	}
	printf("\t stype: ");
	print_stype(le->stype);
	printf("\textra: %x\n", le->extra);

	printf(" active: %d [ ", le->active.i);
	for(i=0; i<le->active.i ; i++)
		printf("%d ", le->active.pid[i]);
	printf("] epochs_skipped: %d runnable: %d\n", 
		le->active.epochs_skipped,
		le->active.runnable);

	print_bc(&le->ctr);
	printf(" eip:0x%x ecx:0x%x edi:0x%x nrxuser:0x%x wchan:0x%x sid:%lld\n", 
		le->eip,
		le->data,
		le->edi,
		le->nrxuser,
		le->wchan,
		le->syscall_id);
}

void print_sysret(struct rlog_header *h, struct rlog_entry_sysret * le) {
	printf("%u:Sysret (0x%x %d %lld) ", h->id, h->type, h->len, h->pos);
	printf("pid:%d syscall:%d nrxuser:%d id:%lld ret: 0x%x\n",
					le->pid, 
					le->syscall, 
					le->nrxuser,
					le->syscall_id,
					le->regs.eax);
#ifdef REPLAY_WHOLE_PTREGS
	printf(" %d %d (%x, %x, %x, %x, %x) @ %x == %x\n",
			le->pid,
			le->syscall,
			le->regs.ebx,
			le->regs.ecx,
			le->regs.edx,
			le->regs.esi,
			le->regs.edi,
			le->regs.eip,
			le->regs.eax);
	//printf(" eip: %x\n", le->regs.eip);
#endif
#ifdef REPLAY_SYSCALL_LOGS_BRANCHCOUNT
	print_bc(&le->ctr);
#endif
}

void print_signal(struct rlog_header *h, struct rlog_entry_signal * le) {
	printf("%u:Signal (0x%x %d %lld) ", h->id, h->type, h->len, h->pos);
#if 0
	printf("pid:%5d sig:%2d infoval:%2d\n",
		le->pid,
		le->sig,
		le->infoval);
#endif

#ifdef REPLAY_WHOLE_SIGINFO
	printf("pid:%5d si_signo:%2d si_errno:%2d si_code:%2d\n",
		le->pid,
		le->info.si_signo,
		le->info.si_errno,
		le->info.si_code);
#else
	printf("pid:%5d si_signo:%2d si_errno:%2d si_code:%2d\n",
		le->pid,
		le->si_signo,
		le->si_errno,
		le->si_code);
#endif

#if 0
	printf(" b: 0x%x %x", 
		le->blocked.sig[0],
		le->blocked.sig[1]);
	printf(" p: 0x%x %x",
		le->pending_signal.sig[0],
		le->pending_signal.sig[1]);
	printf(" d: 0x%x %x",
		le->dfl.sig[0],
		le->dfl.sig[1]);
	si = (struct siginfo *)(((char *)le)+sizeof(le->signal));
	printf("si_signo: %d si_errno: %d si_code: %d\n", 
			si->si_signo,
			si->si_errno,
			si->si_code);
	for(i=0; i<7; i++) {
		printf("%8x %8x %8x %8x\n",
			si->_sifields._pad[i],
			si->_sifields._pad[i+8],
			si->_sifields._pad[i+15],
			si->_sifields._pad[i+22]);
	}
	printf("%8x\n", si->_sifields._pad[7]);
#endif
}

void print_sigframe(struct rlog_header *h, struct rlog_entry_sigframe * le) {
	printf("%u:Sigframe (0x%x %d %lld) ", h->id, h->type, h->len, h->pos);
#ifdef REPLAY_WHOLE_SIGFRAME
	printf("sig %d @ %x\n", le->frame.sig, le->frame.sc.eip);
#else
	printf("sig %d \n", le->frame.sig);
#endif
}

void print_rt_sigframe(struct rlog_header *h, struct rlog_entry_rt_sigframe * le) {
	printf("%u:RT Sigframe (0x%x %d %lld) ", h->id, h->type, h->len, h->pos);
#ifndef REPLAY_WHOLE_SIGFRAME
	printf("pid:%5d sig:%2d trapno:%2d\n",
		le->pid,
		le->frame.sig,
		le->frame.trapno);
#else
	printf("pid:%5d sig:%2d\n",
		le->pid,
		le->frame.sig);
	printf("[don't know how to print whole sigframe yet]\n");
#endif
}

void print_fork(struct rlog_header *h, struct rlog_entry_fork * le) {
	printf("%u:Fork (0x%x %d %d) ", h->id, h->type, h->len, h->pos);
	printf("parent: %d child: %d\n", le->pid, le->child);
}

void print_exit(struct rlog_header *h, struct rlog_exit_entry* le) {
	int i;

	printf("%u:exit (0x%x %d %lld) pid %d", h->id, h->type, h->len, h->pos);
	printf("pid: %d\n", le->pid);

	printf(" active: %d [ ", le->active.i);
	for(i=0; i<le->active.i ; i++)
		printf("%d ", le->active.pid[i]);
	printf("] epochs_skipped: %d runnable: %d\n", 
		le->active.epochs_skipped,
		le->active.runnable);

}

void print_dma(struct rlog_header *h, struct rlog_dma_entry * le) {
	printf("%u:DMA (0x%x %d %lld) pid %d dest 0x%x size 0x%x eip %x seq %d\n", 
		h->id, h->type, h->len, h->pos,
		le->pid, le->dest, (int)le->size, le->eip, le->seq);
}

void print_copyuser(struct rlog_header *h, struct rlog_copy_user_entry * le) {
	printf("%u:Copyuser (0x%x %d %lld) pid %d to 0x%x size 0x%x syscall 0x%x sid %lld\n", 
		h->id, h->type, h->len, h->pos,
		le->pid, le->tto, le->ssize, le->syscall, le->syscall_id);
}

void print_rdtsc(struct rlog_header *h, struct rlog_rdtsc * le) {
#ifdef REPLAY_RDTSC_LOGS_SID
	printf("%u:RDTSC (0x%x %d %lld) @0x%x value 0x%x %x (sid %llx)\n",
		h->id, h->type, h->len, h->pos,
		le->eip, le->hi, le->lo, le->syscall_id);
#else
	printf("%u:RDTSC (0x%x %d %lld) @0x%x value 0x%x %x\n",
		h->id, h->type, h->len, h->pos,
		le->eip, le->hi, le->lo);
#endif
}

void print_marker(struct rlog_header *h, struct rlog_marker_entry *le) {
	printf("%lu:Marker (0x%x %d %lld) mark %3d %10d.%d sec\n",
		h->id, h->type, h->len, h->pos,
		le->mark, le->tv.tv_sec, le->tv.tv_usec);
}

void print_in(struct rlog_header *h, struct rlog_entry_in *le) {
	printf("%lu:IN (0x%x %d %lld) @0x%x in 0x%x == 0x%x\n",
		h->id, h->type, h->len, h->pos,
		le->eip, (unsigned int)le->port, (unsigned int)le->byte);
}

void print_entry(struct rlog_header rh, char * buf) {
		static unsigned int last_id=0;
		if(rh.id - last_id > 1) {
			printf("**** Warning **** id increases by more than one: %d %d\n",
				last_id, rh.id);
		} else if(rh.id - last_id == 0 && last_id) {
			printf("**** Warning *** Duplicate id: %u %u\n", last_id, rh.id);
		}
		last_id=rh.id;
		switch(rh.type) {
		    case ENTRY_TOP:
			print_top(&rh);
			break;
		    case ENTRY_SCHED:
			print_sched(&rh, (struct rlog_schedule_entry *)buf);
			break;
		    case ENTRY_FORK:
			print_fork(&rh, (struct rlog_entry_fork *)buf);
		     	//printf("%u Fork (0x%x %d)\n", rh.id, rh.type, rh.len);
			break;
		    case ENTRY_EXIT:
			print_exit(&rh, (struct rlog_exit_entry *)buf);
		     	//printf("%u Exit (0x%x %d)\n", rh.id, rh.type, rh.len);
			break;
		    case ENTRY_SIGNAL:
			print_signal(&rh, (struct rlog_entry_signal *)buf);
			break;
		    case ENTRY_PUT_USER:
		     	printf("put_user() log: 0x%x %d\n", rh.type, rh.len);
			break;
		    case ENTRY_COPY_USER:
		    	print_copyuser(&rh, (struct rlog_copy_user_entry *)buf);
		     	//printf("copy_user() log: 0x%x %d\n", rh.type, rh.len);
			break;
		    case ENTRY_DMA:
		    	print_dma(&rh, (struct rlog_dma_entry *)buf);
			break;
		    case ENTRY_CLEAR_USER:
		     	printf("clear_user() log: 0x%x %d\n", rh.type, rh.len);
			break;
		    case ENTRY_SYSRET:
			print_sysret(&rh, (struct rlog_entry_sysret *)buf);
			break;
		    case ENTRY_SIGFRAME:
		    	print_sigframe(&rh, (struct rlog_entry_sigframe *)buf);
		     	//printf("%u: Sigframe log: 0x%x %d\n", rh.id, rh.type, rh.len);
			break;
		    case ENTRY_RT_SIGFRAME:
			print_rt_sigframe(&rh, (struct rlog_entry_rt_sigframe *)buf);
			break;
		    case ENTRY_RDTSC:
		    	print_rdtsc(&rh, (struct rlog_rdtsc *)buf);
			break;
		    case ENTRY_MARKER:
		    	print_marker(&rh, (struct rlog_marker_entry *)buf);
			break;
		    case ENTRY_IN:
		    	print_in(&rh, (struct rlog_entry_in *)buf);
			break;
		    default:
			printf("Other log: 0x%x %d\n", rh.type, rh.len);
		}
}

int printdata(char *buf, int l) {	
	int i, j;
	for(i=0; i<l; i+=16) {
	    printf(" %08x:", i);
	    for(j=i; j<i+16 && j<l; j++) {
		if (!(j%4) && j>0)
			printf(" ");
		printf("%02x ", (unsigned int)buf[j]);
                   }
	    if(j<i+16) {
		for(; j<i+16; j++) {
			if(!j%4)
				printf("  ");
			printf("   ");
		}
	    }
	    printf("  ");
	    for(j=i; j<i+16; j++) {
		if(isprint(buf[j]))
			fputc(buf[j], stdout);
		else
			fputc('.', stdout);
	    }
	    printf("\n");
	}
	return 0;
}
