#include <stdlib.h>
#include <stdio.h>

#include <unistd.h>
#include <asm/unistd.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <errno.h>

#include <ctype.h>

#include <linux/autoconf.h>
#include <revirt/rlog.h>
#include <revirt/flags.h>

/*
 * So asm/siginfo.h will compile...
 */
typedef long clock_t;

//#include <asm/siginfo.h>

#define BUFSIZE (1024 * 1024) 
unsigned char buf[BUFSIZE];

struct marklist_struct {
    char type;
    int mi;
    int mark;
    struct {
	int mark;
	int sec;
	int usec;
	double f;
	loff_t off;
    } list[10];
} log = { .mi=0, .mark=0, .type='l' }, 
  checkpoint = { .mi=0 , .mark=0 , .type='c' },
  testperiod = { .mi=0, .mark=0, .type='t' }, 
  ovhdperiod = { .mi=0, .mark=0, .type='o'};

void print_marks(struct marklist_struct * m) {
    int i;
    printf("%c ", m->type);
    for(i=1; i<m->mi; i++) {
    	printf("%lf ", m->list[i].f - m->list[i-1].f);
    }
    for(i=1; i<m->mi; i++) {
    	printf("%lld ", m->list[i].off-m->list[i-1].off);
    }
    printf("\n");
}

int handle_marker_entry(struct rlog_header * rh, struct rlog_marker_entry * le, 
	struct marklist_struct * m) {
    //print_marker(rh, le);

    switch(le->mark) {
    	case 255:
	    m=&testperiod;
	    break;	    
	case 254:
	    m=&ovhdperiod;
	    break;
    }


    if(le->mark < m->mark) {
    	print_marks(m);
	m->mi=0;
    }

    m->mark = le->mark;
    m->list[m->mi].mark = le->mark;
    m->list[m->mi].sec = le->tv.tv_sec;
    m->list[m->mi].usec = le->tv.tv_usec;
    m->list[m->mi].f =(double)le->tv.tv_sec + (le->tv.tv_usec)/(double)1000000;
    m->list[m->mi].off = rh->pos;

    m->mi++;

    return 0;
}

int main(int argc, char * argv[]) {
	int fd;
	int l;
	struct rlog_header rh;
	int logmark=1;
	int checkmark=1;

	if(argc==2) {
		logmark=1;
		checkmark=1;
	} else if(argc==3){
	    if(argv[2][0]=='l') {
	    	checkmark=0;
	    } else if(argv[2][0]=='c') {
	    	logmark=0;
	    } else {
	    	fprintf(stderr, "Unknown cmd: %s\n", argv[2]);
		exit(1);
	    }
	} else  {
		fprintf(stderr, "Usage: %s logfile [c|l]\n", argv[0]);
		exit(1);
	}

	printf("Opening file %s\n", argv[1]);
	if((fd=open(argv[1], O_RDONLY))<0) {
		perror("opening file");
		exit(1);
	}
	while((l=read(fd, &rh, sizeof(rh)))>0) {
		if((l=read(fd, buf, rh.len))<rh.len) {
			if(l<0)
				perror("reading file");
			else
				fprintf(stderr, "Bad file format: wanted %d, got %d\n", rh.len, l);
			exit(1);
		}
		if(rh.type==ENTRY_MARKER && logmark) {
	    	    handle_marker_entry(&rh, (struct rlog_marker_entry *)buf, &log);
		} else if(rh.type==ENTRY_CHKPT_MARKER && checkmark)  {
	    	    handle_marker_entry(&rh, (struct rlog_marker_entry *)buf, &checkpoint);
		}
	}
	if(logmark) {
	    print_marks(&log);
	    print_marks(&testperiod);
	    print_marks(&ovhdperiod);
	} 
	if(checkmark)
	    print_marks(&checkpoint);
	return 0;

}
