#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include <sys/poll.h>
#include <errno.h>
#include <signal.h>

#define __USE_LARGEFILE64
#include <fcntl.h>

#define BUFSIZE (1 << 22)

char * buf [BUFSIZE];
int live=1;
int rfd, wfd;
int len;

unsigned long total = 0;

void sig(int b) {
	struct pollfd pfd;
	int r;

	printf("Interrupt received, polling for unread data...\n");

	pfd.fd = rfd;
	pfd.events = POLLIN;
	r=poll(&pfd, 1, 0);
	if((r > 0) && (pfd.revents & POLLIN)) {
		/*
 		 * FIXME this assumes BUFSIZE is greater than kernel BUFSIZE
		 */
		printf("Calling read...\n");
		if((len=read(rfd, buf, BUFSIZE)) < 0) 
			perror("reading file");
		else {
			total+=len;
			printf("Writing %d bytes (%lu total)\n", len, total);
			if(write(wfd, buf, len)<0) {
			    perror("Writing file");
			}
			
		}
	} else if(r < 0) {
		perror("poll");
	} else
		printf("Poll returns 0.\n");
	printf("Closing files and exiting...\n");
	close(rfd);
	close(wfd);
	exit(1);
}

#define __NR_ugetrlimit 191

int main(int argc, char * argv[]) {
	struct rlimit rl;
	char * ofname_dfl = "/var/demon/log.000", *ofname;

	signal(SIGUSR1, sig);
	signal(SIGINT, sig);

	rl.rlim_cur=(~0UL);
	rl.rlim_max=(~0UL);
	printf("setrlimit: %ul, %ul\n", rl.rlim_cur, rl.rlim_max);
	if(setrlimit(RLIMIT_FSIZE, &rl)) {
		perror("Setting rlimit_fsize to rlim_infinity");
	}

	syscall(__NR_ugetrlimit, RLIMIT_FSIZE, &rl);
	printf("getrlimit: %ul, %ul\n", rl.rlim_cur, rl.rlim_max);

	if((rfd=open("/proc/rlog", O_RDONLY,0664))<0) {
		perror("Opening /proc/rlog");
		exit(1);
	}
	printf("/proc/rlog opened.\n");

	if(argc > 1) 
		ofname=argv[1];
	else
		ofname=ofname_dfl;

//	if((wfd = open(ofname, O_CREAT|O_TRUNC|O_WRONLY|O_SYNC))<0) {
	if((wfd = open(ofname, O_CREAT|O_TRUNC|O_WRONLY|O_LARGEFILE, 0664))<0) {
		perror("Opening output file");
		exit(1);
	}
	printf("File %s opened for writing.\n", ofname);

	while(live) {
	   /*
   	    * Unsync'd data...
	    */
		if((len=read(rfd, buf, BUFSIZE)) < 0)  {
		    if(live) {
			perror("Reading file");
			exit(1);
		    } else
		    	break;
		
		}
		total+=len;
		printf("Writing %d bytes. (%lu)\n", len, total);
		if(write(wfd, buf, len)<0) {
			perror("Writing file");
		}
	}
	printf("Closing files.\n");
	close(rfd);
	close(wfd);
	return 1;
}
