#include "linux/kernel.h"
#include "linux/version.h"
#include "linux/string.h"
#include "linux/pci.h"
#include "linux/types.h"
#include "linux/init.h"
#include "linux/slab.h"
#include "asm/uaccess.h"
#include "linux/slab.h"
#include "linux/vmalloc.h"
#include "linux/list.h"

#include "linux/mm.h"
#include "asm/io.h"
#include "os.h"

#include "../../../../revirt-skas-2.4.18/uml/uml.h"

#define MAX_RW_BUF (1024 * 100)
#define KMALLOC_MAX (PAGE_SIZE * 2)

#if 0
struct rw_node {
	struct list_head link;
	struct uml_rw_struct rw;
};
static unsigned long currSize = 0;

static struct list_head rw_list = LIST_HEAD_INIT(rw_list);

static void rw_hypercall(struct task_struct *tsk, unsigned long dataLen, 
			 char *kAddr) {
	int fd;
	struct openflags flags;
	struct rw_node *node;

	node = kmalloc(sizeof(struct rw_node), GFP_KERNEL);
	if(node == NULL) {
		panic("cannot alloc rw_node");
	}
	currSize += sizeof(struct rw_node);
	if(dataLen > KMALLOC_MAX) {
		currSize += ((dataLen + PAGE_SIZE - 1) & PAGE_MASK);
	} else {
		currSize += dataLen;
	}

	node->rw.pid = tsk->pid;
	memcpy(node->rw.comm, tsk->comm, 16);
	node->rw.dataLen = dataLen;
	node->rw.buf = kAddr;

	list_add_tail(&node->link, &rw_list);

	if(currSize > MAX_RW_BUF) {
		currSize = 0;
		memset(&flags, 0, sizeof(flags));
		flags.r = 1;
		fd = os_open_file("/dev/uml",flags,0666);
		if(fd < 0) {
			panic("could not open /dev/uml");
		}
		while(rw_list.prev != &rw_list) {
			node = list_entry(rw_list.prev, struct rw_node, link);

			ioctl(fd, UML_IOC_RW_CMD, &node->rw);

			list_del(&node->link);
			
			if(node->rw.dataLen > KMALLOC_MAX) {
				vfree(node->rw.buf);
			} else {
				kfree(node->rw.buf);
			}
			kfree(node);
		}

		close(fd);
	}
}


void rw_callback(struct task_struct *tsk, unsigned long dataLen, 
		 char *userAddr) {
  /*
	int ret;
	unsigned char *kData;
	unsigned long idx;
	int isAscii;

	if(dataLen > KMALLOC_MAX) {
	  kData = (unsigned char *) vmalloc(dataLen);
	} else {
	  kData = (unsigned char *) kmalloc(dataLen, GFP_KERNEL);
	}

	if(tsk == current) {
		ret = copy_from_user(kData, userAddr, dataLen);
		if(ret != 0) {
			panic("could not copy rw data");
		}

		// make sure all of the data is ascii
		isAscii = 1;
		for(idx = 0; idx < dataLen; idx++) {
			if(kData[idx] > 127) {
				isAscii = 0;
				break;
			}
		}

		if(isAscii) {
			rw_hypercall(tsk, dataLen, kData);
		} else {
			if(dataLen > KMALLOC_MAX) {
				vfree(kData);
			} else {
				kfree(kData);
			}
		}
	} else {
		panic("task != current");
	}
  */
	
}

#endif


#define MAX_PCI_DEV 100
static struct pci_dev pci_dev_list[MAX_PCI_DEV];

struct local_dma_struct {
  void *kernelVaddr;
  void *vaddrFromHost;
  dma_addr_t dmaHandle;
};

#define NUM_GUEST_DMA 100
static struct local_dma_struct local_dma_list[NUM_GUEST_DMA];

static int low_level_pci_cmd(unsigned long cmd, struct pci_dev *dev,
			     u64 dma_mask, dma_addr_t phys_buff, void *guest_vaddr,
			     ssize_t dma_buff_size, 
			     unsigned int irq,
			     dma_addr_t *phys_buff_out,
			     void **host_vaddr_out) {
  int fd;
  struct openflags flags;
  struct uml_pci_cmd_struct pciCmd;

  memset(&pciCmd, 0, sizeof(pciCmd));
  memset(&flags, 0, sizeof(flags));
  flags.r = 1;
  fd = os_open_file("/dev/uml",flags,0666);
  if(fd < 0) {
    panic("could not open /dev/uml");
  }
  
  pciCmd.cmd_in = cmd;
  if(dev != NULL) {
    pciCmd.vendor_in = dev->vendor;
    pciCmd.device_in = dev->device;
  }
  pciCmd.dma_mask_in = dma_mask;
  pciCmd.phys_buff_in = phys_buff;
  pciCmd.dma_buff_size_in = dma_buff_size;
  pciCmd.guest_vaddr_in = (unsigned long) guest_vaddr;
  pciCmd.irq_in = irq;

  ioctl(fd, UML_IOC_PCI_CMD, &pciCmd);

  if(phys_buff_out != NULL) *phys_buff_out = pciCmd.phys_buff_out;
  if(host_vaddr_out != NULL) *host_vaddr_out = (void *) pciCmd.host_vaddr_out;

  close(fd);

  return 0;
}

int low_level_enable_irq(int irq) {
  // XXX FIXME, pass this parameter down
  low_level_pci_cmd(UML_PCI_CMD_ENABLE_IRQ,NULL, 0, 0, 0, 0, irq, NULL, NULL);
  return 0;
}

int pci_enable_device(struct pci_dev *dev) {
  low_level_pci_cmd(UML_PCI_CMD_ENABLE_DEVICE, dev, 0, 0, 0, 0, 0, NULL, NULL);
  return 0;
}

static int get_real_pci_dev(struct pci_dev *dev, unsigned long *iter){
  struct uml_get_pci_dev_struct getPciDev;
  int fd;
  struct openflags flags;
  unsigned long idx;

  memset(&flags, 0, sizeof(flags));
  flags.r = 1;

  memset(&getPciDev, 0, sizeof(getPciDev));

  getPciDev.pci_dev_iter_inout = (void *) *iter;

  fd = os_open_file("/dev/uml",flags,0666);
  if(fd < 0) {
    panic("could not open /dev/uml");
  }

  ioctl(fd, UML_IOC_GET_PCI_DEV, &getPciDev);

  close(fd);

  if(getPciDev.irq_out) {
    dev->irq = getPciDev.irq_out;
    for(idx=0;idx<DEVICE_COUNT_RESOURCE;idx++) {
      dev->resource[idx].start = getPciDev.resource[idx].start;
      dev->resource[idx].end = getPciDev.resource[idx].end;
      dev->resource[idx].flags = getPciDev.resource[idx].flags;
    }
    dev->vendor = getPciDev.vendor_out;
    dev->device = getPciDev.device_out;
    dev->class = getPciDev.class_out;
    *iter = (unsigned long) getPciDev.pci_dev_iter_inout;
    return 1;
  }

  return 0;
}

static int deviceMatch(struct pci_dev *dev, struct pci_driver *hwdrv,
		       struct pci_device_id **id) {
  unsigned long idx;

  idx = 0;
  while(hwdrv->id_table[idx].vendor != 0) {
    // check the vendor
    if((hwdrv->id_table[idx].vendor == PCI_ANY_ID) ||
       (hwdrv->id_table[idx].vendor == dev->vendor)) {
      
      // check the device
      if((hwdrv->id_table[idx].device == PCI_ANY_ID) ||
	 (hwdrv->id_table[idx].device == dev->device)) {

	// XXX FIXME use the class mask
	// now the class
	if((hwdrv->id_table[idx].class == PCI_ANY_ID) ||
	   (hwdrv->id_table[idx].class == dev->class) ||
	   (hwdrv->id_table[idx].class == 0)) {

	  *id = &(hwdrv->id_table[idx]);
	  return 1;
	}
      }
    }

    idx++;
  }

  return 0;
}


int pci_register_driver(struct pci_driver *hwdrv) {
  unsigned long i;
  int ret;
  struct pci_device_id *id = NULL;

  printk("registering %s pci device\n",hwdrv->name);

  ret = 0;

  // first check the existing drivers
  for(i = 0; i < MAX_PCI_DEV; i++) {
    if(deviceMatch(&pci_dev_list[i],hwdrv,&id)) {
      
      hwdrv->probe(&pci_dev_list[i], id);
      ret = 1;
    }
  }


  if(ret == 0) {
    printk("pci.c: could not find device\n");
  }

  return ret;
}
void pci_unregister_driver(struct pci_driver *hwdrv) {
  // just leave the pci entry in there
}


void *pci_alloc_consistent(struct pci_dev *hwdev, size_t size,
                           dma_addr_t *dma_handle) {
  dma_addr_t dmaRet;
  void *hostVaddr;

  low_level_pci_cmd(UML_PCI_CMD_ALLOC_DMA_BUFFER, hwdev, 0, 0, NULL, size, 0, &dmaRet,&hostVaddr);
  *dma_handle = dmaRet;

  return hostVaddr;
}

void pci_free_consistent(struct pci_dev *hwdev, size_t size,
                         void *vaddr, dma_addr_t dma_handle) {
  low_level_pci_cmd(UML_PCI_CMD_FREE_DMA_BUFFER, hwdev, 0, dma_handle, 0, 
		    size, 0, NULL, NULL);
}

/*
void *pci_alloc_consistent(struct pci_dev *hwdev, size_t size,
                           dma_addr_t *dma_handle) {
  void *ret;
  int gfp = GFP_ATOMIC;
  dma_addr_t dmaRet;

  if (hwdev == NULL || hwdev->dma_mask != 0xffffffff)
    gfp |= GFP_DMA;
  ret = (void *)__get_free_pages(gfp, get_order(size));
  
  if (ret != NULL) {
    memset(ret, 0, size);
    low_level_pci_cmd(UML_PCI_CMD_ALLOC_DMA_BUFFER, hwdev, 0, 0, ret, size, 0, &dmaRet,NULL);
    *dma_handle = dmaRet;
  }
  return ret;
}

void pci_free_consistent(struct pci_dev *hwdev, size_t size,
                         void *vaddr, dma_addr_t dma_handle) {
  free_pages((unsigned long)vaddr, get_order(size));
  low_level_pci_cmd(UML_PCI_CMD_FREE_DMA_BUFFER, hwdev, 0, dma_handle, 0, 
		    size, 0, NULL, NULL);
}
*/


int pci_set_dma_mask(struct pci_dev *dev, u64 mask) {
  dev->dma_mask = mask;
  low_level_pci_cmd(UML_PCI_CMD_SET_DMA_MASK, dev, mask, 0, 0, 0, 0, NULL, NULL);
  return 0;
}

void pci_set_master(struct pci_dev *dev) { 
  low_level_pci_cmd(UML_PCI_CMD_SET_MASTER, dev, 0, 0, 0, 0, 0, NULL, NULL);
}
 
int pcibios_present(void) {
   return 1;
}

void __init uml_pci_bios(void) {
  unsigned long idx;
  unsigned long iter;

  printk("****** uml_pci_bios init *****\n");

  for(idx = 0; idx < MAX_PCI_DEV; idx++) {
    pci_dev_list[idx].vendor = 0;
    pci_dev_list[idx].device = 0;
    pci_dev_list[idx].irq = 0;
  }

  for(idx = 0; idx < NUM_GUEST_DMA; idx++) {
    memset(&(local_dma_list[idx]), 0, sizeof(local_dma_list[idx]));
  }

  iter = 0;
  idx = 0;

#ifdef CONFIG_HOST_USB_PCI
  while(get_real_pci_dev(&pci_dev_list[idx],&iter)) {
    idx++;
    if(idx >= MAX_PCI_DEV) {
      panic("TOO MANY PCI DEVICES\n");
    }
  }
#endif
    
}

__initcall(uml_pci_bios);

static void low_level_iomap_cmd(unsigned long map, 
				unsigned long phys_addr,
				unsigned long size,
				unsigned long vaddr_in, // only used for unmap
				unsigned long *vaddr_out) {

  struct uml_iomap_cmd_struct iomapCmd;
  int fd;
  struct openflags flags;
  memset(&flags, 0, sizeof(flags));
  flags.r = 1;
  fd = os_open_file("/dev/uml",flags,0666);
  if(fd < 0) {
    panic("could not open /dev/uml");
  }

  memset(&iomapCmd, 0, sizeof(iomapCmd));
  iomapCmd.map = map;
  iomapCmd.phys_addr_in = phys_addr;
  iomapCmd.size_in = size;
  iomapCmd.vaddr_in = vaddr_in;

  ioctl(fd, UML_IOC_IOMAP_CMD, &iomapCmd);

  if(vaddr_out != NULL) *vaddr_out = iomapCmd.vaddr_out;

  close(fd);
}
  

void * ioremap (unsigned long offset, unsigned long size) {
  unsigned long vaddr;
  low_level_iomap_cmd(1,offset,size,0,&vaddr);
  return (void *) vaddr;
}

void iounmap(void *addr) {
  low_level_iomap_cmd(0,0,0,(unsigned long) addr, NULL);
}

static struct local_dma_struct *find_local_dma(dma_addr_t dma_handle) {
  unsigned long idx;

  // find the local buffer and copy the data
  for(idx = 0; idx < NUM_GUEST_DMA; idx++) {
    if(local_dma_list[idx].dmaHandle == dma_handle) {
      return &(local_dma_list[idx]);
    }
  }

  panic("could not find dma");
  return NULL;
}

// dma handling for the USB device
static void local_dma_sync(struct pci_dev *hwdev, dma_addr_t dma_handle, 
			   size_t size, int direction) {
  // call down to the driver
  struct uml_sync_dma_struct syncDma;
  int fd;
  struct openflags flags;
  struct local_dma_struct *localDma = find_local_dma(dma_handle);

  memset(&flags, 0, sizeof(flags));
  flags.r = 1;
  fd = os_open_file("/dev/uml",flags,0666);
  if(fd < 0) {
    panic("could not open /dev/uml");
  }

  memset(&syncDma, 0, sizeof(syncDma));
  syncDma.vendor = hwdev->vendor;
  syncDma.device = hwdev->device;
  syncDma.dmaAddr = dma_handle;
  syncDma.size = size;
  syncDma.direction = direction;

  if(direction == PCI_DMA_TODEVICE) {
    //dbgmemcpy(localDma->vaddrFromHost, localDma->kernelVaddr, size);
    memcpy(localDma->vaddrFromHost, localDma->kernelVaddr, size);
  }

  ioctl(fd, UML_IOC_SYNC_DMA, &syncDma);

  if(direction == PCI_DMA_FROMDEVICE) {
    memcpy(localDma->kernelVaddr, localDma->vaddrFromHost, size);
  }

  close(fd);
}

void pci_dma_sync_single(struct pci_dev *hwdev,
			 dma_addr_t dma_handle,
			 size_t size, int direction) {

  local_dma_sync(hwdev, dma_handle, size, direction);
}

static void add_guest_dma_entry(dma_addr_t dmaAddr, void *guestDmaVaddr,
				void *kernelVaddr) {
  unsigned long idx;

  for(idx = 0; idx < NUM_GUEST_DMA; idx++) {
    if(local_dma_list[idx].kernelVaddr == NULL) {
      local_dma_list[idx].kernelVaddr = kernelVaddr;
      local_dma_list[idx].dmaHandle = dmaAddr;
      local_dma_list[idx].vaddrFromHost = guestDmaVaddr;
      return;
    }
  }
  panic("could not find open dma slot");
}

static void del_guest_dma_entry(struct pci_dev *hwdev, dma_addr_t dmaAddr, 
				ssize_t size) {
  struct local_dma_struct *localDma = find_local_dma(dmaAddr);
  if(localDma == NULL) panic("could not find local dma");

  pci_free_consistent(hwdev, size, localDma->vaddrFromHost, localDma->dmaHandle);

  localDma->kernelVaddr = NULL;
  localDma->dmaHandle = 0;
  localDma->vaddrFromHost = NULL;
}

dma_addr_t pci_map_single(struct pci_dev *hwdev, void *ptr,
			  size_t size, int direction) {
  void *guestDmaVaddr;
  dma_addr_t dmaAddr;

  guestDmaVaddr = pci_alloc_consistent(hwdev, size, &dmaAddr);
  if(guestDmaVaddr == NULL) panic("could not allocate dma");

  add_guest_dma_entry(dmaAddr, guestDmaVaddr, ptr);

  if(direction == PCI_DMA_TODEVICE) {
    local_dma_sync(hwdev, dmaAddr, size, direction);
  }

  return dmaAddr;
}

void pci_unmap_single(struct pci_dev *hwdev, dma_addr_t dma_addr,
		      size_t size, int direction) {
  if(direction == PCI_DMA_FROMDEVICE) {
    local_dma_sync(hwdev, dma_addr, size, direction);
  }  

  del_guest_dma_entry(hwdev, dma_addr, size);
}

int pci_write_config_word(struct pci_dev *dev, int where, u16 val) {
  struct uml_pci_config_struct pciConfig;
  struct openflags flags;
  int fd;
  
  memset(&flags, 0, sizeof(flags));
  flags.r = 1;

  memset(&pciConfig, 0, sizeof(pciConfig));

  pciConfig.vendor_in = dev->vendor;
  pciConfig.device_in = dev->device;  
  pciConfig.cmd_in = PCI_WRITE_CONFIG_WORD;
  pciConfig.where_in = where;
  pciConfig.val_in = val;

  fd = os_open_file("/dev/uml",flags,0666);
  if(fd < 0) {
    panic("could not open /dev/uml");
  }

  ioctl(fd, UML_IOC_PCI_CONFIG, &pciConfig);

  close(fd);

  return pciConfig.ret_out;
}


// this is copied from the pci driver code in driver/pci/pci.c
/*
 * Pool allocator ... wraps the pci_alloc_consistent page allocator, so
 * small blocks are easily used by drivers for bus mastering controllers.
 * This should probably be sharing the guts of the slab allocator.
 */

struct pci_pool {	/* the pool */
	struct list_head	page_list;
	spinlock_t		lock;
	size_t			blocks_per_page;
	size_t			size;
	int			flags;
	struct pci_dev		*dev;
	size_t			allocation;
	char			name [32];
	wait_queue_head_t	waitq;
};

struct pci_page {	/* cacheable header for 'allocation' bytes */
	struct list_head	page_list;
	void			*vaddr;
	dma_addr_t		dma;
	unsigned long		bitmap [0];
};

#define	POOL_TIMEOUT_JIFFIES	((100 /* msec */ * HZ) / 1000)
#define	POOL_POISON_BYTE	0xa7

// #define CONFIG_PCIPOOL_DEBUG


/**
 * pci_pool_create - Creates a pool of pci consistent memory blocks, for dma.
 * @name: name of pool, for diagnostics
 * @pdev: pci device that will be doing the DMA
 * @size: size of the blocks in this pool.
 * @align: alignment requirement for blocks; must be a power of two
 * @allocation: returned blocks won't cross this boundary (or zero)
 * @flags: SLAB_* flags (not all are supported).
 *
 * Returns a pci allocation pool with the requested characteristics, or
 * null if one can't be created.  Given one of these pools, pci_pool_alloc()
 * may be used to allocate memory.  Such memory will all have "consistent"
 * DMA mappings, accessible by the device and its driver without using
 * cache flushing primitives.  The actual size of blocks allocated may be
 * larger than requested because of alignment.
 *
 * If allocation is nonzero, objects returned from pci_pool_alloc() won't
 * cross that size boundary.  This is useful for devices which have
 * addressing restrictions on individual DMA transfers, such as not crossing
 * boundaries of 4KBytes.
 */
struct pci_pool *
pci_pool_create (const char *name, struct pci_dev *pdev,
	size_t size, size_t align, size_t allocation, int flags)
{
	struct pci_pool		*retval;

	if (align == 0)
		align = 1;
	if (size == 0)
		return 0;
	else if (size < align)
		size = align;
	else if ((size % align) != 0) {
		size += align + 1;
		size &= ~(align - 1);
	}

	if (allocation == 0) {
		if (PAGE_SIZE < size)
			allocation = size;
		else
			allocation = PAGE_SIZE;
		// FIXME: round up for less fragmentation
	} else if (allocation < size)
		return 0;

	if (!(retval = kmalloc (sizeof *retval, flags)))
		return retval;

#ifdef	CONFIG_PCIPOOL_DEBUG
	flags |= SLAB_POISON;
#endif

	strncpy (retval->name, name, sizeof retval->name);
	retval->name [sizeof retval->name - 1] = 0;

	retval->dev = pdev;
	INIT_LIST_HEAD (&retval->page_list);
	spin_lock_init (&retval->lock);
	retval->size = size;
	retval->flags = flags;
	retval->allocation = allocation;
	retval->blocks_per_page = allocation / size;
	init_waitqueue_head (&retval->waitq);

#ifdef CONFIG_PCIPOOL_DEBUG
	printk (KERN_DEBUG "pcipool create %s/%s size %d, %d/page (%d alloc)\n",
		pdev ? pdev->slot_name : NULL, retval->name, size,
		retval->blocks_per_page, allocation);
#endif

	return retval;
}


static struct pci_page *
pool_alloc_page (struct pci_pool *pool, int mem_flags)
{
	struct pci_page	*page;
	int		mapsize;

	mapsize = pool->blocks_per_page;
	mapsize = (mapsize + BITS_PER_LONG - 1) / BITS_PER_LONG;
	mapsize *= sizeof (long);

	page = (struct pci_page *) kmalloc (mapsize + sizeof *page, mem_flags);
	if (!page)
		return 0;
	page->vaddr = pci_alloc_consistent (pool->dev,
					    pool->allocation,
					    &page->dma);
	if (page->vaddr) {
		memset (page->bitmap, 0xff, mapsize);	// bit set == free
		if (pool->flags & SLAB_POISON)
			memset (page->vaddr, POOL_POISON_BYTE, pool->allocation);
		list_add (&page->page_list, &pool->page_list);
	} else {
		kfree (page);
		page = 0;
	}
	return page;
}


static inline int
is_page_busy (int blocks, unsigned long *bitmap)
{
	while (blocks > 0) {
		if (*bitmap++ != ~0UL)
			return 1;
		blocks -= BITS_PER_LONG;
	}
	return 0;
}

static void
pool_free_page (struct pci_pool *pool, struct pci_page *page)
{
	dma_addr_t	dma = page->dma;

	if (pool->flags & SLAB_POISON)
		memset (page->vaddr, POOL_POISON_BYTE, pool->allocation);
	pci_free_consistent (pool->dev, pool->allocation, page->vaddr, dma);
	list_del (&page->page_list);
	kfree (page);
}


/**
 * pci_pool_destroy - destroys a pool of pci memory blocks.
 * @pool: pci pool that will be destroyed
 *
 * Caller guarantees that no more memory from the pool is in use,
 * and that nothing will try to use the pool after this call.
 */
void
pci_pool_destroy (struct pci_pool *pool)
{
	unsigned long		flags;

#ifdef CONFIG_PCIPOOL_DEBUG
	printk (KERN_DEBUG "pcipool destroy %s/%s\n",
		pool->dev ? pool->dev->slot_name : NULL,
		pool->name);
#endif

	spin_lock_irqsave (&pool->lock, flags);
	while (!list_empty (&pool->page_list)) {
		struct pci_page		*page;
		page = list_entry (pool->page_list.next,
				struct pci_page, page_list);
		if (is_page_busy (pool->blocks_per_page, page->bitmap)) {
			printk (KERN_ERR "pci_pool_destroy %s/%s, %p busy\n",
				pool->dev ? pool->dev->slot_name : NULL,
				pool->name, page->vaddr);
			/* leak the still-in-use consistent memory */
			list_del (&page->page_list);
			kfree (page);
		} else
			pool_free_page (pool, page);
	}
	spin_unlock_irqrestore (&pool->lock, flags);
	kfree (pool);
}


/**
 * pci_pool_alloc - get a block of consistent memory
 * @pool: pci pool that will produce the block
 * @mem_flags: SLAB_KERNEL or SLAB_ATOMIC
 * @handle: pointer to dma address of block
 *
 * This returns the kernel virtual address of a currently unused block,
 * and reports its dma address through the handle.
 * If such a memory block can't be allocated, null is returned.
 */
void *
pci_pool_alloc (struct pci_pool *pool, int mem_flags, dma_addr_t *handle)
{
	unsigned long		flags;
	struct list_head	*entry;
	struct pci_page		*page;
	int			map, block;
	size_t			offset;
	void			*retval;

restart:
	spin_lock_irqsave (&pool->lock, flags);
	list_for_each (entry, &pool->page_list) {
		int		i;
		page = list_entry (entry, struct pci_page, page_list);
		/* only cachable accesses here ... */
		for (map = 0, i = 0;
				i < pool->blocks_per_page;
				i += BITS_PER_LONG, map++) {
			if (page->bitmap [map] == 0)
				continue;
			block = ffz (~ page->bitmap [map]);
			if ((i + block) < pool->blocks_per_page) {
				clear_bit (block, &page->bitmap [map]);
				offset = (BITS_PER_LONG * map) + block;
				offset *= pool->size;
				goto ready;
			}
		}
	}
	if (!(page = pool_alloc_page (pool, mem_flags))) {
		if (mem_flags == SLAB_KERNEL) {
			DECLARE_WAITQUEUE (wait, current);

			current->state = TASK_INTERRUPTIBLE;
			add_wait_queue (&pool->waitq, &wait);
			spin_unlock_irqrestore (&pool->lock, flags);

			schedule_timeout (POOL_TIMEOUT_JIFFIES);

			current->state = TASK_RUNNING;
			remove_wait_queue (&pool->waitq, &wait);
			goto restart;
		}
		retval = 0;
		goto done;
	}

	clear_bit (0, &page->bitmap [0]);
	offset = 0;
ready:
	retval = offset + page->vaddr;
	*handle = offset + page->dma;
done:
	spin_unlock_irqrestore (&pool->lock, flags);

	return retval;
}


static struct pci_page *
pool_find_page (struct pci_pool *pool, dma_addr_t dma)
{
	unsigned long		flags;
	struct list_head	*entry;
	struct pci_page		*page;

	spin_lock_irqsave (&pool->lock, flags);
	list_for_each (entry, &pool->page_list) {
		page = list_entry (entry, struct pci_page, page_list);
		if (dma < page->dma)
			continue;
		if (dma < (page->dma + pool->allocation))
			goto done;
	}
	page = 0;
done:
	spin_unlock_irqrestore (&pool->lock, flags);
	return page;
}


/**
 * pci_pool_free - put block back into pci pool
 * @pool: the pci pool holding the block
 * @vaddr: virtual address of block
 * @dma: dma address of block
 *
 * Caller promises neither device nor driver will again touch this block
 * unless it is first re-allocated.
 */
void
pci_pool_free (struct pci_pool *pool, void *vaddr, dma_addr_t dma)
{
	struct pci_page		*page;
	unsigned long		flags;
	int			map, block;

	if ((page = pool_find_page (pool, dma)) == 0) {
		printk (KERN_ERR "pci_pool_free %s/%s, %p/%x (bad dma)\n",
			pool->dev ? pool->dev->slot_name : NULL,
			pool->name, vaddr, (int) (dma & 0xffffffff));
		return;
	}
#ifdef	CONFIG_PCIPOOL_DEBUG
	if (((dma - page->dma) + (void *)page->vaddr) != vaddr) {
		printk (KERN_ERR "pci_pool_free %s/%s, %p (bad vaddr)/%x\n",
			pool->dev ? pool->dev->slot_name : NULL,
			pool->name, vaddr, (int) (dma & 0xffffffff));
		return;
	}
#endif

	block = dma - page->dma;
	block /= pool->size;
	map = block / BITS_PER_LONG;
	block %= BITS_PER_LONG;

#ifdef	CONFIG_PCIPOOL_DEBUG
	if (page->bitmap [map] & (1UL << block)) {
		printk (KERN_ERR "pci_pool_free %s/%s, dma %x already free\n",
			pool->dev ? pool->dev->slot_name : NULL,
			pool->name, dma);
		return;
	}
#endif
	if (pool->flags & SLAB_POISON)
		memset (vaddr, POOL_POISON_BYTE, pool->size);

	spin_lock_irqsave (&pool->lock, flags);
	set_bit (block, &page->bitmap [map]);
	if (waitqueue_active (&pool->waitq))
		wake_up (&pool->waitq);
	/*
	 * Resist a temptation to do
	 *    if (!is_page_busy(bpp, page->bitmap)) pool_free_page(pool, page);
	 * it is not interrupt safe. Better have empty pages hang around.
	 */
	spin_unlock_irqrestore (&pool->lock, flags);
}

int pci_dma_supported(struct pci_dev *hwdev, u64 mask)
{
  /*                                                                             
   * we fall back to GFP_DMA when the mask isn't all 1s,                         
   * so we can't guarantee allocations that must be                              
   * within a tighter range than GFP_DMA..                                       
   */
  if(mask < 0x00ffffff)
    return 0;

  return 1;
}
