#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>

#include "uml.h"

static int printCheck(int fd, unsigned long checkId) {
  struct uml_ioc_checkinfo_struct cis;
  int isPaused = 1;

  memset(&cis, 0, sizeof(cis));
  cis.checkId = checkId;
  ioctl(fd, UML_IOC_GET_CHECKINFO, &cis);

  if((cis.guestSyscallCount == 0) && (checkId == INVALID_CHECK_ID)) {
    isPaused = 0;
    ioctl(fd, UML_IOC_PAUSE, 0);
    ioctl(fd, UML_IOC_GET_CHECKINFO, &cis);
  }

  if((cis.checkId != INVALID_CHECK_ID) ||
     (checkId == INVALID_CHECK_ID)) {
    printf("checkId:      %lu\n",cis.checkId);
    printf("guestSyscall: %llu\n",cis.guestSyscallCount);
    printf("scheduleId:   %lu\n",cis.revirtId.rhid);
    printf("branchCount:  %lld\n",cis.revirtId.bc);
    printf("\n");
  }

  return isPaused;
}

int main(int argc, char *argv[]) {
  int opt;
  int checkpoint = 0;
  int rollback = 0;
  int restore = 0;
  int foundId = 0;
  int restart=1;
  int ret;
  int fd;
  unsigned long checkId;
  int listCheck = 0;
  unsigned long idx;

  struct uml_ioc_checkpoint_struct cps;
  struct uml_ioc_rollback_struct rbs;

  memset(&cps, 0, sizeof(cps));
  memset(&rbs, 0, sizeof(rbs));

  while((opt = getopt(argc,argv,"qlcr:i:pun")) != -1) {
    switch(opt) {
    case 'n':
	restart=0;
	break;
    case 'c':
      checkpoint = 1;
      break;
    case 'r':
      rollback = 1;
      ret = sscanf(optarg,"%d",&restore); assert(ret == 1);
      break;
    case 'i':
      ret = sscanf(optarg,"%lu",&checkId); assert(ret == 1);
      foundId = 1;
      break;
    case 'l':
      listCheck = 1;
      break;
    case 'p':
      printf("pausing the virtual machine\n");
      fd = open("/dev/uml", O_RDWR);
      if(fd < 0) {
	printf("could not open device, not reset\n");
	return 1;
      }
      ioctl(fd, UML_IOC_PAUSE, 0);
      return 0;
    case 'u':
      printf("unpauseing the virtual machine\n");
      fd = open("/dev/uml", O_RDWR);
      if(fd < 0) {
	printf("could not open device, not reset\n");
	return 1;
      }
      ioctl(fd, UML_IOC_RESTART_UML, 0);
      printf("virtual machine restarted\n");
      return 0;
    case 'q':
      printf("resetting device\n");
      fd = open("/dev/uml", O_RDWR);
      if(fd < 0) {
	printf("could not open device, not reset\n");
	return 1;
      }
      //ioctl(fd, UML_IOC_RESTART_UML, 0);
      ioctl(fd, UML_IOC_ERESET_UML, 0);
      printf("device reset\n");
      return 0;
      break;
    }
  }

  assert(checkpoint || rollback || listCheck);
  assert(!(checkpoint && rollback));
  assert(!(checkpoint && listCheck));
  assert(!(listCheck && rollback));
  if(rollback) {
    assert(foundId);
  }
  
  
  fd = open("/dev/uml",O_RDWR);
  assert(fd >= 0);

  if(checkpoint) {
    ret = read(fd, &cps, sizeof(cps));
    if(ret == -EINTR) return 0;
    assert(ret == sizeof(cps));
    printf("%lu\n",cps.checkId);
    if(restart)
	ioctl(fd,UML_IOC_RESTART_UML,0);
  } else if(rollback) {
    rbs.checkId = checkId;
    rbs.restore = restore;
    ret = write(fd, &rbs, sizeof(rbs));
    if(ret == -EINTR) return 0;
    assert(ret == sizeof(rbs));
    if(restore == 0) {
      printf("checkpoint deleted\n");
    } else {
      printf("uml rolled back\n");
    }
    if(restart)
	ioctl(fd,UML_IOC_RESTART_UML,0);
  } else if(listCheck) {
    int isPaused = printCheck(fd, INVALID_CHECK_ID);
    for(idx = 0; idx < MAX_UML_CHECK; idx++) {
      printCheck(fd, idx);
    }
    if(!isPaused && restart) {
      ioctl(fd, UML_IOC_RESTART_UML, 0);
    }
  } else {
    assert(0);
  }

  return 0;
}
