#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>

#include "uml.h"


int main(int argc, char *argv[]) {
  int opt;
  int checkpoint = 0;
  int restore = 0;
  int delete = 0;
  int reset = 0;
  int fd;
  unsigned long checkId;
  int ret;

  while((opt = getopt(argc,argv,"cr:d:q")) != -1) {
    switch(opt) {
    case 'c':
      checkpoint = 1;
      break;
    case 'r':
      restore = 1;
      ret = sscanf(optarg,"%lu",&checkId); assert(ret == 1);
      break;
    case 'd':
      delete = 1;
      ret = sscanf(optarg,"%lu",&checkId); assert(ret == 1);
      break;
    case 'q':
      reset = 1;
      break;
    default:
      assert(0);
    }
  }

  if(!checkpoint && !restore && !delete && !reset) {
    printf("Usage %s: [-c] [-r checkId] [-d checkId]\n\n",argv[0]);
    printf("-c  take checkpoint\n");
    printf("-r  restore checkpoint checkId\n");
    printf("-d  delete checkpoint checkId\n");
    printf("-q  reset /dev/umlgdb device\n");
    return 1;
  }
  
  fd = open("/dev/umlgdb",O_RDWR);
  assert(fd >= 0);

  if(checkpoint) {
    ioctl(fd,PLATO_IOC_CHECKPOINT,&checkId);
    printf("took checkpoint %lu\n",checkId);
  }
  
  if(restore) {
    printf("restoring checkpoint %lu\n",checkId);    
    ioctl(fd,PLATO_IOC_RESTORE_CHECK,checkId);
  } 

  if(delete) {
    printf("deleting checkpoint %lu\n",checkId);
    ioctl(fd, PLATO_IOC_DELETE_CHECK, checkId);
  }
  
  if(reset) {
    printf("reseting /dev/umlgdb device\n");
    ioctl(fd, UMLGDB_IOC_RESET,0);
  }

  return 0;
}
