#ifndef __INTROSPECT_GDB_H__
#define __INTROSPECT_GDB_H__

#include <asm/ptrace.h>

typedef void (*breakCallback)(void);

#include <string>
#include <vector>
#include <pthread.h>
#include "uml.h"

class introspect_gdb {
 public:
  introspect_gdb(int gdbStdin, int gdbStdout, int gdbStderr);
  int processCmd(std::string *cmd);
  ~introspect_gdb(void);
  void stopThread(void);
  void cont(void);
  void readFdBuffer(int readFd, std::string *str);
  void waitForBreak(std::string *resStr);  
  void waitForBreak(void) {
    waitForBreak(NULL);
  }
 private:
  void mysplit(std::string str, char * delim, std::vector< std::string > *vec);
  void dbgUsrCmd(std::string arg);
  void printTimeDiff(struct timeval *tvStart, struct timeval *tvStop);
  void flushHostFileCache(unsigned long numIter);
  void runTest(void);
  unsigned long getCurrEip(void);
  void checkPerCmd(std::string arg);
  void gotoCmd(std::string arg, std::string *strOut);
  void gotoCmd(unsigned long long guestSyscallCount, std::string *strOut);
  void clearTfAfterSyscall(void);
  void hdWatch(std::string arg);
  static long long abserr(unsigned long checkPos,
			  unsigned long refPos,
			  struct uml_ioc_checkinfo_struct *currInfo,
			  struct uml_ioc_checkinfo_struct *checkList,
			  unsigned long numCheck);
  static void removeOneCheckpoint(introspect_gdb *gdb, 
				  struct uml_ioc_checkinfo_struct *currInfo,
				  struct uml_ioc_checkinfo_struct *checkList,
				  unsigned long numCheck);
  void deleteCheckpoint(introspect_gdb *gdb, unsigned long checkId);
  unsigned long readUnsigned(const char *varName);
  static void takeCheckpoint(introspect_gdb *gdb);
  void setCpCmd(int cmd);
  int isInfoEqual(struct uml_ioc_checkinfo_struct *cis0,
		  struct uml_ioc_checkinfo_struct *cis1);
  unsigned long setBreak(void);
  void deleteBreak(unsigned long breakNo);
  void sortCheckList(struct uml_ioc_checkinfo_struct *checkList);
  int atOrPast(struct uml_ioc_checkinfo_struct *targetInfo);
  void goBack(unsigned long checkId, 
	      struct uml_ioc_checkinfo_struct *finalPosInfo,
	      unsigned long long targetNumBreaks, 
	      unsigned long long *numBreaks,
	      int step,
	      std::string *strOut,
	      unsigned long breakNo);
  void revBreak(int step, std::string arg, std::string *strOut);
  unsigned long getCheckInfo(struct uml_ioc_checkinfo_struct *currInfo,
		    struct uml_ioc_checkinfo_struct **checkList);
  void getCheckInfo(struct uml_ioc_checkinfo_struct *currInfo) {
    getCheckInfo(currInfo, NULL);
  }
  void rollback(unsigned long rollbackCheckId, int step,
		unsigned long long guestSyscallCount);
		
  void startThread(void);
  static void *checkpointThread(void *arg);
  void printCheck(struct uml_ioc_checkinfo_struct *cis);
  void listCheck(void);
  unsigned long currentAccess(char *varName, unsigned long val, int isRead);
  int strMatchEnd(std::string *str, const char *c_str);
  int gdbStdin, gdbStdout, gdbStderr;
  int umlFd;
  pthread_t cpThread;
  int killThread;
  int threadRunning;
  int rollbackCalled;
  int rollbackCmdWaiting;
  int checkpointCmd;
  int checkpointState;
  struct uml_ioc_rollback_struct rbs;
  pthread_mutex_t cmdLock; 
  unsigned long checkPeriod;
};

#endif
