#include <iostream>
#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <sys/select.h>
#include <sys/time.h>
#include <fcntl.h>
#include <sys/ioctl.h>

#include <string>
#include <fstream>

#include "introspect_gdb.h"
#include "../revirt-skas-2.4.18/uml/uml.h"

using namespace std;
pid_t childPid;
introspect_gdb *gdb = NULL;

static void sigHandler(int signo) {
  if(signo == SIGCHLD) {
    if(gdb != NULL) {
      delete gdb;
    }
    exit(0);
  } else if(signo == SIGINT) {
    int fd = open("/dev/umlgdb", O_RDWR);
    if (fd >= 0) {
      ioctl(fd, UMLGDB_IOC_BREAK_NEXT, NULL);
      close(fd);
    }
    kill(childPid, SIGINT);
  } else if(signo == SIGIO) {
    if(getpid() != childPid) {
      kill(SIGIO, childPid);
    }
  }
}

#define PIPE_READ  0
#define PIPE_WRITE 1

int main(int argc, char *argv[]) {
  int ret;
  int gdbStdin[2], gdbStdout[2], gdbStderr[2];
  fd_set readSet;
  int maxFd;
  string str, cmdStr;
  int cmdTaken;
  ifstream inFile;
  char tmpLine[256];
  bool initData;
  string tmpStr;

  signal(SIGINT, sigHandler);
  signal(SIGCHLD, sigHandler);
  signal(SIGIO, sigHandler);

  // fork off a child of gdb
  ret = pipe(gdbStdin); assert(ret == 0);
  ret = pipe(gdbStdout); assert(ret == 0);
  ret = pipe(gdbStderr); assert(ret == 0);

  ret = fork();
  if(ret == 0) {
    // we are the child, become gdb
    dup2(gdbStdin[PIPE_READ],STDIN_FILENO);
    dup2(gdbStdout[PIPE_WRITE],STDOUT_FILENO);
    dup2(gdbStderr[PIPE_WRITE],STDERR_FILENO);
    //ret = execlp("gdb","gdb",NULL);
    ret = execvp("gdb",&(argv[1]));
    /*
    if(argc == 1) {
      ret = execlp("gdb","gdb",NULL);
    } else if(argc == 2) {
      ret = execlp("gdb","gdb",argv[1],NULL);
    } else if (argc == 3) {
      ret = execlp("gdb","gdb",argv[1],argv[2],NULL);
    } else {
      assert(0);
    }
    */
    // this should not return
    assert(0);

  } else {
    childPid = ret;
  }

  gdb = new introspect_gdb(gdbStdin[PIPE_WRITE], gdbStdout[PIPE_READ],
			   gdbStderr[PIPE_READ]);

  inFile.open(".mygdbinit");
  cmdTaken = 0;
  gdb->waitForBreak(&tmpStr);
  cout << tmpStr;
  while(1) {
    maxFd = (gdbStdout[PIPE_READ] > gdbStderr[PIPE_READ]) ? gdbStdout[PIPE_READ] :
      gdbStderr[PIPE_READ];

    maxFd = (STDIN_FILENO > maxFd) ? STDIN_FILENO : maxFd;

    FD_ZERO(&readSet);
    FD_SET(gdbStdout[PIPE_READ], &readSet);
    FD_SET(gdbStderr[PIPE_READ], &readSet);
    FD_SET(STDIN_FILENO, &readSet);
    initData = false;

    if(inFile.eof() || !inFile.is_open()) {
      ret = select(maxFd+1, &readSet, NULL, NULL, NULL);
    } else {
      inFile.getline(tmpLine, sizeof(tmpLine));
      initData = !inFile.eof();
      FD_ZERO(&readSet);
    }

    if((ret > 0) || initData) {
      
      if(FD_ISSET(gdbStderr[PIPE_READ], &readSet)) {
	gdb->readFdBuffer(gdbStderr[PIPE_READ],&str);
	write(STDERR_FILENO,str.c_str(), str.size());
      }
      if(FD_ISSET(gdbStdout[PIPE_READ], &readSet)) {
	gdb->readFdBuffer(gdbStdout[PIPE_READ], &str);
	if(str.find("Program received signal SIGKILL") != string::npos) {
	  gdb->stopThread();
	  gdb->cont();
	} else {
	  write(STDOUT_FILENO, str.c_str(), str.size());
	}
      }
      
      if(FD_ISSET(STDIN_FILENO, &readSet) || initData) {
	if(FD_ISSET(STDIN_FILENO, &readSet)) {
	  gdb->readFdBuffer(STDIN_FILENO, &str);
	} else {
	  assert(initData);
	  str = string(tmpLine) + string("\n");
	  cout << str;
	}
	if(!gdb->processCmd(&str)) {
	  // it was not one of our commands, pass it to gdb
	  if((str.size() == 1) && cmdTaken) {
	    gdb->processCmd(&cmdStr);
	  } else {
	    cmdTaken = 0;
	    write(gdbStdin[PIPE_WRITE],str.c_str(),str.size());
	    if(initData) {
	      tmpStr.clear();
	      gdb->waitForBreak(&tmpStr);
	      cout << tmpStr;
	    }
	  }
	} else {
	  cmdTaken = 1;
	  cmdStr = str;
	}
      }
    }

  }

  return 0;

  /*
  // start the target up
  waitForBreak(gdbStdout[0]);
  ret = write(gdbStdin[1],START_TARGET_STR,strlen(START_TARGET_STR));
  assert(((unsigned) ret) == strlen(START_TARGET_STR));

  // now give the user a chance to set up breakpoints
  waitForBreak(gdbStdout[0]);
  initCb();
  gdbCont();

  // enter the event loop
  string stdoutBuf;
  while(1) {
    stdoutBuf.erase();
    waitForBreak(gdbStdout[0], &stdoutBuf);

    // find the event handler and call it
    callEventHandler(&stdoutBuf);

    if(cont) {
      gdbCont();
    }
  }
  */
}
