#!/usr/bin/perl
use strict;
use Digest::MD5;
use Getopt::Std;

my %opts;

getopts("o:b:c:g:s:u:",\%opts);
if(($opts{"o"} eq "") || ($opts{"b"} eq "") || ($opts{"c"} eq "") ||
   ($opts{"g"} eq "") || ($opts{"s"} eq "")) {
    print STDERR "Usage: -o objectFile -b basename -c break.conf -g globals.conf " .
	"-s structs.conf\n";
    exit(1);
}

my $OBJ_FILE = $opts{"o"};
my $MD5_CACHE = $opts{"b"} . ".md5";
my $BASENAME = $opts{"b"};
my $BREAK_FILE = $opts{"c"};
my $GLOBAL_FILE = $opts{"g"};
my $STRUCT_FILE = $opts{"s"};

my $ctx = Digest::MD5->new;

my $md5;

my $genDbgInfo = 1;
if ((-e $MD5_CACHE) && (-e $OBJ_FILE)) {
    my $cached = `cat $MD5_CACHE`;
    
    open(FILE, "<$OBJ_FILE") || die("could not open $OBJ_FILE");
    $ctx->addfile(*FILE);
    $md5 = $ctx->b64digest();
    
    if($md5 eq $cached) {
	$genDbgInfo = 0;
    }
    close(FILE);
}

if(!(-e $OBJ_FILE)) {
    print STDERR "OBJ_FILE does not exist, compile the guest kenrel!\n";
    exit(1);
}

if($genDbgInfo) {
    print "creating md5 sum of $BASENAME file\n";
    open(FILE, "<$OBJ_FILE") || die("could not open local linux file");
    $ctx = Digest::MD5->new();
    $ctx->addfile(*FILE);
    $md5 = $ctx->b64digest();
    system("echo -n $md5 > $MD5_CACHE");
    close(FILE);

    print "extracting debugging information\n";
    system("readelf -wi $OBJ_FILE > $BASENAME.info.txt");
    system("objdump -D $OBJ_FILE > $BASENAME.S");
}

if($opts{"u"} eq "") {
    print "creating breakpoint information for $BASENAME\n";
    system("./parseBreaks.pl $OBJ_FILE $BREAK_FILE $BASENAME.info.txt $BASENAME.S >> plato.conf");
    print "creating global variable context for $BASENAME\n";
    system("./parseGlobals.pl $OBJ_FILE $GLOBAL_FILE >> plato.conf");
} else {
    print "creating breakpoint information for $BASENAME, comm = " . $opts{"u"} .
	"\n";
    system("./parseBreaks.pl $OBJ_FILE $BREAK_FILE $BASENAME.info.txt $BASENAME.S " .
	   $opts{"u"} . " >> plato.conf");
    print "creating global variable context for $BASENAME, comm = " . $opts{"u"} . 
	"\n";
    system("./parseGlobals.pl $OBJ_FILE $GLOBAL_FILE " . $opts{"u"} . 
	   " >> plato.conf");
}
print "creating guest struct information for $BASENAME\n";
system("./parseStructs.pl $BASENAME.info.txt $STRUCT_FILE $BASENAME >> guestStructs.h");

