#!/usr/bin/perl
use strict;

sub extractEndString {
    my($str) = @_;

    $_ = $str;
    if($str eq "") {return "";}
    if(! /(\S+)\s*$/ ) { die "could not parse key result ($str)\n"; }

    return $1;
}

sub extractEndNum {
    my($str) = @_;
    my $num;
    
    $_ = $str;
    if( ! $_ ) { die "could not find key\n"; }

    if(! /([\d,-]+)\s*$/ ) { die "could not parse key result ($str)\n"; }

    $num = $1;
    if($num < 0) {
      $num = hex('0xffffffff') + $num + 1;
    }

    return $num;
}

sub extractEndParen {
    my($str) = @_;

    $_ = $str;
    if( ! $_ ) { die "could not find paren key\n"; }

    if(! /\((.*)\)\s*$/ ) { die "could not parse key result ($str)\n"; }
    
    return $1;
}

sub getNextFormData($$) {
    my($line,$currPos) = @_;
    my(%map);
    my($id, $dwType, $type);
    my($cont);

    # look for the next form
    do {
	$_ = $line;
	if( /^ \<(\w+)\>\<(\w+)\>:/ ) {
	    $dwType = $1;
	    $id = $2;
	    # now parse the trailing string
	    $type = extractEndString($line);

	    # get as many data items as possible
	    $cont = 1;
	    while ($cont) {
		$line = <INFO_FILE>;
		$_ = $line;
		if( /^     (\w+)/ ) {
		    $map{$1} = $line;
		} else {
		    $cont = 0;
		}
	    }
	    return ($dwType, $id, $type, $line, $currPos, %map);
	} 
	$currPos = tell(INFO_FILE);
    } while ($line = <INFO_FILE>);

     print "eof\n";

     return ("","","","","",0);
 }

 sub getNextAddr {
     my($breakAddr, $asmFile) = @_;
     my($line,$getNext);

     open(VMLINUX_FILE,"<$asmFile") or die "could not open $asmFile\n";

     #look for the address and return the next address
     $getNext = 0;
     while($line = <VMLINUX_FILE>) {
	 $_ = $line;
	 if(/^\s*(\w+)\:/) {
	     if($getNext) {
		 close(VMLINUX_FILE);
		 return $1;
	     }

	     if($1 eq $breakAddr) {
		 $getNext = 1;
	     }
	 }
     }

     die "did not find the break point address $breakAddr in the file $asmFile\n";
     return 0;
 }

 sub findEndBreakAddr {
     my($fileName, $lineNo, $asmFile, $endStr, $gdbCmd, $objFile) = @_;
     my($gdbRet);
     my($line, @lines);
     my($breakAddr, $nextAddr);

     #qurey gdb to see if we can figure it out
     $gdbRet = `./queryGdb $objFile $gdbCmd`;
     @lines = split(/\n/,$gdbRet);
     $breakAddr = "";
     foreach $line (@lines) {

	 $_ = $line;
	 # keep doing this and get the last one
	 if($endStr eq "end") {
	     if( /and ends at 0x(\w+)/ ) {
		 $breakAddr = $1;
	     }
	 } else {
	     if(($endStr eq "begin") || ($endStr eq "func")) {
		 if( /starts at address 0x(\w+)/ ) {
		     $breakAddr = $1;
		 } elsif ($endStr eq "func") {
		     if(/address 0x(\w+)/) {
			 $breakAddr = $1;
		     }
		 }
	     } else {
		 die "end string not valid\n";
	     }
	 }
     }

     if($breakAddr eq "") { die "end = $endStr could not parse the break addr from gdb $gdbRet\n"; }
     $nextAddr = getNextAddr($breakAddr, $asmFile);

     return ($breakAddr, $nextAddr);
 }

 sub addrInRange {
     my($addr,$lowPc, $highPc) = @_;
     $addr = hex($addr);

     return ($addr >= $lowPc) && ($addr <= $highPc);
 }

 sub isMatch($$$) {
     my ($breakAddrArr, $nextAddrArr, $map) = @_;
     my ($lowPc, $highPc);
     my ($num, $idx);

     if($map->{"DW_AT_low_pc"} && $map->{"DW_AT_high_pc"}) {
	 $lowPc = extractEndNum($map->{"DW_AT_low_pc"});
	 $highPc = extractEndNum($map->{"DW_AT_high_pc"});		

	 $num = @$breakAddrArr;
	 for($idx = 0; $idx < $num; $idx++) {
	     if( addrInRange($breakAddrArr->[$idx], $lowPc, $highPc) &&
		 addrInRange($nextAddrArr->[$idx], $lowPc, $highPc) ) {
		 return $idx;
	     }
	 }
     }

     return -1;
 }

 sub getFunctionParameters($$$) {
     my($breakAddrArr, $nextAddrArr,$currPosIn) = @_;
     my($functionName, $frameBase, $prototype, %parameters);
     my($dwType, $id, $type, $line, %map);
     my($lowPc, $highPc);
     my($varName, $varLoc);
     my($matchIdx);
     my($currPosOut, $dummy);

     seek(INFO_FILE, $currPosIn, 0);
     $currPosOut = $currPosIn;
     do {
	($dwType, $id, $type, $line, $currPosOut, %map) = getNextFormData($line, $currPosOut);

	if($line ne "") {
	    if($type eq "(DW_TAG_subprogram)") {
		# see if it is our function
		$matchIdx = isMatch($breakAddrArr, $nextAddrArr, \%map);
		if($matchIdx >= 0) {
		    $lowPc = extractEndNum($map{"DW_AT_low_pc"});
		    $highPc = extractEndNum($map{"DW_AT_high_pc"});		
		    # we found the function
		    
		    $functionName = extractEndString($map{"DW_AT_name"});
		    $frameBase = extractEndParen($map{"DW_AT_frame_base"});
		    
		    # see if we have a non-standard calling convension
		    $prototype = extractEndString($map{"DW_AT_prototyped"});
		    
		    # find all variables and function arguments
		    do {
			($dwType, $id, $type, 
			 $line, $dummy, %map) = getNextFormData($line, $currPosOut);
			
			if(($type eq "(DW_TAG_variable)") ||
			   ($type eq "(DW_TAG_formal_parameter)") ) {
			    # only add variable that define a location
			    if($map{"DW_AT_name"} && $map{"DW_AT_location"}) {
				$varName = extractEndString($map{"DW_AT_name"});
				$varLoc = extractEndParen($map{"DW_AT_location"});
				
				if(($varName eq "") || ($varLoc eq "")) {
				    die "could not get the var name or loc\n";
				}
				if($parameters{$varName}) {
				    die "var $varName already found\n";
				}
				$parameters{$varName} = $varLoc;
			    }
			}
		    } while ($line && (($type ne "(DW_TAG_subprogram)")&&
				       ($type ne "(DW_TAG_lexical_block)")));
		    
		    return ($currPosOut, $matchIdx, $functionName, $frameBase, $prototype, $lowPc, %parameters);
		}
	    }
	}
    } while ($line);

    die "could not find the fuction that the break point is in\n";
}
	    
sub printFunction($$$$$$$$$%) {
    my ($line, $end, $breakAddr, $nextAddr, $functionName, $frameBase, $prototype, $lowPc, $userModeComm, %parameters) = @_;
    my(@keys,$varName,$varLoc,$offset);

    @keys = keys(%parameters);
    print "***BEGIN\n";
    if($userModeComm ne "") {
	if(index($userModeComm,"/") >= 0) {
	    print "userModeSharedLib $userModeComm\n";
	} else {
	    print "userModeComm $userModeComm\n";
	}
    }
    print "functionName $functionName\n";
    if($end eq "func") {
	if(hex($breakAddr) != $lowPc) { die "entryAddr not at func beg"; }
	print "entryAddr $breakAddr\n";
	if($prototype) {
	    print "prototyped $prototype\n";
	}
    } else {
	print "breakId $line\n";
	print "breakAddr $breakAddr\n";
	print "nextAddr $nextAddr\n";
	if(hex($breakAddr) == $lowPc) {
	    print "atFunctionStart\n";
	}
    }
    $_ = $frameBase;
    if( ! /DW\_OP\_reg(\d+)/ ) { die " could not get reg no framebase\n"; }
    print "frameBase DW_OP_reg $1\n";
    
    foreach $varName (@keys) {
	$varLoc = $parameters{$varName};
	#see if there is additional data to extract
	print "variable $varName ";
	if(index($varLoc,"DW_OP_addr") >= 0) {
	    $_ = $varLoc;
	    if( ! /(\w+)\:\s+(\w+)\s*$/ ) { die "could not parse1 $varLoc\n"; }
	    $offset = hex($2);
	    print "$1 $offset\n";
	} else {
	    if(index($varLoc,"DW_OP_fbreg") >= 0) {
		$_ = $varLoc;
		if( ! /(\w+)\:\s+(\-{0,1}\w+)\s*$/ ) {
		    die "could not parse2 $varLoc\n";
		}
		print "$1 $2\n";
	    } else {
		if(index($varLoc,"DW_OP_reg") < 0) { die "unknow var loc\n"; }
		$_ = $varLoc;
		if( ! /DW\_OP\_reg(\d+)/ ) { die " could not get reg no\n"; }
		print "DW_OP_reg $1\n";
	    }
	}
    }
    print "***END\n";
    
}

sub main {
    my($objFile, $confFile, $infoFile, $asmFile, $userModeComm);
    my($line, @breaks, $fileName, $lineNo);
    my($breakAddr, $nextAddr,$lowPc);
    my($functionName, $frameBase, $prototype, %parameters);
    my($end,$gdbCmd);
    my(@lineArr, @endArr, @breakAddrArr, @nextAddrArr,$matchIdx);

    ($objFile, $confFile, $infoFile, $asmFile, $userModeComm) = @ARGV;

    if(!$objFile || !$confFile || !$infoFile || !$asmFile) {
	print "usage: parseBreaks.pl object_file conf_file info_file vmlinux.S " .
	    " [userModeComm]\n";
	exit 1;
    }

    if(!(-e $objFile) || !(-e $confFile) || !(-e $infoFile) || !(-e $asmFile)) {
	die "one of your command line argument file names does not exist\n";
    }

    # open the files config file and process each of the entries
    open(CONF_FILE,"<$confFile") or die "could not open $confFile\n";

    while( $line = <CONF_FILE> ) {
	$_ = $line;
        if(! /^\#.*/ ) {
            $_ = $line;
            if( /\S+/ ) {
		$gdbCmd = "";
		$_ = $line;
		if( /^\s*(\S+)\s+(\".+\")\s*$/ ) {
		    $line = $1;
		    $gdbCmd = $2;
		} else {
		    $_ = $line;
		    if(! /^\s*(\S+)\s*$/ ) { die "could not parse the line"; }
		    $line = $1;
		}
		
		($fileName, $lineNo, $end) = split(/:/,$line);
		if(($fileName eq "") || !$lineNo) {
		    die "malfomed $confFile entry: $line\n";
		}
	
		if(($end eq "end") || ($end eq "begin") || ($end eq "func")) {
		    ($breakAddr, $nextAddr) = findEndBreakAddr($fileName, 
							       $lineNo, 
							       $asmFile,
							       $end,
							       $gdbCmd,
							       $objFile);
		} else {
		    die("$line use queryGdb, no reason for this");
		}

		push(@lineArr, $line);
		push(@endArr, $end);
		push(@breakAddrArr, $breakAddr);
		push(@nextAddrArr, $nextAddr);
            }
        }
    }

    close(CONF_FILE);

    open(INFO_FILE,"<$infoFile") or die "could not open the file $infoFile\n";
    my $currpos = 0;
    my $totalArr = @lineArr;
    my $idx;
    for($idx = 0; $idx < $totalArr; $idx++) {
	($currpos, $matchIdx, $functionName, $frameBase, 
	 $prototype, $lowPc, %parameters) =  getFunctionParameters(\@breakAddrArr, 
								   \@nextAddrArr,
								   $currpos);
	
	printFunction($lineArr[$matchIdx], $endArr[$matchIdx], 
		      $breakAddrArr[$matchIdx], 
		      $nextAddrArr[$matchIdx], $functionName, 
		      $frameBase, $prototype, $lowPc, 
		      $userModeComm, %parameters);
	

	#clear out the old values after we read them in
	$breakAddrArr[$matchIdx] = 0;
	$nextAddrArr[$matchIdx] = 0;
    }
    
    close(INFO_FILE);
}


main();
