#!/usr/bin/perl
use strict;

sub extractEndString {
    my($str) = @_;

    $_ = $str;
    if($str eq "") {return "";}
    if(! /(\S+)\s*$/ ) { die "could not parse key result ($str)\n"; }

    return $1;
}

sub extractEndNum {
    my($str) = @_;

    $_ = $str;
    if( ! $_ ) { die "could not find key\n"; }

    if(! /(\d+)\s*$/ ) { die "could not parse key result ($str)\n"; }

    return $1;
}

sub getNextFormData {
    my($line) = @_;
    my(%map);
    my($id, $dwType, $type);
    my($cont);

    # look for the next form
    do {
	$_ = $line;
	if( /^ \<(\w+)\>\<(\w+)\>:/ ) {
	    $dwType = $1;
	    $id = $2;
	    # now parse the trailing string
	    $type = extractEndString($line);

	    # get as many data items as possible
	    $cont = 1;
	    while ($cont) {
		$line = <ELF>;
		$_ = $line;
		if( /^     (\w+)/ ) {
		    $map{$1} = $line;
		} else {
		    $cont = 0;
		}
	    }

	    return ($dwType, $id, $type, $line, %map);
	} 
    } while ($line = <ELF>);

    print "eof\n";

    return ("","","","",0);
}



sub getStructIdAndSize {
    my($structName) = @_;
    my($dwType, $id, $type, $line, %map);

    do {
	($dwType, $id, $type, $line, %map) = getNextFormData($line);
	if($line) {
	    # check on the name 
	    if( ($type eq "(DW_TAG_structure_type)") &&
		(extractEndString($map{"DW_AT_name"}) eq $structName)) {
		# leave ELF open so we can read the member variables next
		return ($line, $id, extractEndNum($map{"DW_AT_byte_size"}));
	    }
	}
    } while ($line);

    die "could not find struct\n";
    return ("",0,0);
}

sub findStruct {
    my($elfFile, $structName, @structMembers) = @_;
    my($id);
    my($structSize);
    my($member);
    my($dwType, $type, $line, %map);
    my($offset);
    my(%memberMap);
    my($name);

    open(ELF,"<$elfFile") or die "could not open elf file $elfFile\n";

    ($line,$id,$structSize) = getStructIdAndSize($structName);

    do {
	($dwType, $id, $type, $line, %map) = getNextFormData($line);
	
	if($line) {
	    # see if we are at the end of the struct
	    if(($type ne "(DW_TAG_member)") && ($type ne "(DW_TAG_typedef)")) {
		$line = "";
	    } else {
		if($type eq "(DW_TAG_member)") {
		    #check if it is one of our members
		    foreach $member (@structMembers) {
			$name = extractEndString($map{"DW_AT_name"});
			if($name eq $member) {
			    if($memberMap{$member} ne "") { 
				die "member already exists for struct $structName member $member\n";
			    }
			    $_ = $map{"DW_AT_data_member_location"};
			    if( ! /\(DW_OP_plus_uconst\: (\d+)\)\s*$/ ) {
				die "could not parse constant\n";
			    }
			    $memberMap{$member} = $1;
			}
		    }
		}
	    }
	}

    } while ($line);    

    # now make sure we found all of our members
    foreach $member (@structMembers) {
	if($memberMap{$member} eq "") { die "could not find member $member\n"; }
    }

    close(ELF);

    return ($structSize, %memberMap);
}
    

sub main {
    my($elfFile);
    my($confFile);
    my($baseName);
    my(@structs);
    my($line);
    my(%memberMap);
    my(@values);
    my(@keys);
    my($offset,$name,$structName,$structSize);
    my(@structArg);
    my($fill, $currIdx,$dummyIdx,$numPrint);

    $elfFile = $ARGV[0];
    $confFile = $ARGV[1];
    $baseName = $ARGV[2];

    if(!$elfFile || !$confFile || !$baseName) {
	print "usage: parseStructs.pl info_file conf_file basename\n";
	exit 1;
    }
    if(!(-e $elfFile)) {
	print "the file $elfFile does not exists\n";
	exit 1;
    }

    if(!(-e $confFile)) {
	print "the file $confFile does not exist\n";
	exit 1;
    }

    # open the files config file and process each of the entries
    open(CONFIG,"<$confFile") or die "could not open $confFile\n";

    while( $line = <CONFIG> ) {
	$_ = $line;
        if(! /^\#.*/ ) {
            $_ = $line;
            if( /\s*(\w+)\s*/ ) {
                push @structs, $line;
            }
        }
    }

    print "#ifndef __GUEST_" . $baseName . "_STRUCTS_H__\n";
    print "#define __GUEST_" . $baseName . "_STRUCTS_H__\n\n";

    foreach $line (@structs) {
	$_ = $line;
	@structArg = split();
	$structName = $structArg[0];
	($structSize, %memberMap) = findStruct($elfFile,@structArg);

	@values = sort {$a <=> $b} values(%memberMap);
	@keys = keys(%memberMap);

	#print out the struct
	print "struct " . $baseName . "_$structName {\n";
	$currIdx = 0;
	$dummyIdx = 0;
	foreach $offset (@values) {
	    $fill = $offset - $currIdx;
	    if($fill > 0) {
		print "\tunsigned char dummy$dummyIdx\[$fill\];\n";
		$dummyIdx++;
	    }
	    if($fill < 0) {
		die "warning, variables out of order in parseStructs.pl\n";
	    }
	    #find all members that have this value
	    $numPrint = 0;
	    foreach $name (@keys) {
		if($memberMap{$name} == $offset) {
		    print "\tunsigned char $name;\n";
		    $numPrint++;
		}
	    }	    
	    
	    if($numPrint != 1) { die "$numPrint members at offset $offset found\n";}
	    $currIdx = $offset+1;
	}

	$fill = $structSize - $currIdx;
	if($fill > 0) {
	    print "\tunsigned char dummy$dummyIdx\[$fill\];\n";
	    $dummyIdx++;
	}

	print "} __attribute__ ((packed));\n\n";
    }

    #print "\n#define GUEST_UNSIGNED_LONG(st, var) (*((unsigned long *) &st->var))\n";
    #print "#define GUEST_VAR(type,st,var) (*((type *) &st->var))\n\n";

    print "#endif\n";
}


main();
