#include <stdio.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <assert.h>
#include <sys/ioctl.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>

#include <string>
#include <vector>
#include <iostream>

using namespace std;

static pid_t childPid;
static int gdbStdin[2], gdbStdout[2], gdbStderr[2];

int strMatchEnd(string *str,const char *c_str) {
  unsigned idx = str->rfind(string(c_str));

  return idx == (str->size() - strlen(c_str));
}

void waitForBreak(int fd, string *str) {
  char buf[4096];
  int ret;
  do {
    ret = read(fd,buf,sizeof(buf) - 1);
    assert(ret > 0);
    buf[ret] = '\0';

    (*str) += string(buf);
  } while(!strMatchEnd(str,"(gdb) "));
}

int main(int argc, char *argv[]) {
  int ret; 
  string str;
  string cmd, argStr;

  if(argc != 3) {
    cerr << "usage: queryGdb objectFile \"command string1; comand string 2\"" << endl;
    return 1;
  }

  ret = pipe(gdbStdin); assert(ret == 0);
  ret = pipe(gdbStdout); assert(ret == 0);
  ret = pipe(gdbStderr); assert(ret == 0);

  ret = fork();
  if(ret == 0) {
    // we are in the child                                                             
    dup2(gdbStdin[0],STDIN_FILENO);
    dup2(gdbStdout[1],STDOUT_FILENO);
    dup2(gdbStderr[1],STDERR_FILENO);
    ret = execlp("gdb","gdb",argv[1],NULL);
    // this should not return                                                          
    assert(0);
  } else {
    childPid = ret;
  }

  waitForBreak(gdbStdout[0],&str);
  
  argStr = string(argv[2]);
  string::size_type pos;
  while(argStr.size() > 0) {
    // look for ;
    pos = argStr.find(';');
    if(pos == string::npos) {
      // we are at the last command
      cmd = argStr;
      argStr.clear();
    } else {
      cmd = argStr.substr(0,pos);
      argStr = argStr.substr(pos+1);
    }

    cmd.append("\n",1);
    str.clear();

    write(gdbStdin[1],cmd.c_str(),cmd.size());

    waitForBreak(gdbStdout[0],&str);

    cout << str << endl;
  }

  kill(childPid, SIGKILL);
}
