#ifndef __CGI_CC_H_
#define __CGI_CC_H_

#include <string>
#include <map>

// this is copied from the cgicc library, CgiUtils.cpp file
char
hexToChar(char first,
	  char second)
{
  int digit;

  digit = (first >= 'A' ? ((first & 0xDF) - 'A') + 10 : (first - '0'));
  digit *= 16;
  digit += (second >= 'A' ? ((second & 0xDF) - 'A') + 10 : (second - '0'));
  return static_cast<char>(digit);
}

std::string
form_urldecode(const std::string& src)
{
  std::string result;
  std::string::const_iterator iter;
  char c;

  for(iter = src.begin(); iter != src.end(); ++iter) {
    switch(*iter) {
    case '+':
      result.append(1, ' ');
      break;
    case '%':
      // Don't assume well-formed input                                              
      if(std::distance(iter, src.end()) >= 2
	 && std::isxdigit(*(iter + 1)) && std::isxdigit(*(iter + 2))) {
	c = *++iter;
	result.append(1, hexToChar(c, *++iter));
      }
      // Just pass the % through untouched                                           
      else {
	result.append(1, '%');
      }
      break;

    default:
      result.append(1, *iter);
      break;
    }
  }

  return result;
}

void parsePair(const std::string& data, std::map<std::string, std::string> &result) {
  // find the '=' separating the name and value
  std::string::size_type pos = data.find_first_of("=", 0);

  // if no '=' was found, return
  if(pos == std::string::npos)
    return;

  // unescape the data, and add to the form entry list
  std::string name      = form_urldecode(data.substr(0, pos));
  std::string value     = form_urldecode(data.substr(++pos,
                                                     std::string::npos));

  result[name] = value;
}

std::map<std::string, std::string> parseQueryString(std::string &data) {
  std::map<std::string, std::string> result;
  
  std::string::size_type pos;
  std::string::size_type oldPos = 0;

  while(true) {
    // find the '&' separating a name=value pairs                                    
    pos = data.find_first_of("&", oldPos);

    // if no '&' was found, the rest of the string is a pair                         
    if(pos == std::string::npos) {
      parsePair(data.substr(oldPos), result);
      break;
    }

    // otherwise, parse the name=value pair                                          
    parsePair(data.substr(oldPos, pos - oldPos), result);

    // update position                                                               
    oldPos = pos + 1;
  }

  return result;
}

#endif
