#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <vector>
#include <string>

#include <iostream>

#include "plato.h"
#include "../../../revirt-skas-2.4.18/uml/uml.h"

using namespace std;

#define INT3_OPCODE 0xcc

breakpoint::breakpoint() {
  init = 0;
  relEntryAddr = 0;
  atFuncStart = 0;
  umlFd = -1;
  enabled = false;
  defaultMemoryMode = PLATO_MEM_MODE_KERN;
  internalBreak = false;
  remBreak = false;
  breakPointType = KERNEL_MODE_BP;
  numCallbacks = 0;
  savedBreakOp = 0;
  savedNextOp = 0;
}

breakpoint::breakpoint(vector< string > *vec, int fd) {
  string::size_type pos;
  int ret;
  struct breakVar bv;
  string str;

  internalBreak = false;
  umlFd = fd;
  init = 1;
  relEntryAddr = 0;
  atFuncStart = 0;
  relBreakAddr = 0;
  relNextAddr = 0;
  userModeComm = string("");
  userModeSharedLib = string("");
  defaultMemoryMode = PLATO_MEM_MODE_KERN;
  remBreak = false;
  breakPointType = KERNEL_MODE_BP;
  numCallbacks = 0;

  // XXX we could specify this statically, not much savings though
  savedBreakOp = 0;
  savedNextOp = 0;
  //

  for(unsigned long idx = 0; idx < vec->size(); idx++) {
    // parse the arguments
    if(vec->at(idx).find("breakId") == 0) {
      pos = vec->at(idx).find(' '); assert(pos != string::npos);
      breakId = vec->at(idx).substr(pos+1);
    } else if(vec->at(idx).find("breakAddr") == 0) {
      ret = sscanf(vec->at(idx).c_str(),"breakAddr %lx",&relBreakAddr); assert(ret == 1);
    } else if(vec->at(idx).find("nextAddr") == 0) {
      ret = sscanf(vec->at(idx).c_str(),"nextAddr %lx",&relNextAddr); assert(ret == 1);
    } else if(vec->at(idx).find("functionName") == 0) {
      pos = vec->at(idx).find(' '); assert(pos != string::npos);
      functionName = vec->at(idx).substr(pos+1);
    } else if(vec->at(idx).find("userModeComm") == 0) {
      pos = vec->at(idx).find(' '); assert(pos != string::npos);
      userModeComm = vec->at(idx).substr(pos+1);
      breakPointType = USER_MODE_BP;
    } else if(vec->at(idx).find("userModeSharedLib") == 0) {
      pos = vec->at(idx).find(' '); assert(pos != string::npos);
      userModeSharedLib = vec->at(idx).substr(pos+1);
      breakPointType = SHARED_LIB_BP;
    } else if(vec->at(idx).find("frameBase") == 0) {
      pos = vec->at(idx).find(' '); assert(pos != string::npos);
      str = vec->at(idx).substr(pos+1);
      getVarType(&str,&frameBase);
    } else if(vec->at(idx).find("variable") == 0) {
      pos = vec->at(idx).find(' '); assert(pos != string::npos);
      str = vec->at(idx).substr(pos+1);
      pos = str.find(' ');
      bv.name = str.substr(0,pos);
      str = str.substr(pos+1);
      getVarType(&str,&bv.loc);
      varVec.insert(varVec.end(),bv);
    } else if(vec->at(idx).find("entryAddr") == 0) {
      ret = sscanf(vec->at(idx).c_str(),"entryAddr %lx",&relEntryAddr); assert(ret == 1);
    } else if(vec->at(idx).find("atFunctionStart") == 0) {
      atFuncStart = 1;
    } else if(vec->at(idx).find("prototyped") == 0) {
      cerr << "*** function " << functionName << " prototyped, make sure of calling conventions" << endl;
    } else { 
      assert(0);
    }
  }
  
  if(breakPointType == KERNEL_MODE_BP) {
    defaultMemoryMode = PLATO_MEM_MODE_KERN;
  } else {
    defaultMemoryMode = PLATO_MEM_MODE_USER;
  }

  enabled = true;
  callback = NULL;
}


string
breakpoint::getName(void) {
  return breakId;
}

void
breakpoint::getVarType(string *str, struct breakVarLoc *bvl) {
  int ret;

  if(str->find(DW_OP_reg_str) == 0) {
    bvl->type = DW_OP_reg;
    ret = sscanf(str->c_str(),DW_OP_reg_str" %ld",&bvl->offset); assert(ret == 1);
  } else if(str->find(DW_OP_fbreg_str) == 0) {
    bvl->type = DW_OP_fbreg;
    ret = sscanf(str->c_str(),DW_OP_fbreg_str" %ld",&bvl->offset); assert(ret == 1);
  } else if(str->find(DW_OP_addr_str) == 0) {
    bvl->type = DW_OP_addr;
    ret = sscanf(str->c_str(),DW_OP_addr_str" %lu",&bvl->addr); assert(ret == 1);
  } else {
    assert(0);
  }


}

unsigned long 
breakpoint::getRegVal(struct pt_regs *regs, long regno) {
  switch(regno) {
  case FAST_REG_EAX:
    return (unsigned long) regs->eax;
  case FAST_REG_ECX:
    return (unsigned long) regs->ecx;
  case FAST_REG_EDX:
    return (unsigned long) regs->edx;
  case FAST_REG_EBX:
    return (unsigned long) regs->ebx;
  case FAST_REG_ESP:
    return (unsigned long) regs->esp;
  case FAST_REG_EBP:
    return (unsigned long) regs->ebp;
  case FAST_REG_ESI:
    return (unsigned long) regs->esi;
  case FAST_REG_EDI:
    return (unsigned long) regs->edi;
  default:
    assert(0);
  }

  assert(0);
  return 0;
}

void
breakpoint::setRegVal(struct pt_regs *regs, long regno, unsigned long val) {
  switch(regno) {
  case FAST_REG_EAX:
    regs->eax = (long) val;
    return;
  case FAST_REG_ECX:
    regs->ecx = (long) val;
    return;
  case FAST_REG_EDX:
    regs->edx = (long) val;
    return;
  case FAST_REG_EBX:
    regs->ebx = (long) val;
    return;
  case FAST_REG_ESP:
    regs->esp = (long) val;
    return;
  case FAST_REG_EBP:
    regs->ebp = (long) val;
    return;
  case FAST_REG_ESI:
    regs->esi = (long) val;
    return;
  case FAST_REG_EDI:
    regs->edi = (long) val;
    return;
  default:
    assert(0);
  }

  assert(0);
}

unsigned long
breakpoint:: readFuncArg(unsigned long argNo, struct pt_regs *regs) {
  assert(atFuncStart);
  
  return *((unsigned long *) (regs->esp + 4 + (argNo*4)));
}

unsigned long
breakpoint:: readFuncArgUsingFb(unsigned long argNo, struct pt_regs *regs) {
  assert(!atFuncStart);
  // for the offset, the frame pointer and return address are on the 
  // stack before the arguments start
  long offset = 8 + argNo*4;
  
  return *((unsigned long *) (((long) getRegVal(regs,frameBase.offset)) + 
			      offset));
}

void
breakpoint::writeFuncArg(unsigned long argNo, unsigned long val, 
			 struct pt_regs *regs) {
  assert(atFuncStart);

  *((unsigned long *) (regs->esp + 4 + (argNo*4))) = val;
}
	   

unsigned long
breakpoint::readStackStructPtr(char *varStr, struct pt_regs *regs) {
  // figure out which var it is in this context
  vector< struct breakVar >::iterator iter;
  string str = string(varStr);

  for(iter = varVec.begin(); iter != varVec.end(); iter++) {
    if(str == iter->name) {
      assert(iter->loc.type == DW_OP_fbreg); 
      return (unsigned long) ((unsigned long *) 
			      (((long) getRegVal(regs,frameBase.offset)) + 
			       iter->loc.offset));
    }
  }
  
  cerr << "var " << varStr << " not found" << endl;
  assert(0);

  return 0;
}

bool
breakpoint::readVar(char *varStr, struct pt_regs *regs, unsigned long *val) {
  // figure out which var it is in this context
  vector< struct breakVar >::iterator iter;
  string str = string(varStr);
  unsigned long res;
  long addr;

  for(iter = varVec.begin(); iter != varVec.end(); iter++) {
    if(str == iter->name) {
      // figure out where the variable is located
      if(iter->loc.type == DW_OP_reg) {
	*val = getRegVal(regs, iter->loc.offset);
	return true;
      } else if(iter->loc.type == DW_OP_fbreg) {
	addr = ((long) getRegVal(regs,frameBase.offset)) + iter->loc.offset;
	copy_from_guest(&res, addr, sizeof(res));
	*val = res;
	return true;
      } else if(iter->loc.type == DW_OP_addr) {
	copy_from_guest(&res, iter->loc.addr, sizeof(res));
	*val = res;
	return true;
      } else {
	assert(0);
      }
    }
  }

  return false;
}

bool 
breakpoint::writeVar(char *varStr, unsigned long val, struct pt_regs *regs) {
    // figure out which var it is in this context
  vector< struct breakVar >::iterator iter;
  string str = string(varStr);
  unsigned long addr;

  for(iter = varVec.begin(); iter != varVec.end(); iter++) {
    if(str == iter->name) {
      // figure out where the variable is located
      if(iter->loc.type == DW_OP_reg) {
	setRegVal(regs, iter->loc.offset, val);
      } else if(iter->loc.type == DW_OP_fbreg) {
	addr = ((long) getRegVal(regs,frameBase.offset)) + iter->loc.offset;
	copy_to_guest(addr, &val, sizeof(val));
      } else if(iter->loc.type == DW_OP_addr) {
	copy_to_guest(iter->loc.addr, &val, sizeof(val));
      } else {
	assert(0);
      }
      return true;
    }
  }

  return false;
}

string breakpoint::getUserModeSharedLib() {
  assert(breakPointType == SHARED_LIB_BP);
  return userModeSharedLib;
}
string breakpoint::getUserModeComm() {
  assert(breakPointType == USER_MODE_BP);
  return userModeComm;
}
bool breakpoint::isUserMode() {
  return defaultMemoryMode == PLATO_MEM_MODE_USER;
}
BreakPointType_t breakpoint::getBreakType() {
  return breakPointType;
}

unsigned long
breakpoint::getEntryAddr() {
  assert(init);
  return absEntryAddr();
}

string 
breakpoint::getFunctionName() {
  assert(init);
  return functionName;
}

int
breakpoint::canCallFunction() {
  return relEntryAddr != 0;
}

bool breakpoint::isInternalBreak() {
  return internalBreak;
}

void breakpoint::setInternalBreak(bool b) {
  internalBreak = b;
}

void breakpoint::enable() {
  enabled = true;
  if(remBreak) {    
    remBreak = false;
    setBreak();
  }
}
  
void breakpoint::disable(bool removeBreak) {
  remBreak = removeBreak;
  enabled = false;
}

bool breakpoint::isEnabled() {
  return enabled;
}

void breakpoint::clearOffset(pid_t pid) {
  if(offsetMap.find(pid) != offsetMap.end()) {
    offsetMap.erase(offsetMap.find(pid));
  }
}

void breakpoint::setOffset(pid_t pid, unsigned long offset) {
  assert(offsetMap.find(pid) == offsetMap.end());
  offsetMap[pid] = offset;
}

unsigned long breakpoint::getRelBreakAddr() {
  return relBreakAddr;
}

void
breakpoint::setBreak() {
  assert(init);
  if(!remBreak) {
    setBreakPoint(absBreakAddr(), &savedBreakOp);
  }
}

void
breakpoint::setNextBreak() {
  assert(init);
  setBreakPoint(absNextAddr(), &savedNextOp);
}

pid_t breakpoint::getCurrentPid() {
  return ::getCurrentPid(umlFd);
}

unsigned long breakpoint::absAddr(unsigned long relAddr) {
  unsigned long addr;
  if(breakPointType == SHARED_LIB_BP) {
    pid_t pid = getCurrentPid();
    if(offsetMap.find(pid) == offsetMap.end()) {
      return 0;
    }
    addr = offsetMap[pid] + relAddr;
  } else {
    addr = relAddr;
  }

  return addr;
}

unsigned long breakpoint::absBreakAddr() {
  return absAddr(relBreakAddr);
}

unsigned long breakpoint::absNextAddr() {
  return absAddr(relNextAddr);
}

unsigned long breakpoint::absEntryAddr() {
  return absAddr(relEntryAddr);
}

int
breakpoint::atBreak(struct pt_regs *regs) {
  assert(init);
  unsigned long addr = absBreakAddr();
  if(addr == 0) return 0;

  return (regs->eip-1) == (long) addr;
}
int
breakpoint::atNextBreak(struct pt_regs *regs) {
  assert(init);
  unsigned long addr = absNextAddr();
  if(addr == 0) return 0;

  return (regs->eip-1) == (long) addr;
}

void
breakpoint::clearBreak() {
  assert(init);
  clearBreakPoint(absBreakAddr(), savedBreakOp);
}

void
breakpoint::clearNextBreak() {
  assert(init);
  clearBreakPoint(absNextAddr(), savedNextOp);
}

MemMode_t breakpoint::getDefMemMode() {
  return defaultMemoryMode;
}
void breakpoint::copy_from_guest(void *buf, unsigned long guestAddr,
				 unsigned long len, MemMode_t memMode) {
  struct plato_mem_struct pMem;

  pMem.bufferAddr = (unsigned long ) buf;
  pMem.guestAddr = guestAddr;
  pMem.len = len;
  pMem.isKern = (memMode == PLATO_MEM_MODE_KERN) ? 1 : 0;

  ioctl(umlFd, PLATO_IOC_READ_MEM, &pMem);

  if(pMem.ret != 0) {
    throw "copy_from_guest memory exception";
  }
}

void breakpoint::copy_to_guest(unsigned long guestAddr, void *buf,
			       unsigned long len, MemMode_t memMode) {
  struct plato_mem_struct pMem;

  pMem.bufferAddr = (unsigned long ) buf;
  pMem.guestAddr = guestAddr;
  pMem.len = len;
  pMem.isKern = (memMode == PLATO_MEM_MODE_KERN) ? 1 : 0;

  ioctl(umlFd, PLATO_IOC_WRITE_MEM, &pMem);

  if(pMem.ret != 0) {
    throw "copy_to_guest memory exception";
  }
}

void breakpoint::copy_from_guest(void *buf, unsigned long guestAddr,
				 unsigned long len) {
  copy_from_guest(buf, guestAddr, len, defaultMemoryMode);
}

void breakpoint::copy_to_guest(unsigned long guestAddr, void *buf,
			       unsigned long len) {
  copy_to_guest(guestAddr, buf, len, defaultMemoryMode);
}


void
breakpoint::setBreakPoint(unsigned long addr, unsigned char *data) {
  unsigned char c = INT3_OPCODE;
  if(*data == 0) {
    copy_from_guest(data, addr, sizeof(*data));
    assert(*data != 0);
    assert(*data != INT3_OPCODE); // implies already a breakpoint there
  }

  copy_to_guest(addr, &c, sizeof(c));
}

void
breakpoint::clearBreakPoint(unsigned long addr, unsigned char data) {
  assert(data != 0);
  copy_to_guest(addr, &data, sizeof(data));
}  

void breakpoint::printNumBreaks() {
  if(numCallbacks > 0) {
    cerr << "break point " << breakId << " callback called " << numCallbacks 
	 << " times " << endl;
  }
}

void
breakpoint::callCallback() {
  numCallbacks++;
  assert(init);
  if(callback != NULL) callback();
}

void
breakpoint::setCallback(breakCallback cb) {
  assert(init);
  callback = cb;
}

int
breakpoint::isBreakId(char *breakStr) {
  string str = string(breakStr);

  return str == breakId;
}

string breakpoint::getBreakId() {
  return breakId;
}

typedef struct LocalBreak {
  map<pid_t, unsigned long> offsetData;
} LocalBreak_t;

void *breakpoint::checkpoint() {
  LocalBreak_t *bp = new LocalBreak_t;
  bp->offsetData = offsetMap;

  return bp;
}
void breakpoint::rollback(void *data) {
  LocalBreak_t *bp = (LocalBreak_t *) data;

  offsetMap.clear();
  offsetMap = bp->offsetData;
}
void breakpoint::deleteCheckpoint(void *data) {
  LocalBreak_t *bp = (LocalBreak_t *) data;
  delete bp;
}
