#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>

#include <unistd.h>
#include <iostream>
#include <sstream>
#include <iomanip>

#include "plato.h"
#include "guestStructs.h"

Plato plato;

using namespace std;
/* Action modifiers - doesn't change the zoning */
#define __GFP_WAIT      0x10    /* Can wait and reschedule? */
#define __GFP_HIGH      0x20    /* Should access emergency pools? */
#define __GFP_IO        0x40    /* Can start low memory physical IO? */
#define __GFP_HIGHIO    0x80    /* Can start high mem physical IO? */
#define __GFP_FS        0x100   /* Can call down to low-level FS? */

#define GFP_KERNEL      (__GFP_HIGH | __GFP_WAIT | __GFP_IO | __GFP_HIGHIO | __GFP_FS)

bool haveCp = false;
unsigned long checkId = (unsigned long) -1;
bool dontDelete = false;

string getTimeStr(struct timeval *tv) {
  stringstream sStr;

  sStr << tv->tv_sec << "." << setw(6) << setfill('0') << tv->tv_usec;
  return sStr.str();
}

void pageFaultCb() {
  struct timeval tvStart, tvStop, res;
  if(haveCp) {
    gettimeofday(&tvStart, NULL);
    plato.rollback(checkId);
    gettimeofday(&tvStop, NULL);
    timersub(&tvStop, &tvStart, &res);
    cout << "restoring checkpoint " << checkId << " took " 
	 << getTimeStr(&res) << endl;
    
    haveCp = false;  
  }
}

void stkCb() {
  void *guestMem;
  unsigned long mapCount;
  struct timeval tvStart, tvStop, res;

  if(checkId != ((unsigned long) -1)) {
    if(!dontDelete) {
      gettimeofday(&tvStart, NULL);
      plato.deleteCheckpoint(checkId);
      gettimeofday(&tvStop, NULL);
      timersub(&tvStop, &tvStart, &res);
      cout << "deleting checkpoint " << checkId << " took " 
	   << getTimeStr(&res) << endl;
    }
  }
  gettimeofday(&tvStart, NULL);
  checkId = plato.checkpoint();
  gettimeofday(&tvStop, NULL);
  timersub(&tvStop, &tvStart, &res);
  cout << "got checkpoint " << checkId << " took " << getTimeStr(&res) << endl;
  haveCp = true;
  guestMem = (void *) plato.callFunction("kmalloc", 25, GFP_KERNEL);
  
  mapCount = plato.readVar("max_map_count");
}

void initCb() {
  cout << "initCb" << endl;
  plato.registerBreak("sys_stk_syscall:1:end",stkCb);
  plato.registerBreak("handle_page_fault:1:end",pageFaultCb);
}  

int main(int argc, char *argv[]) {
  int opt;

  while((opt = getopt(argc, argv, "d")) != -1) {
    if(opt == 'd') {
      dontDelete = true;
    } else {
      assert(0);
    }
  }

  plato.init("plato.conf");

  plato.startEventLoop(initCb);

  return 0;
}
