#ifndef __LOCAL_PLATO_H__
#define __LOCAL_PLATO_H__

// local classes, predicates should not need to use these, they are used by
// plato
#define  DW_OP_reg_str   "DW_OP_reg"
#define  DW_OP_fbreg_str "DW_OP_fbreg"
#define  DW_OP_addr_str  "DW_OP_addr"

typedef enum {
  DW_OP_reg,
  DW_OP_fbreg,
  DW_OP_addr
} DW_OP_type_t;

#define FAST_REG_EAX 0
#define FAST_REG_ECX 1
#define FAST_REG_EDX 2
#define FAST_REG_EBX 3
#define FAST_REG_ESP 4
#define FAST_REG_EBP 5
#define FAST_REG_ESI 6
#define FAST_REG_EDI 7

struct breakVarLoc {
  DW_OP_type_t type;
  long offset;
  unsigned long addr;
};

struct breakVar {
  struct breakVarLoc loc;
  string name;
};

typedef enum {
  KERNEL_MODE_BP,
  USER_MODE_BP,
  SHARED_LIB_BP
} BreakPointType_t;

class breakpoint {
 public:
  breakpoint(vector< string > *vec, int fd);
  breakpoint();
  bool isInternalBreak();
  void setInternalBreak(bool b);
  void enable();
  void disable(bool removeBreak = false);
  bool isEnabled();
  void setBreak();
  void setNextBreak();
  int atBreak(struct pt_regs *regs);
  int atNextBreak(struct pt_regs *regs);
  void clearBreak();
  void clearNextBreak();
  void callCallback();
  void setCallback(breakCallback cb);
  std::string getBreakId();
  int isBreakId(char *breakStr);
  bool readVar(char *varStr, struct pt_regs *regs, unsigned long *val);
  bool writeVar(char *varStr, unsigned long val, struct pt_regs *regs);
  int canCallFunction();
  void clearOffset(pid_t pid);
  void setOffset(pid_t pid, unsigned long offset);
  unsigned long getRelBreakAddr();
  unsigned long getEntryAddr();
  string getFunctionName();
  unsigned long readStackStructPtr(char *varStr, struct pt_regs *regs);
  MemMode_t getDefMemMode();
  bool isUserMode();
  BreakPointType_t getBreakType();
  string getUserModeComm();
  string getUserModeSharedLib();

  // these functions can only be called when you are at a function start
  unsigned long readFuncArg(unsigned long argNo, struct pt_regs *regs);
  void writeFuncArg(unsigned long argNo, unsigned long val, struct pt_regs *regs);

  // this can be used when you are somewhere inside of the function
  unsigned long readFuncArgUsingFb(unsigned long argNo, struct pt_regs *regs);

  void setRegVal(struct pt_regs *regs, long regno, unsigned long val);
  unsigned long getRegVal(struct pt_regs *regs, long regno);
  
  string getName();
  void *checkpoint();
  void rollback(void *data);
  void deleteCheckpoint(void *data);

  void copy_to_guest(unsigned long guestAddr, void *buf,
		     unsigned long len, MemMode_t memMode);
  void copy_from_guest(void *buf, unsigned long guestAddr, 
		       unsigned long len, MemMode_t memMode);

  // use the default memory mode to decide which one to use
  void copy_to_guest(unsigned long guestAddr, void *buf,
		     unsigned long len);
  void copy_from_guest(void *buf, unsigned long guestAddr, 
		       unsigned long len);
  unsigned long absBreakAddr();
  void printNumBreaks();
 private:
  pid_t getCurrentPid();
  unsigned long absAddr(unsigned long relAddr);
  unsigned long absNextAddr();
  unsigned long absEntryAddr();
  void getVarType(string *str, struct breakVarLoc *bvl);
  void setBreakPoint(unsigned long addr, unsigned char *data);
  void clearBreakPoint(unsigned long addr, unsigned char data);

  unsigned long numCallbacks;
  bool remBreak;
  bool internalBreak; // a breakpoint used by plato, not a predicate
  int atFuncStart;
  string breakId;
  unsigned long relBreakAddr;
  unsigned long relNextAddr;
  unsigned long relEntryAddr;
  map<pid_t, unsigned long> offsetMap;
  string functionName;
  struct breakVarLoc frameBase;
  vector< struct breakVar > varVec;
  unsigned char savedBreakOp;
  unsigned char savedNextOp;
  //unsigned char savedBreakData;
  //unsigned long savedBreakAddr;
  int init;
  breakCallback callback;
  int umlFd;
  bool enabled; // should this be part of the breakpoint checkpoint struct?
  MemMode_t defaultMemoryMode;
  string userModeComm;
  string userModeSharedLib;
  BreakPointType_t breakPointType;
};

class PlatoCheckpoint {
 public:
  PlatoCheckpoint(vector< breakpoint >::iterator bp, struct pt_regs *regs,
		  bool isUserMode);
  void setBreakData(unsigned long breakIdx, void *breakData);
  void *getBreakData(unsigned long breakIdx);
  struct pt_regs getSavedRegs();
  vector<breakpoint>::iterator getSavedBreak();
  void setUmBreakMap(StringMap_t *m);
  StringMap_t *getUmBreakMap();
  void setUmSetBreakMap(PidMap_t *m);
  PidMap_t *getUmSetBreakMap();
  void setUmQueuedBreakMap(PidMap_t *m);
  PidMap_t *getUmQueuedBreakMap();
  bool isUserMode();

 private:
  bool userMode;
  vector< breakpoint >::iterator savedBreak;
  struct pt_regs savedRegs;
  vector< void * > breakCheckVec;
  PidMap_t *savedUmSetBreakMap;
  PidMap_t *savedUmQueuedBreakMap;
};

#endif
