#include <stdio.h>
#include <stdlib.h>

#include <unistd.h>
#include <iostream>

#include "plato.h"
#include "guestStructs.h"

Plato plato;

using namespace std;
/* Action modifiers - doesn't change the zoning */
#define __GFP_WAIT      0x10    /* Can wait and reschedule? */
#define __GFP_HIGH      0x20    /* Should access emergency pools? */
#define __GFP_IO        0x40    /* Can start low memory physical IO? */
#define __GFP_HIGHIO    0x80    /* Can start high mem physical IO? */
#define __GFP_FS        0x100   /* Can call down to low-level FS? */

#define GFP_KERNEL      (__GFP_HIGH | __GFP_WAIT | __GFP_IO | __GFP_HIGHIO | __GFP_FS)

void openCb() {
  struct linux_stat *st;
  struct linux_task_struct *curr;
  unsigned long addrLimit;
  int ret;
  string str;

  // allocate memory for the stat struct
  st = (struct linux_stat *) plato.callFunction("kmalloc",
						sizeof(struct linux_stat),
						GFP_KERNEL);
  plato.guest_memset(st,0,sizeof(struct linux_stat));
  str = plato.guest_strcpy(plato.readVar("tmp"));

  // set the addr_limit so we can call sys_newstat
  curr = (struct linux_task_struct *) plato.getCurrent();  
  addrLimit = plato.guest_read_ulong(&curr->addr_limit);
  plato.guest_write_ulong(&curr->addr_limit, 0xffffffff);

  // now call sys_newstat
  ret = plato.callFunction("sys_newstat",plato.readVar("tmp"),(unsigned long) st);

  cout << "open file " << str << " with flags " << (void *) plato.readVar("flags")
       << " and mode " << (void *) plato.readVar("mode") << ", ret = " 
       << ret << ", inode = " << plato.guest_read_ulong(&st->st_ino)
       << endl;

  // set the address limit back
  plato.guest_write_ulong(&curr->addr_limit, addrLimit);

  // free the kmalloc memory
  plato.callFunction("kfree",(unsigned long) st);
}

void initCb() {
  cout << "initCb" << endl;
  plato.registerBreak("open.c:796:begin",openCb);
}  

int main(int argc, char *argv[]) {
  plato.init("plato.conf");

  plato.startEventLoop(initCb);

  return 0;
}
