#include <stdio.h>
#include <stdlib.h>

#include <string>
#include <vector>

#include <iostream>
#include <fstream>
#include <assert.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <fcntl.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>

#include <sys/user.h>
#include <linux/unistd.h>

#include "plato.h"
#include "../../../revirt-skas-2.4.18/uml/uml.h"
#include "usermode.h"


#define UMLGDB_DEVICE "/dev/umlgdb"
#define GLOBAL_VARS "GLOBAL_VARS"
#define COMM_SIZE 16

#define ZTRACE_RESUME 19
#define __NR_ztrace             276

_syscall4(int,ztrace,int,a,int,b,int,c,int,d);

using namespace std;

StringMap_t *copyStringMap(StringMap_t *m) {
  StringMap_t::iterator iter;
  StringMap_t *ret = new StringMap_t();
  unsigned long idx;
  vector<unsigned long> *vec;

  for(iter = m->begin(); iter != m->end(); iter++) {
    vec = new vector<unsigned long>();
    for(idx = 0; idx < iter->second->size(); idx++) {
      vec->insert(vec->end(), iter->second->at(idx));
    }
    (*ret)[iter->first] = vec;
  }

  return ret;
}

PidMap_t *copyPidMap(PidMap_t *m) {
  PidMap_t::iterator iter;
  PidMap_t *ret = new PidMap_t();
  unsigned long idx;
  vector<unsigned long> *vec;

  for(iter = m->begin(); iter != m->end(); iter++) {
    vec = new vector<unsigned long>();
    for(idx = 0; idx < iter->second->size(); idx++) {
      vec->insert(vec->end(), iter->second->at(idx));
    }
    (*ret)[iter->first] = vec;
  }

  return ret;
}

Plato::Plato() {
  umlFd = -1;
  umBreakMap = new StringMap_t();
  umSetBreakMap = new PidMap_t();
  umQueuedBreakMap = new PidMap_t();
  slNameBreakMap = new StringMap_t();
  slInodeBreakMap = new PidMap_t();
  hasSharedLibBp = false;
  killProc = false;
  hasUserModeCb = false;

  assert(sizeof(pid_t) == sizeof(unsigned long));
}
Plato::~Plato() {
  for(unsigned long idx = 0; idx < breakPointVec.size(); idx++) {
    breakPointVec[idx].printNumBreaks();
  }

  if(umlFd >= 0) {
    close(umlFd);
  }
  
  delBreakMap(umBreakMap);
  delSetBreakMap(umSetBreakMap);
  delSetBreakMap(umQueuedBreakMap);
  delBreakMap(slNameBreakMap);
  delSetBreakMap(slInodeBreakMap);
}

void Plato::killCurrProc() {
  assert(isUserModeTrap());
  killProc = true;
}

void Plato::delBreakMap(StringMap_t *m) {
  StringMap_t::iterator iter;

  for(iter = m->begin(); iter != m->end(); iter++) {
    delete iter->second;
  }

  delete m;
}

void Plato::delSetBreakMap(PidMap_t *m) {
  PidMap_t::iterator iter;

  for(iter = m->begin(); iter != m->end(); iter++) {
    delete iter->second;
  }

  delete m;
}

void Plato::init(char *configFileName) {
  umlFd = open(UMLGDB_DEVICE,O_RDWR);
  assert(umlFd > 0);

  ifstream configFile(configFileName);
  assert(configFile.is_open());
  
  char line[256];
  string str;
  vector< string > vec;
  breakpoint bp;

  // read in the config file to find out about our break points
  while(!configFile.eof()) {
    configFile.getline(line,sizeof(line));
    str = string(line);

    if(str == "***BEGIN") {
      vec.clear();

      while(str != "***END") {
	configFile.getline(line,sizeof(line));
	assert(!configFile.eof());

	str = string(line);

	if(str != "***END") {
	  vec.insert(vec.end(),str);
	}
      }
      
      bp = breakpoint(&vec, umlFd);
      breakPointVec.insert(breakPointVec.end(),bp);
    }
  } 
  
  // look for any global variables
  for(globalVars = breakPointVec.begin(); globalVars != breakPointVec.end();
      globalVars++) {

    if(globalVars->isBreakId(GLOBAL_VARS)) {
      break;
    }
  }
    
}

void Plato::enableBreak(unsigned long id) {
  breakPointVec.at(id).enable();
}
void Plato::disableBreak(unsigned long id) {
  breakPointVec.at(id).disable();
}

unsigned long
Plato::registerBreak(char *breakStr, breakCallback breakCb, bool internalBreak) {
  // first see if it was in the cofig file
  unsigned long idx;

  for(idx = 0; idx < breakPointVec.size(); idx++) {
    if(breakPointVec.at(idx).isBreakId(breakStr)) {
      breakPointVec.at(idx).setCallback(breakCb);
      breakPointVec.at(idx).setInternalBreak(internalBreak);
      if(breakPointVec.at(idx).isUserMode()) {
	registerUmBreak(idx);
      } else {
	breakPointVec.at(idx).setBreak();
      }

      // set our internal breakpoint id
      if(strcmp(breakStr, FORK_BREAK_ID) == 0) {
	assert(internalBreak);
	forkBreak = breakPointVec.begin() + idx;
      } else if(strcmp(breakStr, EXEC_BREAK_ID) == 0) {
	assert(internalBreak);
	execBreak = breakPointVec.begin() + idx;
      } else if(strcmp(breakStr, MAP_BREAK_ID) == 0) {
	assert(internalBreak);
	mapBreak = breakPointVec.begin() + idx;
      } else if(strcmp(breakStr, EXIT_BREAK_ID) == 0) {
	assert(internalBreak);
	exitBreak = breakPointVec.begin() + idx;
      } else if(strcmp(breakStr, SCHED_BREAK_ID) == 0) {
	assert(internalBreak);
	schedBreak = breakPointVec.begin() + idx;
      }

      return idx;
    }
  }

  cerr << "break point " << breakStr << " not found" << endl;
  assert(0);
}

unsigned long numBreak = 0;
unsigned long kernTrap = 0;
unsigned long userTrap = 0;
struct timeval handleTrapTime;
unsigned long numCp = 0;
unsigned long numRb = 0;
unsigned long numDelCp = 0;
struct timeval cpTime;
struct timeval rbTime;
struct timeval delCpTime;

static unsigned long long timeAveUsecs(struct timeval *tv, unsigned long num) {
  unsigned long long usecs = tv->tv_sec;
  usecs *= 1000000;
  usecs += tv->tv_usec;
  return usecs / num;
}

static void sigHandler(int signo) {
  cout << "cntl-c, not clearing breakpoints...." << endl;
  cout << "we got " << numBreak << " breaks and signo = " << signo << endl;
  cout << "we spent " << handleTrapTime.tv_sec << " seconds and " 
       << handleTrapTime.tv_usec << " usecs handling traps" << endl;
  cout << "we got " << kernTrap << " kernel mode traps and " << userTrap
       << " user mode traps" << endl << endl;
  if(numCp > 0) {
    cout << "checkpoint average time (usec) " << timeAveUsecs(&cpTime, numCp) << endl;
  }
  if(numRb > 0) {
    cout << "rollback   average time (usec) " << timeAveUsecs(&rbTime, numRb) << endl;
  }
  if(numDelCp > 0) {
    cout << "delete cp  average time (usec) " << timeAveUsecs(&delCpTime, numDelCp) 
	 << endl;
  }
  exit(0);
}

unsigned long
Plato::readStackStructPtr(char *varStr) {
  return currBreak->readStackStructPtr(varStr, &currRegs);
}

string Plato::guest_strcpy(unsigned long guestAddr, MemMode_t memMode) {
  unsigned long idx;
  string str;
  char c;

  idx = 0;
  do {
    copy_from_guest(&c, guestAddr+idx, sizeof(c), memMode);
    if(c != '\0') str.insert(str.end(), c);
    idx++;
    if(idx > 10000000) {
      cerr << "infinite loop, break" << endl;
      str.clear();
      return str;
    }
  } while(c != '\0');

  return str;
}

unsigned long Plato::guest_read_ulong(void *guestAddr, MemMode_t memMode) {
  unsigned long ret;
  copy_from_guest(&ret, (unsigned long) guestAddr, sizeof(ret), memMode);
  return ret;
}

void Plato::guest_write_ulong(void *guestAddr, unsigned long value, 
			      MemMode_t memMode) {
  copy_to_guest((unsigned long) guestAddr, &value, sizeof(value), memMode);
}

void Plato::guest_memset(void *guestAddr, unsigned char val,
			 unsigned long len, MemMode_t memMode) {
  unsigned long idx;
  for(idx = 0; idx < len; idx++) {
    copy_to_guest(((unsigned long) guestAddr)+idx, &val, sizeof(val), memMode);
  }
}

void Plato::copy_to_guest(unsigned long guestAddr, void *buf,
			  unsigned long len, MemMode_t memMode) {
  currBreak->copy_to_guest(guestAddr, buf, len, memMode);
}

void Plato::copy_from_guest(void *buf, unsigned long guestAddr, 
			    unsigned long len, MemMode_t memMode) {
  currBreak->copy_from_guest(buf, guestAddr, len, memMode);
}


string Plato::guest_strcpy(unsigned long guestAddr) {
  return guest_strcpy(guestAddr, currBreak->getDefMemMode());
}

unsigned long Plato::guest_read_ulong(void *guestAddr) {
  return guest_read_ulong(guestAddr, currBreak->getDefMemMode());
}

void Plato::guest_write_ulong(void *guestAddr, unsigned long value) {
  guest_write_ulong(guestAddr, value, currBreak->getDefMemMode());
}

void Plato::guest_memset(void *guestAddr, unsigned char val,
			 unsigned long len) {
  guest_memset(guestAddr, val, len, currBreak->getDefMemMode());
}

void Plato::copy_to_guest(unsigned long guestAddr, void *buf,
			  unsigned long len) {
  copy_to_guest(guestAddr, buf, len, currBreak->getDefMemMode());
}

void Plato::copy_from_guest(void *buf, unsigned long guestAddr, 
			    unsigned long len) {
  copy_from_guest(buf, guestAddr, len, currBreak->getDefMemMode());
}

unsigned long
Plato::readVar(char *varStr) {
  unsigned long ret;
  if(currBreak->readVar(varStr, &currRegs, &ret)) {
    return ret;
  }
  
  // variable not found at local context, look at global variables
  if(globalVars == breakPointVec.end()) {
    // could throw exception
    assert(0);
    return (unsigned long) -1;
  }

  bool b = globalVars->readVar(varStr, &currRegs, &ret);
  if(!b) {
    assert(0);
    return (unsigned long) -1;
  }

  return ret;
}

void 
Plato::writeVar(char *varStr, unsigned long val) {
  if(currBreak->writeVar(varStr, val, &currRegs)) {
    return;
  }

  // variable not found at local context, look at global variables
  if(globalVars == breakPointVec.end()) {
    assert(0);
    return;
  }

  bool b = globalVars->writeVar(varStr, val, &currRegs);
  if(!b) {
    assert(0);
    return;
  }
}

unsigned long
Plato::readFuncArg(unsigned long argNo) {
  return currBreak->readFuncArg(argNo, &currRegs);
}

unsigned long
Plato::readFuncArgUsingFb(unsigned long argNo) {
  return currBreak->readFuncArgUsingFb(argNo, &currRegs);
}

void
Plato::writeFuncArg(unsigned long argNo, unsigned long val) {
  currBreak->writeFuncArg(argNo, val, &currRegs);
}

void Plato::set_guest_regs(vector< breakpoint >::iterator iter, 
			   struct pt_regs *regs) {
  if(iter->getDefMemMode() == PLATO_MEM_MODE_KERN) {
    ioctl(umlFd,PLATO_IOC_SET_KERN_REGS,regs);
  } else {
    ioctl(umlFd,PLATO_IOC_SET_USER_REGS,regs);
  }
}

bool Plato::isUserModeTrap(void) {
  unsigned long userModeTrap;
  ioctl(umlFd, PLATO_IOC_QUERY_USER_MODE_TRAP, &userModeTrap);
  return userModeTrap != 0;
}

bool Plato::handleTrap(vector<breakpoint>::iterator iter, 
		       struct pt_regs *regs, bool internalOnly) {
  if(iter->atBreak(regs)) {
    currRegs = *regs;
    currBreak = iter;
    if(iter->isEnabled() && (!internalOnly || iter->isInternalBreak())) {
      iter->callCallback();
    }

    if(killProc) {      
      assert(isUserModeTrap());
      killProc = false;
      currRegs.eip = 1;
    } else {
      // use currBreak incase there was a rollback
      currBreak->clearBreak();
      currBreak->setNextBreak();
    }

    *regs = currRegs; // set the regs so any changes are reflected
    regs->eip--;
    set_guest_regs(currBreak, regs);
    return true;
  } else if(iter->atNextBreak(regs)) {
    iter->clearNextBreak();
    iter->setBreak();
    regs->eip--;
    set_guest_regs(iter, regs);
    return true;
  }

  return false;
}

vector<breakpoint>::iterator Plato::handleKernelModeTrap(struct pt_regs *regs, 
							 bool internalOnly) {
  vector< breakpoint >::iterator iter;
  
  // see if it is one of our breakpoints
  for(iter = breakPointVec.begin(); iter != breakPointVec.end(); iter++) {
    if(!iter->isUserMode()) {
      if(handleTrap(iter, regs, internalOnly)) {
	return iter;
      }
    } 
  }
  
  if(iter == breakPointVec.end()) {
    cerr << "unregisterd kernel break point detected at " 
	 << (void *) regs->eip << endl;
  }

  return breakPointVec.end();
}

void
Plato::startEventLoop(breakCallback initCb) {
  struct pt_regs regs;
  int ret;
  int firstEventArrived = 0;

  signal(SIGINT, sigHandler);

  assert(umlFd >= 0);
  ioctl(umlFd, PLATO_IOC_SET_MODE, PLATO_MODE_ON);

  while(1) {
    ret = read(umlFd, &regs, sizeof(regs));
    numBreak++;
    if(ret != sizeof(regs)) {
      cerr << "I am not sure why read returned " << ret << endl;
      exit(0);
    }

    if(!firstEventArrived) {
      firstEventArrived = 1;
      initCb();
      userModeInit(this);
      handleTrapTime.tv_sec = 0;
      handleTrapTime.tv_usec = 0;
      cpTime.tv_sec = 0; cpTime.tv_usec = 0;
      rbTime.tv_sec = 0; rbTime.tv_usec = 0;
      delCpTime.tv_sec = 0; delCpTime.tv_usec = 0;
    } else {
      //struct timeval startTime, stopTime, res;
      //gettimeofday(&startTime, NULL);

      if(isUserModeTrap()) {
	userTrap++;
	handleUserModeTrap(&regs);
      } else {
	kernTrap++;
	handleKernelModeTrap(&regs);
      }

      //gettimeofday(&stopTime, NULL);
      //timersub(&stopTime, &startTime, &res);
      //timeradd(&handleTrapTime, &res, &handleTrapTime);

    }
    
    ioctl(umlFd, PLATO_IOC_RESTART,0);
  }  
}

void Plato::deleteCheckpoint(unsigned long checkId) {
  PlatoCheckpoint *cp;
  struct timeval startTime, stopTime, res;
  gettimeofday(&startTime, NULL);
  numDelCp++;

  if(checkMap.find(checkId) == checkMap.end()) {
    cerr << "invalid checkpoint id" << endl;
  }
  cp = checkMap.find(checkId)->second;
  checkMap.erase(checkId);
  
  // delete breakpoint state
  unsigned long idx;
  for(idx = 0; idx < breakPointVec.size(); idx++) {
    breakPointVec[idx].deleteCheckpoint(cp->getBreakData(idx));
  }

  delSetBreakMap(cp->getUmSetBreakMap());
  delSetBreakMap(cp->getUmQueuedBreakMap());

  delete cp;

  ioctl(umlFd, PLATO_IOC_DELETE_CHECK, checkId);

  gettimeofday(&stopTime, NULL);
  timersub(&stopTime, &startTime, &res);
  timeradd(&delCpTime, &res, &delCpTime);
  
}

unsigned long
Plato::checkpoint(bool pauseRevirt) {
  unsigned long checkId;
  struct timeval startTime, stopTime, res;
  gettimeofday(&startTime, NULL);
  numCp++;

  PlatoCheckpoint *cp = new PlatoCheckpoint(currBreak, &currRegs, 
					    isUserModeTrap());
  assert(cp != NULL);

  // save all of the breakpoint state
  unsigned long idx = 0;
  for(idx = 0; idx < breakPointVec.size(); idx++) {
    cp->setBreakData(idx, breakPointVec[idx].checkpoint());
  }
  
  cp->setUmSetBreakMap(copyPidMap(umSetBreakMap));
  cp->setUmQueuedBreakMap(copyPidMap(umQueuedBreakMap));

  checkId = INVALID_CHECK_ID;
  ioctl(umlFd, PLATO_IOC_CHECKPOINT, &checkId);

  if(checkId != INVALID_CHECK_ID) {
    assert(checkMap.find(checkId) == checkMap.end());
    checkMap[checkId] = cp;
    if(pauseRevirt && (getReplayState() != PLATO_REPLAY_STATE_NONE)) {
      ztrace(ZTRACE_RESUME, getVmHostPid(), 1, 0);
    }      
  } else {
    cerr << "did not take the checkpoint" << endl;
  }  

  gettimeofday(&stopTime, NULL);
  timersub(&stopTime, &startTime, &res);
  timeradd(&cpTime, &res, &cpTime);

  return checkId;
}

unsigned long Plato::getReplayState(void) {
  unsigned long replayState;
  ioctl(umlFd, PLATO_IOC_QUERY_REPLAY_STATE, &replayState);
  return replayState;
}

pid_t Plato::getVmHostPid(void) {
  pid_t pid;
  ioctl(umlFd, PLATO_IOC_QUERY_HOST_PID, &pid);
  return pid;
}

void
Plato::rollback(unsigned long checkId) {
  PlatoCheckpoint *cp;
  struct timeval startTime, stopTime, res;
  gettimeofday(&startTime, NULL);
  numRb++;

  if(checkMap.find(checkId) == checkMap.end()) {
    cerr << "invalid checkpoint id" << endl;
  }
  cp = checkMap.find(checkId)->second;

  if(cp->isUserMode() != isUserModeTrap()) {
    cerr << "cannot restore a checkpoint from different mode (user v kernel)"
	 << endl;
    assert(0);
  }

  ioctl(umlFd, PLATO_IOC_RESTORE_CHECK, checkId);

  // restore breakpoint state
  unsigned long idx;
  for(idx = 0; idx < breakPointVec.size(); idx++) {
    breakPointVec[idx].rollback(cp->getBreakData(idx));
  }
  
  // restore local state
  currRegs = cp->getSavedRegs();
  currBreak = cp->getSavedBreak();

  delSetBreakMap(umSetBreakMap);
  umSetBreakMap = copyPidMap(cp->getUmSetBreakMap());
  delSetBreakMap(umQueuedBreakMap);
  umQueuedBreakMap = copyPidMap(cp->getUmQueuedBreakMap());

  gettimeofday(&stopTime, NULL);
  timersub(&stopTime, &startTime, &res);
  timeradd(&rbTime, &res, &rbTime);
  // XXX FIXME return the breakpoint id for the restored context
}

unsigned long
Plato::callFunction(char *funcName, vector< unsigned long> *args, unsigned long funcAddr) {
  return callFunction(funcName, args, true, funcAddr);
}

unsigned long
Plato::callFunctionNoStack(char *funcName, unsigned long arg0) {
  vector<unsigned long> vec;
  
  vec.insert(vec.begin(), arg0);
  
  return callFunction(funcName, &vec, false);
}

bool Plato::atFunctionReturn(struct pt_regs *regs, unsigned long numArgs,
			     pid_t pid) {
  long stackLoc = regs->esp;
  stackLoc = stackLoc + ((long) (numArgs * sizeof(unsigned long)));

  if(pid && (pid != getCurrentPid())) return false;

  if((regs->eip == currRegs.eip) && (stackLoc == currRegs.esp)) {
    return true;
  }

  return false;
}

unsigned long Plato::callFunction(unsigned long funcAddr, unsigned long arg0) {
  vector<unsigned long> args;
  args.insert(args.end(), arg0);
  return callFunction("",&args,true,funcAddr);
}

unsigned long 
Plato::callFunction(char *funcName, vector< unsigned long> *args,
				 bool useStack, unsigned long funcAddr) {
  // first find the function break point
  vector< breakpoint >::iterator iter, breakIter;
  struct pt_regs regs;
  unsigned long idx, tmpArg;
  int ret;
  pid_t pid;

  if(funcAddr == 0) {
    for(iter = breakPointVec.begin(); iter != breakPointVec.end(); iter++) {
      if(iter->canCallFunction() && iter->getFunctionName() == string(funcName)) {
	assert(currBreak->isUserMode() == iter->isUserMode());
	funcAddr = iter->getEntryAddr();
	break;
      }
    }
  } else {
    // reuse the current context
    iter = currBreak;
  }
  assert(funcAddr != 0);
  regs = currRegs;
  regs.eip = funcAddr;
  assert(regs.eip != 0);
  // set the call stack
  if(useStack) {
    for(idx = 0; idx < args->size(); idx++) {
      regs.esp -= sizeof(unsigned long);
      tmpArg = args->at(args->size() - idx - 1);
      copy_to_guest(regs.esp, &tmpArg, sizeof(tmpArg));
    }
  } else {
    assert(args->size() == 1);
    regs.eax = args->at(0);
  }
      
  // setup the return address
  regs.esp -= sizeof(unsigned long);
  tmpArg = currRegs.eip - 1;
  copy_to_guest(regs.esp, &tmpArg, sizeof(tmpArg));

  set_guest_regs(iter, &regs);

  // only set the pid if we are calling a user mode function
  pid = (iter->isUserMode()) ? getCurrentPid() : 0;
  vector< breakpoint > ::iterator savedBreak = currBreak;
  struct pt_regs savedRegs = currRegs;

  // now wait for the return break point
  while(1) {
    ioctl(umlFd, PLATO_IOC_RESTART, 0);
    ret = read(umlFd, &regs, sizeof(regs));
    if(ret != sizeof(regs)) assert(0);
    
    breakIter = breakPointVec.end();
    if(atFunctionReturn(&regs, args->size(), pid)) {
      // dont worry about cleaning up stack, it will get reset when
      // the user returns from the callback function
      
      // dont worry about actual stack values, COW mem should take
      // care of this for us if we need it
      return regs.eax;
    } else {
      if(isUserModeTrap()) {
	handleUserModeTrap(&regs, true);
      } else {
	breakIter = handleKernelModeTrap(&regs, true);
      }
      
      if(breakIter == exitBreak) {
	// check to see if it is our process that is dying
	if(pid && (pid == getCurrentPid())) {
	  throw "process exiting before returning";
	}
      }	  
      
      currBreak = savedBreak;
      currRegs = savedRegs;
    } 
  }
  
  cerr << "********error, function " << funcName << " not found" << endl;

  return 0;
}    
unsigned long
Plato::callFunction(char *funcName) {
  vector<unsigned long> args;
  return callFunction(funcName, &args);
}

unsigned long
Plato::callFunction(char *funcName, unsigned long arg0) {
  vector<unsigned long> args;
  args.insert(args.end(),arg0);
  return callFunction(funcName, &args);
}

unsigned long
Plato::callFunction(char *funcName, unsigned long arg0,
		    unsigned long arg1) {
  vector<unsigned long> args;
  args.insert(args.end(),arg0);
  args.insert(args.end(),arg1);
  return callFunction(funcName, &args);
}

unsigned long
Plato::callFunction(char *funcName, unsigned long arg0,
		    unsigned long arg1, unsigned long arg2) {
  vector<unsigned long> args;
  args.insert(args.end(),arg0);
  args.insert(args.end(),arg1);
  args.insert(args.end(),arg2);
  return callFunction(funcName, &args);
}

unsigned long getCurrent(struct pt_regs *kernRegs) {
  return kernRegs->esp & 0xffffc000;
}

unsigned long
Plato::getCurrent() {
  return ::getCurrent(&currRegs);
} 

unsigned long
Plato::getRegVal(long regno) {
  return currBreak->getRegVal(&currRegs, regno);
}

void
Plato::setRegVal(long regno, unsigned long val) {
  currBreak->setRegVal(&currRegs, regno, val);
}

string Plato::getCurrentComm() {
  string comm;
  char commBuf[COMM_SIZE+1];
  struct linux_task_struct *tsk;
  tsk = (struct linux_task_struct *) getCurrent();

  // not a fundamental limitation, could be fixed but should not be needed
  assert(!isUserModeTrap());

  memset(commBuf,0, sizeof(commBuf));
  this->copy_from_guest(commBuf, &tsk->comm, COMM_SIZE);
  comm = string(commBuf);
  return comm;
}
