#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>

#include "uml.h"

int main(int argc, char *argv[]) {
  int opt;
  int checkpoint = 0;
  int rollback = 0;
  int restore = 0;
  int foundId = 0;
  int ret;
  int fd;
  unsigned long checkId;

  struct uml_ioc_checkpoint_struct cps;
  struct uml_ioc_rollback_struct rbs;

  memset(&cps, 0, sizeof(cps));
  memset(&rbs, 0, sizeof(rbs));

  while((opt = getopt(argc,argv,"cr:i:")) != -1) {
    switch(opt) {
    case 'c':
      checkpoint = 1;
      break;
    case 'r':
      rollback = 1;
      ret = sscanf(optarg,"%d",&restore); assert(ret == 1);
    case 'i':
      ret = sscanf(optarg,"%lu",&checkId); assert(ret == 1);
      foundId = 1;
      break;
    }
  }

  assert(checkpoint || rollback);
  assert(!(checkpoint && rollback));
  if(rollback) {
    assert(foundId);
  }

  
  fd = open("/dev/uml",O_RDWR);
  assert(fd >= 0);

  if(checkpoint) {
    ret = read(fd, &cps, sizeof(cps));
    if(ret == -EINTR) return 0;
    assert(ret == sizeof(cps));
    printf("uml checkpoint id = %lu\n",cps.checkId);
    ioctl(fd,UML_IOC_RESTART_UML,0);
  } else if(rollback) {
    rbs.checkId = checkId;
    rbs.restore = restore;
    ret = write(fd, &rbs, sizeof(rbs));
    if(ret == -EINTR) return 0;
    assert(ret == sizeof(rbs));
    if(restore == 0) {
      printf("checkpoint deleted\n");
    } else {
      printf("uml rolled back\n");
    }
    ioctl(fd,UML_IOC_RESTART_UML,0);
  } else {
    assert(0);
  }

  return 0;
}
