/*
 * chkpt_interval.c
 */
#include <stdlib.h>
#include <stdio.h>
#include <asm/unistd.h>
#include <errno.h>
#include <time.h>
#include <string.h>

#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>

#include <revirt/rlog.h>
#include <revirt/ztrace.h>
#include <revirt/syscalls.h>
#include <revirt/flags.h>

#include <sys/vfs.h>

#include "library.h"
#include "lib-checkpoint.h"

#include "uml.h"

#define MAX_DATAPOINTS 30

struct datapoint_struct {
    int interval;
    int ncheckpoints;
    int max_chkpts;
    struct timeval run_start;
    struct timeval run_end;
    struct timeval restore_start;
    struct timeval restore_end;
    long long freeblocks_start, freeblocks_end, blocks_deleted;
} dp[MAX_DATAPOINTS];

int intervals[] = { 6 } ;

void print_datapoint(struct datapoint_struct * d) {
    double run1, run2, res1, res2;
    long long bytes;
    long long blocks;
    run1=d->run_start.tv_sec + d->run_start.tv_usec/1000000.0;
    run2=d->run_end.tv_sec + d->run_end.tv_usec/1000000.0;
    res1=d->restore_start.tv_sec + d->restore_start.tv_usec/1000000.0;
    res2=d->restore_end.tv_sec + d->restore_end.tv_usec/1000000.0;
    bytes=d->freeblocks_start - d->freeblocks_end + d->blocks_deleted;
    blocks = bytes / 1024;
    fprintf(stderr, "Interval %d checkpoints %d (%d) 1k-b %lld (%lldD) run %lf restore %lf\n",
		d->interval,
		d->max_chkpts,
		d->ncheckpoints,
		blocks,
		d->blocks_deleted,
		run2-run1,
		res2-res1);
    fprintf(stdout, "Interval %d checkpoints %d (%d) 1k-b %lld (%lldD) run %lf restore %lf\n",
		d->interval,
		d->max_chkpts,
		d->ncheckpoints,
		blocks,
		d->blocks_deleted,
		run2-run1,
		res2-res1);
}

int first_cid;
int max_time=0;

int step_forward(void);

int checkpoint_periodic(struct datapoint_struct * d) {
    int ret;
    int max_chkpts, i, last_cid;
    struct timespec ts;
    struct timeval pre, post;
    int lag=0;
    //struct statfs sfs;

    printf("checkpoint_periodic, interval %d\n", d->interval);

    deleted_checkpoint_blocks=0;

    if(d->interval) {
	max_chkpts=max_time/d->interval;
	fprintf(stderr, "number of checkpoints to take: %d\n", max_chkpts);
	if(max_chkpts + 1 > MAX_UML_CHECK) {
	    fprintf(stderr, "max_chkpts > UML_CHECK!\n");
	}
	ts.tv_sec = d->interval;
	ts.tv_nsec = 0;
    } else {
    	fprintf(stderr, "d->interval == 0, getting total interval...\n");
    	max_chkpts = 0;
    }

    cleanup_intermediate();
#if 0
    fprintf(stderr, "reclaiming extra disk blocks\n");
    dbraw_reclaim();
#endif

    /* Want a fresh run, so take a non-counted checkpoint that gets deleted */
    fprintf(stderr, "taking 'sync' checkpoint...");
    step_forward();
    take_checkpoint_d();
    fprintf(stderr, "done.\n");


    gettimeofday(&d->run_start, NULL);

    //statfs("/checkpoint/", &sfs);
    //d->freeblocks_start=(long long)sfs.f_bfree * sfs.f_bsize;
    d->freeblocks_start=get_free_blocks();

    ioctl(umlfd, UML_IOC_RESTART_UML, 0);

    /* interval 0 -> max_chkpts==0 -> this never gets executed */
    for(i=0; i<max_chkpts ; i++) {
	/* If the last checkpoint didn't lag too long, wait a little longer */
	if(d->interval - lag > 0) {
	    ts.tv_sec = d->interval - lag;
	    ts.tv_nsec = 0;
	    nanosleep(&ts, NULL);
	
	} else {
	    fprintf(stderr, "checkpoint_interval: lag %d, interval %d\n",
	    		lag, d->interval);
	    if((lag / d->interval) > 1) {
		fprintf(stderr, "compressing %d checkpoints into one\n",
			lag / d->interval);
		max_chkpts -= (lag / d->interval) - 1;
	    }
	}

	/* This is a mild race condition (between checking for an event and
	 * not) but it's better than having nothing... */
	if(waitevent_is_waiting()) {
	    fprintf(stderr, "waitevent_is_waiting returns true!  setting max_chkpts = i\n");
	    max_chkpts=i;
	    break;
	} else {
	/* Wait for the right time to take a checkpoint, noting lag... */
	gettimeofday(&pre, NULL);
	ret = ioctl(umlfd, UML_IOC_PAUSE, 0);
	gettimeofday(&post, NULL);
	}

	lag = post.tv_sec - pre.tv_sec;
	if(lag > 1) {
	    printf("lag %d\n", lag);
	}

	if(ret) {
    	    fprintf(stderr, "uml_pause failed: %d\n", ret);
    	    exit(1);
       	}
	take_checkpoint_d();
	ioctl(umlfd, UML_IOC_RESTART_UML, 0);
    }

    /* And wait for the last marker... */
    waitevent_wait();

    if(waitevent_is_waiting()) {
    	waitevent_cont();
    }

    /* And stop things again */
    ret = ioctl(umlfd, UML_IOC_PAUSE, 0);
    if(ret) {
	fprintf(stderr, "uml_pause failed: %d\n", ret);
	exit(1);
    }

    last_cid=take_checkpoint_d();
    gettimeofday(&d->run_end, NULL);

    //statfs("/checkpoint/", &sfs);
    //d->freeblocks_end=(long long)sfs.f_bfree * sfs.f_bsize;
    d->freeblocks_end=get_free_blocks();
    d->blocks_deleted = deleted_checkpoint_blocks;

#if 0
    gettimeofday(&d->restore_start, NULL);
    restore_checkpoint(first_cid);
    gettimeofday(&d->restore_end, NULL);
#endif
    if(!d->interval) {
	/* Fix this fudge factor... */
	max_time = (d->run_end.tv_sec - d->run_start.tv_sec)*0.95-30;
	fprintf(stderr, "Max time for checkpoints: %d seconds\n", max_time);
    }
    /* Delete all checkpoints but first... */
    get_checkpoint_list();

    d->ncheckpoints = nr_checkpoints-1;
    d->max_chkpts = max_chkpts + 1;
#if 0
    /* Deleting backwards is faster now... */
    //for(i=0; i<MAX_UML_CHECK; i++) {
    for(i=MAX_UML_CHECK-1; i>=0; i--) {
    	int cid=checkpoint_list[i].checkId;
    	if(cid != INVALID_CHECK_ID && cid != first_cid) {
		fprintf(stderr, "Deleting checkpoint id %d\n", cid);
		delete_checkpoint(cid);
	}
    }
#endif
    print_datapoint(d);
    return 0;
}

void do_run(int count) {
    int i;

    count_deleted_blocks=1;

    for(i=0; i<count; i++) {
    	checkpoint_periodic(dp+i);
    }
}

int main(int argc, char * argv[]) {
    int i, astart=1, count=argc-1;

    if(argc<2) {
    	fprintf(stderr, "Usage: %d [-maxtime] interval [interval ...]\n", 
			argv[0]);
	exit(1);
    }

    if(argv[1][0]=='-') {
    	astart++;
	count--;
    	max_time=atoi(argv[1]+1);
	fprintf(stderr, "Setting max_time to %d seconds\n", max_time);
    } else if(atoi(argv[1])!=0) {
    	fprintf(stderr, "You must specify -maxtime if the first interval is not zero.\n");
	exit(1);
    }

    if(!count) {
    	fprintf(stderr, "No data points!\n");
    	exit(1);
    }

    for(i=0; i<count; i++) {
	dp[i].interval=atoi(argv[i+astart]);
	fprintf(stderr, "Setting dp[%d] interval to %d\n", i, dp[i].interval);
    }

    setup_stuff();

    do_run(count);

    waitevent_enable(0);

    return 0;
}
