#include <stdlib.h>
#include <stdio.h>

#include <asm/unistd.h>
#include <errno.h>
#include <time.h>
#include <string.h>

#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>

#include <revirt/rlog.h>
#include <revirt/ztrace.h>
#include <revirt/syscalls.h>
#include <revirt/flags.h>

#include <sys/vfs.h>

#include "library.h"
#include "uml.h"
#include "lib-checkpoint.h"

#include "../revirt-skas-2.4.18/uml/dbraw.h"

int umlfd=0;
int dbrawfd=0;
#define DBRAW_FILE "/dev/raw/dbraw11"
int first_cid=-1;
int nr_checkpoints=0;
struct uml_ioc_checkinfo_struct checkpoint_list[MAX_UML_CHECK];


int open_umlfd(void) {
    if(umlfd)
	return -EBUSY;
    if((umlfd = open("/dev/uml",O_RDWR))<0) {
	perror("opening /dev/uml");
    }

    if(!dbrawfd) {
	if((dbrawfd = open(DBRAW_FILE, O_RDONLY))<0) {
	    fprintf(stderr, "Opening %s: %s\n", DBRAW_FILE, strerror(errno));
	    dbrawfd = 0;
	}
    }
    return umlfd;
}

long long dbraw_free_space(void) {
    int ret;
    long long val;
    struct dbraw_size_struct size;

    ret=ioctl(dbrawfd, DBRAW_GET_SIZES, &size);
    if(ret < 0) {
	perror("dbraw_get_sizes");
	return ret;
    }
    val = (long long)size.numFree * size.blockSize;
    fprintf(stderr, "dbraw_free_space: %lld (%d * %d)\n",
    	val, size.numFree, size.blockSize);
    return val;
}

int dbraw_reclaim(void) {
    int ret;
    if(!dbrawfd) {
	fprintf(stderr, "dbraw_reclaim: dbrawfd not open!\n");
	return -ENOENT;
    }
    ret = ioctl(dbrawfd, DBRAW_RECLAIM_FREE_BLOCKS, NULL);
    return ret;
}

int ztrace_waitevent(int cmd, int mark) {
	return ztrace(ZTRACE_WAITEVENT, 0, (unsigned long)cmd, (void *)mark);
}

int waitevent_enable(int enable) {
	return ztrace_waitevent(WAITEVENT_ENABLE, enable);
}

int waitevent_marker(int mark) {
	return ztrace_waitevent(WAITEVENT_SETMARK, mark);
}

int waitevent_wait(void) {
	return ztrace_waitevent(WAITEVENT_WAIT, 0);
}

int waitevent_cont(void) {
	return ztrace_waitevent(WAITEVENT_CONT, 0);
}

int waitevent_is_waiting(void) {
	return ztrace_waitevent(WAITEVENT_IS_WAITING, 0);
}


void print_checkpoint(struct uml_ioc_checkinfo_struct *cis) {
    printf("%d: [ sid %lld rhid %lu bc %lld eip 0x%x ]\n",
    	cis->checkId, cis->guestSyscallCount, cis->revirtId.rhid,
	cis->revirtId.bc, cis->eip);
}

void print_checkpoint_list(void) {
    int i;
    for(i=0; i<MAX_UML_CHECK; i++)  {
    	if(checkpoint_list[i].checkId == INVALID_CHECK_ID)
	    continue;

	print_checkpoint(checkpoint_list+i);
    }
}

int get_checkpoint_list(void) {
    int i;
    nr_checkpoints=0;
    for(i=0; i<MAX_UML_CHECK; i++) {
    	checkpoint_list[i].checkId=i;
    	ioctl(umlfd, UML_IOC_GET_CHECKINFO, checkpoint_list+i);
	if(checkpoint_list[i].checkId != INVALID_CHECK_ID) {
	    nr_checkpoints++;	    
	}
    }
    return nr_checkpoints;
}

int get_current_info(struct uml_ioc_checkinfo_struct * cur) {
	cur->checkId=INVALID_CHECK_ID;
	return ioctl(umlfd, UML_IOC_GET_CHECKINFO, cur);

}

#include <sys/vfs.h>

long long deleted_checkpoint_blocks=0;
int count_deleted_blocks=0;

long long get_free_blocks(void) {
	struct statfs sfs;
	long long hostfs, dbraw, val;
	statfs("/checkpoint/", &sfs);
	hostfs = (long long)sfs.f_bfree * sfs.f_bsize;
	dbraw=dbraw_free_space();
	val = hostfs + dbraw;
	fprintf(stderr, "get_free_blocks: %lld (h %lld d %lld)\n",
		val, hostfs, dbraw);
	return val;
}

int delete_checkpoint(int checkId) {
    struct uml_ioc_rollback_struct rbs;
    struct statfs sfs;
    long long pre, post, d;
    int ret;

    if(count_deleted_blocks) {
	pre=get_free_blocks();
    }
    
    memset(&rbs, 0, sizeof(rbs));
    rbs.checkId=checkId;
    rbs.restore=0;
    ret= write(umlfd, &rbs, sizeof(rbs));

    if(count_deleted_blocks) {
    	post=get_free_blocks();
	/* 
	 * Duh, this is FREE blocks... so the amount deleted is the FREE
	 * blocks after minus the FREE blocks before...
	 */
	d=post-pre;
	fprintf(stderr, "pre %lld, post %lld, d %lld\n", pre, post, d);
	if(d > 0)
	    deleted_checkpoint_blocks += d;
    }
    return ret;
}

int take_checkpoint(void) {
    int ret;
    struct uml_ioc_checkpoint_struct cps;
    ret = read(umlfd, &cps, sizeof(cps));
    if(ret!=sizeof(cps)) {
    	perror("reading /dev/uml");
    	exit(1);
    }
    return cps.checkId;
}

int restore_checkpoint(int checkId) {
    struct uml_ioc_rollback_struct rbs;
    memset(&rbs, 0, sizeof(rbs));
    rbs.checkId=checkId;
    rbs.restore=1;
    return write(umlfd, &rbs, sizeof(rbs));
}

/* Assumes you're already stopped; steps forward at least one checkpoitnable
 * place */
int step_forward(void) {
    struct uml_ioc_checkinfo_struct cur;
    struct timespec ts;
    int last;

    get_current_info(&cur);
    last = cur.guestSyscallCount;

    do {
	ioctl(umlfd, UML_IOC_RESTART_UML, 0);
	ts.tv_sec = 2;
	ts.tv_nsec=0;
	nanosleep(&ts, NULL);
	ioctl(umlfd, UML_IOC_PAUSE, 0);
	get_current_info(&cur);
    } while(cur.guestSyscallCount == last);

    return cur.guestSyscallCount - last;
}

#include <sys/time.h>

int wait_checkpoint_interval(int interval) {
    struct timespec ts;
    struct uml_ioc_checkinfo_struct cur;
    int notdone=0;
    int j, ret;

    ts.tv_sec=interval;
    ts.tv_nsec=0;

    /* 
     * Make sure moved forward at least one checkpoint period
    */
    do {
	ioctl(umlfd, UML_IOC_RESTART_UML, 0);
	nanosleep(&ts, NULL);
	ret = ioctl(umlfd, UML_IOC_PAUSE, 0);
	if(ret) {
	    fprintf(stderr, "uml_pause failed: %d\n", ret);
	    exit(1);
	}
	get_checkpoint_list();
	get_current_info(&cur);
	for(j=0; j<nr_checkpoints; j++) {
	    if(checkpoint_list[j].checkId == INVALID_CHECK_ID)
		continue;
	    if(checkpoint_list[j].guestSyscallCount 
			    == cur.guestSyscallCount) {
		    notdone=1;
		    break;
	    }
	}
    }  while(notdone);
    return 0;
}

int take_checkpoint_d(void) {
    int n;

    if(get_checkpoint_list()==MAX_UML_CHECK) {
	while(1) {
	    n=rand() % MAX_UML_CHECK;
	    if(checkpoint_list[n].checkId != INVALID_CHECK_ID) 
	    	break;
	}
	fprintf(stderr, "Max checkpoints taken, deleting checkpoint %d\n",
		checkpoint_list[n].checkId);
	delete_checkpoint(checkpoint_list[n].checkId);
    }
    return take_checkpoint();
}

int take_checkpoints(int interval, int checkpoints) {
    int i;

    for(i=0; i<checkpoints ; i++) {
	wait_checkpoint_interval(interval);
	fprintf(stderr, "Taking checkpoint %d...\n", i);
	take_checkpoint_d();
    }
    return 0;
}

int find_current_checkpoint(int sid) {
    int i;
    for(i=0; i<nr_checkpoints; i++) {
    	if(checkpoint_list[i].guestSyscallCount == sid)
		return i;
    }
    return -1;
}

void delete_additional_checkpoints() {
    int i;
    fprintf(stderr, "Cleaning up additional checkpoints.\n");
    //for(i=1; i<nr_checkpoints; i++) {
    /* Deleting backwards is faster now. */
    for(i=nr_checkpoints-1; i>=1; i--) {
    	if(checkpoint_list[i].checkId == INVALID_CHECK_ID ) {	
    		fprintf(stderr, "WARNING! i %d has invalid checkId!\n", i);
       	} else {
    		fprintf(stderr, "Deleting checkpoint id %d\n", 
    		    checkpoint_list[i].checkId);
    		delete_checkpoint(checkpoint_list[i].checkId);
       	}
    }
}

void setup_stuff() {
    int ret;
    int uml_did_pause=0;

    if(open_umlfd()<0)
    	exit(1);

    /* See if we're attaching to an already-run system... */
    fprintf(stderr, "Trying to attach to running UML...\n");
    ret = ioctl(umlfd, UML_IOC_PAUSE, 15);
    if(ret==0) {
    	uml_did_pause=1;
    	fprintf(stderr, "ioctl didn't time out, getting checkpoints...\n");
	get_checkpoint_list();
	fprintf(stderr, "get_checkpoint_list: %d checkpoints\n",
			nr_checkpoints);
    } else {
    	uml_did_pause=0;
	nr_checkpoints=0;
    	fprintf(stderr, "ioctl timed out, Waiting for events...\n");
    }

    /* Enable event waiting */
    fprintf(stderr, "Enabling event waiting\n");
    if(waitevent_enable(1)<0) {
	    perror("waitevent_enable");
    	    exit(1);
    }
    fprintf(stderr, "Setting marker for 255...\n");
    if(waitevent_marker(255)<0) {
	    perror("waitevent_marker");
	    exit(1);
    }


    /* 
     * If UML wasn't running, or there weren't any checkpoints yet, 
     * wait for the first logmarker...
     */
    if(nr_checkpoints==0) {

	if(uml_did_pause) {
	    fprintf(stderr, "ioctl didn't timeout, but no chkpts; resuming\n");
	    ioctl(umlfd, UML_IOC_RESTART_UML, 0);
	}

	fprintf(stderr, "Waiting for marker...\n");
	waitevent_wait();

	if(waitevent_is_waiting()) {
	    fprintf(stderr, "Continuing...\n");
	    waitevent_cont();
	}
	/*
	* Then take a checkpoint.
	*/
	ret = ioctl(umlfd, UML_IOC_PAUSE, 0);
	if(ret) {
	    fprintf(stderr, "uml_pause failed: %d\n", ret);
	    exit(1);
	}

	/*
	 * Wait a random amt, in case we can't chkpt right there...
	 */
	wait_checkpoint_interval(rand() % 2);

	/* Take a checkpoint */
	fprintf(stderr, "Taking first checkpoint\n");
	first_cid=take_checkpoint();
    } else {
	struct uml_ioc_checkinfo_struct cur;
    	/*
	 * WARNING, assumptions galore here...
	 */
	fprintf(stderr, "Hm, %d checkpoints, seems to be already running.  Let's see if we can restore...\n",
		nr_checkpoints);
	if(checkpoint_list[0].checkId == INVALID_CHECK_ID) {
	    fprintf(stderr, "check_id[0].checkId invalid, don't know what to do...\n");
	    exit(1);
	}
	first_cid=checkpoint_list[0].checkId;

	get_current_info(&cur);

	/*
	 * If we're not at the first checkpoint, take a checkpoint here, and then
	 * resture to the first one.
	 * (We take a checkpoint here to make sure that after deleting checkpoints,
	 * the disk will start out in "logging" mode)
	 */ 
	if(cur.guestSyscallCount != checkpoint_list[0].guestSyscallCount) {
    	    fprintf(stderr, "Not at first_cid! Taking a checkpoint...\n");
	    take_checkpoint_d();
	    fprintf(stderr, "Restoring to checkpoint %d\n", first_cid);
	    restore_checkpoint(first_cid);
	    get_checkpoint_list();
	}

	/*
	 * Delete all other checkpoints (whether we had to restore or not)
	 */
	delete_additional_checkpoints();

	if(waitevent_is_waiting()) {
		waitevent_cont();
	}
    }
}

void cleanup_intermediate(void) {
    struct uml_ioc_checkinfo_struct cur;

    get_checkpoint_list();
    get_current_info(&cur);
    if(cur.guestSyscallCount != checkpoint_list[0].guestSyscallCount) {
	fprintf(stderr, "Restoring to initial checkpoint...\n");
	restore_checkpoint(checkpoint_list[0].checkId);
	delete_additional_checkpoints();
    }

    fprintf(stderr, "reclaiming extra disk blocks\n");
    dbraw_reclaim();
 
}

void cleanup_stuff(void) {
    /* Let the next process do this if it wants to... */
#if 0
    get_checkpoint_list();
    get_current_info(&cur);
    if(cur.guestSyscallCount != checkpoint_list[0].guestSyscallCount) {
	fprintf(stderr, "Restoring to initial checkpoint...\n");
	restore_checkpoint(checkpoint_list[0].checkId);
    }
 
    delete_additional_checkpoints();
#endif

    close(umlfd);

    waitevent_enable(0);
}

