#include <stdio.h>
#include <linux/autoconf.h>
#include <revirt/rlog.h>
#include <revirt/flags.h>
#include <revirt/ztrace.h>
#include <revirt/syscalls.h>
#include <unistd.h>
#include <linux/limits.h>
#include "printlog.h"

static const char *type_string[]= {
#define X(x) [ENTRY_##x] #x
  X(SCHED),
  X(FORK),
  X(EXIT),
  X(TOP),
  X(SYSRET),
#undef X
};

void print_schedlog(struct rlog_schedule_entry *e) {
#ifdef REPLAY_RECORD_ALL_PERFCTRS
	printf("pid:%d flags:%x\n [%016llx %08x %08x %08x]\n",
	       e->pid, e->flags, e->ctr.hw_branch, 
	       e->ctr.hw_intr, e->ctr.ex_pf, e->ctr.ex_dev);
#else
	printf("pid:%d flags:%x\n [%016llx]\n",
	       e->pid, e->flags, e->ctr.hw_branch);
#endif
	printf("eip:0x%x\n", e->eip);
}

int ztrace(int command,int pid,unsigned long addr,void *data) {
      int ret;
      ret= syscall(__NR_ztrace, command, pid, addr, data);
      return ret;
}


char * flag_names[] = {
"LOG",
"REPLAY",
"SSTEP", 
"FIRST_AFTER_SIG",
"PROCESS_IN_UMODE", 
"DMA",
"INT3", 
"WAIT", 
"AT_EXIT", 
"DELAY_PERF",
"SCHED", 
"IN_COPY_USER", 
"DATASQUASH", 
"XUSER", 
"ZOMBIE",
"SYSCALL",
"PTRACE", 
"NO_YIELD_ON_BLOCK",
"SUPPRESS_SCHED",
"SIGBLOCK",
"SYNC_BLOCK", 
"EMULATE_RDTSC",
"ZTRACE_CAN_CHECKPOINT",
"USER_TRAP",};

void print_replayflags(unsigned int replayflags) {
    int i;
    printf("revirt_flags: 0x%x (", replayflags);
    if(replayflags & 0xffffff) {
	for(i=0; i<24; i++) {
	    if(replayflags & (1 << i)) {
		printf("%s ", flag_names[i]);
	    }
	}
    } else {
	printf("NONE ");
    }
#if 0
	if(replayflags & REPLAY_LOG)
		printf("LOG ");
	if(replayflags & REPLAY_REPLAY)
		printf("REPLAY ");
	if(replayflags & REPLAY_SSTEP)
		printf("SSTEP ");
	if(replayflags & REPLAY_FIRST_AFTER_SIG)
		printf("FIRST_AFTER_SIG ");
	if(replayflags & REPLAY_PROCESS_IN_UMODE)
		printf("PROCESS_IN_UMODE ");
	if(replayflags & REPLAY_WAIT)
		printf("WAIT ");
	if(replayflags & REPLAY_AT_EXIT)
		printf("AT_EXIT ");
#if 0
	if(replayflags & REPLAY_NOTDONE)
		printf("NOTDONE ");
#endif
	if(replayflags & REPLAY_SCHED)
		printf("SCHED ");
	if(replayflags & REPLAY_IN_COPY_USER)
		printf("IN_COPY_USER ");
	if(replayflags & REPLAY_DATASQUASH)
		printf("DATASQUASH ");
	if(replayflags & REPLAY_XUSER)
		printf("XUSER ");
	if(replayflags & REPLAY_ZOMBIE)
		printf("ZOMBIE ");
	if(replayflags & REPLAY_SYSCALL)
		printf("SYSCALL ");
	if(replayflags & REPLAY_PTRACE)
		printf("PTRACE ");
	if(replayflags == 0)
		printf("NONE ");
#endif
	printf(")\n");
}

void print_ztflags(unsigned int zt_flags) {
	printf("zt_flags: 0x%x (", zt_flags);
	if(zt_flags & ZTRACE_XUSER_STOP)
		printf("XUSER_STOP ");
	if(zt_flags & ZTRACE_SYSENTER_STOP)
		printf("SYSENTER_STOP ");
	if(zt_flags & ZTRACE_SYSEXIT_STOP)
		printf("SYSEXIT_STOP ");
	if(zt_flags & ZTRACE_BKPT)
		printf("BKPT ");
	if(zt_flags & ZTRACE_BKPT_IN_KERNEL)
		printf("BKPT_IN_KERNEL ");
	if(zt_flags & ZTRACE_WAITEXIT)
		printf("WAITEXIT ");
	if(zt_flags & ZTRACE_BLOCK_STOP)
		printf("BLOCK_STOP ");
	if(zt_flags & ZTRACE_SIGNAL_STEP)
		printf("SIGNAL_STEP ");
	if(zt_flags & ZTRACE_PERF)
		printf("PERF ");
	if(zt_flags & ZTRACE_EPOCH_STOP)
		printf("EPOCH_STOP ");
	printf(")\n");
}

void print_state(struct ztrace_state_struct * zt_s) {
    	int i;
#define zt_state (*zt_s)
	printf( "eax:0x%8x\teip:0x%8x\n" \
		"ebx:0x%8x\tesp:0x%8x\n" \
		"ecx:0x%8x\tebp:0x%8x\n" \
		"edx:0x%8x\txds:0x%8x\n" \
		"esi:0x%8x\txes:0x%8x\n" \
		"edi:0x%8x\txcs:0x%8x\n" \
		"efl:0x%8x\txss:0x%8x\n" \
		"orig_eax: 0x%8x\n",
		zt_state.regs.eax, zt_state.regs.eip,
		zt_state.regs.ebx, zt_state.regs.esp,
		zt_state.regs.ecx, zt_state.regs.ebp,
		zt_state.regs.edx, zt_state.regs.xds,
		zt_state.regs.esi, zt_state.regs.xes,
		zt_state.regs.edi, zt_state.regs.xcs,
		zt_state.regs.eflags, zt_state.regs.xss,
		zt_state.regs.orig_eax );
	print_replayflags(zt_state.revirt_flags);
	print_ztflags(zt_state.zt_ctl.flags);
	printf( "zt_rununtil: eip:0x%08x wchan:0x%08x flags:0x%08x count:0x%llx\n",
		zt_state.zt_ctl.brk.eip,
		zt_state.zt_ctl.brk.wchan,
		zt_state.zt_ctl.brk.flags,
		zt_state.zt_ctl.brk.count);

	printf(" zt_rununtil.active: %d [ ", zt_state.zt_ctl.active.i);
	for(i=0; i<zt_state.zt_ctl.active.i ; i++)
		printf("%d ", zt_state.zt_ctl.active.pid[i]);
	printf("] epochs_skipped: %d\n", 
		zt_state.zt_ctl.active.epochs_skipped);

	printf(" rlog.active: %d [ ", zt_state.active.i);
	for(i=0; i<zt_state.active.i ; i++)
		printf("%d ", zt_state.active.pid[i]);
	printf("] epochs_skipped: %d\n", 
		zt_state.active.epochs_skipped);
	printf("replay_pos: %lld log_pos: %lld log_id: %lu\n", zt_state.replay_pos,
		zt_state.log_pos, zt_state.log_id);

	printf("zt_pid: %d zt_rhid: %u\n", zt_state.zt_ctl.pid, zt_state.zt_ctl.rhid);
	printf("cr2:0x%8x\ttrap_no:0x%8x\terror_code:0x%8x\n",
		zt_state.cr2, zt_state.trap_no, zt_state.error_code);
	for(i=0; i<8; i++) {
		printf("debugreg[%d]:0x%x\n", i, zt_state.debugreg[i]);
	}
	printf("code: ");
	for(i=0; i<16; i++) {
		printf("%x ", (int) zt_state.code[i]);
	}
	printf("\npcode: ");
	for(i=0; i<16; i++) {
		printf("%x ", (int) zt_state.pcode[i]);
	}
	printf("\n");
	printf("count: 0x%llx [%llx]\tsyscall_id: %lld\n",
		ADJ_BC(zt_state.ctr),
		zt_state.ctr.hw_branch, zt_state.syscall_id);
#ifdef REPLAY_RECORD_ALL_PERFCTRS
	printf("[%x %x %x | ",
		zt_state.ctr.hw_intr, 
		zt_state.ctr.ex_noerror, 
		zt_state.ctr.ex_error);
	for(i=0; i<REPLAY_NR_PERFCTRS; i++) {
		printf("%llx ", zt_state.ctr.counters[i]);
	}
	printf("]\n");
#endif

#ifdef REPLAY_RECORD_ALL_EXCEPTIONS
	for(i=0; i<REVIRT_NR_EX; i++) 
		if(zt_state.ctr.exceptions[i]>0)
			printf("[%d]:%x ", i, zt_state.ctr.exceptions[i]);
	printf("\n");
#endif
	printf("zt_perferr: %llx\n", zt_state.zt_ctl.perferr);
#undef zt_state
}

int ztrace_getstate(int pid, struct ztrace_state_struct * s) {
    return ztrace(ZTRACE_GETSTATE, pid, 0, (void *)s);
}

int ztrace_setopt(int opt, int val) {
	return ztrace(ZTRACE_SET_OPT, getpid(), opt, (void *)val);
}

int clear_fds(void) {
    int i, j=0;
    for(i=3; i<OPEN_MAX; i++) {
    	if(close(i)>=0) {
	    j++;
	    fprintf(stderr, "clear_fds:  closed fd %d.\n", i);
	}
    }
    return j;
}
