#include <stdlib.h>
#include <stdio.h>
#include <asm/unistd.h>
#include <errno.h>
#include <time.h>
#include <string.h>

#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>

#include <revirt/rlog.h>
#include <revirt/ztrace.h>
#include <revirt/syscalls.h>
#include <revirt/flags.h>

#include "library.h"

#include "lib-checkpoint.h"
#include "uml.h"

extern int umlfd;
extern int first_cid;

int max_check=MAX_UML_CHECK;

struct datapoint_struct {
    enum {
    	DP_CHECKPOINT,
	DP_DELETE
    } op;
    struct timeval start;
    struct timeval end;
} dp[1000];

int main(int argc, char * argv[]) {
    int interval;
    int i, total, x, cid, ret;

    double time;

    if(argc < 3 || argc > 4) {
    	fprintf(stderr, "Usage: %s interval datapoints [max_check]\n", argv[0]);
	exit(1);
    }
    interval=atoi(argv[1]);
    total=atoi(argv[2]);
    fprintf(stderr, "interval: %d total: %d\n", interval, total);

    if(argc == 4)
    	max_check=atoi(argv[3]);

    /*
     * Get, somehow, to the first 255 marker...
     */
    fprintf(stderr, "setup_stuff()\n");
    setup_stuff();

    fprintf(stderr, "Doing %d datapoints\n", total);

    for(i=0; i<total; i++) {
	wait_checkpoint_interval(interval);

	x=rand() % max_check;
	cid=checkpoint_list[x].checkId;

	if(cid == INVALID_CHECK_ID) {
	    struct uml_ioc_checkpoint_struct cps;
	    double t1, t2;

	    dp[i].op=DP_CHECKPOINT;

	    sync();
	    gettimeofday(&dp[i].start, NULL);
	    ret = read(umlfd, &cps, sizeof(cps));
	    sync();
	    gettimeofday(&dp[i].end, NULL);
	    if(ret != sizeof(cps)) {
	    	perror("reading /dev/uml");
		exit(1);
	    }

	    t1=dp[i].start.tv_sec + dp[i].start.tv_usec/1000000.0;
	    t2=dp[i].end.tv_sec + dp[i].end.tv_usec/1000000.0;
	    fprintf(stderr, "%d C %d %lf\n", i, nr_checkpoints, t2-t1);

	} else {
	    struct uml_ioc_rollback_struct rbs;
	    double t1, t2;


	    memset(&rbs, 0, sizeof(rbs));

	    rbs.checkId=cid;
	    rbs.restore=0;

	    dp[i].op=DP_DELETE;

	    sync();
	    gettimeofday(&dp[i].start, NULL);
	    ret=write(umlfd, &rbs, sizeof(rbs));
	    sync();
	    gettimeofday(&dp[i].end, NULL);

	    if(ret != sizeof(rbs)) {
	    	perror("writing /dev/uml");
		exit(1);
	    }

	    t1=dp[i].start.tv_sec + dp[i].start.tv_usec/1000000.0;
	    t2=dp[i].end.tv_sec + dp[i].end.tv_usec/1000000.0;
	    fprintf(stderr, "%d\tD%d %lf\n", i, nr_checkpoints, t2-t1);
	}

    }

    cleanup_stuff();

    printf("Checkpoint times:\n");
    time=0;
    x=0;
    for(i=0; i<total; i++) {
	if(dp[i].op==DP_CHECKPOINT) {
	    double t1, t2;
	    t1=dp[i].start.tv_sec + dp[i].start.tv_usec/1000000.0;
	    t2=dp[i].end.tv_sec + dp[i].end.tv_usec/1000000.0;
	    printf("%d: %lf\n", i, t2-t1);
	    time += (t2-t1);
	    x++;
	}
    }
    printf("avg: %lf\n", time / x);

    printf("\nDelete times:\n");
    time=0;
    x=0;
    for(i=0; i<total; i++) {
	if(dp[i].op==DP_DELETE) {
	    double t1, t2;
	    t1=dp[i].start.tv_sec + dp[i].start.tv_usec/1000000.0;
	    t2=dp[i].end.tv_sec + dp[i].end.tv_usec/1000000.0;
	    printf("%d: %lf\n", i, t2-t1);
	    time += (t2-t1);
	    x++;
	}
    }
    printf("avg: %lf\n", time / x);

    return 0;
}
