#include <stdlib.h>
#include <stdio.h>
#include <asm/unistd.h>
#include <errno.h>
#include <time.h>
#include <string.h>

#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <fcntl.h>

#include <revirt/rlog.h>
#include <revirt/ztrace.h>
#include <revirt/syscalls.h>
#include <revirt/flags.h>

#include "library.h"
#include "lib-checkpoint.h"

#include "uml.h"

int max_check=MAX_UML_CHECK;

struct datapoint_struct {
    enum {
    	DP_CHECKPOINT,
	DP_DELETE
    } op;
    struct timeval start;
    struct timeval end;
} dp[1000];

int main(int argc, char * argv[]) {
    struct timespec ts;
    int i, total, x, cid, ret;

    double time;

    if(argc < 3 || argc > 4) {
    	fprintf(stderr, "Usage: %s interval datapoints [max_check]\n", argv[0]);
    }
    ts.tv_sec=atoi(argv[1]);
    ts.tv_nsec=0;
    total=atoi(argv[2]);

    if(argc == 4)
    	max_check=atoi(argv[3]);

    if(open_umlfd()<0)
    	exit(1);

    //ztrace_setopt(REPLAY_OPT_STK_NOWRITE, REPLAY_OPT_STK_NOWRITE);

    waitevent_enable(1);
    fprintf(stderr, "Setting marker for 255...\n");
    waitevent_marker(255);

    fprintf(stderr, "Waiting for marker...\n");
    waitevent_wait();

    fprintf(stderr, "Continuing...\n");
    waitevent_cont();

    fprintf(stderr, "Doing %d datapoints\n", total);
//    ts.tv_sec=5;
//    ts.tv_nsec=100000;

    ret = ioctl(umlfd, UML_IOC_PAUSE, 10);
    if(ret) {
	fprintf(stderr, "uml_pause failed: %d\n", ret);
	exit(1);
    }

    for(i=0; i<total; i++) {
    	int notdone=0, j;
	struct uml_ioc_checkinfo_struct cur;

	/* 
	 * Make sure moved forward at least one checkpoint period
	 */
    	do {
	    ioctl(umlfd, UML_IOC_RESTART_UML, 0);
	    nanosleep(&ts, NULL);
	    ret = ioctl(umlfd, UML_IOC_PAUSE, 10);
	    if(ret) {
		fprintf(stderr, "uml_pause failed: %d\n", ret);
		exit(1);
	    }
	    get_checkpoint_list();
	    get_current_info(&cur);
	    for(j=0; j<max_check; j++) {
		if(checkpoint_list[j].checkId == INVALID_CHECK_ID)
		    continue;
		if(checkpoint_list[j].guestSyscallCount 
				== cur.guestSyscallCount) {
			notdone=1;
			break;
		}
	    }
	}  while(notdone);

	x=rand() % max_check;
	//x=MAX_UML_CHECK-1;
	cid=checkpoint_list[x].checkId;
	if(cid == INVALID_CHECK_ID) {
	    struct uml_ioc_checkpoint_struct cps;
	    double t1, t2;

	    dp[i].op=DP_CHECKPOINT;

	    gettimeofday(&dp[i].start, NULL);
	    ret = read(umlfd, &cps, sizeof(cps));
	    gettimeofday(&dp[i].end, NULL);
	    if(ret != sizeof(cps)) {
	    	perror("reading /dev/uml");
		exit(1);
	    }
	    t1=dp[i].start.tv_sec + dp[i].start.tv_usec/1000000.0;
	    t2=dp[i].end.tv_sec + dp[i].end.tv_usec/1000000.0;
	    fprintf(stderr, "%d C %d %lf\n", i, nr_checkpoints, t2-t1);

	} else {
	    struct uml_ioc_rollback_struct rbs;
	    double t1, t2;


	    memset(&rbs, 0, sizeof(rbs));

	    rbs.checkId=cid;
	    rbs.restore=0;

	    dp[i].op=DP_DELETE;

	    gettimeofday(&dp[i].start, NULL);
	    ret=write(umlfd, &rbs, sizeof(rbs));
	    gettimeofday(&dp[i].end, NULL);

	    if(ret != sizeof(rbs)) {
	    	perror("writing /dev/uml");
		exit(1);
	    }

	    t1=dp[i].start.tv_sec + dp[i].start.tv_usec/1000000.0;
	    t2=dp[i].end.tv_sec + dp[i].end.tv_usec/1000000.0;
	    fprintf(stderr, "%d\tD%d %lf\n", i, nr_checkpoints, t2-t1);
	}

    }
   ioctl(umlfd, UML_IOC_RESTART_UML, 0);


    fprintf(stderr, "Now I'm done -- disabling wait and letting it run to the end...\n");
    waitevent_enable(0);
    //waitevent_cont();

    //ztrace_setopt(REPLAY_OPT_STK_NOWRITE, 0);

    close(umlfd);

    printf("Checkpoint times:\n");
    time=0;
    x=0;
    for(i=0; i<total; i++) {
	if(dp[i].op==DP_CHECKPOINT) {
	    double t1, t2;
	    t1=dp[i].start.tv_sec + dp[i].start.tv_usec/1000000.0;
	    t2=dp[i].end.tv_sec + dp[i].end.tv_usec/1000000.0;
	    printf("%d: %lf\n", i, t2-t1);
	    time += (t2-t1);
	    x++;
	}
    }
    printf("avg: %lf\n", time / x);

    printf("\nDelete times:\n");
    time=0;
    x=0;
    for(i=0; i<total; i++) {
	if(dp[i].op==DP_DELETE) {
	    double t1, t2;
	    t1=dp[i].start.tv_sec + dp[i].start.tv_usec/1000000.0;
	    t2=dp[i].end.tv_sec + dp[i].end.tv_usec/1000000.0;
	    printf("%d: %lf\n", i, t2-t1);
	    time += (t2-t1);
	    x++;
	}
    }
    printf("avg: %lf\n", time / x);

    return 0;
}
