#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <linux/unistd.h>
#include <asm/unistd.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "readlog.h"
#include "misc.h"
#include <linux/autoconf.h>
#include <revirt/rlog.h>
#include <string.h>
#include <errno.h>


#define BUFSIZE (4096*2)

static int fd;
char buf[BUFSIZE];
/* FIXME -- long files */
loff_t zr_pos;
unsigned long next_rhid=1;

/* gcc may align things.. kernel prints packed */
#define DATAOF(x) ((void*) (((char*) (x))+sizeof(struct rlog_header)))

void zr_open_log(char *fn) {
	fprintf(stderr, "about to open\n");
	try((fd= open(fn,O_RDONLY))>0);
	fprintf(stderr, "about to return\n");
	zr_pos=0;
}

void zr_release_log(void) {
	close(fd);
}

loff_t lseek64(int, loff_t, int);

/*
 * Sam's checkpointing doesn't restore file pos state, so hack it by storing it in mem...
 */
int __zr_read(int fd, char * buf, int len) {
	long int r;
	loff_t result;

	if((result=lseek64(fd, zr_pos, SEEK_SET))<0) {
	    fprintf(stderr, "Tried to seek to zr_pos %lld\n\n", zr_pos);
	    perror("lseek");	    
	    goto out;
	} else if(zr_pos != result) {
	    fprintf(stderr, "Tried to seek to pos %lld, got %lld!\n", 
	    	zr_pos, result);
	    r=-EINVAL;
	    goto out;
	}

	if((r=read(fd, buf, len))<0) {
	    perror("read");
	    goto out;
	}
	zr_pos += r;
    out: 
    	return r;
}

struct rlog_header * zr_next_schedrec(void) {
	struct rlog_header * rh = (__typeof(rh)) buf;
	char * data = buf + sizeof(*rh);
	int len;

    while(1) {
	len=__zr_read(fd, (char *)rh, sizeof(*rh));
	if(len != sizeof(*rh)) {
	    fprintf(stderr, "tried to read %d bytes, got %d\n", sizeof(*rh), len);
	    exit(1);
	}

	/* 
	 * What the heck is this next_rhid stuff about?
	 */
	if(IS_SCHED(rh->type) && (rh->id >= next_rhid)) {
    	    try(rh->len +sizeof(*rh) <= BUFSIZE);
    	    len = __zr_read(fd, data, rh->len);
    	    if(len != rh->len) {
    		fprintf(stderr, "tried to read %d bytes, got %d\n", rh->len, len);
    	    }
	    break;
	} else {
	    zr_pos += rh->len;
	}

    }
    next_rhid = rh->id;
    return rh;
}
#if 0
struct rlog_header * zr_next_schedrec(void) {
	struct rlog_header * sr;
	sr=(struct rlog_header *)buf;
	char * data = buf + sizeof(*sr);
	/*
	 * If in a log we have the following setup:
	 *  [sysret] [sched] [sysret]
	 * Using only replay_pos, we can't tell if we were before or after
	 * the sched in the middle.  So we use the rhid zR had passed in before.
	 *
	 * During logging, we log the current id; this should always work right.
	 */
	while(1) {
    	    sr = zr_next_entry();
	    if(IS_SCHED(sr->type)
	} while(!IS_SCHED(sr->type) || sr->id < next_rhid);
	next_rhid = sr->id;
	try(sr);
	return sr;
}
#endif

