#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <fcntl.h>

#include <string.h>

#include <revirt/ztrace.h> 
#include <revirt/wraw.h>
#include <asm/unistd.h>

#define O_LARGEFILE 0100000

loff_t lseek64(int fd, loff_t offset, int cmd);

/*
 * wrestore-patchgen mapfile /dev/wraw/wrawN newbase
 */
int main(int argc, char * argv[]) {
	int wfd, omapfd, nmapfd, writtenfd, datafd;
	char fname[256];
	char * buf, * buf_area;

	loff_t disksize;
	int dmapsize, dmapbytes;
	int *datamap;

	loff_t offset;

	struct wraw_grab_arg_s grab_arg;

	int c, i;

	if(argc != 4) {
		fprintf(stderr, "Usage: %s mapfile wrawdev newbase\n",
				argv[0]);
		exit(1);
	}

	/*
	 * Get a page-aligned buffer
	 */
	if((buf_area = malloc(4096*2))==NULL) {
		fprintf(stderr, "malloc for buf_area failed\n");
		exit(1);
	}
	buf = (char *) ((( (unsigned int)buf_area + 4096)) & (~4095));

	/* 
	 * Open all the files 
	 */
	if((omapfd=open(argv[1], O_RDONLY))<0) {
		perror("opening old mapfile");
		exit(1);
	}

	/* Open the wraw device and determine its size */
	if((wfd=open(argv[2], O_RDONLY | O_LARGEFILE))<0) {
		perror("Opening wraw device");
		exit(1);
	}
	/*
	 * FIXME -- get this info w/o grabbing the device...
	 */
	grab_arg.maxoffset = 1;
	if(ioctl(wfd, WRAW_GRAB, &grab_arg)<0) {
		perror("ioctl WRAW_GRAB");
		fprintf(stderr, "errno: %d\n", errno);
		exit(1);
	}
	disksize=grab_arg.maxoffset;
	dmapsize=OFF2BYTE(disksize-1);
	dmapbytes = dmapsize * sizeof(unsigned int) ;
	printf("disksize: %lld dmapsize: %d dmapbytes: %d\n",
		disksize, dmapsize, dmapbytes);

	if(ioctl(wfd, WRAW_RELEASE)<0) {
		perror("WRAW_RELEASE");
		exit(1);
	}

	/* Allocate the buffer and read it in */
	if((datamap=malloc(dmapbytes))==NULL) {
		fprintf(stderr, "datamap malloc(%d) failed.\n",
			dmapbytes);
		exit(1);
	}

	if((c=read(omapfd, datamap, dmapbytes))!=dmapbytes) {
		fprintf(stderr, "datamap read %d bytes, wanted %d\n",
				c, dmapbytes);
		exit(1);
	}

	/* Open the output files */
	sprintf(fname, "%s.map", argv[3]);
	if((nmapfd=open(fname, O_CREAT | O_WRONLY | O_TRUNC, 0644))<0) {
                perror("opening map");
                exit(1);
        }

        sprintf(fname, "%s.written", argv[3]);
        if((writtenfd=open(fname, O_CREAT | O_WRONLY | O_TRUNC, 0644))<0) {
                perror("opening written");
                exit(1);
        }

        sprintf(fname, "%s.data", argv[3]);
        if((datafd=open(fname, O_CREAT | O_WRONLY | O_TRUNC, 0644))<0) {
                perror("opening data");
                exit(1);
        } 

#define Bit2Byte(x) ((x) >> 5)
#define Bit2Bitmask(x) (1 << ((x) & 31))
#define Bit2Off(x) (((long long)x) << (12))

	for(i=0; i < ((disksize-1)>> 12) ; i++) {
		int bi, bm, len;

		bi = Bit2Byte(i);
		bm=Bit2Bitmask(i);

		if(datamap[bi] & bm) {
			//printf("d[%d] == %x & %x\n", bi, datamap[bi], bm);
			/* Read the data */
			offset = Bit2Off(i);
			len=4096;
			if(offset + len > disksize) {
				len = disksize-offset;
			}
			printf("Reading offset 0x%llx, %d bytes\n", 
				offset, len);

			if(lseek64(wfd, offset, SEEK_SET) <0) {
				perror("lseek wfd");
				exit(1);
			}
			if(read(wfd, buf, len) < 0) {
				perror("reading wraw device");
				exit(1);
			}
			if(write(writtenfd, &offset, sizeof(offset))<0) {
				perror("writting writtenfd");
				exit(1);
			}
			if(write(datafd, buf, 4096)<0) {
				perror("writting writtenfd");
				exit(1);
			}
		}
	}
	if(write(nmapfd, datamap, dmapbytes) < 0) {
		perror("writing datamap");
	}
	close(writtenfd);
	close(datafd);
	close(nmapfd);
	close(omapfd);
	close(wfd);
	return 0;
}
