#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <errno.h>
#include <revirt/ztrace.h>
#include <asm/unistd.h>
#include <string.h>

#include <revirt/wraw.h>

#define BUFSIZE 1024 * 1024 * 10

#define O_LARGEFILE 0100000

//#define DISKSIZE ((loff_t)1024 * 3381651)
//#define DMAPSIZE  ((unsigned int)OFF2BYTE(DISKSIZE))
//#define DMAPBYTES (DMAPSIZE*sizeof(unsigned int))

#define dprintf(x...)

#define PGS 256


#if 0
int llseek(unsigned int fd, unsigned long offset_high, 
	    unsigned long offset_low, loff_t * result, unsinged int whence) {
    return syscall(__NR__llseek, fd, offset_high, offset_low, result, whence);
}
#endif

unsigned long long lseek64(int fd, unsigned long long offset, unsigned int whence);

int main(int argc, char * argv []) {
	int cfd, mapfd, writtenfd, datafd, r, c;
	unsigned int i;
	long long total=0;
	loff_t start, result;
	char fname[256];

	char * databuf;
	loff_t * written;

	loff_t disksize;

	unsigned int * datamap;
	unsigned int dmapsize, dmapbytes;
	struct wraw_grab_arg_s grab_arg;


	if(argc != 3) {
		fprintf(stderr, "Usage: %s undobase file\n",
				argv[0]);
		exit(1);
	}

	if((written=malloc(PGS * sizeof(loff_t)))==NULL) {
		fprintf(stderr, "Couldn't allocate written[%d]\n",
				PGS*sizeof(loff_t));
		exit(1);
	}

	if((databuf = mmap(0, PGS*4096, PROT_READ|PROT_WRITE, 
					MAP_PRIVATE | MAP_ANON, 0, 0))<0) {
		perror("mmap'ing databuf");
		exit(1);
	}
	printf("databuf at 0x%x\n", databuf);

	if((cfd=open(argv[2], O_WRONLY | O_LARGEFILE))<0){
		perror("Opening copy file");
		exit(1);
	}

	/*
         * Find the size of the disk.  We need to find a better way
         * to do this...
         */
        if(ioctl(cfd, WRAW_GRAB, &grab_arg)<0) {
                        perror("ioctl WRAW_GRAB");
                        fprintf(stderr, "errno: %d\n", errno);
                        exit(1);
        }
        if(ioctl(cfd, WRAW_RELEASE, NULL)<0) {
                        perror("ioctl WRAW_RELEASE");
                        fprintf(stderr, "errno: %d. WARNING! release failed\n", errno);
                        exit(1);
        }

        disksize=grab_arg.maxoffset;
        dmapsize=OFF2BYTE(disksize-1);
        dmapbytes=dmapsize * sizeof(unsigned int);

        printf("Device size: %lld\n", disksize);

        if((datamap=malloc(dmapbytes+2))==NULL) {
                fprintf(stderr, "datamap malloc(%d) failed.\n",
                                dmapbytes);
                exit(1);
        }
        bzero(datamap, dmapbytes);

	sprintf(fname, "%s.map", argv[1]);
	if((mapfd=open(fname, O_RDONLY))<0){
		perror("Opening map file");
		exit(1);
	}
	if((c=read(mapfd, datamap, dmapbytes))<0) {
		perror("Reading datamap\n");
		exit(1);
	}
	close(mapfd);

	sprintf(fname, "%s.written", argv[1]);
	if((writtenfd=open(fname, O_RDONLY))<0) {
		perror("Opening written");
		exit(1);
	}

	sprintf(fname, "%s.data", argv[1]);
	if((datafd=open(fname, O_RDONLY))<0) {
		perror("Opening data");
		exit(1);
	}



	do {
		r=read(writtenfd, written, PGS * sizeof(loff_t));
		if(r<0) {
			perror("read(writtenfd)");
			exit(1);
		}
		c = r / sizeof(loff_t);
		fprintf(stderr, "Read %d records from writtenfile\n", c);
		r=read(datafd, databuf, 4096 * c);
		if(r<0) {
			perror("read(datafd)");
			exit(1);
		} else if(r < 4096*c) {
			fprintf(stderr, "short read: %d 0x%x\n",
					c, r);
			exit(1);
		}
		for(i=0; i<c; i++) {
		    int bit, byte, bitmask;
		    char * d;

		    start=written[i];
		    d = databuf + 4096 * i;

		    bit=start >> 12;
		    byte = bit >> 5;
		    bitmask = (1 << (bit & 31));


		    //printf("%d: %llx, %x\n", i, start, d);
		    if(byte > dmapsize) {
 		     fprintf(stderr, "offset %llx is greater than disk size %llx (at least, byte %d > dmapsize %d\n",
				start, disksize, byte, dmapsize);
		    } else if((datamap[byte] & bitmask) == 0) {
		     fprintf(stderr, "offset %llx, but bit %d not set: %x %x\n",
				     	start, byte, datamap[byte], bitmask);
		     exit(1);
		    }

		    result=lseek64(cfd, start, SEEK_SET);
		    if(result<0) {
			    perror("lseek");
			    exit(1);
		    } else if(result!=start) {
			    fprintf(stderr, "lseke64: wanted %llx, got %llx\n",
					   start, result);
			    exit(1);
		    }
		    r=write(cfd, d, 4096);
		    total += 4096;
		    if(r<0) {
			    perror("write cfd");
			    exit(1);
		    } else if(r<4096) {
			    fprintf(stderr, "Short write: %d\n", r);
		    }
		}
		

	} while (c==PGS);

	printf("%lld bytes copied\n", total);
	close(datafd);
	close(writtenfd);
	close(cfd);
	return 0;
}
