#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <fcntl.h>

#include <string.h>

#include <revirt/ztrace.h>
#include <revirt/wraw.h>
#include <asm/unistd.h>

int interrupted=0;

void sigint(int s) {
	interrupted=1;
}

//#define DISKSIZE ((loff_t)1024 * 3381651)
//#define DMAPSIZE ((unsigned int)OFF2BYTE(DISKSIZE))
//#define DMAPBYTES ((DMAPSIZE)*sizeof(unsigned int))

#define PGS (256)

unsigned int * datamap;
char * databuf;
loff_t * written;

int main(int argc, char * argv[]) {
	int wfd, mapfd, writtenfd, datafd, c;
	struct wraw_grab_arg_s grab_arg;
	struct wraw_listen_arg_s listen_arg;
	char fname[256];

	loff_t disksize;
	unsigned int dmapbytes, dmapsize;

	if(argc < 3) {
		fprintf(stderr ,"Usage: %s mapfile wrawdev\n", argv[0]);
		exit(1);
	}

	if((written=malloc(PGS * sizeof(loff_t)))==NULL) {
		fprintf(stderr, "datamap malloc(%d) failed.\n",
				PGS * sizeof(loff_t));
		exit(1);
	}
	bzero(written, PGS * sizeof(loff_t));

	if((databuf=mmap(0, PGS * 4096, PROT_READ|PROT_WRITE,
				MAP_PRIVATE|MAP_ANON, -1, 0))<0) {
		perror("mmap'ing databuf");
		exit(1);
	}
	bzero(databuf, PGS * 4096);
	
	signal(SIGINT, sigint);

	sprintf(fname, "%s.map", argv[1]);
	if((mapfd=open(fname, O_CREAT | O_WRONLY | O_TRUNC, 0644))<0) {
		perror("opening map");
		exit(1);
	}

	sprintf(fname, "%s.written", argv[1]);
	if((writtenfd=open(fname, O_CREAT | O_WRONLY | O_TRUNC, 0644))<0) {
		perror("opening written");
		exit(1);
	}

	sprintf(fname, "%s.data", argv[1]);
	if((datafd=open(fname, O_CREAT | O_WRONLY | O_TRUNC, 0644))<0) {
		perror("opening data");
		exit(1);
	}

	if((wfd=open(argv[2], O_RDONLY))<0) {
		perror("opening wraw device");
		exit(1);
	}

	grab_arg.maxoffset = 0xDEADBEEF;

	if(ioctl(wfd, WRAW_GRAB, &grab_arg)<0) {
			perror("ioctl WRAW_GRAB");
			fprintf(stderr, "errno: %d\n", errno);
			exit(1);
	}
	disksize=grab_arg.maxoffset;
	dmapsize = OFF2BYTE(disksize -1);
	dmapbytes = dmapsize * sizeof(unsigned int);

	if((datamap=malloc(dmapbytes))==NULL) {
		fprintf(stderr, "datamap malloc(%d) failed.\n",
				dmapbytes);
		goto out_free;
	}
	bzero(datamap, dmapbytes);

	listen_arg.datamap_u=datamap;
	listen_arg.written_u=written;
	listen_arg.databuf_u=databuf;
	listen_arg.maxwr=PGS;
	listen_arg.curwr=0;

	do {
		//fprintf(stderr, "Calling ioctl...\n");
		c=syscall(__NR_ioctl, wfd, WRAW_LISTEN, &listen_arg);
		if(c<0 && errno != EINTR) { 
			fprintf(stderr, "errno==%d\n", errno);
			break;	
		}
		fprintf(stderr, "liseten_arg.curwr == %d, errno=%d\n", 
				listen_arg.curwr, errno);
		if(write(writtenfd, written, 
				listen_arg.curwr * sizeof(loff_t))<0) {
			perror("Writing writtenfd");
		}
		if(write(datafd, databuf, listen_arg.curwr * 4096)<0) {
			perror("Writing datafd");
		}
	} while(!interrupted);

	perror("ioctl WRAW_LISTEN");

	if((c=write(mapfd, (char *)datamap, dmapbytes))<0) {
		perror("writing dmapfile");
		goto out_free;
	}
	else if(c<dmapbytes) {
		fprintf(stderr, "Warning: write returned %d, wanted %d\n",
				c, dmapbytes);
	}
      out_free:
	if(ioctl(wfd, WRAW_RELEASE)<0) {
		perror("ioctl WRAW_RELEASE");
	}
	close(wfd);
	close(mapfd);
	close(writtenfd);
	close(datafd);
	return 0;
}
