#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#include <string.h>

#include <revirt/ztrace.h>
#include <revirt/wraw.h>
#include <asm/unistd.h>

int interrupted=0;

void sigint(int s) {
	interrupted=1;
}

//#define GUESSDISKSIZE ((loff_t)1024 * 3381651)
#define GUESSDISKSIZE 5
//#define DMAPSIZE ((unsigned int)OFF2BYTE(DISKSIZE))
//#define DMAPBYTES ((DMAPSIZE)*sizeof(unsigned int))

unsigned int * datamap;

int main(int argc, char * argv[]) {
	int wfd, ofd, c;
	struct wraw_grab_arg_s grab_arg;
	struct wraw_listen_arg_s listen_arg;

	loff_t disksize;
	unsigned int dmapsize, dmapbytes;

	if(argc < 3) {
		fprintf(stderr ,"Usage: %s mapfile wrawdev\n", argv[0]);
		exit(1);
	}

	
	signal(SIGINT, sigint);
	if((ofd=open(argv[1], O_CREAT | O_WRONLY | O_TRUNC, 0644))<0) {
		perror("opening mapfile");
		exit(1);
	}
	if((wfd=open(argv[2], O_RDONLY))<0) {
		perror("opening wraw device");
		exit(1);
	}

	grab_arg.maxoffset = GUESSDISKSIZE-1;

	if(ioctl(wfd, WRAW_GRAB, &grab_arg)<0) {
			perror("ioctl WRAW_GRAB");
			fprintf(stderr, "errno: %d\n", errno);
			exit(1);
	}
	disksize=grab_arg.maxoffset;
	dmapsize=OFF2BYTE(disksize-1);
	dmapbytes=dmapsize * sizeof(unsigned int);

	printf("Device size: %lld\n", disksize);

	if((datamap=malloc(dmapbytes))==NULL) {
		fprintf(stderr, "datamap malloc(%d) failed.\n",
				dmapbytes);
		goto out_free;
	}
	bzero(datamap, dmapbytes);

	listen_arg.datamap_u=datamap;
	listen_arg.written_u=NULL;
	listen_arg.databuf_u=NULL;
	listen_arg.maxwr=256;
	listen_arg.curwr=0;

	do {
		//fprintf(stderr, "Calling ioctl...\n");
		c=syscall(__NR_ioctl, wfd, WRAW_LISTEN, &listen_arg);
		if(c<0 && errno != EINTR) { 
			fprintf(stderr, "errno==%d\n", errno);
			break;	
		}
		//fprintf(stderr, "liseten_arg.curwr == %d, errno=%d\n", 
		//		listen_arg.curwr, errno);
	} while(!interrupted);

	perror("ioctl WRAW_LISTEN");
	if((c=write(ofd, (char *)datamap, dmapbytes))<0) {
		perror("writing dmapfile");
		goto out_free;
	}
	else if(c<dmapbytes) {
		fprintf(stderr, "Warning: write returned %d, wanted %d\n",
				c, dmapbytes);
	}
     out_free:
	if(ioctl(wfd, WRAW_RELEASE)<0) {
		perror("ioctl WRAW_RELEASE");
	}
	close(wfd);
	return 0;
}
