#include <asm/ptrace.h> 
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <asm/unistd.h>
#include <stdlib.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#include <time.h>

#include <linux/autoconf.h>
#include <revirt/rlog.h>
#include <revirt/ztrace.h>
#include <revirt/syscalls.h>
#include <revirt/flags.h>

/* from linux/include/linux/sched.h */
#define TASK_RUNNING            0
#define TASK_INTERRUPTIBLE      1
#define TASK_UNINTERRUPTIBLE    2
#define TASK_ZOMBIE             4
#define TASK_STOPPED            8

#include "library.h"
#include "readlog.h"
#include "printlog.h"
#include "decode.h"

#define ZTRACE_BREAKPOINT_METHOD
#define ZTRACE_OVERFLOW_METHOD
#define ZTRACE_FAST_REPSTRING_HANDLING
//#define SET_NETWORK_THROTTLE 1
#define ZR_CHECKPOINT -512

/* Buffer to deal with overflow imprecision */
#define OVF_BUFFER (0x100)

#define RECORD_MAX(x, y) \
	if((x) > (y)) { \
	    (y)=(x); \
	}

struct {
    int max_ovf_error;
    __s64 max_bc;
    int max_intr;
    int max_pf;
    __s64 max_real;
    int nr_breakpoint;
    int max_breakpoint;
    int ztrace_call[32];
    int replen[8];

    int nr_repstring, cnt_repstring, nr_repstring_sstep,
	cnt_repstring_sstep;
    int skipped_as_null;
    int no_action;
    int did_something;
} rstat;

int logcount[6]= {0, 0, 0, 0, 0, 0};
unsigned long log_id;
unsigned long last_sched_id;

int slow_replay=0;

unsigned long debug_from=0, debug_to=0;
int debug_on=0;
int debuglevel_save=-1;


#define try(x) do { if (!(x)) die(__FILE__,__LINE__,#x);  } while(0)

int debuglevel=0;

#ifndef ZTRACE_SSTEP_METHOD 
# ifndef ZTRACE_BREAKPOINT_METHOD
#  error "define a method first"
# endif
#endif

#define DEBUG0 1
#define DEBUG1 (1 << 1)
#define DEBUG2 (1 << 2)
#define DEBUG3 (1 << 3)
#define THROTTLE (1 << 4)

/* None of this should change between checkpoints */
struct task {
	int has_run;
	int pid;
};

extern loff_t zr_pos;
extern unsigned long next_rhid;

#define MAX_TASKS 10
struct task tasklist[MAX_TASKS];
int mapl;
int stop_request=0;

#ifdef ZR_KEEP_STATEREC
#define STATE_HISTORY 100
#define STATEREC(x) (staterec[(x)%STATE_HISTORY])
int staterec_index=0, staterec_lastlog=0;
struct ztrace_state_struct staterec[STATE_HISTORY];

int writestaterec(void) {
    int i, fd;
    if((fd = creat("staterec", 0644))<0) {
	perror("creat");
	return -1;
    }
    for(i=staterec_index; i<staterec_index + STATE_HISTORY; i++) {
	if( write(fd, (char *)&STATEREC(i), sizeof(struct ztrace_state_struct))
		    	<0) {
		perror("write");
		return -1;
	}
    }
    return 0;
}

void sigusr1(int sig) {
	fprintf(stderr, "zR: Writing staterec\n");
	writestaterec();
}
#endif

void sigint(int sig) {
	static int c=0;
	struct ztrace_state_struct zt_state;
	if(c) {
		fprintf(stderr, "Second interrupt received, dying.\n");
		exit(1);
	} else {
		c=1;
#if 0
		stop_request=1;
		printf("Interrupt received, putting in stop request.\n");
#else
		/* UMLinux specific!  Host-based needs last proc run */
		if(ztrace(ZTRACE_GETSTATE, tasklist[0].pid,
			0, &zt_state)) {
			perror("getstate");
		}
		if(zt_state.revirt_flags & REPLAY_SYSCALL) {
		 printf("Stop request received, but still in syscall.  Calling SYSCALL_EXIT...\n");
		 ztrace(ZTRACE_SYSCALL_EXIT,tasklist[0].pid,0,0);
		} 
		printf("Stop request received, exiting\n");
		exit(1);
		
#endif
	}
}

void debug_pause(void) {
	fprintf(stderr, "Pausing with SIGSTOP\n");
	fflush(stdout);
	kill(getpid(), SIGSTOP);
}

void die(char *file,int line,char *ass) {
	fprintf(stderr, "file %s, line %d: assertion %s failed",
		file,line,ass);
	perror("system error is");
	*((int*)0)= 0;
}

void print_yield(int r) {
	printf(" ztrace yielded at ");
	switch(r) {
		case ZTRACE_SSTEP:
			printf("ZTRACE_SSTEP\n");
			break;
		case ZTRACE_PERF:
			printf("ZTRACE_PERF\n");
			break;
		case ZTRACE_BLOCK:
			printf("ZTRACE_BLOCK\n");
			break;
		case ZTRACE_SYSCALL_ENTER:
			printf("ZTRACE_SYSCALL_ENTER\n");
			break;
		case ZTRACE_SYSCALL_EXIT:
			printf("ZTRACE_SYSCALL_EXIT\n");
			break;
		case ZTRACE_XSTOP:
			printf("ZTRACE_XUSER_STOP\n");
			break;
		case ZTRACE_SIGNALQUEUE:
			printf("ZTRACE_SIGNALQUEUE\n");
			break;
		default:
			printf("unknown\n");
			break;
	}
}



struct task * new_task(int pid) {
	/*
	 * FIXME -- make this more general (i.e., arbitrary num processes)
	 */
	try(mapl<MAX_TASKS);
	tasklist[mapl]= (struct task) {
		.pid= pid, 
		.has_run= 0 , 
		};
	mapl++;
	if(debuglevel & DEBUG0)
		printf("New task %d\n",pid);
	return tasklist+mapl-1;

}

void unmap_task(struct task *tsk) {
	try(mapl>0);
	if(debuglevel & DEBUG0)
		printf("No longer tracing task %d\n",tsk->pid);
	/*
  	 * move last task in list to deleted task's space
	 */
	mapl--;
	*tsk= tasklist[mapl];
}

struct task *find_by_pid(int pid) {
	int i;
	for(i= 0;i<mapl;i++)
		if (tasklist[i].pid==pid)
			return tasklist+i;
	fprintf(stderr, "Cannot find pid %d\n", pid);
	try(0);
	return 0;
}

void singlestep_task(struct task *task) {
	struct ztrace_state_struct state; 
	int ret;
	int last_yield;

	ztrace(ZTRACE_GETSTATE, task->pid, 0, &state);

	last_yield=ztrace(ZTRACE_SSTEP_WAIT,task->pid,0,0);
	if(debuglevel & DEBUG1)
	    print_yield(last_yield);
	if(ret==ZTRACE_SIGNALQUEUE) {
		if(debuglevel & DEBUG1)
		  printf(" sstep interrupted by signal!\n");
	}
}

/*
 * Single-step through repstring until we get to the next instruction
 *
 * Well, OK, now we decode the instruction, set a bkpt for the next eip, and
 * let it run.  
 */
void repstring_step(struct task * task) {
	struct ztrace_state_struct state;
	unsigned long this_eip;
	struct ztrace_ctl_struct ztc;
	struct ztrace_breturn bbuf;
	int d;
	__s64 count, xret;
	int last_yield;

	rstat.nr_repstring_sstep++;
	ztrace(ZTRACE_GETSTATE, task->pid, 0, &state);
#ifdef SSTEP_THRU_REPSTRING 
	this_eip = state.regs.eip;
	while(1) {
		ztrace(ZTRACE_GETSTATE, task->pid, 0, &state);
		if(this_eip) {
			if(this_eip!=state.regs.eip)
				break;
		}
#if 0
 		else
			this_eip=state.regs.eip;
#endif
		if(singlestep_task(task)) {
			fprintf(stderr, "Wierd... singlestep_task returns non-zero\n");
			debug_pause();
		}
		rstat.cnt_repstring_sstep++;
	}
	d=state.regs.eip - this_eip;
	if(d >0 && d < 8) {
		rstat.replen[d]++;
	} else
		rstat.replen[0]++;
#else /* sstep_thru_repstring */
	/*
 	 * At this point, it's already established that we've got
	 * a repstring; we just need to know how long it is.  For now
	 * we assume that the actual opcode is only one byte long; 
  	 * the rest comes from prefixes (including the 'rep' prefix)
	 */
	 /* FIXME -- check to make sure this covers all cases */
	count=ADJ_BC(state.ctr);
	xret = parse_prefixes(state.code, &d);
	this_eip = state.regs.eip + d + 1;
	if(debuglevel & DEBUG1) {
		printf("parse_prefixes returns xret = 0x%x, offset = %d\n", 
				xret, d);
		printf("repstring_step: stepping from eip %x to %x\n",
			state.regs.eip, this_eip);
		if(debuglevel & DEBUG2)
			debug_pause();
	}

	ztc.flags = ZTRACE_BKPT;
	ztc.yield=0;
	ztc.pid = task->pid;
	ztc.rhid =log_id;
	ztc.last_sched_id = last_sched_id;

	ztc.perferr=0;

	ztc.brk.eip=this_eip;
	ztc.brk.wchan=0;
	ztc.brk.flags=0;
	ztc.brk.count=count;
	ztc.active.i=1;
	ztc.active.pid[0]=task->pid;

	ztc.bret = &bbuf;

	if((last_yield=ztrace(ZTRACE_CTL,task->pid,(unsigned long)&ztc,0))<0){
		perror("ztrace_rununtil");
		exit(1);
	 }
	if(last_yield== ZTRACE_WAITEXIT) {
	    fprintf(stderr, "ZTRACE_CTL BKPT returned WAITEXIT, exiting...\n");
    	    exit(1);
       	}
	 if(debuglevel & DEBUG1)
	    print_yield(last_yield);

	 if(last_yield == ZTRACE_RESTORE_CHKPT) {
	 	fprintf(stderr, "repstring_step: ztrace returned RESTORE_CHKPT!\n");
	 }
	
	 xret=ADJ_BC(bbuf.ctr);
	 if(xret - count > 0) {
	 	printf("decode & breakpoint returned %d!\n", xret);
	 }
#endif
}

/*
 * Step through a repz instruction until current ecx == logged ecx (data).
 */
void handle_repstring(struct task *task,unsigned int data) {
	struct ztrace_state_struct state;
	unsigned long this_eip=0, eflags=0;
	long ecx=0;
	unsigned long real_ecx=0;
	long delta;
	rstat.nr_repstring++;

	ztrace(ZTRACE_GETSTATE, task->pid, 0, &state);
	real_ecx=state.regs.ecx;
	this_eip=state.regs.eip;
	eflags = state.regs.eflags;

	/*
	 * FIXME -- both of these methods have the disadvantage that they
	 * modify ecx.  This is fine if we're only ever running repstring
	 * instructions; but what if the repstring is interrupted by a 
	 * guest page fault?  The pf handler might see a different ecx...
	 */
#ifdef ZTRACE_FAST_REPSTRING_HANDLING
	/*
	 * FIXME -- is this a problem with 16 bit addr-size attributes?
	 */
	ecx = state.regs.ecx;
	delta = ecx - data ;

	ztrace(ZTRACE_PUTREG, task->pid, ECX, (void *)delta);
	/*
	 * Step through the instruction
	 */
	repstring_step(task);
	/*
	 * Calculate 'real' value of ecx...
	 */
	ztrace(ZTRACE_GETSTATE, task->pid, 0, &state);
	real_ecx = ecx - (delta - state.regs.ecx);

	/*
	 * Check to see if everything's ok...
	 */
	if(state.regs.ecx != 0) {
		fprintf(stderr, "hmm, repstring_step, ecx != 0?\n");
	}

	if(real_ecx != data) {
		fprintf(stderr, "real_ecx != data: ecx %d delta %d state.regs.ecx %d data %d real_ecx %d\n",
			ecx, delta, state.regs.ecx, data, real_ecx);
		fprintf(stdout, "real_ecx != data: ecx %d delta %d state.regs.ecx %d data %d real_ecx %d\n",
			ecx, delta, state.regs.ecx, data, real_ecx);
		debug_pause();
	}

	/*
	 * And reset our EIP and ECX back to normal
	 */
	if(eflags != state.regs.eflags) {
		if( ztrace(ZTRACE_PUTREG, task->pid, EFL, (void *)eflags) ) {
			perror("ztrace_putreg: eflags");
		}
	}
	if( ztrace(ZTRACE_PUTREG, task->pid, EIP, (void *)this_eip) ) {
		perror("ztrace_putreg: eip");
	}
	if(ztrace(ZTRACE_PUTREG, task->pid, ECX, (void *)real_ecx) ) {
		perror("ztrace_putreg: ecx");
	}
#else
	while(1) {
		/*
		 * Pentum 4 has fast string instructions, which if > a
		 * certain amount jump down by 128 (or something like that)
		 * rather than one at a time if greater than a certain
		 * amount.
		 * 
		 * So we make it small (2) so that it only goes up / down
		 * by 1.  We maintain real_ecx in parallel, moving it up
		 * or down as we see state.regs.ecx move up or down from 2.
		 */
		ztrace(ZTRACE_PUTREG, task->pid, ECX, (void *)2);
		/* FIXME -- This getstate is redundant. */
		ztrace(ZTRACE_GETSTATE, task->pid, 0, &state);
		if(state.regs.ecx != 2) {
		    fprintf(stderr, "ZTRACE_PUTREG didn't change ecx from %d to 2.\n",
			    	state.regs.ecx);
		    debug_pause();
		}

		/*
		 * Make sure we haven't drifted or anything wierd
		 * like that...
		 */
		if(this_eip) {
		    if(this_eip!=state.regs.eip) {
			fprintf(stderr, "Wrong eip! want %x, got %x\n",
				this_eip, state.regs.eip);
			debug_pause();
		    }
		}

		/*
		 * Are we there yet?
		 */
		if (real_ecx == data)
			break;
		else if(debuglevel & DEBUG1)
			printf(" repstring: ecx==%d [%d].  want %d\n",
					real_ecx, state.regs.ecx, data);

		ecx=state.regs.ecx;
		if(singlestep_task(task)) {
		   fprintf(stderr, "That's wierd... singlestep_task returns non-zero.\n");
		   debug_pause();
		}
		ztrace(ZTRACE_GETSTATE, task->pid, 0, &state);
		switch(state.regs.ecx - ecx) {
		   case 1:
		       fprintf(stderr, "ecx goes up!\n");
		       real_ecx+=1;
		       break;
		   case -1:
		       real_ecx-=1;
		       	break;
		   default:
			fprintf(stderr, "ecx jumps from %d to %d!\n",
					ecx, state.regs.ecx);
			real_ecx += (state.regs.ecx-ecx);
			//debug_pause();
		}
		rstat.cnt_repstring++;
	}
	ztrace(ZTRACE_PUTREG, task->pid, ECX, (void *)real_ecx);
#endif
}

int parse_prefixes(unsigned char * code, int * cstart);

void sync_with_checkpoint(struct task *task) {
    struct ztrace_state_struct state;
   /*
     * replay_pos will be the file position of the *next* log that the 
     * kernel will read; log_pos is the file position of the last log that
     * the kernel read.  For checkpoints taken during replay, we need to 
     * fast-forward to the right rhid anyway, so being a little slower won't
     * matter; and during logging, that's all we have.
     * 
     * So, copy log_pos from the state struct, and rhid from the zt_ctl struct.
     */
    ztrace(ZTRACE_GETSTATE,task->pid,0,&state);
    printf("  last_yield returned RESTORE_CHKPT, checkpoint restore in progress.\n   setting zr_pos from %lld to %lld, returning to top level.\n",
    	zr_pos, state.log_pos);

    printf("  setting next_rhid from %lu to %lu\n", next_rhid,
    		state.zt_ctl.rhid);
    printf("  setting last_sched_id from %lu to %lu\n", last_sched_id,
    		state.zt_ctl.last_sched_id);
    last_sched_id = state.zt_ctl.last_sched_id;
    next_rhid = state.zt_ctl.rhid;
    zr_pos=state.log_pos;
}


/*
 * We should only stop for system calls if we want to -- i.e., 
 * REPLAY_SSTEP should be set *only* when single-stepping through
 */
__s64 breakpoint(struct task *task,unsigned eip, int flags, unsigned long wchan, 
			__s64 real_count, struct rlog_schedule_entry * log) {
	struct ztrace_breturn bbuf;
	struct ztrace_state_struct state;
	//struct ztrace_rununtil_struct ru;
	struct ztrace_ctl_struct ztc;
	__s64 ret,xret;
	int last_yield;
	//unsigned long task_state = REPLAY_TSK_GETSTATE(flags);

	ret= 0;

	/* 
 	 * First, if we're already at the bkpt in userland, step once.  
	 * If this is a string instruction, step through to the end.
	 *
	 * If we're in the kernel, the fact that we're already at the
	 * breakpoint means that the bkpt is at our return address.
	 * If we want to stop right there, single-step will confound
	 * us.
	 */
        ztrace(ZTRACE_GETSTATE, task->pid, 0, &state);
	/* 
	 * UMLinux tailoring: for zero branch counts, don't do single-stepping.
	 * This allows us to go to usermode and back again -- useful if we
	 * need to deliver two signals in a row.
	 */
        if(state.regs.eip==eip 
		&& state.revirt_flags & REPLAY_PROCESS_IN_UMODE
		&& real_count > 0) {

		if(debuglevel & DEBUG1) 
			printf("Single-stepping over ip...\n");
                //task->was_taken= 0;
                //task->syscall_pending= 0;

		/*
		 * Step over this eip.  If a repstring, decode & bkpt; 
		 * otherwise, just use the TRAP flag.
		 *
		 * ...unless there's been a signal delivered that hasn't
		 * been handled yet.  In that case, repstring_step will break
		 * at totally the wrong place...
		 */
		if(is_repstring(state.code) 
			&& !(state.revirt_flags & REPLAY_FIRST_AFTER_SIG)){
		    if(debuglevel & DEBUG1)
			printf(" instruction is repstr, stepping through...\n");
		    repstring_step(task);
		} else {
                	singlestep_task(task);
		}
#if 0
                if (task->syscall_pending) {
			printf("Calling ZTRACE_SYSCALL_EXIT\n");
                        ztrace(ZTRACE_SYSCALL_EXIT,task->pid,0,0);
		} 
#endif
        }

	/* 
	 * FIXME -- I'm pretty sure this shouldn't be here... it should be
	 * next to the 'XUSER' stop stuff.
	 *
	 * Anyway, if the process blocked during logging, wait for it
	 * to block in the same place...
	 *
	 * Hmm... no, block is actually used *in place of* the breakpoint.
	 * Eh...this needs some more thought.
	 *
	 * OK, and sometimes a process may block 'in user mode' -- i.e., 
	 * waiting for a page.  Those we're not worried about; we only want
	 * the wchan for blocking in kernel mode.
	 */
	if(wchan 
		&& (flags & REPLAY_PROCESS_IN_UMODE) 
		&& !(flags & REPLAY_SIGBLOCK)
		&& (debuglevel & DEBUG0)) {
		printf("Warning: skipping wchan, process in umode (flags %x)\n",
			flags);
	}
	if(wchan 
	    && (!(flags & REPLAY_PROCESS_IN_UMODE) 
	         || (flags & REPLAY_SIGBLOCK))) {

#if 0
	 ru.eip=eip;
	 ru.wchan=wchan;
	 ru.flags=flags;
	 ru.rhid=log_id;
	 ru.active = log->active;
#endif
	 ztc.flags = ZTRACE_BLOCK_STOP;
	 ztc.yield=0;
	 ztc.pid=task->pid;
	 ztc.rhid=log_id;
	ztc.last_sched_id = last_sched_id;

	 ztc.perferr=0;

	 ztc.brk.eip=eip;
	 ztc.brk.wchan=wchan;
	 ztc.brk.flags=flags;
	 ztc.brk.count=real_count;

	 ztc.active.i=1;
	 ztc.active.pid[0]=task->pid;

	 ztc.bret = &bbuf;

	 if(flags & REPLAY_SIGBLOCK
	    && (debuglevel & DEBUG0)) {
	 	printf("wchan, SIGBLOCK\n");
	 }

	 if(debuglevel & DEBUG1)
		printf("Waiting for block 0x%x:0x?%x\n", eip, wchan);

	 if((last_yield=ztrace(ZTRACE_CTL, task->pid, (unsigned long)&ztc, 0))<0) {
	 	perror("ztrace_block");
		exit(1);
	 } else if(last_yield == ZTRACE_RESTORE_CHKPT) {
	 	fprintf(stderr, "bkpt-wchan: ztrace returned RESTORE_CHKPT!\n");
		sync_with_checkpoint(task);
		return ZR_CHECKPOINT;
	 }
	if(last_yield== ZTRACE_WAITEXIT) {
	    fprintf(stderr, "ZTRACE_CTL BLOCK_STOP returned WAITEXIT, exiting...\n");
    	    exit(1);
       	}

	 if(debuglevel & DEBUG1)
	    print_yield(last_yield);
	} else {
	 /*
	  * Now set the breakpoint and run.
	  */
	 if(debuglevel & DEBUG1)
		printf("Setting breakpoint for 0x%x\n", eip);
	ztc.flags = ZTRACE_BKPT;
	ztc.pid=task->pid;
	ztc.rhid=log_id;
	ztc.last_sched_id = last_sched_id;

	ztc.yield=0;
	ztc.perferr=0;

	ztc.brk.eip=eip;
	ztc.brk.wchan=0;
	ztc.brk.flags=flags;
	ztc.brk.count=real_count;

	ztc.active.i=1;
	ztc.active.pid[0]=task->pid;

	ztc.bret=&bbuf;
#if 0
	 ru.eip=eip;
	 ru.flags=flags;
	 ru.wchan=0;
	 ru.count = real_count;
	 ru.rhid=log_id;
	 ru.active = log->active;
#endif

	 /*
	  * FIXME -- this is a hack, until we re-write zR for proper signal delivery
	  *
	  * This should make it block at ztrace_tsk_exits_syscall, but not 
	  * ztrace_rexec_done().
	  */
	 if(!(flags & REPLAY_PROCESS_IN_UMODE) && (log->stype == STYPE_SIGNAL) )
	 	ztc.brk.flags  |= REPLAY_PROCESS_IN_UMODE;

	if(!(ztc.brk.flags & REPLAY_PROCESS_IN_UMODE))
		ztc.flags |= ZTRACE_BKPT_IN_KERNEL;

	 if((last_yield=ztrace(ZTRACE_CTL, task->pid, (unsigned long)&ztc, 0))<0) {
		perror("ztrace_rununtil");
		exit(1);
	 } else if(last_yield == ZTRACE_RESTORE_CHKPT) {
	 	fprintf(stderr, "bkpt-rununtil: ztrace returned RESTORE_CHKPT!\n");
		sync_with_checkpoint(task);
		return ZR_CHECKPOINT;
	 }
	 if(last_yield== ZTRACE_WAITEXIT) {
	    fprintf(stderr, "ZTRACE_CTL BKPT returned WAITEXIT, exiting...\n");
    	    exit(1);
       	 }
	 if(debuglevel & DEBUG1)
	    print_yield(last_yield);
	}

	xret = ADJ_BC(bbuf.ctr);
	if (bbuf.eip!=eip) {
		printf("eips differ. new=%x, orig= %x\n",bbuf.eip, eip);
		if(last_yield==ZTRACE_SIGNALQUEUE)
		    printf("Returned from signalqueue, ignoring...\n");
		else
		    try(bbuf.eip==eip);
	}
	if(debuglevel & DEBUG1) {
		printf("task %d returns with eip= %x bc= %x\n",task->pid,
			bbuf.eip, xret);
#ifdef REPLAY_RECORD_ALL_PERFCTRS
		printf("returned counts: hw=%x i=%x pf=%x def=%x\n",
			bbuf.ctr.hw_branch, bbuf.ctr.hw_intr, bbuf.ctr.ex_pf,
			 bbuf.ctr.ex_dev);
#else
		printf("returned counts: hw=%x\n", bbuf.ctr.hw_branch);
#endif
	}
	ret = xret;

	return ret;
}

int do_epochs(struct task * task, struct revirt_active_pid_st * active, 
			int eip, int flags, __s64 count, int stype) {
    struct ztrace_ctl_struct ztc;
    struct ztrace_state_struct state;
    int last_yield;
//    struct task* t;
    //loff_t pos=0;

    ztc.flags = ZTRACE_EPOCH_STOP;
    ztc.pid = task->pid;
    ztc.rhid=log_id;
	ztc.last_sched_id = last_sched_id;
    
    ztc.yield=0;
    ztc.perferr=0;
    
    ztc.brk.eip=eip;
    ztc.brk.flags = flags;
    ztc.brk.wchan=0;
    ztc.brk.count=count;
    
    ztc.active = *active;
#if 0
    if(stype == STYPE_END) {
    	fprintf(stderr, "do_epochs: STYPE_END...\n");
     	debuglevel_save=debuglevel;
    	debuglevel=3;
	ztrace(ZTRACE_SET_OPT, getpid(), 0x300, (void *)0x300);
    }
#endif
    
    ztc.bret=NULL;

    ztrace(ZTRACE_GETSTATE,task->pid,0,&state);

    while(state.active.epochs_skipped < active->epochs_skipped) {

	if(debuglevel & DEBUG0) {
	    printf("Epoch %d, want %d, calling ZTRACE_EPOCH...\n", 
		state.active.epochs_skipped, active->epochs_skipped);
	}

	if((last_yield=ztrace(ZTRACE_CTL, task->pid, (unsigned long)&ztc, 0))<0)
		perror("ZTRACE_EPOCH");
	else if(last_yield== ZTRACE_WAITEXIT) {
	    fprintf(stderr, "ZTRACE_CTL EPOCH_STOP returned WAITEXIT, exiting...\n");
    	    exit(1);
       	 }
	else if(last_yield == ZTRACE_RESTORE_CHKPT) {
		printf("  last_yield returned RESTORE_CHECKPOINT, chkpt restore in progress.\n");
	    sync_with_checkpoint(task);
	    return ZR_CHECKPOINT; 
	} else if(debuglevel & DEBUG1)
		printf("  epoch returned %d\n", last_yield); 
	if(stype == STYPE_END) {
	    if(last_yield != ZTRACE_RESTORE_CHKPT) {
		fprintf(stderr, "WARNING: STYPE_END, but yield returned %x!\n", last_yield);
		print_yield(last_yield);
	    } 
#if 0
	    else {
		debuglevel=debuglevel_save;
		debuglevel_save=-1;
		ztrace(ZTRACE_SET_OPT, getpid(), 0x300, (void *)0x000);
	    }
#endif
	}

	ztrace(ZTRACE_GETSTATE,task->pid,0,&state);
    }
    return last_yield;
}

void do_schedule_end_entry(struct rlog_schedule_entry *log, struct task * task) {
    struct ztrace_ctl_struct ztc;
    int last_yield;
//    struct task* t;
    //loff_t pos=0;

    ztc.flags = ZTRACE_CHECKPOINT_WAIT;
    ztc.pid = task->pid;
    ztc.rhid=log_id;
    ztc.last_sched_id = last_sched_id;
    
    ztc.yield=0;
    ztc.perferr=0;
    
    ztc.brk.eip=log->eip;
    ztc.brk.flags = log->flags;
    ztc.brk.wchan=0;
    ztc.brk.count=ADJ_BC(log->ctr);
    
    ztc.active = log->active;

    if((last_yield=ztrace(ZTRACE_CTL, task->pid, (unsigned long)&ztc, 0))<0)
		perror("ZTRACE_EPOCH");
    if(last_yield== ZTRACE_WAITEXIT) {
    	fprintf(stderr, "ZTRACE_CTL CHECKPOINT_WAIT returned WAITEXIT, exiting...\n");
	exit(1);
    }
    if(last_yield !=  ZTRACE_RESTORE_CHKPT) {
    	fprintf(stderr, "do_schedule_end_entry: last yield %x!\n", last_yield);
    } else {
	sync_with_checkpoint(task);
#if 0
	debuglevel=debuglevel_save;
	debuglevel_save=-1;
	ztrace(ZTRACE_SET_OPT, getpid(), 0x300, (void *)0x000);
#endif
    }
}

void do_schedule_entry(struct rlog_schedule_entry *log) {
	struct task *task;
	__s64 real_count;
	/* FIXME these are "local vars" for the overflow method... */
	__s64 ovf_remaining, ovf_taken, task_counter;
	int iterations;
	struct ztrace_state_struct state;
	int errors=0, flag_doone=0, overflow_iterations=0;
	int last_yield;
	int did_something=0;

	if(debuglevel & DEBUG0) {
		 printf("do_schedule_log: %d\n", log->pid);
		print_schedlog(log);
	}

	if(log->stype == STYPE_SYNCBLOCK
		|| log->stype == STYPE_NULL) {
	    if(debuglevel & DEBUG0)
		printf(" STYPE_{SYNCBLOCK|NULL}, skipping\n");
    	    return;
	}

	if(!(task=find_by_pid(log->pid)))  {
		printf("WARNING, can't find pid %d!\n", log->pid);
		debug_pause();
		return;
	}

#if 0
	/*
	 * If we've never run this pid before, attach to it
	 */
	if(!task->has_run) {
	    try(!ztrace(ZTRACE_ATTACH, task->pid,0,0));
	    if(debuglevel & DEBUG0) 
		printf("Attached to pid %d\n", task->pid);
	    task->has_run = 1;
	}
#endif

	real_count = ADJ_BC(log->ctr);

	/*
	 * Log some statistics...
	 */
	RECORD_MAX(real_count, rstat.max_real);
	RECORD_MAX(log->ctr.hw_branch, rstat.max_bc);
#ifdef REPLAY_RECORD_ALL_PERFCTRS
	RECORD_MAX(log->ctr.hw_intr, rstat.max_intr);
	RECORD_MAX(log->ctr.ex_pf, rstat.max_pf);
#endif

	if(debuglevel & DEBUG1) {
#ifdef REPLAY_RECORD_ALL_PERFCTRS
	    printf("real count: %llx [%llx %x %x]\n",
			real_count,
			log->ctr.hw_branch,
			log->ctr.hw_intr,
			log->ctr.ex_pf);
#else
	    printf("real count: %llx [%llx]\n",
			real_count);
#endif
	}

	if(real_count < 0) {
		printf("Warning: real_count < 0, treating as zero...\n");
		errors++;
		real_count=0;
	}

	/*
	 * First, step to the right epoch.
	 */
	if( (last_yield=do_epochs(task, &log->active, log->eip, log->flags, real_count, log->stype) )
								== ZR_CHECKPOINT) {
	    printf("do_epochs returned ZR_CHECKPOINT, returning\n");
	    return;
	}

	if(log->stype == STYPE_END) {
	    do_schedule_end_entry(log, task);
	    return;
	}

	ztrace(ZTRACE_GETSTATE,task->pid,0,&state);
	task_counter= state.ctr.hw_branch;
	if(debuglevel & DEBUG1) {
	    printf("task_counter: %lld\n", task_counter);
	}

	/*
 	 * Do some sanity checks, see if we can get away with skipping
	 * this log.
	 *
	 * Never skip signal logs, or sigblocks that are actually blocks
	 */
#define FLAGMASK ((0xffffff) & ~(REPLAY_LOG|REPLAY_REPLAY|REPLAY_WAIT|REPLAY_FIRST_AFTER_SIG|REPLAY_ZTRACE_CAN_CHECKPOINT|REPLAY_USER_TRAP))
       {
	unsigned int compare_eip=state.regs.eip;
	unsigned int compare_flags=state.revirt_flags;
	if(debuglevel & DEBUG1)
		printf(" compare{ eip: 0x%x flags: 0x%x } log->flags %x\n",
				compare_eip,
				compare_flags,
				log->flags);

	if(!(compare_flags & REPLAY_FIRST_AFTER_SIG)
	//!task->first_after_sig
  	   && (real_count - task_counter)== 0 
	   && log->eip == compare_eip
	   && log->stype != STYPE_SIGNAL) {

	   if(debuglevel & DEBUG0)
	    printf("real_count-task_counter==0, log->eip == compare_eip, checking for more..\n");
	   if((log->flags&FLAGMASK) == (compare_flags&FLAGMASK)) {
	    if(debuglevel & DEBUG0) 
		printf("flags are the same\n");
	    if((log->flags & REPLAY_SIGBLOCK)){
	    	if((log->wchan==0)) {
		    if(debuglevel & DEBUG0)
			printf("SIGBLOCK, but wchan 0! treating as null...\n");
		    rstat.skipped_as_null++;
		    goto out;
		} else if(1 || debuglevel & DEBUG0)
		    printf("SIGBLOCK, non-null wchan, running...\n");
	    } else if(log->flags & REPLAY_PROCESS_IN_UMODE) {
		if(debuglevel & DEBUG0)
		    printf("process in umode");
		/* 
		 * This is kind of tough.  We need to keep logs in which only the
		 * ecx changes.  But in this case, we're already there.  I guess we
		 * can check the actual state... 
		 *
		 * The problem is if a signal is sent from an interrupt that happened
		 * before the process' frame was set up ; in that case, the logged ecx
		 * might not be the actual value of the ecx at that point.
		 * 
		 * We could also check the eip to see if this is actually a string 
		 * instruction.
		 */
		/* This is spurious right now... */
		//ztrace(ZTRACE_GETSTATE,task->pid,0,&state);
		if((log->data == state.regs.ecx)) {
			if(debuglevel & DEBUG0)
			  printf("Zero count, no change in eip, same flags," \
					" at ecx. Treating as null...\n");
			if(debuglevel & DEBUG2) debug_pause();
			rstat.skipped_as_null++;
			goto out;
		} else if(debuglevel & DEBUG2) {
			fprintf(stderr, "WARNING: only ecx changes; " \
				"suspect multiple interruption of string" \
				" instruction\n");
			errors++;
		}
	     } else { // log->flags & REPLAY_PROCESS_IN_UMODE
		if(log->nrxuser == state.nrxuser) {
		/* 
		 * FIXME -- need to deal w/ multiple interruptions of same copyuser.. 
		 * This information should be in zt_ctl.xuser_stop, I think
		 */
		//&& log->edi == task->lastlog.edi) {
			if(debuglevel & DEBUG0)
			   printf("Still in kernel, same nrxuser & edi."\
				" Treating as null.\n");
			rstat.skipped_as_null++;
			goto out;
		} else {
		    printf("log->nrxusr %d state.nrxuser %d\n",
		    		log->nrxuser, state.nrxuser);
		    if(1 && debuglevel & DEBUG0)
		      printf("Still in kernel, nrxuser or edi changes. Continuing.\n");
		}
	     }
	    } else { // log->flags == task->lastlog.flags
	    	if(debuglevel & DEBUG0)
		    printf("flags differ: log %x state %x (flagmask %x)\n",
		    	log->flags, compare_flags, FLAGMASK);
	    }
	}
       }

#ifdef ZR_KEEP_STATEREC
	ztrace(ZTRACE_GETSTATE,task->pid, 0, &STATEREC(staterec_index));
	STATEREC(staterec_index).debugreg[1]=logcount[0];
	STATEREC(staterec_index).debugreg[2]=0;
	staterec_index++;
#endif

	/* 
	 * Each of these "methods" should stop with the next eip as prescribed
	 * in the log, and the # of branches.  If the next eip is on the 
	 * return from a syscall, they should stop on the entry to
	 * the syscall.  After that we decide whether to stop in the middle
	 * of the syscall or go to the end.
	 */

#ifdef ZTRACE_SSTEP_METHOD
#error ZTRACE_SSTEP_METHOD no longer supported
#endif

#ifdef ZTRACE_OVERFLOW_METHOD
    	/*
	 * First get as close as possible using performance counter overflow
	 * then we step closer.
	 *
	 * FIXME for some reason, we're getting OVF interrupts waaay early.  I'll
	 * look into the reason later, but for now, keep using the overflow method
	 * until we're within range.
	 */

	for(ovf_remaining=real_count - task_counter; 
		ovf_remaining > OVF_BUFFER; 
		ovf_remaining = real_count - ovf_taken) {
	    __s64 b1, b2, ocount;
	    struct ztrace_breturn bret;
//	    struct ztrace_rununtil_struct rununtil;
	    struct ztrace_ctl_struct ztc;

	    did_something=1;

	    /* First, wake it up if it's not already... */
	ztrace(ZTRACE_GETSTATE,task->pid, 0, &state);
	while(REPLAY_TSK_GETSTATE(state.revirt_flags) == TASK_INTERRUPTIBLE
	      || REPLAY_TSK_GETSTATE(state.revirt_flags) == TASK_STOPPED) {
	    if(debuglevel & DEBUG2)
		printf("Pid %d in state %d, doing WAKEUP\n",
	    		task->pid, REPLAY_TSK_GETSTATE(state.revirt_flags));
	    ztrace(ZTRACE_WAKEUP, task->pid, 0, 0);
    	    ztrace(ZTRACE_GETSTATE,task->pid,0,&state);
	}

	    if(overflow_iterations > 0) {
	    	fprintf(stderr, "Warning: overflow_iterations == %d\n",
			overflow_iterations);
	    }
	    overflow_iterations++;

	    ztrace(ZTRACE_GETSTATE,task->pid,0,&state);
	    b1=state.ctr.hw_branch;

	    ocount = ovf_remaining - OVF_BUFFER;
	    if(debuglevel & DEBUG1) {
		printf("%d:setting performance overflow for %lld branches\n", 
				task->pid, ocount);
	    }
#if 0
	    rununtil.eip=0x0;
	    rununtil.wchan=0x0;
	    rununtil.flags = log->flags;
	    rununtil.count = ocount;
	    rununtil.rhid = log_id;
	    rununtil.active = log->active;
#endif
	    ztc.flags = ZTRACE_PERF;
	    ztc.pid=task->pid;
	    ztc.rhid=log_id;
	    ztc.last_sched_id = last_sched_id;

	    /* FIXME eventually, do we want to set this here? */
	    ztc.yield=ztc.perferr=0;

	    ztc.brk.eip=0;
	    ztc.brk.wchan=0;
	    ztc.brk.flags=log->flags;
	    ztc.brk.count=ocount;

	    ztc.active.i=1;
	    ztc.active.pid[0]=task->pid;

	    ztc.bret = &bret;

	    if((last_yield=ztrace(ZTRACE_CTL, task->pid, 
	    				(unsigned long)&ztc, 0)) < 0) {
		perror("performance overflow");
	    } else if(last_yield== ZTRACE_WAITEXIT) {
		fprintf(stderr, "ZTRACE_CTL PERF returned WAITEXIT, exiting...\n");
		exit(1);
	    }
	    if(debuglevel & DEBUG1)
	       	print_yield(last_yield);

	    ovf_taken=ADJ_BC(bret.ctr);
	    task_counter=ovf_taken;

	    if(ovf_taken == ovf_remaining) {
		printf("ovf_taken==ovf_remaining!\n");
	    }

	    b2=bret.ctr.hw_branch - b1 - ocount;

	    RECORD_MAX(b2, rstat.max_ovf_error);

	    if(debuglevel&DEBUG1) {
		printf("%d: overflow returned %lld (error %lld)\n",
			task->pid, ovf_taken, b2);
	    }

	    //if(ovf_taken > ovf_remaining) {
	    if(ovf_taken > real_count) {
#ifdef REPLAY_RECORD_ALL_PERFCTRS
		fprintf(stderr, "Overflow overcounted! wanted %lld (%lld) [%llx %x %x], got %lld [%llx %x %x]\n",
				real_count, 
				ovf_remaining,
				log->ctr.hw_branch,
				log->ctr.hw_intr,
				log->ctr.ex_pf,
				ovf_taken,
				bret.ctr.hw_branch,
				bret.ctr.hw_intr,
				bret.ctr.ex_pf);
		printf("Overflow overcounted! wanted %lld (%lld) [%llx %x %x], got %lld [%llx %x %x]\n",
				real_count, 
				ovf_remaining,
				log->ctr.hw_branch,
				log->ctr.hw_intr,
				log->ctr.ex_noerror,
				ovf_taken,
				bret.ctr.hw_branch,
				bret.ctr.hw_intr,
				bret.ctr.ex_pf);
#else
		fprintf(stderr, "Overflow overcounted! wanted %lld (%lld) [%llx], got %lld [%llx]\n",
				real_count, 
				ovf_remaining,
				log->ctr.hw_branch,
				ovf_taken,
				bret.ctr.hw_branch);
		printf("Overflow overcounted! wanted %lld (%lld) [%llx], got %lld [%llx ]\n",
				real_count, 
				ovf_remaining,
				log->ctr.hw_branch,
				ovf_taken,
				bret.ctr.hw_branch);
#endif
		debug_pause();
	    }
	}

#endif
	
#ifdef ZTRACE_BREAKPOINT_METHOD
	if(real_count - task_counter == 0) {

	/*
	 * If we're in a system call and the log says we're not, step out of it.
	 *
	 * (Note that if an unlogged task blocks during logging and not in replay, it may
	 * stop in ztrace_tsk_exits_syscall(), which right now has UMODE set.  So for
	 * now we only do doone if the log wants to be in umode and we're not.)
	 */
	    if(log->stype == STYPE_END) {
		fprintf(stderr, "log->type == STYPE_END, setting flag_doone!\n");
		flag_doone=1;
	    }

	    ztrace(ZTRACE_GETSTATE,task->pid,0,&state);
	    if(!(state.revirt_flags&REPLAY_PROCESS_IN_UMODE) 
		&& (log->flags & REPLAY_PROCESS_IN_UMODE)) {
	    	if(debuglevel & DEBUG0)
		    printf("state !UMODE, log wanted UMODE, setting flag_doone (%x %x)\n",
			state.revirt_flags, log->flags);
		flag_doone=1;
	    }
	/*
	 * FIXME
	 * zero count means we put two
	 * signals sequentially on the stack; but then we need to call
	 * do_signal again, which means we need to return to user-mode
	 * again... anyway, set the breakpoint.
	 *
	 * OK, now we know if it's a signal or not; if so...
	 *
	 * And of course, this only counts if there were no epochs in between.
	 * Er, no, that's not quite right... this only counts if *this task*
	 * hasn't run in between.
	 *
	 * lastlog.sype==STYPE_SIGNAL and task->epochs_skipped=0 
	 * should both be implied by FIRST_AFTER_SIG.
	 *
	 * If we're getting a signal, sometimes that has other side effects,
	 * like STOPping all other processes.  This is important if one of
	 * the next logs is an EPOCH with several procs running.  So if it's
	 * a signal, and we want to move from SYSCALL to UMODE, do it...
	 */
	    if(log->stype == STYPE_SIGNAL
	    	&& state.revirt_flags & REPLAY_FIRST_AFTER_SIG) {
	    	if(debuglevel & DEBUG0)
	    	    printf("real_count - task_counter ==0, stype SIGNAL, FIRST_AFTER_SIG;  setting flag_doone.\n");
	    	flag_doone=1;
	    }

	/* 
	 * If we're waiting on a sigblock, (i.e., SIGSTOP or ptrace redirection)
	 * we always need to roll forward, to get the block...
	 *
	 * Well, we don't need to roll forward if we're already there. =)
	 * ...unless we got two signals in a row; that's what the FIRST_AFTER_SIG check
	 * is for.
	 *
	 * Hmm... it appears that there *ARE* spurious SIGBLOCK logs; we only want to
	 * do this if the process in question was actually blocked.  So check wchan
	 * first.
	 *
	 * OK, so... apparently there's multiple SIGBLOCKs in a row; it seems to happen
	 * naturally -- you can tell by the fact that the pid in question isn't actually
	 * blocked... I think...
	 */
	    if(log->wchan && log->flags & REPLAY_SIGBLOCK) {
		if(!(state.revirt_flags&REPLAY_SIGBLOCK)
		 	|| (REPLAY_TSK_GETSTATE(state.revirt_flags)==TASK_RUNNING)
			|| (log->flags & REPLAY_FIRST_AFTER_SIG)) {
		 if(state.revirt_flags & REPLAY_SYNC_BLOCK) {
		   fprintf(stderr, "WARNING, setting flag_doone from SYNC_BLOCK! (#%lu)\n",
		   		log_id);
		 }
		 if(REPLAY_TSK_GETSTATE(state.revirt_flags)==TASK_RUNNING) {
		    if(debuglevel & DEBUG0)
		   fprintf(stderr, "WARNING, log SIGBLOCK (%x), state SIGBLOCK but not running (%x), setting flag_doone (#%lu)\n", log->flags, state.revirt_flags, log_id);
		 } else if(log->flags & REPLAY_FIRST_AFTER_SIG) {
		   fprintf(stderr, "WARNING, log SIGBLOCK|FIRST_AFTER_SIG (%x), state SIGBLOCK (%x), setting flag_doone (#%lu)\n", log->flags, state.revirt_flags, log_id);
		 }
	    	 if(debuglevel & DEBUG0)
	    	     printf("log->flags & REPLAY_SIGBLOCK. setting flag_doone\n");
		 flag_doone=1;
		} else {
		  fprintf(stderr, "WARNING, log SIGBLOCK (%x), state SIGBLOCK (%x), leaving flag_doone (#%lu)\n", log->flags, state.revirt_flags, log_id);
		}
	    }

	/*
	 * Not sure about this one... blocks are not likely to be spurious.  Is it
	 * possible?  Well for now...
	 *
	 * We don't want to force a block we've already completed the system call, it
	 * will go to the next stoppping point (syscall_exit for rexec_done, the next
	 * syscall for syscall_exit), and we may want to stop before then (in the case of
	 * another xuser interruption).  But only if we're blocking from within the
	 * system call; if we're doing a SIGBLOCK, we want to go to the next block even
	 * if we're starting from rexec_done() or syscall_exit.
	 *
	 * [IOW, do one if either we're NOT at rexec_done or syscall_exit, OR if 
	 *  we're doing a sigblock.]
	 *
	 * Actually, this should be caught above... strike that, do some more
	 * investigation...
	 *
	 * OK, it looks like it has to do with ptrace stuff, which doesn't have the
	 * sigblock flag.  What we really want is REPLAY_SYSCALL.  Hmm... I wonder
	 * if this means we'll have trouble that I took out that "REPLAY_SYSCALL" flag
	 * in rexec_done...  yes, I think we might.  Hmm...
	 *
	 * Added check for current state IN_UMODE.
	 */
	    if(log->wchan 
		&& !(log->flags & REPLAY_PROCESS_IN_UMODE)
		&& !(state.revirt_flags & REPLAY_PROCESS_IN_UMODE)
		&& log->edi==0
		&& !(((state.zt_ctl.yield == ZTRACE_XSTOP) 
		      ||(state.zt_ctl.yield == ZTRACE_SYSCALL_EXIT)) 
		      && log->flags & REPLAY_SYSCALL) )
	    {
	    	if(debuglevel & DEBUG0)
		    printf("log->wchan %x log->flags %x state.zt_ctl.yield %x\n",
				log->wchan, log->flags, state.zt_ctl.yield);
		flag_doone=1;
	    }
	}
	/*
	 * In the case of a signal schedule and a zero count, the eip may be different;
	 * that's acceptable.  Ditto for SIGNALQUEUE returns; it means during logging
	 * the signal was put in the queue between the time the last signal was
	 * dequeued and the time the eip was changed.  (I.e., the logged eip is "wrong").
	 */
	iterations=0;
	while(task_counter!=real_count 
		|| state.regs.eip!=log->eip 
		|| flag_doone) {
		__s64 taken;
		iterations++;
	
	    did_something=1;
	    /* First, wake it up if it's not already... */
       if(iterations == 1) {
	while(REPLAY_TSK_GETSTATE(state.revirt_flags) == TASK_INTERRUPTIBLE
	      || REPLAY_TSK_GETSTATE(state.revirt_flags) == TASK_STOPPED) {
	    if(debuglevel & DEBUG2)
		printf("Pid %d in state %d, doing WAKEUP\n",
	    		task->pid, REPLAY_TSK_GETSTATE(state.revirt_flags));
	    ztrace(ZTRACE_WAKEUP, task->pid, 0, 0);
    	    ztrace(ZTRACE_GETSTATE,task->pid,0,&state);
	}
	}


		if( real_count == task_counter
			&& iterations > 0
			&& state.zt_ctl.yield == ZTRACE_SIGNALQUEUE) {
			fprintf(stderr, "Warning: right count, >1 iteration, signalqueue; eip %x want %x, continuing anyway\n",
				state.regs.eip, log->eip);
			break;
		}

		flag_doone=0; /* FIXME */

#ifdef ZR_KEEP_STATEREC
		ztrace(ZTRACE_GETSTATE,task->pid, 0, &STATEREC(staterec_index));
		STATEREC(staterec_index).debugreg[1]=logcount[0];
		STATEREC(staterec_index).debugreg[2]=iterations;
		staterec_index++;
#endif

		taken= breakpoint(task,log->eip, log->flags, log->wchan, real_count, log);
		if(taken == ZR_CHECKPOINT) {
		    printf("breakpoint returned ZR_CHECKPOINT, returning\n");
		    return;
		}
		task_counter= taken; 

		if(debuglevel & DEBUG1)
		 printf("%d:Breakpoint returns %lld.  Total now %lld; want %lld.\n", 
				logcount[0], taken, task_counter, real_count);

		ztrace(ZTRACE_GETSTATE, task->pid, 0, &state);
		if (taken > real_count) {
		    fprintf(stderr, "%d: Breakpoint returns %lld [%llx], wanted %lld[%llx]\n",
				logcount[0],
				taken,
				state.ctr.hw_branch,
				real_count, 
				log->ctr.hw_branch);
		    printf("%d: Breakpoint returns %lld [%llx], wanted %lld[%llx]\n",
				logcount[0],
				taken,
				state.ctr.hw_branch,
				real_count, 
				log->ctr.hw_branch);
#if 0
		    printf("last_log:\n");
		    print_state(&STATEREC(staterec_lastlog));
#endif
#ifdef ZR_KEEP_STATEREC
		    if(staterec_index>0) {
			printf("last_bkpt:\n");
			print_state(&STATEREC(staterec_index-1));
		    }
#endif
		    printf("current:\n");
		    print_state(&state);
		    debug_pause();
		    errors++;
		    break;
		} else if (taken == real_count-1 && (debuglevel & DEBUG3)) {
		    printf("Close shave: %lld [%llx], log %lld[%llx]\n",
				taken,
				state.ctr.hw_branch,
				real_count, 
				log->ctr.hw_branch);
		}

		if(debuglevel & DEBUG1)
			printf("  eip=0x%x ecx=%d rep=%c\n",
				state.regs.eip, state.regs.ecx,
				is_repstring(state.code)?'y':'n');
		if(state.code[0]==0xcc) {
			printf("Warning: bkpt on int3\n");
		    	print_state(&state);
			//debug_pause();
		}
	}

#ifdef ZR_KEEP_STATEREC
	ztrace(ZTRACE_GETSTATE,task->pid, 0, &STATEREC(staterec_index));
	STATEREC(staterec_index).debugreg[1]=logcount[0];
	STATEREC(staterec_index).debugreg[2]=iterations+1;
	staterec_index++;
#endif

	ztrace(ZTRACE_GETSTATE,task->pid,0,&state);
#endif /* ZTRACE_BREAKPOINT_METHOD */

	/* 
	 * Now we're to the eip in question.  If we stopped halfway thru a
	 * syscall, walk to the right place; If we stopped in the middle of a
	 * repstring, walk until we're at the right iteration.
	 */
	if(!(log->flags & REPLAY_PROCESS_IN_UMODE)) {
		struct ztrace_ctl_struct ztc;

		if(log->nrxuser || log->edi) {
		 ztc.flags = ZTRACE_XUSER_STOP;
		 ztc.pid=task->pid;
		 ztc.rhid=log_id;
		ztc.last_sched_id = last_sched_id;
		 
		 ztc.yield=ztc.perferr=0;

		 ztc.xstop.nrxuser=log->nrxuser;
		 ztc.xstop.edi=log->edi;

		 if(log->edi) {
		    fprintf(stderr, "WARNING: log %lu has non-zero edi %lx\n",
		    		log_id, log->edi);
		 }

		 ztc.active.i=1;
		 ztc.active.pid[0]=task->pid;

		 ztc.bret=NULL;

		did_something=1;
		 if(1/*debuglevel & DEBUG1*/) {
		  printf("Log %d interrupted in syscall.  calling ZTRACE_XSTOP.\n",
		  		log_id);
		  printf("    edi:0x%x nrxuser:%d\n", ztc.xstop.edi, 
				ztc.xstop.nrxuser);
		 }

		 if((last_yield=ztrace(ZTRACE_CTL, task->pid, (unsigned long)&ztc,0))<0)
		 { 
			perror("XSTOP");
		  } else if(last_yield== ZTRACE_WAITEXIT) {
		    fprintf(stderr, "ZTRACE_CTL XUSER_STOP returned WAITEXIT, exiting...\n");
		    exit(1);
		 } else if(last_yield == ZTRACE_RESTORE_CHKPT) {
		    sync_with_checkpoint(task);
		    return ; 
		 }else if(1/*debuglevel & DEBUG1*/)
		    print_yield(last_yield);
		} else if(debuglevel & DEBUG1) {
		 printf("Stopped in kernel, but before any xusers.\n");
		}
	} else if (is_repstring(state.code) && (state.regs.ecx != log->data)) {
	    if(REPLAY_TSK_GETSTATE(state.revirt_flags)!= TASK_RUNNING) {
	    	fprintf(stderr, "WARNING, handle_repstring but state.revirt_flags==%x!\n",
			REPLAY_TSK_GETSTATE(state.revirt_flags));
	    }
	    if(real_count == 0) {	
		fprintf(stderr, "handle_repstring for 0 count (%x %x) (#%lu)\n",
			state.regs.ecx, log->data, log_id);
	    }
		if(debuglevel & DEBUG0)
		 printf(" Handling repeat string instruction, from %x to %x\n",
		 		state.regs.ecx, log->data);
		did_something=1;
		handle_repstring(task,log->data);
		ztrace(ZTRACE_GETSTATE,task->pid,0,&state);
	}

    out:
	if(!did_something) {
	    rstat.no_action++;
	} else {	
	    rstat.did_something++;
	}
	last_sched_id = log_id;

	//task->epochs_skipped=0;

	/*
	 * FIXME -- we need to add this check back; it's just tough with
	 * the signalqueue wierdness...
	 */
	//errors += ((unsigned)state.regs.eip != log->eip);
	errors += (task_counter != real_count);
	if((state.regs.ecx != log->data) 
		&& (state.revirt_flags & REPLAY_PROCESS_IN_UMODE)) {
		if(state.zt_ctl.yield == ZTRACE_SYSCALL_EXIT)
			fprintf(stderr, "ecx differs; ignoring due to SYSCALL_EXIT\n");
		else
		    errors ++;
	}

	if((debuglevel & DEBUG3) || errors)
		printf("log done. pid=%d\n"
		 "          eip= %08x count= %08llx ecx= %08x\n" 
		 "          eip= %08x count= %08llx ecx= %08x (logged)\n",
		 task->pid, 
		 (unsigned) state.regs.eip, task_counter, state.regs.ecx,
		 log->eip, real_count, log->data);
	if(errors) {
		printf("%d warnings.\n", errors);
#ifdef ZR_KEEP_STATEREC
		printf("Writing staterec to disk...");
		writestaterec();
		printf("done.\n");
#endif
		debug_pause();
	}
	//task->lastlog=*log;
	if(ztrace(ZTRACE_CLEARBRANCH, task->pid, 1, 0)) {
		perror("ZTRACE_CLEARBRANCH!\n");
		try(0);
	}

	if(log->stype == STYPE_DMA) {
	    ztrace(ZTRACE_DMA, task->pid, 0, 0);
	}
}

void print_sigset_t(__kernel_sigset_t * t) {
	int i, j, count=1;

	for(i=0; i<_NSIG_WORDS_K; i++) {
	    for(j=0; j<_NSIG_BPW_K; j++, count++) {
		if(t->sig[i] & (1 << j)) 
			fprintf(stderr, "%d ", i*_NSIG_BPW_K + j);
	    }
	}

	fputc('\n', stderr);
}

void deliver_signals(struct rlog_entry_signal * s) {
	struct task * task;

	if(!(task=find_by_pid(s->pid))) 
		return;

	//task->first_after_sig=1;
	if(debuglevel & DEBUG3) {
	 fprintf(stderr, "Delivering signal to pid %d\n", s->pid);
	}
	if(ztrace(ZTRACE_SIGNALQUEUE, s->pid, 0, s)<0) {
		perror("signalqueue");
		exit(1);
	}

}

void do_exit_entry(struct rlog_exit_entry * ee) {
	struct task *task;
	struct ztrace_ctl_struct ztc; 
	int epochs;
	int i;

	if(!(task=find_by_pid(ee->pid))) 
		return;

	if(debuglevel & DEBUG0) 
		printf("Exit Log -- pid %d\n", task->pid);

	/*
	 * First, step to the right epoch.
	 */
	epochs = do_epochs(task, &ee->active, 0, 0, 0, 0);
	if(epochs == ZR_CHECKPOINT) {
		printf("do_epochs returned ZR_CHECKPOINT, returning\n");
		return;
	}

	/*
	 * Now wait for the exit
	 */
	ztc.flags = ZTRACE_WAITEXIT;
	ztc.pid=task->pid;
	ztc.rhid=log_id;
	ztc.last_sched_id = last_sched_id;

	ztc.yield=ztc.perferr=0;

	ztc.brk.eip=0;
	ztc.brk.wchan=0;
	ztc.brk.flags=0;
	ztc.brk.count=0;

	//ztc.active = ee->active;
	ztc.active.i=1;
	ztc.active.pid[0]=task->pid;

	ztc.bret=NULL;

	ztrace(ZTRACE_CTL, task->pid,(unsigned long )&ztc,0);

	ztrace(ZTRACE_DETACH, task->pid,0,0);

	unmap_task(task);

	/* 
	 * And if there were any other procs, clear their branchcounts
	 */
	for(i=0; i<ee->active.i ; i++) {
	    if(ee->active.pid[i] != task->pid) {
		int opid=ee->active.pid[i];

		fprintf(stderr, "zR do_exit: CLEARBRANCH pid %d\n", 
	    		opid);
		if(ztrace(ZTRACE_CLEARBRANCH, opid, 1, 0)) {
	    	    perror("ZTRACE_CLEARBRANCH!\n");
	    	    try(0);
		}
	    }
	}

	return ;
}

void do_fork_entry(struct rlog_entry_fork * fe) {
	if(debuglevel & DEBUG0)
		printf("Creating new task for pid %d\n", fe->child);
	new_task(fe->child);
}

void do_replay(void) {
	struct rlog_header * rh;
	struct rlog_data * rd;

	while(1) {

	    if(slow_replay) {
	    	struct timespec t = { .tv_sec=0, .tv_nsec=500000};
		nanosleep(&t, NULL); 
	    }

	    rh=zr_next_schedrec();
	    rd=DATAOF(rh);
	    log_id=rh->id;

	
	    if(!debug_on) {
		if(rh->id >= debug_from && rh->id <= debug_to) {
		    fprintf(stderr, "id %lu, turning on extra debugging\n", rh->id);
		    debug_on=1;
		    debuglevel_save=debuglevel;
		    debuglevel=3;
		    ztrace(ZTRACE_SET_OPT, getpid(), 0x300, (void *)0x300);
		}
	    } else {
	    	if(rh->id < debug_from || rh->id > debug_to) {
		    fprintf(stderr, "id %lu, turning off extra debugging\n", rh->id);
		    debug_on=0;
		    debuglevel=debuglevel_save;
		    debuglevel_save=-1;
		    ztrace(ZTRACE_SET_OPT, getpid(), 0x300, (void *)0x000);
		}
	    }
#if 0
	    if(notdone==1 && (rh->id > 139100)) {
	     	printf("Turning on debugging\n");
		notdone=2;
		debuglevel = 3;
		ztrace(ZTRACE_SET_OPT, getpid(), 0x300, (void *)0x300);
	    } else if(notdone==2 && rh->id > 140000){
	    	printf("Turning off debugging\n");
		debuglevel = 0;
		ztrace(ZTRACE_SET_OPT, getpid(), 0x300, (void *)0x000);
		notdone=0;
	    }
#endif
	    switch(rh->type) {
	    	case ENTRY_SCHED:
			if(debuglevel & DEBUG0)
			 printf("--- Schedule entry id %d [#%d] --- \n", 
			 		rh->id, logcount[0]++);
			do_schedule_entry((struct rlog_schedule_entry *)rd);
			//last_sched_id = rh->id;
			break;
		case ENTRY_SIGNAL:
			if(debuglevel & DEBUG0)
			 printf("--- Signal entry id %d [#%d] ---\n", 
			 		rh->id, logcount[2]++);
			deliver_signals((struct rlog_entry_signal *)rd);
			if(stop_request) {
				struct ztrace_state_struct zt_state;
				/* UML specific! */
				if(ztrace(ZTRACE_GETSTATE, tasklist[0].pid,
						0, &zt_state)) {
					perror("getstate");
				}
				if(zt_state.revirt_flags & REPLAY_SYSCALL) {
				 printf("Stop request received, but still in syscall.  Carrying on.\n");
				} else {
				 printf("Stop request received, exiting after signal entry.\n");
				 exit(1);
				}
			}
			break;
		case ENTRY_EXIT:
			if(debuglevel & DEBUG0)
			 printf("--- Exit entry id %d [#%d]--- \n", 
			 		rh->id, logcount[3]++);
			do_exit_entry((struct rlog_exit_entry *)rd);
			break;
		case ENTRY_FORK:
			if(debuglevel & DEBUG0)
			 printf("--- Fork entry id %d [#%d]---\n", 
			 	rh->id, logcount[5]++);
			do_fork_entry((struct rlog_entry_fork *)rd);
			break;
		case ENTRY_TOP:
		case ENTRY_EXECPARM:
			if(debuglevel & DEBUG0)
			 printf("Silly entry %d, skipping.\n", logcount[4]++);
			break;
		default:
			fprintf(stderr, "Unknown log: 0x%x\n", rh->type);
			debug_pause();
			break;
	    }
    	    if (!mapl)
		break;
	}
}

int attach_to_group(int pid) {
    int i, err;
    struct revirt_active_pid_st active;
    struct task * task;

    mapl= 0;
    printf("Trying to attach to process %d\n", pid);
    err=ztrace(ZTRACE_ATTACH, pid,(unsigned long)&active,0);
    if(err<0) {
    	perror("ztrace_attach");
	exit(1);
    }
    for(i=0; i<active.i; i++) {
    	printf("Creating task struct for pid %d\n", active.pid[i]);
	task=new_task(active.pid[i]);
    }
    sync_with_checkpoint(task);
    return 0;
}

int main(int argc,char **argv) {
	int i;

	debuglevel=DEBUG0;
	switch(argc) {
	    case 6:
		debug_from = atol(argv[4]);
		debug_to=atol(argv[5]);
		fprintf(stderr, "Extra debugging from %lu to %lu\n", debug_from, debug_to);
	      /* FALLTHRU */
	    case 4:
		debuglevel=atoi(argv[3]);
	      /* FALLTHRU */
	    case 3:
		break;
	    case 5:
	    	fprintf(stderr, "Focused debugging requires both a from and a to.\n");
	    	/* FALLTHRU */
	    default:
		fprintf(stderr, "usage: %s <pid> log [debugmask] [from to]\n",argv[0]);
		exit(1);
	}

	if(signal(SIGINT, sigint)==SIG_ERR) {
		perror("installing sigint handler");
	}
#ifdef ZR_KEEP_STATEREC
	if(signal(SIGUSR1, sigusr1)==SIG_ERR) {
		perror("installing sigusr1 handler");
	}
#endif
	printf("zr pid: %d\n", getpid());

	bzero(&rstat, sizeof(rstat));
	bzero(tasklist, sizeof(struct task) * MAX_TASKS);
	//bzero(staterec, sizeof(struct ztrace_state_struct) * STATE_HISTORY);

	zr_open_log(argv[2]);

	attach_to_group(atoi(argv[1]));

	do_replay();

	printf("ztrace calls:\n");
	for(i=0; i<8; i++) {
		printf("[%2d] %6d [%2d] %6d [%2d] %6d [%2d] %6d\n",
			i, rstat.ztrace_call[i],
			i+8, rstat.ztrace_call[i+8],
			i+16, rstat.ztrace_call[i+16],
			i+24, rstat.ztrace_call[i+24]);
	}
	printf("repstring nr: %d cnt:%d\nrepstring sstep nr: %d cnt: %d\n", 
			rstat.nr_repstring, rstat.cnt_repstring,
			rstat.nr_repstring_sstep, rstat.cnt_repstring_sstep);
	printf("repstring distribution: %d %d %d %d %d %d %d %d\n",
			rstat.replen[0],
			rstat.replen[1],
			rstat.replen[2],
			rstat.replen[3],
			rstat.replen[4],
			rstat.replen[5],
			rstat.replen[6],
			rstat.replen[7]);
	printf("Highest overflow error: %d\n", rstat.max_ovf_error);
	printf("Max hw_branch: %lld\nMax hw_intr: %d\nMax real_count: %lld\n",
			rstat.max_bc,
			rstat.max_intr,
			rstat.max_real);
	printf("Logs where nothing was done: %d  skipped early: %d did something: %d\n",
		rstat.no_action, rstat.skipped_as_null, rstat.did_something);
	return 0;
}
